/*
Copyright (C) 1997-2001 Id Software, Inc.
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// cl_ents.c -- entity parsing and management

#include "client.h"

/*
=========================================================================

FRAME PARSING

=========================================================================
*/

/*
=================
CL_ParseEntityBits

Returns the entity number and the header bits
=================
*/

int CL_ParseEntityBits (unsigned *bits)
{
	unsigned	b, total;
	int			number;

	total = MSG_ReadByte (&net_message);
	if (total & U_MOREBITS1)
	{
		b = MSG_ReadByte (&net_message);
		total |= b<<8;
	}
	if (total & U_MOREBITS2)
	{
		b = MSG_ReadByte (&net_message);
		total |= b<<16;
	}
	if (total & U_MOREBITS3)
	{
		b = MSG_ReadByte (&net_message);
		total |= b<<24;
	}

	if (total & U_NUMBER16)
		number = MSG_ReadShort (&net_message);
	else
		number = MSG_ReadByte (&net_message);

	*bits = total;

	return number;
}

/*
==================
CL_ParseDelta

Can go from either a baseline or a previous packet_entity
==================
*/
void CL_ParseDelta (entity_state_t *from, entity_state_t *to, int number, unsigned bits)
{
	// set everything to the state we are delta'ing from
	*to = *from;

	VectorCopy (from->origin, to->old_origin);
	to->number = number;

	if (bits & U_SOLID)
		to->solid = MSG_ReadShort (&net_message);

	if (bits & U_MODEL)
		to->modelindex = MSG_ReadByte (&net_message);
	if (bits & U_MODEL2)
		to->modelindex2 = MSG_ReadByte (&net_message);
	if (bits & U_MODEL3)
		to->modelindex3 = MSG_ReadByte (&net_message);
		
	if (bits & U_FRAME8)
		to->frame = MSG_ReadByte (&net_message);
	if (bits & U_FRAME16)
		to->frame = MSG_ReadShort (&net_message);

	if ((bits & U_SKIN8) && (bits & U_SKIN16))		//used for laser colors
		to->skinnum = MSG_ReadLong(&net_message);
	else if (bits & U_SKIN8)
		to->skinnum = MSG_ReadByte(&net_message);
	else if (bits & U_SKIN16)
		to->skinnum = MSG_ReadShort(&net_message);

	if ( (bits & (U_EFFECTS8|U_EFFECTS16)) == (U_EFFECTS8|U_EFFECTS16) )
		to->effects = MSG_ReadLong(&net_message);
	else if (bits & U_EFFECTS8)
		to->effects = MSG_ReadByte(&net_message);
	else if (bits & U_EFFECTS16)
		to->effects = MSG_ReadShort(&net_message);

	if ( (bits & (U_RENDERFX8|U_RENDERFX16)) == (U_RENDERFX8|U_RENDERFX16) )
		to->renderfx = MSG_ReadLong(&net_message);
	else if (bits & U_RENDERFX8)
		to->renderfx = MSG_ReadByte(&net_message);
	else if (bits & U_RENDERFX16)
		to->renderfx = MSG_ReadShort(&net_message);

	if (bits & U_ORIGIN1)
		to->origin[0] = MSG_ReadCoord (&net_message);
	if (bits & U_ORIGIN2)
		to->origin[1] = MSG_ReadCoord (&net_message);
	if (bits & U_ORIGIN3)
		to->origin[2] = MSG_ReadCoord (&net_message);
		
	if ( (bits & U_ANGLE1) && (to->solid == SOLID_BMODEL) )
		to->angles[0] = MSG_ReadAngle16(&net_message);
	else if (bits & U_ANGLE1)
		to->angles[0] = MSG_ReadAngle(&net_message);

	if ( (bits & U_ANGLE2) && (to->solid == SOLID_BMODEL) )
		to->angles[1] = MSG_ReadAngle16(&net_message);
	else if (bits & U_ANGLE2)
		to->angles[1] = MSG_ReadAngle(&net_message);

	if ( (bits & U_ANGLE3) && (to->solid == SOLID_BMODEL) )
		to->angles[2] = MSG_ReadAngle16(&net_message);
	else if (bits & U_ANGLE3)
		to->angles[2] = MSG_ReadAngle(&net_message);

	if (bits & U_OLDORIGIN)
		MSG_ReadPos (&net_message, to->old_origin);

	if (bits & U_SOUND)
		to->sound = MSG_ReadByte (&net_message);

	if (bits & U_EVENT)
		to->event = MSG_ReadByte (&net_message);
	else
		to->event = 0;

	if (bits & U_LIGHT)
		to->light = MSG_ReadLong (&net_message);
}

/*
==================
CL_DeltaEntity

Parses deltas from the given base and adds the resulting entity
to the current frame
==================
*/
void CL_DeltaEntity (frame_t *frame, int newnum, entity_state_t *old, unsigned bits)
{
	centity_t	*ent;
	entity_state_t	*state;

	ent = &cl_entities[newnum];

	state = &cl_parse_entities[cl.parse_entities & (MAX_PARSE_ENTITIES-1)];
	cl.parse_entities++;
	frame->num_entities++;

	CL_ParseDelta (old, state, newnum, bits);

	// some data changes will force no lerping
	if (state->modelindex != ent->current.modelindex
		|| state->modelindex2 != ent->current.modelindex2
		|| state->modelindex3 != ent->current.modelindex3
		|| abs(state->origin[0] - ent->current.origin[0]) > 512
		|| abs(state->origin[1] - ent->current.origin[1]) > 512
		|| abs(state->origin[2] - ent->current.origin[2]) > 512
		|| state->event == EV_TELEPORT
		)
	{
		ent->serverframe = -99;
	}

	if (ent->serverframe != cl.frame.serverframe - 1)
	{	// wasn't in last update, so initialize some things
		ent->trailcount = 1024;		// for diminishing rocket / grenade trails

		// duplicate the current state so lerping doesn't hurt anything
		ent->prev = *state;

		if (state->event == EV_TELEPORT)
		{
			VectorCopy (state->origin, ent->prev.origin);
			VectorCopy (state->origin, ent->lerp_origin);
		}
		else
		{
			VectorCopy (state->old_origin, ent->prev.origin);
			VectorCopy (state->old_origin, ent->lerp_origin);
		}
	}
	else
	{	// shuffle the last state to previous
		ent->prev = ent->current;
	}

	ent->serverframe = cl.frame.serverframe;
	ent->current = *state;
}

/*
==================
CL_ParsePacketEntities

An svc_packetentities has just been parsed, deal with the
rest of the data stream.
==================
*/
void CL_ParsePacketEntities (frame_t *oldframe, frame_t *newframe)
{
	int			newnum;
	unsigned	bits;
	entity_state_t	*oldstate;
	int			oldindex, oldnum;

	newframe->parse_entities = cl.parse_entities;
	newframe->num_entities = 0;

	// delta from the entities present in oldframe
	oldindex = 0;
	if (!oldframe)
		oldnum = 99999;
	else
	{
		if (oldindex >= oldframe->num_entities)
			oldnum = 99999;
		else
		{
			oldstate = &cl_parse_entities[(oldframe->parse_entities+oldindex) & (MAX_PARSE_ENTITIES-1)];
			oldnum = oldstate->number;
		}
	}

	while (1)
	{
		newnum = CL_ParseEntityBits (&bits);
		if (newnum >= MAX_EDICTS)
			Com_Error (ERR_DROP, "CL_ParsePacketEntities: bad number:%i", newnum);

		if (net_message.readcount > net_message.cursize)
			Com_Error (ERR_DROP, "CL_ParsePacketEntities: end of message");

		if (!newnum)
			break;

		while (oldnum < newnum)
		{	// one or more entities from the old packet are unchanged
			if (cl_shownet->value == 3)
				Com_Printf ("   unchanged: %i\n", oldnum);
			CL_DeltaEntity (newframe, oldnum, oldstate, 0);
			
			oldindex++;

			if (oldindex >= oldframe->num_entities)
				oldnum = 99999;
			else
			{
				oldstate = &cl_parse_entities[(oldframe->parse_entities+oldindex) & (MAX_PARSE_ENTITIES-1)];
				oldnum = oldstate->number;
			}
		}

		if (bits & U_REMOVE)
		{	// the entity present in oldframe is not in the current frame
			if (cl_shownet->value == 3)
				Com_Printf ("   remove: %i\n", newnum);
			if (oldnum != newnum)
				Com_Printf ("U_REMOVE: oldnum != newnum\n");

			oldindex++;

			if (oldindex >= oldframe->num_entities)
				oldnum = 99999;
			else
			{
				oldstate = &cl_parse_entities[(oldframe->parse_entities+oldindex) & (MAX_PARSE_ENTITIES-1)];
				oldnum = oldstate->number;
			}
			continue;
		}

		if (oldnum == newnum)
		{	// delta from previous state
			if (cl_shownet->value == 3)
				Com_Printf ("   delta: %i\n", newnum);
			CL_DeltaEntity (newframe, newnum, oldstate, bits);

			oldindex++;

			if (oldindex >= oldframe->num_entities)
				oldnum = 99999;
			else
			{
				oldstate = &cl_parse_entities[(oldframe->parse_entities+oldindex) & (MAX_PARSE_ENTITIES-1)];
				oldnum = oldstate->number;
			}
			continue;
		}

		if (oldnum > newnum)
		{	// delta from baseline
			if (cl_shownet->value == 3)
				Com_Printf ("   baseline: %i\n", newnum);
			CL_DeltaEntity (newframe, newnum, &cl_entities[newnum].baseline, bits);
			continue;
		}

	}

	// any remaining entities in the old frame are copied over
	while (oldnum != 99999)
	{	// one or more entities from the old packet are unchanged
		if (cl_shownet->value == 3)
			Com_Printf ("   unchanged: %i\n", oldnum);
		CL_DeltaEntity (newframe, oldnum, oldstate, 0);
		
		oldindex++;

		if (oldindex >= oldframe->num_entities)
			oldnum = 99999;
		else
		{
			oldstate = &cl_parse_entities[(oldframe->parse_entities+oldindex) & (MAX_PARSE_ENTITIES-1)];
			oldnum = oldstate->number;
		}
	}

	if ( cls.state == ca_connected ) {
		CL_Spawn ();
	}
}



/*
===================
CL_ParsePlayerstate
===================
*/
void CL_ParsePlayerstate (frame_t *oldframe, frame_t *newframe)
{
	int			flags;
	player_state_t	*state;
	int			i;
	int			statbits;

	state = &newframe->playerstate;

	// clear to old value before delta parsing
	if (oldframe)
		*state = oldframe->playerstate;
	else
		memset (state, 0, sizeof(*state));

	flags = MSG_ReadShort (&net_message);

	//
	// parse the pmove_state_t
	//
	if (flags & PS_M_TYPE)
		state->pmove.pm_type = MSG_ReadByte (&net_message);

	if (flags & PS_M_ORIGIN)
	{
		state->pmove.origin[0] = MSG_ReadInt3 (&net_message);
		state->pmove.origin[1] = MSG_ReadInt3 (&net_message);
		state->pmove.origin[2] = MSG_ReadInt3 (&net_message);
	}

	if (flags & PS_M_VELOCITY0)
		state->pmove.velocity[0] = MSG_ReadInt3 (&net_message);
	if (flags & PS_M_VELOCITY1)
		state->pmove.velocity[1] = MSG_ReadInt3 (&net_message);
	if (flags & PS_M_VELOCITY2)
		state->pmove.velocity[2] = MSG_ReadInt3 (&net_message);

	if (flags & PS_M_TIME)
		state->pmove.pm_time = MSG_ReadByte (&net_message);

	if (flags & PS_M_FLAGS)
		state->pmove.pm_flags = MSG_ReadByte (&net_message);

	if (flags & PS_M_GRAVITY)
		state->pmove.gravity = MSG_ReadShort (&net_message);

	if (flags & PS_M_DELTA_ANGLES)
	{
		state->pmove.delta_angles[0] = MSG_ReadShort (&net_message);
		state->pmove.delta_angles[1] = MSG_ReadShort (&net_message);
		state->pmove.delta_angles[2] = MSG_ReadShort (&net_message);
	}

	if (cl.attractloop)
		state->pmove.pm_type = PM_FREEZE;		// demo playback

	//
	// parse the rest of the player_state_t
	//
	if (flags & PS_VIEWOFFSET)
	{
		state->viewoffset[0] = MSG_ReadChar (&net_message) * 0.25;
		state->viewoffset[1] = MSG_ReadChar (&net_message) * 0.25;
		state->viewoffset[2] = MSG_ReadChar (&net_message) * 0.25;
	}

	if (flags & PS_VIEWANGLES)
	{
		state->viewangles[0] = MSG_ReadAngle16 (&net_message);
		state->viewangles[1] = MSG_ReadAngle16 (&net_message);
		state->viewangles[2] = MSG_ReadAngle16 (&net_message);
	}

	if (flags & PS_KICKANGLES)
	{
		state->kick_angles[0] = MSG_ReadChar (&net_message) * 0.25;
		state->kick_angles[1] = MSG_ReadChar (&net_message) * 0.25;
		state->kick_angles[2] = MSG_ReadChar (&net_message) * 0.25;
	}

	if (flags & PS_WEAPONINDEX)
	{
		state->gunindex = MSG_ReadByte (&net_message);
	}

	if (flags & PS_WEAPONFRAME)
	{
		state->gunframe = MSG_ReadShort (&net_message);
		state->gunangles[0] = MSG_ReadChar (&net_message)*0.25;
		state->gunangles[1] = MSG_ReadChar (&net_message)*0.25;
		state->gunangles[2] = MSG_ReadChar (&net_message)*0.25;
	}

	if (flags & PS_BLEND)
	{
		state->blend[0] = MSG_ReadByte (&net_message)/255.0;
		state->blend[1] = MSG_ReadByte (&net_message)/255.0;
		state->blend[2] = MSG_ReadByte (&net_message)/255.0;
		state->blend[3] = MSG_ReadByte (&net_message)/255.0;
	}

	if (flags & PS_FOV)
		state->fov = MSG_ReadByte (&net_message);

	// parse stats
	statbits = MSG_ReadLong (&net_message);
	for (i=0 ; i<MAX_STATS ; i++)
		if (statbits & (1<<i) )
			state->stats[i] = MSG_ReadShort(&net_message);
}


/*
==================
CL_FireEntityEvents

==================
*/
void CL_FireEntityEvents (frame_t *frame)
{
	entity_state_t		*state;
	int					pnum, num;

	for (pnum = 0 ; pnum<frame->num_entities ; pnum++)
	{
		num = (frame->parse_entities + pnum)&(MAX_PARSE_ENTITIES-1);
		state = &cl_parse_entities[num];

		if (state->event)
			CL_EntityEvent (state);
	}
}


/*
================
CL_ParseFrame
================
*/
void CL_ParseFrame (void)
{
	int			cmd;
	int			len;
	frame_t		*old;

	memset (&cl.frame, 0, sizeof(cl.frame));

	cl.frame.serverframe = MSG_ReadLong (&net_message);
	cl.frame.deltaframe = MSG_ReadLong (&net_message);
	cl.frame.servertime = cl.frame.serverframe*100;
	cl.suppressCount = MSG_ReadByte (&net_message);

	if (cl_shownet->value == 3)
		Com_Printf ("   frame:%i  delta:%i\n", cl.frame.serverframe,
		cl.frame.deltaframe);

	// If the frame is delta compressed from data that we
	// no longer have available, we must suck up the rest of
	// the frame, but not use it, then ask for a non-compressed
	// message 
	if (cl.frame.deltaframe <= 0)
	{
		cl.frame.valid = qtrue;		// uncompressed frame
		old = NULL;
		cls.demowaiting = qfalse;	// we can start recording now
	}
	else
	{
		old = &cl.frames[cl.frame.deltaframe & UPDATE_MASK];
		if (!old->valid)
		{	// should never happen
			Com_Printf ("Delta from invalid frame (not supposed to happen!).\n");
		}
		if (old->serverframe != cl.frame.deltaframe)
		{	// The frame that the server did the delta from
			// is too old, so we can't reconstruct it properly.
			Com_Printf ("Delta frame too old.\n");
		}
		else if (cl.parse_entities - old->parse_entities > MAX_PARSE_ENTITIES-128)
		{
			Com_Printf ("Delta parse_entities too old.\n");
		}
		else
			cl.frame.valid = qtrue;	// valid delta parse
	}

	// clamp time 
	if (cl.time > cl.frame.servertime)
		cl.time = cl.frame.servertime;
	else if (cl.time < cl.frame.servertime - 100)
		cl.time = cl.frame.servertime - 100;

	// read areabits
	len = MSG_ReadByte (&net_message);
	MSG_ReadData (&net_message, &cl.frame.areabits, len);

	// read playerinfo
	cmd = MSG_ReadByte (&net_message);
	SHOWNET(svc_strings[cmd]);
	if (cmd != svc_playerinfo)
		Com_Error (ERR_DROP, "CL_ParseFrame: not playerinfo");
	CL_ParsePlayerstate (old, &cl.frame);

	// read packet entities
	cmd = MSG_ReadByte (&net_message);
	SHOWNET(svc_strings[cmd]);
	if (cmd != svc_packetentities)
		Com_Error (ERR_DROP, "CL_ParseFrame: not packetentities");
	CL_ParsePacketEntities (old, &cl.frame);

	// save the frame off in the backup array for later delta comparisons
	cl.frames[cl.frame.serverframe & UPDATE_MASK] = cl.frame;

	if (cl.frame.valid)
	{
		// getting a valid frame message ends the connection process
		if (cls.state != ca_active)
		{
			CL_SetClientState (ca_active);
			cl.force_refdef = qtrue;
			cl.predicted_origin[0] = cl.frame.playerstate.pmove.origin[0]*0.125;
			cl.predicted_origin[1] = cl.frame.playerstate.pmove.origin[1]*0.125;
			cl.predicted_origin[2] = cl.frame.playerstate.pmove.origin[2]*0.125;
			VectorCopy (cl.frame.playerstate.viewangles, cl.predicted_angles);
			if (cls.disable_servercount != cl.servercount
				&& cl.refresh_prepped)
				SCR_EndLoadingPlaque ();	// get rid of loading plaque
		}
		cl.sound_prepped = qtrue;	// can start mixing ambient sounds
	
		// fire entity events
		CL_FireEntityEvents (&cl.frame);
		CL_CheckPredictionError ();

		//splitmodels
		CL_PlayerModels_UpdateAnimationBuffer (&cl.frame);
	}
}

/*
==========================================================================

INTERPOLATE BETWEEN FRAMES TO GET RENDERING PARMS

==========================================================================
*/

/*
===============
CL_AddPacketEntities

===============
*/
void CL_AddPacketEntities (void)
{
	entity_t			ent;
	entity_state_t		*state;
	vec3_t				ent_angles, autorotate;
	mat3_t				autorotate_axis;
	int					i;
	int					pnum;
	centity_t			*cent;
	int					autoanim;
//	clientinfo_t		*ci;	//splitmodels
	unsigned int		effects, renderfx;
	int					msec;

	// bonus items rotate at a fixed rate
	VectorSet ( autorotate, 0, anglemod(cl.time*0.1), 0 );
	AnglesToAxis ( autorotate, autorotate_axis );

	// brush models can auto animate their frames
	autoanim = 2*cl.time/1000;

	memset (&ent, 0, sizeof(ent));

	for (pnum = 0 ; pnum<cl.frame.num_entities ; pnum++)
	{
		state = &cl_parse_entities[(cl.frame.parse_entities+pnum)&(MAX_PARSE_ENTITIES-1)];
		cent = &cl_entities[state->number];

		effects = state->effects;
		renderfx = state->renderfx;

		// set frame
		if (effects & EF_ANIM01)
			ent.frame = autoanim & 1;
		else if (effects & EF_ANIM23)
			ent.frame = 2 + (autoanim & 1);
		else if (effects & EF_ANIM_ALL)
			ent.frame = autoanim;
		else if (effects & EF_ANIM_ALLFAST)
			ent.frame = cl.time / 100;
		else
			ent.frame = state->frame;

		if ( state->number < MAX_CLIENTS+1 ) {
			renderfx |= RF_MINLIGHT;
		}

		//CHASEHACKv2 & splitmodels //moved below 

		ent.shaderTime = 0;
		ent.oldframe = cent->prev.frame;
		ent.backlerp = 1.0 - cl.lerpfrac;

		if (renderfx & (RF_FRAMELERP|RF_BEAM|RF_PORTALSURFACE))
		{	// step origin discretely, because the frames
			// do the animation properly
			VectorCopy (cent->current.origin, ent.origin);
			VectorCopy (cent->current.old_origin, ent.oldorigin);
		}
		else
		{	// interpolate origin
			for (i=0 ; i<3 ; i++)
			{
				ent.origin[i] = ent.oldorigin[i] = cent->prev.origin[i] + cl.lerpfrac * 
					(cent->current.origin[i] - cent->prev.origin[i]);
			}
		}

		// create a new entity
	
		// tweak the color of beams
		if ( renderfx & RF_BEAM )
		{	// the four beam colors are encoded in 32 bits of skinnum (hack)
			CL_AddLaser ( state->origin, state->old_origin, state->frame * 0.5f, state->skinnum, cl.media.shaderLaser );
			continue;
		}
		else if ( renderfx & RF_PORTALSURFACE ) 
		{
			ent.type = ET_PORTALSURFACE;
			ent.scale = state->frame / 256.0f;

			if ( state->modelindex3 )
				ent.frame = state->modelindex2 ? state->modelindex2 : 50;
			else
				ent.frame = 0;

			ent.skinnum = state->skinnum;
			ent.model = NULL;
			V_AddEntity ( &ent );
			continue;
		}	
		else
		{
			ent.type = ET_MODEL;

			//-------------------
			//splitmodels [start]

			// set skin
			if (state->modelindex != 255)
			{
				//not a player model
				ent.skinnum = state->skinnum;
				ent.customShader = NULL;
				ent.customSkin = NULL;
				ent.model = cl.model_draw[state->modelindex];
			}
			//splitmodels [end]
			//--------------------
		}

		// render effects
		if ( renderfx & (RF_SHELL_GREEN | RF_SHELL_RED | RF_SHELL_BLUE) )
		{
			if (renderfx & RF_MINLIGHT)
				ent.flags = RF_MINLIGHT;	// renderfx go on color shell entity
			else
				ent.flags = 0;
		}
		else
		{
			ent.flags = renderfx;
		}

		// glow if light is set
		if ( state->light ) {
			V_AddLight ( ent.origin, 
				COLOR_A ( state->light ) * 4.0, 
				COLOR_R ( state->light ) * (1.0/255.0), 
				COLOR_G ( state->light ) * (1.0/255.0),	
				COLOR_B ( state->light ) * (1.0/255.0) );
		}

		// respawning items
		if ( cent->respawnTime ) {
			msec = cl.time - cent->respawnTime;
		} else {
			msec = ITEM_RESPAWN_TIME;
		}
		
		if ( msec >= 0 && msec < ITEM_RESPAWN_TIME  ) {
			ent.scale = (float)msec / ITEM_RESPAWN_TIME;
		} else {
			ent.scale = 1.0f;
		}

		if ( renderfx & RF_SCALEHACK ) {
			ent.scale *= 1.5;
		}

		if ( !ent.scale ) {
			continue;
		}

		// bobbing items
		if (effects & EF_BOB) 
		{
			float scale = 0.005f + state->number * 0.00001f;
			float bob = 4 + cos( (cl.time + 1000) * scale ) * 4;
			ent.oldorigin[2] += bob;
			ent.origin[2] += bob;
		}

		VectorCopy (ent.origin, ent.lightingOrigin);

		// calculate angles
		if (effects & EF_ROTATE)
		{	// some bonus items auto-rotate
			Matrix3_Copy ( autorotate_axis, ent.axis );
		}
		else
		{	// interpolate angles
			float	a1, a2;

			for (i=0 ; i<3 ; i++)
			{
				a1 = cent->current.angles[i];
				a2 = cent->prev.angles[i];
				ent_angles[i] = LerpAngle (a2, a1, cl.lerpfrac);
			}

			if ( ent_angles[0] || ent_angles[1] || ent_angles[2] ) {
				AnglesToAxis ( ent_angles, ent.axis );
			} else {
				Matrix3_Copy ( axis_identity, ent.axis );
			}
		}

		//JAL CHASEHACKv2 & splitmodels [begin]
		if ( (state->number == cl.playernum+1) || ((cl.chasednumber) && (cl.chasednumber == state->number)) )
		{
			//update view weapon buffer & more
			if (!cl.chasednumber){
				cl.vweap.latchedweapon = ( state->skinnum >> 8 );
				cl.effects = effects;
				if ( !cl_thirdPerson->value ) {
					ent.flags |= RF_VIEWERMODEL;	// only draw from mirrors
					cl.thirdperson = qfalse;
				} else { 
					cl.thirdperson = ( state->modelindex != 0 );
				}
			}
			else if (cl.chasednumber == state->number){
				cl.vweap.latchedweapon = ( state->skinnum >> 8 );
				cl.effects = effects;
				if ( !cl_thirdPerson->value ) {
					ent.flags |= RF_VIEWERMODEL;	// only draw from mirrors
					cl.thirdperson = qfalse;
				} else { 
					cl.thirdperson = ( state->modelindex != 0 );
				}
			}
		}
		//splitmodels [end]

		// if set to invisible, skip
		if (!state->modelindex)
			continue;

		//----------------------------
		//splitmodels [begin]
		if (state->modelindex == 255)
		{	
			CL_AddPlayerEntities (&ent, state, effects, renderfx);

		} else {
			
			// add to refresh list
			V_AddEntity (&ent);
			
			CL_AddShellEffects (&ent, effects);	// Shell Effects
			CL_AddColorShell (&ent, renderfx);	// color shells
		}

		ent.customSkin = NULL;	////restore color and skins

		//splitmodels [end]
		//----------------------------

		//restore color and skins
		Vector4Set ( ent.color, 255, 255, 255, 255 );

		ent.customShader = NULL;		// never use a custom skin on others
		ent.skinnum = 0;
		

		if ( ent.flags & RF_VIEWERMODEL )	// only draw from mirrors
			ent.flags = RF_VIEWERMODEL;
		else
			ent.flags = 0;

		// duplicate for linked models
		//----------------------------
		//splitmodels [begin] weapon model + barrel

		if (state->modelindex2 && state->modelindex2 != 255){

			//jal: at this point ent still has the modelindex model
			orientation_t	tag;
			if (R_LerpAttachment( &tag, ent.model, ent.frame, ent.oldframe, ent.backlerp, "tag_barrel" ))
			{
				CL_PlaceModelOnTag(&ent, &ent, &tag);
				
			}//jal[end]

			ent.model = cl.model_draw[state->modelindex2];
			V_AddEntity (&ent);
			
			CL_AddShellEffects (&ent, effects);	// Shell Effects
			CL_AddColorShell (&ent, renderfx);	// color shells
		} 
		//splitmodels [end]
		//-----------------

		if (state->modelindex3)
		{
			int frame, oldframe;

			frame = ent.frame;
			oldframe = ent.oldframe;

			if ( effects & (EF_FLAG1|EF_FLAG2) ) {
				ent.frame = 0;
				ent.oldframe = 0;
			}

			ent.model = cl.model_draw[state->modelindex3];
			V_AddEntity (&ent);

			ent.frame = frame;
			ent.oldframe = oldframe;
		}

		if ( effects & EF_POWERSCREEN )
		{
			ent.model = cl.media.modPowerScreen;
			ent.oldframe = 0;
			ent.frame = 0;
			V_AddEntity (&ent);
		}

		// add automatic particle trails
		if ( (effects&~EF_ROTATE) )
		{
			if (effects & EF_BLASTER)
			{
				CL_BlasterTrail (cent->lerp_origin, ent.origin);
				V_AddLight (ent.origin, 300, 1, 1, 0);
			}
			else if (effects & EF_HYPERBLASTER)
			{
				V_AddLight (ent.origin, 300, 1, 1, 0);
			}
			else if (effects & EF_GIB)
			{
				CL_BloodTrail (cent->lerp_origin, ent.origin, &cent->trailcount);
			}
			if (effects & EF_ROCKET)
			{
				CL_RocketTrail (cent->lerp_origin, ent.origin);
				V_AddLight (ent.origin, 300, 1, 1, 0);
			} 
			else if (effects & EF_GRENADE)
			{
				CL_GrenadeTrail (cent->lerp_origin, ent.origin);
			}
			else if (effects & EF_FLIES)
			{
				CL_FlyEffect (cent, ent.origin);
			}
			else if (effects & EF_BFG)
			{
				static int bfg_lightramp[6] = {400, 500, 700, 400, 250, 175};

				if (effects & EF_ANIM_ALLFAST)
				{
					CL_BfgParticles (&ent);
					i = 300;
				}
				else
				{
					i = bfg_lightramp[state->frame];
				}
				V_AddLight (ent.origin, i, 0, 1, 0);
			}
			else if ((effects & (EF_FLAG1|EF_FLAG2)) == EF_FLAG1)
			{
				CL_FlagTrail (cent->lerp_origin, ent.origin, EF_FLAG1);
				V_AddLight (ent.origin, 225, 1, 0.1, 0.1);
			}
			else if ((effects & (EF_FLAG1|EF_FLAG2)) == EF_FLAG2)
			{
				CL_FlagTrail (cent->lerp_origin, ent.origin, EF_FLAG2);
				V_AddLight (ent.origin, 225, 0.1, 0.1, 1);
			}
			else if ((effects & (EF_FLAG1|EF_FLAG2)) == (EF_FLAG1|EF_FLAG2))
			{
				CL_FlagTrail (cent->lerp_origin, ent.origin, EF_FLAG1);
				CL_FlagTrail (cent->lerp_origin, ent.origin, EF_FLAG2);
				V_AddLight (ent.origin, 225, 1, 0.2, 1);
			}
		}

		VectorCopy (ent.origin, cent->lerp_origin);
	}
}


/*
==============
CL_AddViewWeapon - SPLITMODELS
==============
*/
void CL_AddViewWeapon (player_state_t *ps, player_state_t *ops)
{
	//buffer gunindex
	cl.vweap.latchedgunindex = ps->gunindex;
	if (!cl.vweap.currentgunindex)
		cl.vweap.currentgunindex = ps->gunindex;

	//do not draw
	if (ps->fov > 120 || cl.thirdperson || !cl_gun->value)
		return;

	CL_AddFirstViewWeapon (ps, ops);
}

/*
===============
CL_CalcViewValues

Sets cl.refdef view values
===============
*/
void CL_CalcViewValues (void)
{
	int			i;
	float		lerp, backlerp;
	centity_t	*ent;
	frame_t		*oldframe;
	player_state_t	*ps, *ops;

	// find the previous frame to interpolate from
	ps = &cl.frame.playerstate;
	i = (cl.frame.serverframe - 1) & UPDATE_MASK;
	oldframe = &cl.frames[i];
	if (oldframe->serverframe != cl.frame.serverframe-1 || !oldframe->valid)
		oldframe = &cl.frame;		// previous frame was dropped or invalid
	ops = &oldframe->playerstate;

	//JAL CHASEHACKv2 & splitmodels
	CL_ChaseHack (ps);

	// see if the player entity was teleported this frame
	if ( abs(ops->pmove.origin[0] - ps->pmove.origin[0]) > 256*8
		|| abs(ops->pmove.origin[1] - ps->pmove.origin[1]) > 256*8
		|| abs(ops->pmove.origin[2] - ps->pmove.origin[2]) > 256*8)
		ops = ps;		// don't interpolate

	ent = &cl_entities[cl.playernum+1];
	lerp = cl.lerpfrac;

	// calculate the origin
	if (cl_predict->value && !(cl.frame.playerstate.pmove.pm_flags & PMF_NO_PREDICTION)
		&& !cl.thirdperson)
	{	// use predicted values
		unsigned	delta;

		backlerp = 1.0 - lerp;
		for (i=0 ; i<3 ; i++)
		{
			cl.refdef.vieworg[i] = cl.predicted_origin[i] + ops->viewoffset[i] 
				+ lerp * (ps->viewoffset[i] - ops->viewoffset[i])
				- backlerp * cl.prediction_error[i];
		}

		// smooth out stair climbing
		delta = cls.realtime - cl.predicted_step_time;
		if (delta < 150)
			cl.refdef.vieworg[2] -= cl.predicted_step * (150 - delta) / 150;
	}
	else
	{	// just use interpolated values
		for (i=0 ; i<3 ; i++)
			cl.refdef.vieworg[i] = ops->pmove.origin[i]*0.125 + ops->viewoffset[i] 
				+ lerp * (ps->pmove.origin[i]*0.125 + ps->viewoffset[i] 
				- (ops->pmove.origin[i]*0.125 + ops->viewoffset[i]) );
	}

	// if not running a demo or on a locked frame, add the local angle movement
	if ( cl.frame.playerstate.pmove.pm_type < PM_DEAD )
	{	// use predicted values
		for (i=0 ; i<3 ; i++)
			cl.refdef.viewangles[i] = cl.predicted_angles[i];
	}
	else
	{	// just use interpolated values
		for (i=0 ; i<3 ; i++)
			cl.refdef.viewangles[i] = LerpAngle (ops->viewangles[i], ps->viewangles[i], lerp);
	}

	for (i=0 ; i<3 ; i++)
		cl.refdef.viewangles[i] += LerpAngle (ops->kick_angles[i], ps->kick_angles[i], lerp);

	AngleVectors (cl.refdef.viewangles, cl.v_forward, cl.v_right, cl.v_up);

	// interpolate field of view
	cl.refdef.fov_x = ops->fov + lerp * (ps->fov - ops->fov);

	// don't interpolate blend color
	for (i=0 ; i<4 ; i++)
		cl.refdef.blend[i] = ps->blend[i];

	// add the weapon
	CL_AddViewWeapon (ps, ops);
}

/*
===============
CL_AddEntities

Emits all entities, particles, and lights to the refresh
===============
*/
void CL_AddEntities (void)
{
	if (cls.state != ca_active)
		return;

	if (cl.time > cl.frame.servertime)
	{
		if (cl_showclamp->value)
			Com_Printf ("high clamp %i\n", cl.time - cl.frame.servertime);
		cl.time = cl.frame.servertime;
		cl.lerpfrac = 1.0;
	}
	else if (cl.time < cl.frame.servertime - 100)
	{
		if (cl_showclamp->value)
			Com_Printf ("low clamp %i\n", cl.frame.servertime-100 - cl.time);
		cl.time = cl.frame.servertime - 100;
		cl.lerpfrac = 0;
	}
	else
		cl.lerpfrac = 1.0 - (cl.frame.servertime - cl.time) * 0.01;

	if (cl_timedemo->value)
		cl.lerpfrac = 1.0;

	//CL_CalcViewValues ();	//jal: backup of original

	cl_numBeamEnts = 0;

	CL_AddPacketEntities ();	

	CL_CalcViewValues ();	//splitmodels: I reordered this so I have the clientinfo
							//values updated before adding the viewweapon (jal).
	CL_AddBeams ();
	CL_AddLocalEntities ();
	CL_AddDecals ();
	CL_AddParticles ();
	CL_AddDLights ();
}



/*
===============
CL_GetEntitySoundOrigin

Called to get the sound spatialization origin
===============
*/
void CL_GetEntitySoundOrigin (int entnum, vec3_t org)
{
	centity_t	*ent;

	if (entnum < 0 || entnum >= MAX_EDICTS)
		Com_Error (ERR_DROP, "CL_GetEntitySoundOrigin: bad entnum");
	ent = &cl_entities[entnum];

	if ( ent->current.solid == 31 ) {	// bmodel
		VectorAdd ( cl.model_clip[ent->current.modelindex]->maxs,
			cl.model_clip[ent->current.modelindex]->mins, org );
		VectorMA ( ent->lerp_origin, 0.5f, org, org );
	} else {
		VectorCopy ( ent->lerp_origin, org );
	}
}
