/*
Copyright (C) 1997-2001 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// cl_parse.c  -- parse a message received from the server

#include "client.h"
#include "cl_players.h"	//splitmodels

char *svc_strings[256] =
{
	"svc_bad",

	"svc_muzzleflash",
	"svc_muzzleflash2",
	"svc_temp_entity",
	"svc_layout",
	"svc_inventory",

	"svc_nop",
	"svc_disconnect",
	"svc_reconnect",
	"svc_sound",
	"svc_print",
	"svc_stufftext",
	"svc_serverdata",
	"svc_configstring",
	"svc_spawnbaseline",	
	"svc_centerprint",
	"svc_download",
	"svc_playerinfo",
	"svc_packetentities",
	"svc_deltapacketentities",
	"svc_frame",
	"svc_obituary",
	"svc_stringcmd"
};

//=============================================================================

void CL_DownloadFileName(char *dest, int destlen, char *fn)
{
	//splitmodels (models/players)
	//jal testme: I'm not sure of this being 100% working
	if (strncmp(fn, "models/players", 7) == 0)
		Com_sprintf (dest, destlen, "%s/%s", BASEDIRNAME, fn);
	else
		Com_sprintf (dest, destlen, "%s/%s", FS_Gamedir(), fn);
}

/*
===============
CL_CheckOrDownloadFile

Returns true if the file exists, otherwise it attempts
to start a download from the server.
===============
*/
qboolean	CL_CheckOrDownloadFile (char *filename)
{
	FILE *fp;
	char	name[MAX_OSPATH];

	if (strstr (filename, ".."))
	{
//		Com_Printf ("Refusing to download a path with ..\n");
		Com_Printf ("Refusing to download a path with ..: %s\n", filename);
		return qtrue;
	}

	if ( FS_FileExists ( filename ) )
	{	// it exists, no need to download
		return qtrue;
	}

	strcpy (cls.downloadname, filename);

	// download to a temp name, and only rename
	// to the real name when done, so if interrupted
	// a runt file wont be left
	COM_StripExtension (cls.downloadname, cls.downloadtempname);
	strcat (cls.downloadtempname, ".tmp");

//ZOID
	// check to see if we already have a tmp for this file, if so, try to resume
	// open the file if not opened yet
	CL_DownloadFileName(name, sizeof(name), cls.downloadtempname);

//	FS_CreatePath (name);

	fp = fopen (name, "r+b");
	if (fp) { // it exists
		int len;
		fseek(fp, 0, SEEK_END);
		len = ftell(fp);

		cls.download = fp;

		// give the server an offset to start the download
		Com_Printf ("Resuming %s\n", cls.downloadname);
		MSG_WriteByte (&cls.netchan.message, clc_stringcmd);
		MSG_WriteString (&cls.netchan.message,
			va("download %s %i", cls.downloadname, len));
	} else {
		Com_Printf ("Downloading %s\n", cls.downloadname);
		MSG_WriteByte (&cls.netchan.message, clc_stringcmd);
		MSG_WriteString (&cls.netchan.message,
			va("download %s", cls.downloadname));
	}

	cls.downloadnumber++;

	return qfalse;
}

/*
===============
CL_Download_f

Request a download from the server
===============
*/
void	CL_Download_f (void)
{
	char filename[MAX_OSPATH];

	if (Cmd_Argc() != 2) {
		Com_Printf("Usage: download <filename>\n");
		return;
	}

	Com_sprintf(filename, sizeof(filename), "%s", Cmd_Argv(1));

	if (strstr (filename, ".."))
	{
		Com_Printf ("Refusing to download a path with ..: %s\n", filename);
		return;
	}

	if ( FS_FileExists (filename) )
	{	// it exists, no need to download
		Com_Printf("File already exists.\n");
		return;
	}

	strcpy (cls.downloadname, filename);
	Com_Printf ("Downloading %s\n", cls.downloadname);

	// download to a temp name, and only rename
	// to the real name when done, so if interrupted
	// a runt file wont be left
	COM_StripExtension (cls.downloadname, cls.downloadtempname);
	strcat (cls.downloadtempname, ".tmp");

	MSG_WriteByte (&cls.netchan.message, clc_stringcmd);
	MSG_WriteString (&cls.netchan.message,
		va("download %s", cls.downloadname));

	cls.downloadnumber++;
}

/*
======================
CL_RegisterSounds
======================
*/
void CL_RegisterSounds (void)
{
	int		i;

	CL_LoadingString ("sounds");

	S_BeginRegistration ();

	CL_RegisterMediaSounds ();

	if ( cl.configstrings[CS_AUDIOTRACK][0] )
		cl.music_precache = S_RegisterSound (cl.configstrings[CS_AUDIOTRACK]);
	else
		cl.music_precache = NULL;

	for (i=1 ; i<MAX_SOUNDS ; i++)
	{
		if (!cl.configstrings[CS_SOUNDS+i][0])
			break;
		cl.sound_precache[i] = S_RegisterSound (cl.configstrings[CS_SOUNDS+i]);
		Sys_SendKeyEvents ();	// pump message loop
	}

	S_EndRegistration ();
}


/*
=====================
CL_ParseDownload

A download message has been received from the server
=====================
*/
void CL_ParseDownload (void)
{
	int		size, percent;
	char	name[MAX_OSPATH];
	int		r;

	// read the data
	size = MSG_ReadShort (&net_message);
	percent = MSG_ReadByte (&net_message);
	if (size == -1)
	{
		Com_Printf ("Server does not have this file.\n");
		if (cls.download)
		{
			// if here, we tried to resume a file but the server said no
			fclose (cls.download);
			cls.download = NULL;
		}
		CL_RequestNextDownload ();
		return;
	}

	// open the file if not opened yet
	if (!cls.download)
	{
		CL_DownloadFileName(name, sizeof(name), cls.downloadtempname);

		FS_CreatePath (name);

		cls.download = fopen (name, "wb");
		if (!cls.download)
		{
			net_message.readcount += size;
			Com_Printf ("Failed to open %s\n", cls.downloadtempname);
			CL_RequestNextDownload ();
			return;
		}
	}

	fwrite (net_message.data + net_message.readcount, 1, size, cls.download);
	net_message.readcount += size;

	if (percent != 100)
	{
		// request next block
// change display routines by zoid
		cls.downloadpercent = percent;

		MSG_WriteByte (&cls.netchan.message, clc_stringcmd);
		SZ_Print (&cls.netchan.message, "nextdl");
	}
	else
	{
		char	oldn[MAX_OSPATH];
		char	newn[MAX_OSPATH];

		fclose (cls.download);

		// rename the temp file to its final name
		CL_DownloadFileName(oldn, sizeof(oldn), cls.downloadtempname);
		CL_DownloadFileName(newn, sizeof(newn), cls.downloadname);
		r = rename (oldn, newn);
		if (r)
			Com_Printf ("failed to rename.\n");

		cls.download = NULL;
		cls.downloadpercent = 0;

		// get another file if needed

		CL_RequestNextDownload ();
	}
}


/*
=====================================================================

  SERVER CONNECTING MESSAGES

=====================================================================
*/

/*
==================
CL_ParseServerData
==================
*/
void CL_ParseServerData (void)
{
	extern cvar_t	*fs_gamedirvar;
	char	*str, *message;
	int		i;
	
	Com_DPrintf ("Serverdata packet received.\n");
//
// wipe the client_state_t struct
//
	CL_ClearState ();
	CL_SetClientState (ca_connected);

// parse protocol version number
	i = MSG_ReadLong (&net_message);

	if (i != PROTOCOL_VERSION)
		Com_Error (ERR_DROP, "Server returned version %i, not %i", i, PROTOCOL_VERSION);

	cl.servercount = MSG_ReadLong (&net_message);
	cl.attractloop = MSG_ReadByte (&net_message);

	// game directory
	str = MSG_ReadString (&net_message);
	Q_strncpyz (cl.gamedir, str, sizeof(cl.gamedir));

	// set gamedir
	if ((*str && (!fs_gamedirvar->string || !*fs_gamedirvar->string || strcmp(fs_gamedirvar->string, str))) || (!*str && (fs_gamedirvar->string || *fs_gamedirvar->string))) {
		Cvar_Set("fs_game", str);
		UI_Init ();	
	}

	// parse player entity number
	cl.playernum = MSG_ReadShort (&net_message);

	// get the full level name
	message = MSG_ReadString (&net_message);
	
	if (cl.playernum == -1)
	{	// playing a cinematic or showing a pic, not a level
		SCR_PlayCinematic (message);
	}
	else
	{
		// separate the printfs so the server message can have a color
		Com_Printf("\n\n\35\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\37\n\n");
		Com_Printf ("%s%s\n", S_COLOR_RED, message);

		// need to prep refresh at next oportunity
		cl.refresh_prepped = qfalse;
	}
}

/*
==================
CL_ParseBaseline
==================
*/
void CL_ParseBaseline (void)
{
	entity_state_t	*es;
	unsigned		bits;
	int				newnum;
	entity_state_t	nullstate;

	memset (&nullstate, 0, sizeof(nullstate));

	newnum = CL_ParseEntityBits (&bits);
	es = &cl_entities[newnum].baseline;
	CL_ParseDelta (&nullstate, es, newnum, bits);
}


/*
================
CL_CopyAnimation - SPLITMODELS
================
*/
void CL_CopyAnimation (clientinfo_t *ci, int put, int pick)
{
	ci->anim.firstframe[put] = ci->anim.firstframe[pick];
	ci->anim.lastframe[put] = ci->anim.lastframe[pick];
	ci->anim.loopingframes[put] = ci->anim.loopingframes[pick];
	ci->anim.frametime[put] = ci->anim.frametime[pick];
}

/*
================
CL_LoadPlayerAnims - SPLITMODELS (store anims in player)
 Reads the animation config file.

 Note: The animation count begins at 1, not 0. I preserve zero for "no animation change"
 ---------------
 New keywords:
 nooffset: Uses the first frame value at the script, and no offsets, for the animation.
 alljumps: Uses 3 different animations (bunnyhoping) instead of copying jumpback inside jump1 & jump2.

 Q3 keywords:
 sex m/f : sets gender

 Q3 Unsupported:
 headoffset <value> <value> <value>
 foosteps
================
*/
qboolean CL_LoadPlayerAnims (clientinfo_t *ci)
{
	qbyte	*buf;
	char	*ptr, *token;
	int		rounder, counter, i, offset, nooffset, alljumps;
	int		anim_data[4][PM_MAX_ANIMS];
	//0 = first frame
	//1 = number of frames
	//2 = looping frames
	//3 = frame time
	

	rounder = nooffset = alljumps = 0;
	counter = 1; //reseve 0 for 'no animation change'
	

	if (!FS_FileExists(ci->anim.filename)){
		Com_Printf ("ERROR: Couldn't load animation: %s\n", ci->anim.filename);
		return qfalse;
	}
	
	FS_LoadFile ( ci->anim.filename, (void **)&buf );

	if ( !buf ) {
		FS_FreeFile ( buf );
		Com_Printf ("ERROR: Couldn't load animation: %s\n", ci->anim.filename);
		return qfalse;
	}
	
	//proceed
	ptr = ( char * )buf;
	while ( ptr )
	{
		token = COM_ParseExt ( &ptr, qtrue );
		if ( !token )
			break;
		
		//see if it is keyword or number
		if (*token < '0' || *token > '9') {
			
			//headoffset (ignored)
			if ( !Q_stricmp (token, "headoffset") ) 
			{
				if (developer->value)
					Com_DPrintf ("%s:", token);
				
				token = COM_ParseExt ( &ptr, qfalse );
				if ( !token ){
					break;
				}
				if (developer->value)
					Com_DPrintf ("%s - ", token);
				
				token = COM_ParseExt ( &ptr, qfalse );
				if ( !token ){
					break;
				}
				if (developer->value)
					Com_DPrintf ("%s - ", token);
				
				token = COM_ParseExt ( &ptr, qfalse );
				if ( !token ){
					break;
				}
				if (developer->value)
					Com_DPrintf ("%s\n", token);
				
				//gender
			} else if ( !Q_stricmp (token, "sex") ) {
				
				if (developer->value)
					Com_DPrintf ("%s:", token);
				
				token = COM_ParseExt ( &ptr, qfalse );
				if ( !token ){
					break;
				}
				
				if ( token[0] == 'm' || token[0] == 'M' ) {
					ci->gender = GENDER_MALE;
					if (developer->value)
						Com_DPrintf ("%s -Gender set to MALE\n", token);
					
				} else if ( token[0] == 'f' || token[0] == 'F' ) {
					
					ci->gender = GENDER_FEMALE;
					if (developer->value)
						Com_DPrintf ("%s -Gender set to FEMALE\n", token);
				} else {
					ci->gender = GENDER_NEUTRAL;
					if (developer->value)
						Com_DPrintf ("%s -Gender set to NEUTRAL\n", token);
				}
				
				//nooffset
			} else if ( !Q_stricmp (token, "nooffset") ) {
				nooffset = 1;
				if (developer->value)
					Com_DPrintf ("No offset on lower frames for this script\n");
				//alljumps
			} else if ( !Q_stricmp (token, "alljumps") ) {
				alljumps = 1;
				if (developer->value)
					Com_DPrintf ("Using all jump animations\n");
			} else
				if (developer->value)
					Com_DPrintf ("ignored: %s\n", token);
				
		} else {
			
			//frame & animation values
			i = (int)atoi(token);
			if (developer->value)
				Com_Printf ("%i - ", i);
			anim_data[rounder][counter] = i;
			rounder++;
			if (rounder > 3){
				rounder = 0;
				if (developer->value)
					Com_DPrintf (" anim: %i\n", counter);
				counter++;
			}
		}
	}
	
	FS_FreeFile ( buf );
	
	//it must contain at least as many animations as a Q3 script to be valid
	if (counter-1 < LEGS_TURN)
	{
		Com_Printf ("ERROR: Not enough animations(%i) at script: %s\n", counter, ci->anim.filename);
		return qfalse;
	}
	else if (developer->value)
		Com_DPrintf ("LOADING: %s - %s\n", ci->name, ci->anim.filename);
	
	//animation HEAD_NONE (0) is always at frame 0, and it's never 
	//received from the game, but just used on the client when none
	//animation was ever set for a model (head).
	
	anim_data[0][HEAD_NONE] = 0;
	anim_data[1][HEAD_NONE]	= 0;
	anim_data[2][HEAD_NONE]	= 1;
	anim_data[3][HEAD_NONE] = 10;
	
	//reorganize to make my life easier
	for (i=0 ; i<counter ; i++)
	{
		ci->anim.firstframe[i] = anim_data[0][i];
		ci->anim.lastframe[i] = ((anim_data[0][i]) + (anim_data[1][i]));
		ci->anim.loopingframes[i] = anim_data[2][i];
		
		//HACK: make weapon switching animations stuck at 10 fps
		if (i == TORSO_FLIPOUT || i == TORSO_FLIPIN || anim_data[3][i] < 10)//never allow less than 10 fps
			anim_data[3][i] = 10;
		
		ci->anim.frametime[i] = 1000/anim_data[3][i];
	}
	
	if (!nooffset)
	{
		//find offset (Why did they do this???)  
		offset = ci->anim.firstframe[LEGS_CRWALK] - ci->anim.firstframe[TORSO_GESTURE];
		
		//Remove offset from the lower & extra animations
		for (i = LEGS_CRWALK; i < counter; ++i){
			ci->anim.firstframe[i] -= offset;
			ci->anim.lastframe[i] -= offset;
		}

		if (developer->value)
			Com_DPrintf ("Fixing offset on lower frames (Q3 format)\n");
	}
	
	//Fix the rates to fit my looping calculations
	for (i = 0; i < counter; ++i)
	{
		if (ci->anim.loopingframes[i])
			ci->anim.loopingframes[i] -= 1;
		if (ci->anim.lastframe[i])
			ci->anim.lastframe[i] -= 1;
	}
	
	//Alljumps:	 I use 3 jump animations for bunnyhoping
	//animation support. But they will only be loaded as
	//bunnyhoping animations if the keyword "alljumps" is
	//present at the animation.cfg. Otherwise, LEGS_JUMP1
	//will be used for all the jump styles.
	
	if (!alljumps)
	{
		CL_CopyAnimation (ci, LEGS_JUMP3, LEGS_JUMP1);
		CL_CopyAnimation (ci, LEGS_JUMP3ST, LEGS_JUMP1ST);
		CL_CopyAnimation (ci, LEGS_JUMP2, LEGS_JUMP1);
		CL_CopyAnimation (ci, LEGS_JUMP2ST, LEGS_JUMP1ST);
	}
	
	//HACK Uncomplete scripts (for Q3 player models)
	//----------------------------------------------------
	counter--;
	
	if (counter < TORSO_RUN)
		CL_CopyAnimation (ci, TORSO_RUN, TORSO_STAND);
	if (counter < TORSO_DROPHOLD)
		CL_CopyAnimation (ci, TORSO_DROPHOLD, TORSO_STAND);
	if (counter < TORSO_DROP)
		CL_CopyAnimation (ci, TORSO_DROP, TORSO_ATTACK2);
	if (counter < TORSO_PAIN1)
		CL_CopyAnimation (ci, TORSO_PAIN1, TORSO_STAND2);
	if (counter < TORSO_PAIN2)
		CL_CopyAnimation (ci, TORSO_PAIN2, TORSO_STAND2);
	if (counter < TORSO_PAIN3)
		CL_CopyAnimation (ci, TORSO_PAIN3, TORSO_STAND2);
	if (counter < TORSO_CRSTAND)
		CL_CopyAnimation (ci, TORSO_CRSTAND, TORSO_STAND);
	if (counter < TORSO_CRWALK)
		CL_CopyAnimation (ci, TORSO_CRWALK, TORSO_STAND);
	if (counter < TORSO_SWIM)
		CL_CopyAnimation (ci, TORSO_SWIM, TORSO_STAND);
	
	if (counter < LEGS_WALKBACK)
		CL_CopyAnimation (ci, LEGS_WALKBACK, LEGS_RUNBACK);
	if (counter < LEGS_WALKLEFT)
		CL_CopyAnimation (ci, LEGS_WALKLEFT, LEGS_WALKFWD);
	if (counter < LEGS_WALKRIGHT)
		CL_CopyAnimation (ci, LEGS_WALKRIGHT, LEGS_WALKFWD);
	if (counter < LEGS_RUNLEFT)
		CL_CopyAnimation (ci, LEGS_RUNLEFT, LEGS_RUNFWD);
	if (counter < LEGS_RUNRIGHT)
		CL_CopyAnimation (ci, LEGS_RUNRIGHT, LEGS_RUNFWD);
	if (counter < LEGS_SWIM)
		CL_CopyAnimation (ci, LEGS_SWIM, LEGS_SWIMFWD);
	
	//update the frames for the new animation script
	if (ci->anim.current[UPPER] < 1 || ci->anim.current[LOWER] < 1 || ci->anim.current[HEAD] < 1)
	{
		ci->anim.current[UPPER] = TORSO_STAND;
		ci->anim.current[LOWER] = LEGS_STAND;
		ci->anim.current[HEAD] = HEAD_NONE;
	}
	
	ci->anim.oldframe[UPPER] = ci->anim.frame[UPPER] = ci->anim.firstframe[ci->anim.current[UPPER]];
	ci->anim.oldframe[LOWER] = ci->anim.frame[LOWER] = ci->anim.firstframe[ci->anim.current[LOWER]];
	ci->anim.oldframe[HEAD] = ci->anim.frame[HEAD] = ci->anim.firstframe[ci->anim.current[HEAD]];
	
	return qtrue;
}

/*
================
CL_LoadModelClientinfo - SPLITMODELS
================
*/
qboolean CL_LoadModelClientinfo (clientinfo_t *ci, char *model_name)
{
	char	scratch_filename[MAX_QPATH];
	char	short_filename[MAX_QPATH];
	int		i;

	//torso
	Com_sprintf (scratch_filename, sizeof(scratch_filename), "models/players/%s/upper.md3", model_name);
	ci->model[UPPER] = R_RegisterModel (scratch_filename);

	//didn't work, try with skm (untested)
	if (!ci->model[UPPER])
	{
		COM_StripExtension ( scratch_filename, short_filename );
		Com_sprintf (scratch_filename, sizeof(scratch_filename), "%s.skm", short_filename);
		ci->model[UPPER] = R_RegisterModel (scratch_filename);
	}
	
	//legs
	Com_sprintf (scratch_filename, sizeof(scratch_filename), "models/players/%s/lower.md3", model_name);
	ci->model[LOWER] = R_RegisterModel (scratch_filename);

	//didn't work, try with skm (untested)
	if (!ci->model[LOWER])
	{
		COM_StripExtension ( scratch_filename, short_filename );
		Com_sprintf (scratch_filename, sizeof(scratch_filename), "%s.skm", short_filename);
		ci->model[LOWER] = R_RegisterModel (scratch_filename);
	}

	//head
	Com_sprintf (scratch_filename, sizeof(scratch_filename), "models/players/%s/head.md3", model_name);
	ci->model[HEAD] = R_RegisterModel (scratch_filename);

	//didn't work, try with skm (untested)
	if (!ci->model[HEAD])
	{
		COM_StripExtension ( scratch_filename, short_filename );
		Com_sprintf (scratch_filename, sizeof(scratch_filename), "%s.skm", short_filename);
		ci->model[HEAD] = R_RegisterModel (scratch_filename);
	}


	for (i=0; i<PLAYERMODEL_PARTS; i++)
	{
		if (!ci->model[i])
			goto cleanup;
	}

	//animation.cfg
	Com_sprintf (scratch_filename, sizeof(scratch_filename), "models/players/%s/animation.cfg", model_name);

	if (!Q_stricmp (ci->anim.filename, scratch_filename))
		return qtrue;//no need to update

	strcpy (ci->anim.filename, scratch_filename);
	if (!CL_LoadPlayerAnims(ci))
		goto cleanup;

	return qtrue;

cleanup:
	for (i=0; i<PLAYERMODEL_PARTS; i++)
		ci->model[i] = NULL;

	return qfalse;
}

/*
================
CL_LoadSkinClientinfo - SPLITMODELS (skin files)
================
*/
qboolean CL_LoadSkinClientinfo (clientinfo_t *ci, char *model_name, char *skin_name)
{
	char		scratch[MAX_QPATH];
	int			i;
	
	Com_sprintf (scratch, sizeof(scratch), "models/players/%s/upper_%s.skin", model_name, skin_name);
	ci->customSkin[UPPER] = R_RegisterSkinFile (scratch);

	Com_sprintf (scratch, sizeof(scratch), "models/players/%s/lower_%s.skin", model_name, skin_name);
	ci->customSkin[LOWER] = R_RegisterSkinFile (scratch);

	Com_sprintf (scratch, sizeof(scratch), "models/players/%s/head_%s.skin", model_name, skin_name);
	ci->customSkin[HEAD] = R_RegisterSkinFile (scratch);


	for (i=0; i<PLAYERMODEL_PARTS; i++)
	{
		if (!ci->customSkin[i])
		{
			int	j;
			for (j=0; j<PLAYERMODEL_PARTS; j++)
				ci->customSkin[j] = NULL;
			return qfalse;
		}
	}

	//icon
	Com_sprintf (ci->iconname, sizeof(ci->iconname), "models/players/%s/icon_%s", model_name, skin_name);
	ci->icon = R_RegisterPic (ci->iconname);

	return qtrue;
}

/*
================
CL_LoadClientinfo - SPLITMODELS
================
*/
void CL_LoadClientinfo (clientinfo_t *ci, char *s)
{
	int 		i;
	char		*t;
	char		model_name[MAX_QPATH];
	char		skin_name[MAX_QPATH];
	char		weapon_filename[MAX_QPATH];
	char		weaponshort_filename[MAX_QPATH];
	char		scratch[MAX_QPATH];
	qboolean 	loaded_model, loaded_skin;
	
	
	Q_strncpyz (ci->cinfo, s, sizeof(ci->cinfo));
	
	//isolate the player's name
	Q_strncpyz (ci->name, s, sizeof(ci->name));
	t = strstr (s, "\\");
	if (t)
	{
		ci->name[t-s] = 0;
		s = t+1;
	}
	
	//I should rename cl_noskins to cl_forceplayermodels
	if (cl_noskins->value || *s == 0)
	{
		//use defaults
		strcpy (model_name, DEFAULT_PLAYERMODEL);
		strcpy (skin_name, DEFAULT_PLAYERSKIN);
		
		//load model & skin
		loaded_model = CL_LoadModelClientinfo(ci, model_name);
		loaded_skin = CL_LoadSkinClientinfo(ci, model_name, skin_name);

	}
	else
	{
		//isolate the model name
		strcpy (model_name, s);
		t = strstr(model_name, "/");
		if (!t)
			t = strstr(model_name, "\\");
		if (!t)
			t = model_name;
		*t = 0;
		
		//isolate the skin name
		strcpy (skin_name, s + strlen(model_name) + 1);
		
		//try model
		loaded_model = (CL_LoadModelClientinfo(ci, model_name));

		//didn't work so back to DEFAULT_PLAYERMODEL.
		if (!loaded_model){
			strcpy (model_name, DEFAULT_PLAYERMODEL);
			loaded_model = (CL_LoadModelClientinfo(ci, model_name));
		}
	
		//skin file
		loaded_skin = (CL_LoadSkinClientinfo(ci, model_name, skin_name));
	
		//if we don't have the skin and the model wasn't DEFAULT_PLAYERMODEL,
		//see if the DEFAULT_PLAYERMODEL has it (this is for CTF's skins)
		if (!loaded_skin && Q_stricmp(model_name, DEFAULT_PLAYERMODEL))
		{
			//try on the default model with the custom skin name
			strcpy(model_name, DEFAULT_PLAYERMODEL);
			
			loaded_skin = CL_LoadSkinClientinfo(ci, model_name, skin_name);
			loaded_model = CL_LoadModelClientinfo(ci, model_name);
		}

		//if we still don't have a skin, it means that the DEFAULT_PLAYERMODEL
		//didn't have it, so back to DEFAULT_PLAYERSKIN
		if (!loaded_skin){
			strcpy(skin_name, DEFAULT_PLAYERSKIN);
			loaded_skin = (CL_LoadSkinClientinfo(ci, model_name, skin_name));
		}
	}
	
	//WEAPON file for everyone: these models are also used for the 1st person view attached weapon
	for (i = 0; i < num_cl_weaponmodels; i++) 
	{
		//playermodel folder
		Com_sprintf (weapon_filename, sizeof(weapon_filename), "models/players/%s/%s", model_name, cl_weaponmodels[i]);
		ci->weaponinfo[i].model[WEAPON] = R_RegisterModel(weapon_filename);
		 
		COM_StripExtension ( weapon_filename, weaponshort_filename );//remove .md3

		Com_sprintf (scratch, sizeof(scratch), "%s_barrel.md3", weaponshort_filename);
		ci->weaponinfo[i].model[BARREL] = R_RegisterModel(scratch);

		Com_sprintf (scratch, sizeof(scratch), "%s_flash.md3", weaponshort_filename);
		ci->weaponinfo[i].model[FLASH] = R_RegisterModel(scratch);
		
		//common weapons folder (models/weapons2 folder)
		if (!ci->weaponinfo[i].model[WEAPON]) 
		{		
			Com_sprintf (weapon_filename, sizeof(weapon_filename), "models/weapons2/%s", cl_weaponmodels[i]);
			ci->weaponinfo[i].model[WEAPON] = R_RegisterModel(weapon_filename);
		}
		if (!ci->weaponinfo[i].model[BARREL])
		{
			Com_sprintf (weapon_filename, sizeof(weapon_filename), "models/weapons2/%s", cl_weaponmodels[i]);
			COM_StripExtension ( weapon_filename, weaponshort_filename );
			Com_sprintf (scratch, sizeof(scratch), "%s_barrel.md3", weaponshort_filename);
			ci->weaponinfo[i].model[BARREL] = R_RegisterModel(scratch);
		}
		if (!ci->weaponinfo[i].model[FLASH]) 
		{
			Com_sprintf (weapon_filename, sizeof(weapon_filename), "models/weapons2/%s", cl_weaponmodels[i]);
			COM_StripExtension ( weapon_filename, weaponshort_filename );
			Com_sprintf (scratch, sizeof(scratch), "%s_flash.md3", weaponshort_filename);
			ci->weaponinfo[i].model[FLASH] = R_RegisterModel(scratch);
		}

		if (!cl_vwep->value)
						break; // only do weapon 0 (weapon.md3)
	}

	// must have loaded model & skin to be valid
	if (!loaded_model || !loaded_skin)
	{
		ci->icon = NULL;
		for (i = 0; i < PLAYERMODEL_PARTS; i++) {
		ci->customSkin[i] = NULL;
		ci->model[i] = NULL;
		}
		return;
	}
}


/*
================
CL_ParseClientinfo

Load the skin, icon, and model for a client
================
*/
void CL_ParseClientinfo (int player)
{
	CL_LoadClientinfo (&cl.clientinfo[player], cl.configstrings[player+CS_PLAYERSKINS]);
}

/*
================
CL_LoadStatusbar
================
*/
void CL_LoadStatusbar ( char *s )
{
	int length;
	char *buffer;
	char path[MAX_QPATH], shortname[MAX_QPATH];

	if ( !s ) {
		return;
	}

	// strip extension and add local path
	COM_StripExtension ( s, shortname );
	Com_sprintf ( path, sizeof(path), "huds/%s.hud", shortname );

	// load the file
	length = FS_LoadFile ( path, (void **)&buffer );
	if ( !buffer ) {
		Com_DPrintf ( "Bad hud file %s\n", path );
		return;
	}
	if ( !length ) {
		Com_DPrintf ( "Empty hud file %s\n", path );
		FS_FreeFile ( buffer );
		return;
	}

	// free old status bar if present
	if ( cl.statusbar ) { 
		Z_Free ( cl.statusbar );
	}

	// copy file contents to statusbar string
	cl.statusbar = Z_Malloc ( length );
	memcpy ( cl.statusbar, buffer, length );

	FS_FreeFile ( buffer );
}

/*
================
CL_ParseConfigString
================
*/
void CL_ParseConfigString (void)
{
	int		i;
	char	*s;
	char	olds[MAX_QPATH];

	i = MSG_ReadShort (&net_message);
	if (i < 0 || i >= MAX_CONFIGSTRINGS)
		Com_Error (ERR_DROP, "configstring > MAX_CONFIGSTRINGS");

	s = MSG_ReadString(&net_message);

	Q_strncpyz (olds, cl.configstrings[i], sizeof(olds));
	Q_strncpyz (cl.configstrings[i], s, sizeof(cl.configstrings[i]));

	// do something apropriate 

	if ( i == CS_MAPNAME ) {
		Com_sprintf ( cl.levelshot, sizeof(cl.levelshot), "levelshots/%s.jpg", cl.configstrings[CS_MAPNAME] );

		if ( !FS_FileExists ( cl.levelshot ) ) 
			Com_sprintf ( cl.levelshot, sizeof(cl.levelshot), "levelshots/%s.tga", cl.configstrings[CS_MAPNAME] );

		if ( !FS_FileExists ( cl.levelshot ) ) 
			Com_sprintf ( cl.levelshot, sizeof(cl.levelshot), "menu/art/unknownmap", cl.configstrings[CS_MAPNAME] );
	} else if ( i == CS_AUDIOTRACK ) {
		if (cl.refresh_prepped)
			cl.music_precache = S_RegisterSound (cl.configstrings[i]);
	} else if ( i == CS_STATUSBAR ) { 
		CL_LoadStatusbar ( cl.configstrings[i] );
	} else if ( i >= CS_MODELS && i < CS_MODELS+MAX_MODELS ) {
		if (cl.refresh_prepped)
		{
			cl.model_draw[i-CS_MODELS] = R_RegisterModel (cl.configstrings[i]);
			if (cl.configstrings[i][0] == '*')
				cl.model_clip[i-CS_MODELS] = CM_InlineModel (cl.configstrings[i]);
			else
				cl.model_clip[i-CS_MODELS] = NULL;
		}
	} else if ( i >= CS_SOUNDS && i < CS_SOUNDS+MAX_MODELS ) {
		if (cl.refresh_prepped)
			cl.sound_precache[i-CS_SOUNDS] = S_RegisterSound (cl.configstrings[i]);
	} else if ( i >= CS_IMAGES && i < CS_IMAGES+MAX_MODELS ) {
		if (cl.refresh_prepped)
			cl.image_precache[i-CS_IMAGES] = R_RegisterPic (cl.configstrings[i]);
	} else if (i >= CS_PLAYERSKINS && i < CS_PLAYERSKINS+MAX_CLIENTS) {
		if (cl.refresh_prepped && strcmp(olds, s))
			CL_ParseClientinfo (i-CS_PLAYERSKINS);
	}
}


/*
================
CL_ParseInventory
================
*/
void CL_ParseInventory (void)
{
	int		i;

	for (i=0 ; i<MAX_ITEMS ; i++)
		cl.inventory[i] = MSG_ReadShort (&net_message);
}

/*
=====================================================================

ACTION MESSAGES

=====================================================================
*/

/*
==================
CL_ParseStartSoundPacket
==================
*/
void CL_ParseStartSoundPacket(void)
{
    vec3_t  pos_v;
	float	*pos;
    int 	channel, ent;
    int 	sound_num;
    float 	volume;
    float 	attenuation;  
	int		flags;
	float	ofs;

	flags = MSG_ReadByte (&net_message);
	sound_num = MSG_ReadByte (&net_message);

    if (flags & SND_VOLUME)
		volume = MSG_ReadByte (&net_message) / 255.0;
	else
		volume = DEFAULT_SOUND_PACKET_VOLUME;
	
    if (flags & SND_ATTENUATION)
		attenuation = MSG_ReadByte (&net_message) / 64.0;
	else
		attenuation = DEFAULT_SOUND_PACKET_ATTENUATION;	

    if (flags & SND_OFFSET)
		ofs = MSG_ReadByte (&net_message) / 1000.0;
	else
		ofs = 0;

	if (flags & SND_ENT)
	{	// entity reletive
		channel = MSG_ReadShort(&net_message); 
		ent = channel>>3;
		if (ent > MAX_EDICTS)
			Com_Error (ERR_DROP,"CL_ParseStartSoundPacket: ent = %i", ent);

		channel &= 7;
	}
	else
	{
		ent = 0;
		channel = 0;
	}

	if (flags & SND_POS)
	{	// positioned in space
		MSG_ReadPos (&net_message, pos_v);
 		pos = pos_v;
	}
	else	// use entity number
		pos = NULL;

	if (!cl.sound_precache[sound_num])
		return;

	S_StartSound (pos, ent, channel, cl.sound_precache[sound_num], volume, attenuation, ofs);
}


void SHOWNET(char *s)
{
	if (cl_shownet->value>=2)
		Com_Printf ("%3i:%s\n", net_message.readcount-1, s);
}

/*
==================
CL_ParseObituary
==================
*/
void CL_ParseObituary (void)
{
	char message[128];
	char message2[128];
	clientinfo_t *victim, *attacker;
	int victimNum = MSG_ReadShort (&net_message);
	int attackerNum = MSG_ReadShort (&net_message);
	int mod = MSG_ReadByte (&net_message);

	victim = &cl.clientinfo[victimNum - 1];

	if ( attackerNum ) {
		attacker = &cl.clientinfo[attackerNum - 1];
	} else {
		attacker = NULL;
	}

	Q_Obituary ( victim, victim->gender, attacker, mod, message, message2 );

	if ( attackerNum ) {
		if ( victimNum != attackerNum ) {
			Com_Printf ( "%s %s%s %s%s%s\n", victim->name, S_COLOR_WHITE, message, attacker->name, S_COLOR_WHITE, message2 );
		} else {
			Com_Printf ( "%s %s%s\n", victim->name, S_COLOR_WHITE, message );
		}
	} else {
		Com_Printf ( "%s %s%s\n", victim->name, S_COLOR_WHITE, message );
	}
}


void CL_Reconnect_f (void);
void CL_Changing_f (void);
void CL_Precache_f (void);
void CL_ForwardToServer_f (void);

typedef struct {
	char	*name;
	void	(*func) (void);
} svcmd_t;

svcmd_t svcmds[] =
{
	{"reconnect", CL_Reconnect_f},
	{"changing", CL_Changing_f},
	{"precache", CL_Precache_f},
	{"cmd", CL_ForwardToServer_f},
	{NULL}
};

/*
==================
CL_ParseStringCmd
==================
*/
void CL_ParseStringCmd (void)
{
	char *text, *s;
	svcmd_t *cmd;

	text = MSG_ReadString (&net_message);
	Cmd_TokenizeString (text, qfalse);
	s = Cmd_Argv (0);

	for (cmd = svcmds; cmd->name; cmd++)
		if (!strcmp (s, cmd->name) ) {
			cmd->func ();
			return;
		}

	Com_Printf ("CL_ParseStringCmd: bad cmd \"%s\"\n", s);
}

/*
=====================
CL_ParseServerMessage
=====================
*/
void CL_ParseServerMessage (void)
{
	int			cmd;
	char		*s;
	int			i;

//
// if recording demos, copy the message out
//
	if (cl_shownet->value == 1)
		Com_Printf ("%i ",net_message.cursize);
	else if (cl_shownet->value >= 2)
		Com_Printf ("------------------\n");


//
// parse the message
//
	while (1)
	{
		if (net_message.readcount > net_message.cursize)
		{
			Com_Error (ERR_DROP,"CL_ParseServerMessage: Bad server message");
			break;
		}

		cmd = MSG_ReadByte (&net_message);

		if (cmd == -1)
		{
			SHOWNET("END OF MESSAGE");
			break;
		}

		if (cl_shownet->value>=2)
		{
			if (!svc_strings[cmd])
				Com_Printf ("%3i:BAD CMD %i\n", net_message.readcount-1,cmd);
			else
				SHOWNET(svc_strings[cmd]);
		}
	
	// other commands
		switch (cmd)
		{
		default:
			Com_Error (ERR_DROP,"CL_ParseServerMessage: Illegible server message");
			break;
			
		case svc_nop:
//			Com_Printf ("svc_nop\n");
			break;
			
		case svc_disconnect:
			Com_Error (ERR_DISCONNECT,"Server disconnected");
			break;

		case svc_reconnect:
			Com_Printf ("Server disconnected, reconnecting\n");
			if (cls.download) {
				//ZOID, close download
				fclose (cls.download);
				cls.download = NULL;
			}
			CL_SetClientState (ca_connecting);
			cls.connect_time = -99999;	// CL_CheckForResend() will fire immediately
			break;

		case svc_print:
			i = MSG_ReadByte (&net_message);

			if (i == PRINT_CHAT)
				S_StartLocalSound ("sound/player/talk.wav");

			Com_Printf ("%s", MSG_ReadString (&net_message));
			break;
			
		case svc_centerprint:
			SCR_CenterPrint (MSG_ReadString (&net_message));
			break;

		case svc_obituary:
			CL_ParseObituary ();
			break;
	
		case svc_stufftext:
			s = MSG_ReadString (&net_message);
			Com_DPrintf ("stufftext: %s\n", s);
			Cbuf_AddText (s);
			break;
			
		case svc_serverdata:
			Cbuf_Execute ();		// make sure any stuffed commands are done
			CL_ParseServerData ();
			break;
			
		case svc_configstring:
			CL_ParseConfigString ();
			break;
			
		case svc_sound:
			CL_ParseStartSoundPacket();
			break;
			
		case svc_spawnbaseline:
			CL_ParseBaseline ();
			break;

		case svc_temp_entity:
			CL_ParseTEnt ();
			break;

		case svc_muzzleflash:
			CL_ParseMuzzleFlash ();
			break;

		case svc_muzzleflash2:
			CL_ParseMuzzleFlash2 ();
			break;

		case svc_download:
			CL_ParseDownload ();
			break;

		case svc_frame:
			CL_ParseFrame ();
			break;

		case svc_inventory:
			CL_ParseInventory ();
			break;

		case svc_layout:
			s = MSG_ReadString (&net_message);
			Q_strncpyz (cl.layout, s, sizeof(cl.layout));
			break;

		case svc_playerinfo:
		case svc_packetentities:
		case svc_deltapacketentities:
			Com_Error (ERR_DROP, "Out of place frame data");
			break;

		case svc_stringcmd:
			CL_ParseStringCmd ();
			break;
		}
	}

	CL_AddNetgraph ();

	//
	// we don't know if it is ok to save a demo message until
	// after we have parsed the frame
	//
	if (cls.demorecording && !cls.demowaiting)
		CL_WriteDemoMessage ();
}


