/*
Copyright (C) 1997-2001 Id Software, Inc.
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// cl_players.c - Adding the Player models and the View Weapon in split pieces

#include "client.h"
#include "cl_players.h"

/*
==========================================================================

					 - SPLITMODELS -

==========================================================================
*/

/*
===============
CL_ChaseHackChooseClient - CHASEHACKv2
===============
*/
clientinfo_t *CL_ChaseHackChooseClient()
{
	clientinfo_t *ci;

	if (cl.chasednumber){
		ci = &cl.clientinfo[cl.chasednumber-1];//get the info from the chased player
	} else {
		ci = &cl.clientinfo[cl.playernum];//use your own info (no chase)
	}
	return ci;
}

/*
===============
CL_ChaseHackChooseCent - CHASEHACKv2 - (actually not used)
===============
*/
centity_t *CL_ChaseHackChooseCent()
{
	centity_t *cent;

	if (cl.chasednumber){
		cent = &cl_entities[cl.chasednumber];//get the info from the chased player
	} else {
		cent = &cl_entities[cl.playernum+1];//use your own info (no chase)
	}
	return cent;
}

/*
===============
JAL CHASEHACKv2

 Only used when reproducing a demo. Fixes some problems with
 synchronization of packetents & player by setting the chasehack
 again by brute force.

 Not needed anymore with the inverted order at CL_AddEntities.
===============
*/
void CL_ChaseDemoHack (void)
{
	entity_state_t		*state;
	int					pnum;

	//track the whole entity list
	for (pnum = 0 ; pnum<cl.frame.num_entities ; pnum++)
	{
		//find the clients on the list and reset viewermodel
		state = &cl_parse_entities[(cl.frame.parse_entities+pnum)&(MAX_PARSE_ENTITIES-1)];

		if ( state->number < MAX_CLIENTS+1 )
		{
			if ((cl.chasednumber == state->number) || ((cl.playernum + 1) == state->number))
				state->renderfx |= RF_VIEWERMODEL;
			else
				state->renderfx &= ~RF_VIEWERMODEL;
		}
	}
}

/*
===============
CL_ChaseHack - IN-EYES chasecam.

 jal: I disabled the demo enforcing. I think it is not
 needed anymore. Try reenabling it if the view weapon
 appears in very quick flashes while reproducing demos.
===============
*/
void CL_ChaseHack (player_state_t *ps)
{
	if ((ps->pmove.pm_type == PM_FREEZE) && (ps->pmove.pm_flags & PMF_CHASEHACK))
	{
		//gravity stores the chased player number. There the chasehack name.
		cl.chasednumber = ps->pmove.gravity;

		//if (cl.attractloop)
		//	CL_ChaseDemoHack ();
	} else
		cl.chasednumber = 0;
}

/*
===============
CL_MoveToEntityPosition
===============
*/
void CL_MoveToEntityPosition (entity_t *ent, entity_t *dest)
{
	int i;

	VectorCopy (dest->origin, ent->origin);
	VectorCopy (dest->oldorigin, ent->oldorigin);
	VectorCopy (dest->lightingOrigin, ent->lightingOrigin);

	for ( i = 0 ; i < 3 ; i++ )
		VectorCopy (dest->axis[i], ent->axis[i]);
}

/*
===============
CL_PlaceModelOnTag
===============
*/
void CL_PlaceModelOnTag (entity_t *ent, entity_t *dest, orientation_t *tag)
{
	int i;

	//move to destiny model origin
	VectorCopy (dest->origin, ent->origin);
	VectorCopy (dest->lightingOrigin, ent->lightingOrigin);

	//move to tag inside destiny model
	for ( i = 0 ; i < 3 ; i++ )
		VectorMA( ent->origin, tag->origin[i], dest->axis[i], ent->origin );

	VectorCopy (ent->origin, ent->oldorigin);

	Matrix3_Multiply( tag->axis, dest->axis, ent->axis );
}

/*
===============
CL_AddShellEffects
===============
*/
void CL_AddQuadShell(entity_t *ent)
{
	ent->customSkin = NULL;

	if ( ent->flags & RF_WEAPONMODEL )
		ent->customShader = cl.media.shaderQuadWeapon;
	else
		ent->customShader = cl.media.shaderPowerupQuad;

	V_AddEntity (ent);

	ent->customShader = NULL; //restore
}

void CL_AddPentShell (entity_t *ent)
{
	ent->customSkin = NULL;

	if ( ent->flags & RF_WEAPONMODEL )
		return;//jalfixme: look for a PENTA shader for view weapon
	else
		ent->customShader = cl.media.shaderPowerupPenta;

	V_AddEntity (ent);

	ent->customShader = NULL; //restore
}

void CL_AddShellEffects (entity_t *ent, int effects)
{
	// quad and pent can do different things on client

	if (effects & EF_QUAD)
	{
		CL_AddQuadShell (ent);
	}

	if (effects & EF_PENT)
	{
		CL_AddPentShell (ent);
	}
}

/*
===============
CL_AddColorShell
===============
*/
void CL_AddColorShell_Do (entity_t *ent, int renderfx)
{
	entity_t	shell;
	int			i;
	vec4_t shadelight = { 0, 0, 0, 0.3 };

	memset (&shell, 0, sizeof(shell));

	shell = *ent;
	shell.customSkin = NULL;

	if ( renderfx & RF_SHELL_RED )
		shadelight[0] = 1.0;
	if ( renderfx & RF_SHELL_GREEN )
		shadelight[1] = 1.0;
	if ( renderfx & RF_SHELL_BLUE )
		shadelight[2] = 1.0;

	for (i=0 ; i<4 ; i++)
		shell.color[i] = shadelight[i] * 255;

	if ( ent->flags & RF_WEAPONMODEL )
		return;//jalfixme: try the shell shader for viewweapon, or build a good one
	else
		shell.customShader = cl.media.shaderShellEffect;

	V_AddEntity (&shell);
}

void CL_AddColorShell (entity_t *ent, int renderfx)
{
	if ( !(renderfx & (RF_SHELL_GREEN | RF_SHELL_RED | RF_SHELL_BLUE)) )
		return;

	CL_AddColorShell_Do (ent, renderfx);
}

/*
===============
CL_AnimToServerFrame

 cl_playermodel_use_fps "0" : version clamped to Serverframe rate (10 fps)
===============
*/
void CL_AnimToServerFrame (animationinfo_t *anim)
{
	int i;

	//backlerp
	for (i=0; i<PLAYERMODEL_PARTS; i++)
		anim->backlerp[i] = 1.0 - cl.lerpfrac;

	//delay 1 server frame
	if (cl.frame.servertime < anim->frame_delay)
		return;
	anim->frame_delay = (cl.frame.servertime + 1);

	for (i=0; i<PLAYERMODEL_PARTS; i++)
	{
		//advance frames
		anim->oldframe[i] = anim->frame[i];
		anim->frame[i]++;

		//looping
		if (anim->frame[i] > anim->lastframe[anim->current[i]])
			anim->frame[i] = (anim->lastframe[anim->current[i]] - anim->loopingframes[anim->current[i]]);

		//new animation incoming (buffered)
		if (anim->buffer.newanim[i] && (anim->bufferframe_updated[i] != anim->buffer.serverframe[i]))
		{
			anim->current[i] = anim->buffer.newanim[i];
			anim->frame[i] = anim->firstframe[anim->buffer.newanim[i]];
			anim->bufferframe_updated[i] = anim->buffer.serverframe[i];
			anim->buffer.newanim[i] = 0;
		}
	}
}

/*
===============
CL_AnimToFrameRate

  Transforms the values inside state->frame into 3 animation values
  for head, torso and legs (head is always zero right now). It also
  uses those values to handle the animation frames clientside, as
  also does the backlerp calculation for variable framerates.

  The frames and other data is stored inside animationinfo_t. Actually
  only clientinfo has structs of this type inside, but they could also
  be applied to a client-side monsterinfo handler.

  It works like this:
  When a new animation is recieved, the new animation is launched.
  While 0 is recieved, it advances the frames in last recieved animation.
  When the frame gets to the end of the animation, it looks for the
  looping value and sends the frame back to (lastframe - loopingframes)
  and so on. If the looping value is 0, it holds the animation in the
  last frame until a new animation is recieved.
===============
*/
void CL_AnimToFrameRate (animationinfo_t *anim)
{
	int i;

	for (i=0; i<PLAYERMODEL_PARTS; i++)
	{
		//timer
		if (cl.time > anim->nextframetime[i])
		{
			anim->prevframetime[i] = cl.time;
			anim->nextframetime[i] = cl.time + anim->frametime[anim->current[i]];

			anim->backlerp[i] = 1.0f;

			//advance frames
			anim->oldframe[i] = anim->frame[i];
			anim->frame[i]++;

			//looping
			if (anim->frame[i] > anim->lastframe[anim->current[i]]){
				anim->frame[i] = (anim->lastframe[anim->current[i]] - anim->loopingframes[anim->current[i]]);
			}

			//new animation incoming (buffered)
			if (anim->buffer.newanim[i] && (anim->bufferframe_updated[i] != anim->buffer.serverframe[i]))
			{
				anim->current[i] = anim->buffer.newanim[i];
				anim->frame[i] = anim->firstframe[anim->buffer.newanim[i]];
				anim->bufferframe_updated[i] = anim->buffer.serverframe[i];
				anim->buffer.newanim[i] = 0;
			}

		} else {

			//interpolate
			anim->backlerp[i] = 1.0f - ((cl.time - anim->prevframetime[i])/(anim->nextframetime[i] - anim->prevframetime[i]));
			if (anim->backlerp[i] > 1)
				anim->backlerp[i] = 1;
			else if (anim->backlerp[i] < 0)
				anim->backlerp[i] = 0;
		}
	}
}

/*
===============
CL_AnimToFrame

  Choose framerate system
===============
*/
void CL_AnimToFrame (animationinfo_t *anim)
{
	if (!cl_ppmUseFramerates->value)
		CL_AnimToServerFrame(anim);
	else
		CL_AnimToFrameRate(anim);
}

/*
===============
CL_AnimUpdateBuffer
===============
*/
void CL_AnimUpdateBuffer(animationbuffer_t	*buffer, int newanim[PLAYERMODEL_PARTS], int serverframe)
{
	int					i;

	for ( i = 0 ; i < PLAYERMODEL_PARTS ; i++ )
	{
		if (newanim[i]){
			buffer->newanim[i] = newanim[i];
			buffer->serverframe[i] = serverframe;
		}
	}
}

/*
==================
CL_PlayerModels_UpdateAnimationBuffer

  Called just after parsing a new frame from the server
==================
*/
void CL_PlayerModels_UpdateAnimationBuffer (frame_t *frame)
{
	entity_state_t		*state;
	int					pnum, num;
	clientinfo_t		*ci;
	int					newanim[PLAYERMODEL_PARTS];

	for (pnum = 0 ; pnum<frame->num_entities ; pnum++)
	{
		num = (frame->parse_entities + pnum)&(MAX_PARSE_ENTITIES-1);
		state = &cl_parse_entities[num];

		if ( state->number < MAX_CLIENTS+1 ) {
			if (state->modelindex == 255 && !(state->renderfx & RF_DEADBODY)){

				ci = &cl.clientinfo[state->skinnum & 0xff];

				//extract the different animations from s.frame
				newanim[UPPER] = (int)(state->frame &0x3F);
				newanim[LOWER] = (int)((state->frame >>6)&0x3F);
				newanim[HEAD] = (int)((state->frame>>12)&0xF);

				CL_AnimUpdateBuffer(&ci->anim.buffer, newanim, frame->serverframe);

			}
		}
	}
}

/*
===============
CL_GetClientWeaponInfo
===============
*/
weaponinfo_t *CL_GetClientWeaponInfo(clientinfo_t *ci, int currentweapon)
{
	weaponinfo_t	*weaponinfo;

	if ((!cl_vwep->value ) || (currentweapon > MAX_CLIENTWEAPONMODELS - 1))
		currentweapon = 0; // 0 is generic weapon model replacement

	weaponinfo = &ci->weaponinfo[currentweapon];

	//use baseclientinfo if there isn't a weapon model in clientinfo
	if (!weaponinfo->model[WEAPON])
		weaponinfo = &cl.baseclientinfo.weaponinfo[currentweapon];

	//use weapon 0 if still didn't find a weapon model
	if (!weaponinfo->model[WEAPON])
		weaponinfo = &cl.baseclientinfo.weaponinfo[0];


	//store flash info
	weaponinfo->flash = qfalse;

	if ( (ci->anim.frame[UPPER] == ci->anim.firstframe[TORSO_ATTACK1]) && (ci->anim.backlerp[UPPER] > 0.3) )
		weaponinfo->flash = qtrue;

	return weaponinfo;
}

/*
===============
CL_AddWeaponOnTagWeapon

  Adds the weapon model(s) inside weaponinfo_t, positioned at
  the tag_weapon inside the entity_t.
===============
*/
void CL_AddWeaponOnTagWeapon (entity_t *ent, weaponinfo_t *weaponinfo, int effects)
{
	orientation_t		tag;
	entity_t	weapon;


	if ( !weaponinfo->model[WEAPON] || !ent->model ) //don't try without basic models
		return;

	//weapon

	////don't try without a tag_weapon in the holder model
	if (!R_LerpAttachment( &tag, ent->model, ent->frame, ent->oldframe, ent->backlerp, "tag_weapon" ))
		return;

	memset (&weapon, 0, sizeof(weapon));
	weapon.scale = ent->scale;
	weapon.flags = ent->flags;
	weapon.frame = 0;					//add weapon animations synched to torso?
	weapon.oldframe = 0;
	weapon.model = weaponinfo->model[WEAPON];

	CL_PlaceModelOnTag (&weapon, ent, &tag);

	V_AddEntity (&weapon);

	CL_AddShellEffects (&weapon, effects);
	//CL_AddColorShell (&weapon, renderfx);

	//barrel
	if (!weapon.model) //don't try barrel & flash without a weapon
		return;

	//is there a barrel model waiting?
	if ( weaponinfo->model[BARREL] )
	{
		if (R_LerpAttachment( &tag, weapon.model, weapon.frame, weapon.oldframe, weapon.backlerp, "tag_barrel" ))
		{
			entity_t	barrel;
			memset (&barrel, 0, sizeof(barrel));
			barrel.model = weaponinfo->model[BARREL];
			barrel.scale = ent->scale;
			barrel.flags = ent->flags;
			barrel.frame = 0;
			barrel.oldframe = 0;

			CL_PlaceModelOnTag (&barrel, &weapon, &tag);

			V_AddEntity (&barrel);

			CL_AddShellEffects (&barrel, effects);
			//CL_AddColorShell (&barrel, renderfx);
		}
	}

	//flash

	if (!weaponinfo->flash)
		return;

	if ( (!cl_weaponFlashes->value) || (cl_weaponFlashes->value < 2 && (ent->flags & RF_WEAPONMODEL)) )
		return;

	if ( (weaponinfo->model[FLASH]) && (R_LerpAttachment(&tag, weapon.model, weapon.frame, weapon.oldframe, weapon.backlerp, "tag_flash" )) )
	{
		entity_t	flash;
		memset (&flash, 0, sizeof(flash));
		flash.model = weaponinfo->model[FLASH];
		flash.scale = ent->scale;
		flash.flags = ent->flags | RF_NOSHADOW;
		flash.frame = 0;
		flash.oldframe = 0;

		CL_PlaceModelOnTag (&flash, &weapon, &tag);

		V_AddEntity (&flash);
	}
}

/*
===============
CL_ModelPart_Special

  Add rotations, etc.
===============
*/
void CL_ModelPart_Special (entity_t *model, int i)
{
	switch ( i )
	{
	case LOWER:
		break;
	case UPPER:
		break;
	case HEAD:
		break;
	}
}

/*
===============
CL_AddPlayerEntities

===============
*/
static char *tagnames[] = { "tag_torso", "tag_head", NULL };

void CL_AddPlayerEntities (entity_t *ent, entity_state_t *state, int effects, int renderfx)
{
	clientinfo_t		*ci;
	orientation_t		tag;
	entity_t			ppm[PLAYERMODEL_PARTS];
	int					i, prev;
	weaponinfo_t		*weaponinfo;
	int					currentweapon;


	ci = &cl.clientinfo[state->skinnum & 0xff];
	memset (&ppm, 0, sizeof(ppm));


	//transform animation values in frames
	if (state->renderfx & RF_DEADBODY){

		ppm[LOWER].frame = ci->anim.lastframe[(int)((state->frame >>6)&0x3F)];
		ppm[LOWER].oldframe = ppm[LOWER].frame;

		ppm[UPPER].frame = ci->anim.lastframe[((int)state->frame & 0x3F)];
		ppm[UPPER].oldframe = ppm[UPPER].frame;

		ppm[HEAD].frame = 0;
		ppm[HEAD].oldframe = ppm[HEAD].frame;

	} else {

		CL_AnimToFrame( &ci->anim );

		for ( i = 0 ; i < PLAYERMODEL_PARTS ; i++ )
		{
			ppm[i].backlerp = ci->anim.backlerp[i];
			ppm[i].frame = ci->anim.frame[i];
			ppm[i].oldframe = ci->anim.oldframe[i];
		}
	}


	// Add playermodel models

	//LOWER needs special handling because of not being tagged
	CL_MoveToEntityPosition (&ppm[LOWER], ent);

	//make all the model parts at once
	for ( i = 0 ; i < PLAYERMODEL_PARTS ; i++ )
	{
		prev = i - 1;
		if (i && (ppm[prev].model))	//don't try moving the first model to tag
		{
			if (R_LerpAttachment( &tag, ppm[prev].model, ppm[prev].frame, ppm[prev].oldframe, ppm[prev].backlerp, tagnames[prev] ))
			{
				CL_PlaceModelOnTag(&ppm[i], &ppm[prev], &tag);
			}
		}

		//set model & skin
		ppm[i].scale = ent->scale;
		ppm[i].flags = ent->flags;
		ppm[i].type = ET_MODEL;
		ppm[i].customShader = NULL;
		ppm[i].customSkin = ci->customSkin[i];
		ppm[i].model = ci->model[i];

		if (!ppm[i].customSkin || !ppm[i].model)
		{
			ppm[i].customSkin = cl.baseclientinfo.customSkin[i];
			ppm[i].model = cl.baseclientinfo.model[i];
		}

		CL_ModelPart_Special (&ppm[i], i);

		V_AddEntity (&ppm[i]);

		if (!ppm[i].model)	//don't continue without a model
			break;

		CL_AddShellEffects (&ppm[i], effects);
		CL_AddColorShell (&ppm[i], renderfx);
	}


	// Add weapon model
	//-------------------

	if (state->modelindex2 != 255) //player doesn't have a weapon
		return;

	currentweapon = (state->skinnum >> 8);

	weaponinfo = CL_GetClientWeaponInfo(ci, currentweapon);

	CL_AddWeaponOnTagWeapon (&ppm[UPPER], weaponinfo, 0);
}




//======================================================================
//						ViewWeapon
//======================================================================

#define ANIMWEAP_NONE		0
#define ANIMWEAP_FLIPIN		1
#define ANIMWEAP_FLIPOUT	2
#define ANIMWEAP_ATTACK		3

/*
===============
CL_PositionViewWeapon

  Custom gun position
===============
*/
void CL_PositionViewWeapon (entity_t *gun, player_state_t *ps, player_state_t *ops)
{
	vec3_t			gun_angles;
	int				i;
	vec3_t			anglemovex, anglemovey, anglemovez;
	float			gunx, guny, gunz;

	extern cvar_t *hand;

	//set up gun position
	for (i=0 ; i<3 ; i++)
	{
		gun->origin[i] = cl.refdef.vieworg[i];
		gun_angles[i] = cl.refdef.viewangles[i] + LerpAngle (ops->gunangles[i], ps->gunangles[i], cl.lerpfrac);
	}

	//Move to client-defined position
	gunx = (cl_gunx->value);
	guny = (cl_guny->value);
	gunz = (cl_gunz->value);

	//move hand 2 to the left (center) by default
	if (hand->value == 2)
		gunx -= 3;

	//Move the gun
	AngleVectors(gun_angles, anglemovez, anglemovex, anglemovey);

	VectorScale(anglemovex, gunx, anglemovex);
	VectorScale(anglemovey, guny, anglemovey);
	VectorScale(anglemovez, gunz, anglemovez);
	VectorAdd(gun->origin, anglemovex, gun->origin);
	VectorAdd(gun->origin, anglemovey, gun->origin);
	VectorAdd(gun->origin, anglemovez, gun->origin);

	AnglesToAxis ( gun_angles, gun->axis );
}

/*
===============
CL_ViewWeaponSetAnimation

  These are hardcoded animations for Q3 *hand models*.
  It should be easy to expand and make scriptable anima-
  tions, or simply add more of them.
===============
*/
void CL_ViewWeaponSetAnimation (clientinfo_t *ci)
{

	cl.vweap.frame_updated = cl.frame.serverframe;


	//changeweapon animation
	if (cl.vweap.anim_priority == ANIMWEAP_FLIPOUT)
	{
		//hold on until a new weapon is received from both packets and playerstate
		if ((cl.vweap.currentweapon != cl.vweap.latchedweapon) && (cl.vweap.currentgunindex != cl.vweap.latchedgunindex))
			goto changeweapon;
	}
	else if ((!cl.vweap.currentweapon))
	{
		//No weapon. Try showing up one
		goto changeweapon;
	}
	else if ((cl.vweap.currentweapon != cl.vweap.latchedweapon) || (cl.vweap.currentgunindex != cl.vweap.latchedgunindex))
	{
		//Wrong Weapon: Try rebooting it
		goto changeweapon;
	}

	//begin flipout
	if ((ci->anim.current[UPPER] == TORSO_FLIPOUT) && !cl.vweap.anim_priority)
	{
		cl.vweap.anim_priority = ANIMWEAP_FLIPOUT;
		cl.vweap.frame = 6;
		return;
	}

	//attack
	if ( (ci->anim.current[UPPER] == TORSO_ATTACK1) && (ci->anim.frame[UPPER] == (ci->anim.firstframe[TORSO_ATTACK1])) )
	{
			cl.vweap.frame = 1;

		cl.vweap.anim_priority = ANIMWEAP_ATTACK;
		return;
	}

	return;

changeweapon:

	if (cl.vweap.latchedweapon){
		cl.vweap.currentweapon = cl.vweap.latchedweapon;
		cl.vweap.currentgunindex = cl.vweap.latchedgunindex;
		cl.vweap.anim_priority = ANIMWEAP_FLIPIN;
		cl.vweap.frame = 11;
		cl.vweap.oldframe = cl.vweap.frame;//don't lerp
		cl.vweap.backlerp = 0;
	}
}

/*
===============
CL_ViewWeaponSetFrame

  These are hardcoded animations for Q3 *hand models*.
  It should be easy to expand and make scriptable anima-
  tions, or simply add more of them.
===============
*/
void CL_ViewWeaponSetFrame (clientinfo_t *ci)
{
	cl.vweap.oldframe = cl.vweap.frame;

	if (cl.vweap.anim_priority == ANIMWEAP_FLIPOUT)
	{
		//run the weapon flipout animation (6-10)
		cl.vweap.frame++;
		if (cl.vweap.frame > 10)
			cl.vweap.frame = 10;//wait at frame 10
	}
	else if (cl.vweap.anim_priority == ANIMWEAP_FLIPIN)
	{
		//run the weapon flipin animation(11-15)
		cl.vweap.frame++;
		if (cl.vweap.frame > 15){
			cl.vweap.frame = 0;
			cl.vweap.anim_priority = ANIMWEAP_NONE;//finish
		}
	}
	else if (cl.vweap.anim_priority == ANIMWEAP_ATTACK)
	{
		//run the weapon attack animation (1-5)
		cl.vweap.frame++;
		if (cl.vweap.frame > 5){
			cl.vweap.frame = 0;
			cl.vweap.anim_priority = ANIMWEAP_NONE;//finish
		}
	}
	else
		cl.vweap.frame = 0;
}

/*
===============
CL_ViewWeaponServerFrame
===============
*/
void CL_ViewWeaponServerFrame (clientinfo_t	*ci)
{

	cl.vweap.backlerp = 1.0 - cl.lerpfrac;

	if (cl.frame.servertime < cl.vweap.fps_delay) //delay 1 serverframe (10 fps)
		return;

	cl.vweap.fps_delay = (cl.frame.servertime + 1);

	CL_ViewWeaponSetFrame (ci);

	CL_ViewWeaponSetAnimation(ci);
}

/*
===============
CL_ViewWeaponFramerate

  Same as CL_VieWeaponFrame, but with frame times based on TORSO model.
===============
*/
void CL_ViewWeaponFrameRate (clientinfo_t *ci)
{
	//timer (using torso fps)
	if (cl.time > cl.vweap.nextframetime)
	{
		//change frame
		cl.vweap.backlerp = 1.0f;
		cl.vweap.prevframetime = cl.time;
		cl.vweap.nextframetime = cl.time + ci->anim.frametime[ci->anim.current[UPPER]];

		CL_ViewWeaponSetFrame (ci);

		if (cl.vweap.frame_updated != cl.frame.serverframe)
			CL_ViewWeaponSetAnimation(ci);

	} else {

		//interpolate between frames
		cl.vweap.backlerp = 1.0f - ((cl.time - cl.vweap.prevframetime)/(cl.vweap.nextframetime - cl.vweap.prevframetime));
		if (cl.vweap.backlerp > 1)
			cl.vweap.backlerp = 1.0f;
		else if (cl.vweap.backlerp < 0)
			cl.vweap.backlerp = 0;
	}
}

/*
===============
CL_ViewWeaponFramerate

  Choose Framerate system
===============
*/
void CL_ViewWeaponFrame(clientinfo_t *ci)
{
	if (!cl_ppmUseFramerates->value)
	{
		CL_ViewWeaponServerFrame(ci);
	}
	else
		CL_ViewWeaponFrameRate(ci);
}

/*
==============
CL_AddViewWeapon
==============
*/
void CL_AddFirstViewWeapon (player_state_t *ps, player_state_t *ops)
{
	entity_t		gun;	// md2 hand+weap, or md3 model holder: let's say HAND model
	clientinfo_t	*ci;
	orientation_t	tag;
	weaponinfo_t	*weaponinfo;


	//hand entity & model
	memset (&gun, 0, sizeof(gun));
	gun.model = cl.model_draw[cl.vweap.currentgunindex];	//[ps->gunindex] is buffered
	if (!gun.model || !ps->gunindex)
	{
		//clean up for rebooting
		cl.vweap.currentgunindex = 0;
		cl.vweap.anim_priority = ANIMWEAP_NONE;
		cl.vweap.currentweapon = 0;
		return;
	}

	//setup
	ci = CL_ChaseHackChooseClient();
	CL_PositionViewWeapon (&gun, ps, ops);
	CL_ViewWeaponFrame(ci);//Q3 Animation

	//setup common parameters
	gun.flags = RF_MINLIGHT | RF_DEPTHHACK | RF_WEAPONMODEL;
	gun.scale = 1.0;
	VectorCopy (gun.origin, gun.oldorigin);	// don't lerp origin
	VectorCopy (gun.origin, gun.lightingOrigin);


	//proceed

	//Q2
	if (!R_LerpAttachment( &tag, gun.model, 0, 0, 0, "tag_weapon" ))
	{
		gun.frame = ps->gunframe;
		if (gun.frame == 0)
			gun.oldframe = 0;	// just changed weapons, don't lerp from old
		else
			gun.oldframe = ops->gunframe;

		gun.backlerp = 1.0 - cl.lerpfrac;

		V_AddEntity (&gun);

		CL_AddShellEffects(&gun, cl.effects);

		cl.vweap.currentweapon = 0;//remove attached weapon

		return;
	}

	//Q3
	gun.frame = cl.vweap.frame;
	gun.oldframe = cl.vweap.oldframe;
	gun.backlerp = cl.vweap.backlerp;

	V_AddEntity (&gun);

	CL_AddShellEffects(&gun, cl.effects);

	// add attached weapon
	weaponinfo = CL_GetClientWeaponInfo(ci, cl.vweap.currentweapon);

	CL_AddWeaponOnTagWeapon (&gun, weaponinfo, cl.effects);
}
