/*
Copyright (C) 1997-2001 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// cl_tent.c -- client side temporary entities

#include "client.h"

#define	MAX_BEAMS			32
#define	MAX_LOCAL_ENTITIES	512

typedef enum
{
	LE_FREE,
	LE_NO_FADE,
	LE_RGB_FADE,
	LE_ALPHA_FADE,
	LE_SCALE_ALPHA_FADE,
	LE_LASER
} letype_t;

typedef struct lentity_s
{
	struct lentity_s *prev, *next;

	letype_t	type;

	entity_t	ent;

	float		start;

	float		light;
	vec3_t		lightcolor;

	vec3_t		velocity;
	vec3_t		accel;

	int			trailcount;
	int			frames;
} lentity_t;

typedef struct
{
	int		entity;
	int		dest_entity;
	struct model_s	*model;
	int		endtime;
	vec3_t	offset;
	vec3_t	start, end;
} beam_t;

lentity_t	cl_localents[MAX_LOCAL_ENTITIES];
lentity_t	cl_localents_headnode, *cl_free_lents;

beam_t		cl_beams[MAX_BEAMS];

cl_beament_t	cl_beamEnts[MAX_BEAMENTS];
int				cl_numBeamEnts;

void CL_BlasterParticles (vec3_t org, vec3_t dir);
void CL_BFGExplosionParticles (vec3_t org);
int	 CL_PMpointcontents (vec3_t point);

/*
=================
CL_RegisterMediaSounds
=================
*/
void CL_RegisterMediaSounds (void)
{
	int		i;
	char	name[MAX_QPATH];

	cl.media.sfxRic1 = S_RegisterSound ( "sound/weapons/machinegun/ric1.wav" );
	cl.media.sfxRic2 = S_RegisterSound ( "sound/weapons/machinegun/ric2.wav" );
	cl.media.sfxRic3 = S_RegisterSound ( "sound/weapons/machinegun/ric3.wav" );

	cl.media.sfxLashit = S_RegisterSound ( "sound/weapons/lashit.wav" );

	cl.media.sfxSpark5 = S_RegisterSound ( "sound/world/spark5.wav" );
	cl.media.sfxSpark6 = S_RegisterSound ( "sound/world/spark6.wav" );
	cl.media.sfxSpark7 = S_RegisterSound ( "sound/world/spark7.wav" );

	cl.media.sfxRailg = S_RegisterSound ( "sound/weapons/railgun/railgf1a.wav" );

	cl.media.sfxRockexp = S_RegisterSound ( "sound/weapons/rocket/rocklx1a.wav" );
	cl.media.sfxGrenexp = S_RegisterSound ( "sound/weapons/grenlx1a.wav" );
	cl.media.sfxWatrexp = S_RegisterSound ( "sound/weapons/xpld_wat.wav" );

	cl.media.sfxItemRespawn = S_RegisterSound ( "sound/items/respawn1.wav" );
	cl.media.sfxTeleportIn = S_RegisterSound ( "sound/world/telein.wav" );
	cl.media.sfxTeleportOut = S_RegisterSound ( "sound/world/teleout.wav" );
	cl.media.sfxJumpPad = S_RegisterSound ( "sound/world/jumppad.wav" );
	cl.media.sfxLand = S_RegisterSound ( "sound/player/land1.wav" );
	cl.media.sfxGibSound = S_RegisterSound ( "sound/player/gibsplt1.wav" );

	for ( i = 0; i < 4; i++ ) {
		Com_sprintf ( name, sizeof(name), "sound/player/footsteps/step%i.wav", i+1 );
		cl.media.sfxFootsteps[FOOTSTEP_NORMAL][i] = S_RegisterSound ( name );

		Com_sprintf ( name, sizeof(name), "sound/player/footsteps/boot%i.wav", i+1 );
		cl.media.sfxFootsteps[FOOTSTEP_BOOT][i] = S_RegisterSound ( name );

		Com_sprintf ( name, sizeof(name), "sound/player/footsteps/flesh%i.wav", i+1 );
		cl.media.sfxFootsteps[FOOTSTEP_FLESH][i] = S_RegisterSound ( name );

		Com_sprintf ( name, sizeof(name), "sound/player/footsteps/mech%i.wav", i+1 );
		cl.media.sfxFootsteps[FOOTSTEP_MECH][i] = S_RegisterSound ( name );

		Com_sprintf ( name, sizeof(name), "sound/player/footsteps/energy%i.wav", i+1 );
		cl.media.sfxFootsteps[FOOTSTEP_ENERGY][i] = S_RegisterSound ( name );

		Com_sprintf ( name, sizeof(name), "sound/player/footsteps/splash%i.wav", i+1 );
		cl.media.sfxFootsteps[FOOTSTEP_SPLASH][i] = S_RegisterSound ( name );

		Com_sprintf ( name, sizeof(name), "sound/player/footsteps/clank%i.wav", i+1 );
		cl.media.sfxFootsteps[FOOTSTEP_METAL][i] = S_RegisterSound ( name );
	}

	for ( i = 0; i < 4; i++ ) {
		Com_sprintf ( name, sizeof(name), "sound/weapons/machinegun/machgf%ib.wav", i+1 );
		cl.media.sfxMachinegunSplashes[i] = S_RegisterSound ( name );
	}

	cl.media.sfxHyperblasterSplash = S_RegisterSound("sound/weapons/hyprbf1a.wav");

	cl.media.sfxLightning = S_RegisterSound ( "sound/weapons/tesla.wav");
	cl.media.sfxDisrexp = S_RegisterSound ( "sound/weapons/disrupthit.wav");

	cl.media.sfxGrenBounce1 = S_RegisterSound ( "sound/weapons/grenade/hgrenb1a.wav" );
	cl.media.sfxGrenBounce2 = S_RegisterSound ( "sound/weapons/grenade/hgrenb2a.wav" );
}

/*
=================
CL_RegisterMediaModels
=================
*/
void CL_RegisterMediaModels (void)
{
	//qbism md3 replace... in progress
	//cl.media.modExplode = R_RegisterModel ("models/objects/explode/tris.md2");
	cl.media.modBulletExplode = R_RegisterModel ("models/weaphits/bullet.md3");  //qbism md3 replace
	cl.media.modFlash = R_RegisterModel ("models/objects/flash/tris.md2");
	cl.media.modParasiteSegment = R_RegisterModel ("models/monsters/parasite/segment/tris.md2");
	cl.media.modGrappleCable = R_RegisterModel ("models/ctf/segment/tris.md2");
	cl.media.modParasiteTip = R_RegisterModel ("models/monsters/parasite/tip/tris.md2");
	cl.media.modExplo4 = R_RegisterModel ("models/objects/r_explode/tris.md2");
	cl.media.modBfgExplo = R_RegisterModel ("sprites/s_bfg2.sp2");
	cl.media.modPowerScreen = R_RegisterModel ("models/items/armor/effect/tris.md2");
	cl.media.modLightning = R_RegisterModel ("models/proj/lightning/tris.md2");
	cl.media.modMeatyGib = R_RegisterModel ("models/gibs/intestine.md3"); //qbism md3 replace
	cl.media.modTeleportEffect = R_RegisterModel ( "models/misc/telep.md3" );


	//qbism md3 replace
	R_RegisterModel ("models/powerups/health/small_sphere_2.md3");
	R_RegisterModel ("models/ammo/grenade1.md3");
	R_RegisterModel ("models/gibs/forearm.md3");
	R_RegisterModel ("models/gibs/foot.md3");
	R_RegisterModel ("models/gibs/abdomen.md3"); //qbism added
}

/*
=================
CL_RegisterMediaShaders
=================
*/
void CL_RegisterMediaShaders (void)
{
	cl.media.shaderBulletExplosion = R_RegisterShader ( "bulletExplosion" );
	cl.media.shaderGrenadeExplosion = R_RegisterShader ( "grenadeExplosion" );
	cl.media.shaderRocketExplosion = R_RegisterShader ( "rocketExplosion" );
	cl.media.shaderWaterBubble = R_RegisterShader ( "waterBubble" );
	cl.media.shaderTeleportEffect = R_RegisterShader ( "teleportEffect" );
	cl.media.shaderSmokePuff = R_RegisterShader ( "smokePuff" );
	cl.media.shaderBulletMark = R_RegisterShader( "gfx/damage/bullet_mrk" );
	cl.media.shaderExplosionMark = R_RegisterShader ( "gfx/damage/burn_med_mrk" );
	cl.media.shaderEnergyMark = R_RegisterShader ( "gfx/damage/plasma_mrk" );
	cl.media.shaderLaser = R_RegisterShader ( "laser" );

	cl.media.shaderPowerupQuad = R_RegisterSkin ( "powerups/quad" );
	cl.media.shaderQuadWeapon = R_RegisterSkin ( "powerups/quadWeapon" );
	cl.media.shaderPowerupPenta = R_RegisterSkin ( "powerups/blueflag" );
	cl.media.shaderShellEffect = R_RegisterSkin ( "shellEffect" );
}

/*
=================
CL_ClearLocalEntities
=================
*/
void CL_ClearLocalEntities (void)
{
	int i;

	memset ( cl_beams, 0, sizeof(cl_beams) );
	memset ( cl_localents, 0, sizeof(cl_localents) );

	// link local entities
	cl_free_lents = cl_localents;
	cl_localents_headnode.prev = &cl_localents_headnode;
	cl_localents_headnode.next = &cl_localents_headnode;
	for ( i = 0; i < MAX_LOCAL_ENTITIES - 1; i++ ) {
		cl_localents[i].next = &cl_localents[i+1];
	}
}

/*
=================
CL_AllocLocalEntity
=================
*/
lentity_t *CL_AllocLocalEntity ( int type )
{
	lentity_t *le;

	if ( cl_free_lents ) {	// take a free decal if possible
		le = cl_free_lents;
		cl_free_lents = le->next;
	} else {				// grab the oldest one otherwise
		le = cl_localents_headnode.prev;
		le->prev->next = le->next;
		le->next->prev = le->prev;
	}

	memset ( le, 0, sizeof (*le) );
	le->type = type;
	le->start = cl.frame.servertime - 100;

	// put the decal at the start of the list
	le->prev = &cl_localents_headnode;
	le->next = cl_localents_headnode.next;
	le->next->prev = le;
	le->prev->next = le;

	return le;
}

/*
=================
CL_FreeLocalEntity
=================
*/
void CL_FreeLocalEntity ( lentity_t *le )
{
	// remove from linked active list
	le->prev->next = le->next;
	le->next->prev = le->prev;

	// insert into linked free list
	le->next = cl_free_lents;
	cl_free_lents = le;
}

/*
=================
CL_AllocPoly
=================
*/
lentity_t *CL_AllocPoly ( letype_t type, vec3_t origin, vec3_t angles, int frames,
				float light, float lr, float lg, float lb, struct model_s *model, struct shader_s *shader  )
{
	lentity_t	*le;

	le = CL_AllocLocalEntity ( type );
	le->frames = frames;
	le->light = light;
	le->lightcolor[0] = lr;
	le->lightcolor[1] = lg;
	le->lightcolor[2] = lb;

	le->ent.type = ET_MODEL;
	le->ent.flags = RF_NOSHADOW;
	le->ent.model = model;
	le->ent.customShader = shader;
	le->ent.shaderTime = cl.time * 0.001f;
	le->ent.scale = 1.0f;
	le->ent.color[0] = 255;
	le->ent.color[1] = 255;
	le->ent.color[2] = 255;
	le->ent.color[3] = 255;

	AnglesToAxis ( angles, le->ent.axis );
	VectorCopy ( origin, le->ent.origin );

	return le;
}

/*
=================
CL_AllocSprite
=================
*/
lentity_t *CL_AllocSprite ( letype_t type, vec3_t origin, float radius, int frames,
				float light, float lr, float lg, float lb, struct shader_s *shader  )
{
	lentity_t	*le;

	le = CL_AllocLocalEntity ( type );
	le->frames = frames;
	le->light = light;
	le->lightcolor[0] = lr;
	le->lightcolor[1] = lg;
	le->lightcolor[2] = lb;

	le->ent.type = ET_SPRITE;
	le->ent.flags = RF_NOSHADOW;
	le->ent.radius = radius;
	le->ent.customShader = shader;
	le->ent.shaderTime = cl.time * 0.001f;
	le->ent.scale = 1.0f;
	le->ent.color[0] = 255;
	le->ent.color[1] = 255;
	le->ent.color[2] = 255;
	le->ent.color[3] = 255;

	Matrix3_Copy ( mat3_identity, le->ent.axis );
	VectorCopy ( origin, le->ent.origin );

	return le;
}

/*
=================
CL_AllocLaser
=================
*/
lentity_t *CL_AllocLaser ( vec3_t start, vec3_t end, float radius, int frames, int color, struct shader_s *shader  )
{
	lentity_t	*le;

	le = CL_AllocLocalEntity ( LE_LASER );
	le->frames = frames;

	le->ent.radius = radius;
	le->ent.customShader = shader;
	le->ent.skinnum = color;

	VectorCopy ( start, le->ent.origin );
	VectorCopy ( end, le->ent.oldorigin );

	return le;
}

/*
=================
CL_AddLaser
=================
*/
void CL_AddLaser ( vec3_t start, vec3_t end, float radius, int colors, struct shader_s *shader )
{
	int				i;
	vec3_t			perpvec;
	vec3_t			direction, normalized_direction;
	vec3_t			start_points[NUM_BEAM_SEGS], end_points[NUM_BEAM_SEGS];
	vec3_t			oldorigin, origin;
	poly_t			*poly;
	cl_beament_t	*beamEnt;

	if ( cl_numBeamEnts >= MAX_BEAMENTS ) {
		return;
	}

	VectorCopy ( start, origin );
	VectorCopy ( end, oldorigin );
	VectorSubtract ( oldorigin, origin, direction );

	if ( VectorNormalize2( direction, normalized_direction ) == 0 )
		return;

	PerpendicularVector( perpvec, normalized_direction );
	VectorScale( perpvec, radius, perpvec );

	for ( i = 0; i < NUM_BEAM_SEGS; i++ )
	{
		RotatePointAroundVector( start_points[i], normalized_direction, perpvec, (360.0/NUM_BEAM_SEGS)*i );
		VectorAdd( start_points[i], origin, start_points[i] );
		VectorAdd( start_points[i], direction, end_points[i] );
	}

	beamEnt = &cl_beamEnts[cl_numBeamEnts++];

	for ( i = 0, poly = beamEnt->polys; i < NUM_BEAM_SEGS; i++, poly++ )
	{
		poly->numverts = 4;
		poly->shader = shader;
		poly->colors = beamEnt->colors[i];
		poly->stcoords = beamEnt->stcoords[i];
		poly->verts = beamEnt->verts[i];

		VectorCopy ( start_points[i], poly->verts[0] );
		poly->stcoords[0][0] = 0;
		poly->stcoords[0][1] = 0;
		*(int *)poly->colors[0] = colors;

		VectorCopy ( end_points[i], poly->verts[1] );
		poly->stcoords[1][0] = 0;
		poly->stcoords[1][1] = 1;
		*(int *)poly->colors[1] = colors;

		VectorCopy ( end_points[(i+1)%NUM_BEAM_SEGS], poly->verts[2] );
		poly->stcoords[2][0] = 1;
		poly->stcoords[2][1] = 1;
		*(int *)poly->colors[2] = colors;

		VectorCopy ( start_points[(i+1)%NUM_BEAM_SEGS], poly->verts[3] );
		poly->stcoords[3][0] = 1;
		poly->stcoords[3][1] = 0;
		*(int *)poly->colors[3] = colors;

		V_AddPoly ( poly );
	}
}

/*
=================
CL_BulletExplosion
=================
*/
void CL_BulletExplosion ( vec3_t origin, vec3_t dir ) //qbism from Jalisko splitmod 1.1
{
	lentity_t	*le;

	le = CL_AllocPoly ( LE_NO_FADE, origin, vec3_origin, 2, 0, 0, 0, 0, cl.media.modFlash, NULL );

	le = CL_AllocPoly (LE_NO_FADE, origin, vec3_origin, 4, 0, 0, 0, 0, cl.media.modBulletExplode,
		cl.media.shaderBulletExplosion);
		if ( !dir || VectorCompare(dir, vec3_origin) ) {
		Matrix3_Identity ( le->ent.axis );
	} else {
		vec3_t v;

		VectorMA ( le->ent.origin, -8, dir, le->ent.origin );
		VectorCopy ( dir, le->ent.axis[0] );
		PerpendicularVector ( v, le->ent.axis[0] );
		RotatePointAroundVector ( le->ent.axis[1], le->ent.axis[0], v, rand() % 360 );
		CrossProduct ( le->ent.axis[0], le->ent.axis[1], le->ent.axis[2] );
	}
}

/*
=================
CL_AddBeam
=================
*/
void CL_AddBeam (int ent, vec3_t start, vec3_t end, vec3_t offset, struct model_s *model )
{
	int		i;
	beam_t	*b;

	if ( !model ) {
		return;
	}

// override any beam with the same entity
	for ( i = 0, b = cl_beams; i < MAX_BEAMS; i++, b++ ) {
		if ( b->entity != ent ) {
			continue;
		}

		b->entity = ent;
		b->model = model;
		b->endtime = cl.time + 200;
		VectorCopy (start, b->start);
		VectorCopy (end, b->end);
		VectorCopy (offset, b->offset);
		return;
	}

// find a free beam
	for ( i = 0, b = cl_beams; i < MAX_BEAMS; i++, b++ ) {
		if ( b->model || b->endtime >= cl.time ) {
			continue;
		}

		b->entity = ent;
		b->model = model;
		b->endtime = cl.time + 200;
		VectorCopy (start, b->start);
		VectorCopy (end, b->end);
		VectorCopy (offset, b->offset);
		return;
	}

	Com_DPrintf ("Beam list overflow!\n");
}

/*
=================
CL_AddLightning
=================
*/
void CL_AddLightning (int srcEnt, int destEnt, vec3_t start, vec3_t end, struct model_s *model)
{
	int		i;
	beam_t	*b;

	if ( !model ) {
		return;
	}

// override any beam with the same source AND destination entities
	for ( i = 0, b = cl_beams; i < MAX_BEAMS; i++, b++ ) {
		if ( b->entity != srcEnt || b->dest_entity != destEnt ) {
			continue;
		}

		b->entity = srcEnt;
		b->dest_entity = destEnt;
		b->model = model;
		b->endtime = cl.time + 200;
		VectorCopy (start, b->start);
		VectorCopy (end, b->end);
		VectorClear (b->offset);
		return;
	}

// find a free beam
	for ( i = 0, b = cl_beams; i < MAX_BEAMS; i++, b++ ) {
		if ( b->model || b->endtime >= cl.time ) {
			continue;
		}

		b->entity = srcEnt;
		b->dest_entity = destEnt;
		b->model = model;
		b->endtime = cl.time + 200;
		VectorCopy (start, b->start);
		VectorCopy (end, b->end);
		VectorClear (b->offset);
		return;
	}

	Com_DPrintf ("Beam list overflow!\n");
}

/*
===============
CL_BubbleTrail
===============
*/
void CL_BubbleTrail ( vec3_t start, vec3_t end, int dist )
{
	int			i;
	float		len;
	vec3_t		move, vec;
	lentity_t	*le;

	VectorCopy ( start, move );
	VectorSubtract ( end, start, vec );
	len = VectorNormalize ( vec );
	VectorScale ( vec, dist, vec );

	for ( i = 0; i < len; i += dist )
	{
		le = CL_AllocSprite ( LE_ALPHA_FADE, move, 3, 10, 0, 0, 0, 0, cl.media.shaderWaterBubble );
		VectorSet ( le->velocity, crandom()*5, crandom()*5, crandom()*5 + 6 );
		VectorAdd ( move, vec, move );
	}
}

/*
===============
CL_RocketTrail
===============
*/
void CL_RocketTrail (vec3_t start, vec3_t end)
{
	lentity_t	*le;
	float		len;
	vec3_t		vec;
	int			contents, oldcontents;

	if ( random() < 0.5f ) {
		return;
	}

	VectorSubtract ( end, start, vec );
	len = VectorNormalize ( vec );
	if ( !len ) {
		return;
	}

	contents = CM_PointContents (end, NULL);
	oldcontents = CM_PointContents (start, NULL);

	if ( contents & MASK_WATER ) {
		if ( oldcontents & MASK_WATER ) {
			CL_BubbleTrail ( start, end, 8 );
		}
		return;
	}

	le = CL_AllocSprite ( LE_SCALE_ALPHA_FADE, end, 5, 9, 0, 0, 0, 0, cl.media.shaderSmokePuff );
	VectorSet ( le->velocity, -vec[0] * 5 + crandom()*5, -vec[1] * 5 + crandom()*5, -vec[2] * 5 + crandom()*5 + 3 );
}

/*
===============
CL_GrenadeTrail
===============
*/
void CL_GrenadeTrail (vec3_t start, vec3_t end)
{
	lentity_t	*le;
	float		len;
	vec3_t		vec;
	int			contents, oldcontents;

	if ( random() < 0.4f ) {
		return;
	}

	VectorSubtract ( end, start, vec );
	len = VectorNormalize ( vec );
	if ( !len ) {
		return;
	}

	contents = CM_PointContents (end, NULL);
	oldcontents = CM_PointContents (start, NULL);

	if ( contents & MASK_WATER ) {
		if ( oldcontents & MASK_WATER ) {
			CL_BubbleTrail ( start, end, 8 );
		}
		return;
	}

	le = CL_AllocSprite ( LE_SCALE_ALPHA_FADE, end, 8, 8, 0, 0, 0, 0, cl.media.shaderSmokePuff );
	VectorSet ( le->velocity, -vec[0] * 5 + crandom()*5, -vec[1] * 5 + crandom()*5, -vec[2] * 5 + crandom()*5 + 3 );
	le->ent.color[0] = 64;
	le->ent.color[1] = 64;
	le->ent.color[2] = 64;
}

/*
===============
CL_TeleportEffect
===============
*/
void CL_TeleportEffect ( vec3_t org )
{
	lentity_t *le;

	le = CL_AllocPoly ( LE_RGB_FADE, org, vec3_origin, 5, 0, 0, 0, 0, cl.media.modTeleportEffect, cl.media.shaderTeleportEffect );
	le->ent.origin[2] -= 24;
}

/*
=================
CL_ParseTEnt
=================
*/
static vec3_t splash_color[] =
{
	{ 0, 0, 0 },
	{ 1, 0.67, 0 },
	{ 0.47, 0.48, 0.8 },
	{ 0.48, 0.37, 0.3 },
	{ 0, 1, 0 },
	{ 1, 0.67, 0 },
	{ 0.61, 0.1, 0 }
};

void CL_ParseTEnt (void)
{
	int			type;
	vec3_t		pos, pos2, dir, angles;
	vec3_t		color;
	int			cnt, r;
	int			ent, ent2;
	lentity_t	*le;

	type = MSG_ReadByte (&net_message);

	switch (type)
	{
	case TE_BLOOD:			// bullet hitting flesh
		MSG_ReadPos (&net_message, pos);
		MSG_ReadDir (&net_message, dir);
		CL_ParticleEffect (pos, dir, 0.61, 0.1, 0, 60);
		break;

	case TE_GUNSHOT:			// bullet hitting wall
	case TE_SPARKS:
	case TE_BULLET_SPARKS:
		MSG_ReadPos (&net_message, pos);
		MSG_ReadDir (&net_message, dir);

		if (type == TE_GUNSHOT)
			CL_ParticleEffect (pos, dir, 0, 0, 0, 40);
		else
			CL_ParticleEffect (pos, dir, 1, 0.67, 0, 6);

		if (type != TE_SPARKS)
		{
			CL_BulletExplosion(pos, dir);

			// impact sound
			cnt = rand() & 3;
			if (cnt < 2)
				S_StartSound (pos, 0, 0, cl.media.sfxRic1, 1, ATTN_NORM, 0);
			else if (cnt == 2)
				S_StartSound (pos, 0, 0, cl.media.sfxRic2, 1, ATTN_NORM, 0);
			else
				S_StartSound (pos, 0, 0, cl.media.sfxRic3, 1, ATTN_NORM, 0);

			CL_SpawnDecal ( pos, dir, random()*360, 8, 1, 1, 1, 1, 8, 1, qfalse, cl.media.shaderBulletMark );
		}

		break;

	case TE_SCREEN_SPARKS:
	case TE_SHIELD_SPARKS:
		MSG_ReadPos (&net_message, pos);
		MSG_ReadDir (&net_message, dir);

		if (type == TE_SCREEN_SPARKS)
			CL_ParticleEffect (pos, dir, 0, 1, 0, 40);
		else
			CL_ParticleEffect (pos, dir, 0.47, 0.48, 0.8, 40);
		//FIXME : replace or remove this sound
		S_StartSound (pos, 0, 0, cl.media.sfxLashit, 1, ATTN_NORM, 0);
		break;

	case TE_SHOTGUN:			// bullet hitting wall
		MSG_ReadPos (&net_message, pos);
		MSG_ReadDir (&net_message, dir);

		CL_ParticleEffect (pos, dir, 0, 0, 0, 20);
		CL_BulletExplosion (pos, dir);

		CL_SpawnDecal ( pos, dir, random()*360, 8, 1, 1, 1, 1, 8, 1, qfalse, cl.media.shaderBulletMark );
		break;

	case TE_SPLASH:			// bullet hitting water
		cnt = MSG_ReadByte (&net_message);
		MSG_ReadPos (&net_message, pos);
		MSG_ReadDir (&net_message, dir);
		r = MSG_ReadByte (&net_message);

		if (r > 6)
			VectorClear (color);
		else
			VectorCopy (splash_color[r], color);

		CL_ParticleEffect (pos, dir, color[0], color[1], color[2], cnt);

		if (r == SPLASH_SPARKS)
		{
			r = rand() & 3;
			if (r == 0)
				S_StartSound (pos, 0, 0, cl.media.sfxSpark5, 1, ATTN_STATIC, 0);
			else if (r == 1)
				S_StartSound (pos, 0, 0, cl.media.sfxSpark6, 1, ATTN_STATIC, 0);
			else
				S_StartSound (pos, 0, 0, cl.media.sfxSpark7, 1, ATTN_STATIC, 0);
		}
		break;

	case TE_LASER_SPARKS:
		cnt = MSG_ReadByte (&net_message);
		MSG_ReadPos (&net_message, pos);
		MSG_ReadDir (&net_message, dir);
		r = MSG_ReadLong (&net_message);

		CL_ParticleEffect2 (pos, dir,
			(( r	   ) & 0xFF) * (1.0 / 255.0),
			(( r >> 8  ) & 0xFF) * (1.0 / 255.0),
			(( r >> 16 ) & 0xFF) * (1.0 / 255.0),
			cnt);
		break;

	case TE_BLASTER:			// blaster hitting wall
		MSG_ReadPos (&net_message, pos);
		MSG_ReadDir (&net_message, dir);

		CL_BlasterParticles (pos, dir);

		angles[0] = RAD2DEG( acos(dir[2]) );
		if ( dir[0] ) {
			angles[1] = RAD2DEG( atan2(dir[1], dir[0]) );
		} else if ( dir[1] > 0 ) {
			angles[1] = 90;
		} else if ( dir[1] < 0 ) {
			angles[1] = 270;
		} else {
			angles[1] = 0;
		}
		angles[2] = 0;

		le = CL_AllocPoly ( LE_ALPHA_FADE, pos, angles, 4, 150, 1, 1, 0, cl.media.modExplode, NULL  );
		S_StartSound (pos,  0, 0, cl.media.sfxLashit, 1, ATTN_NORM, 0);

		CL_SpawnDecal ( pos, dir, random()*360, 16, 1, 0.8, 0, 1, 8, 2, qtrue, cl.media.shaderEnergyMark );
		break;

	case TE_RAILTRAIL:			// railgun effect
		MSG_ReadPos (&net_message, pos);
		MSG_ReadPos (&net_message, pos2);

		CL_RailTrail (pos, pos2);
		S_StartSound (pos2, 0, 0, cl.media.sfxRailg, 1, ATTN_NORM, 0);
		break;

	case TE_GRENADE_EXPLOSION:
	case TE_GRENADE_EXPLOSION_WATER:
		MSG_ReadPos (&net_message, pos);
		MSG_ReadDir (&net_message, dir);

		le = CL_AllocSprite ( LE_NO_FADE, pos, 64, 5, 350, 1, 1, 0, cl.media.shaderGrenadeExplosion );

		if (type == TE_GRENADE_EXPLOSION_WATER)
			S_StartSound (pos, 0, 0, cl.media.sfxWatrexp, 1, ATTN_NORM, 0);
		else
			S_StartSound (pos, 0, 0, cl.media.sfxGrenexp, 1, ATTN_NORM, 0);

		CL_SpawnDecal ( pos, dir, random()*360, 64, 1, 1, 1, 1, 10, 1, qfalse, cl.media.shaderExplosionMark );
		break;

	case TE_ROCKET_EXPLOSION:
	case TE_ROCKET_EXPLOSION_WATER:
		MSG_ReadPos (&net_message, pos);
		MSG_ReadDir (&net_message, dir);

		le = CL_AllocSprite ( LE_NO_FADE, pos, 64, 8, 350, 1, 0.75, 0, cl.media.shaderRocketExplosion  );

		if (type == TE_ROCKET_EXPLOSION_WATER)
			S_StartSound (pos, 0, 0, cl.media.sfxWatrexp, 1, ATTN_NORM, 0);
		else
			S_StartSound (pos, 0, 0, cl.media.sfxRockexp, 1, ATTN_NORM, 0);

		CL_SpawnDecal ( pos, dir, random()*360, 64, 1, 1, 1, 1, 10, 1, qfalse, cl.media.shaderExplosionMark );
		break;

	case TE_EXPLOSION1:
		MSG_ReadPos (&net_message, pos);
		le = CL_AllocSprite ( LE_NO_FADE, pos, 64, 8, 350, 1, 0.75, 0, cl.media.shaderRocketExplosion  );
		S_StartSound (pos, 0, 0, cl.media.sfxRockexp, 1, ATTN_NORM, 0);
		break;

	case TE_EXPLOSION2:
		MSG_ReadPos (&net_message, pos);
		le = CL_AllocSprite ( LE_NO_FADE, pos, 62, 5, 350, 1, 1, 0, cl.media.shaderGrenadeExplosion );
		S_StartSound (pos, 0, 0, cl.media.sfxGrenexp, 1, ATTN_NORM, 0);
		break;

	case TE_BFG_EXPLOSION:
		MSG_ReadPos (&net_message, pos);

		le = CL_AllocPoly ( LE_NO_FADE, pos, vec3_origin, 4, 350, 0, 1.0, 0, cl.media.modBfgExplo, NULL  );
		break;

	case TE_BFG_BIGEXPLOSION:
		MSG_ReadPos (&net_message, pos);

		CL_BFGExplosionParticles (pos);
		break;

	case TE_BFG_LASER:
		MSG_ReadPos (&net_message, pos);
		MSG_ReadPos (&net_message, pos2);
		le = CL_AllocLaser ( pos, pos2, 2, 2, COLOR_RGBA(0, 220, 0, 74), cl.media.shaderLaser );
		break;

	case TE_BUBBLETRAIL:
		MSG_ReadPos (&net_message, pos);
		MSG_ReadPos (&net_message, pos2);

		CL_BubbleTrail (pos, pos2, 32);
		break;

	case TE_PARASITE_ATTACK:
	case TE_MEDIC_CABLE_ATTACK:
		ent = MSG_ReadShort (&net_message);
		MSG_ReadPos (&net_message, pos);
		MSG_ReadPos (&net_message, pos2);

		CL_AddBeam (ent, pos, pos2, vec3_origin, cl.media.modParasiteSegment);
		break;

	case TE_BOSSTPORT:			// boss teleporting to station
		MSG_ReadPos (&net_message, pos);

		CL_BigTeleportParticles (pos);
		S_StartSound (pos, 0, 0, S_RegisterSound ("sound/misc/bigtele.wav"), 1, ATTN_NONE, 0);
		break;

	case TE_GRAPPLE_CABLE:
		ent = MSG_ReadShort (&net_message);
		MSG_ReadPos (&net_message, pos);
		MSG_ReadPos (&net_message, pos2);
		MSG_ReadPos (&net_message, dir);

		CL_AddBeam (ent, pos, pos2, dir, cl.media.modGrappleCable);
		break;

	case TE_PLAYER_TELEPORT_IN:
		MSG_ReadPos (&net_message, pos);

		S_StartSound (pos, 0, 0, cl.media.sfxTeleportIn, 1, ATTN_NORM, 0);
		CL_TeleportEffect (pos);
		break;

	case TE_PLAYER_TELEPORT_OUT:
		MSG_ReadPos (&net_message, pos);

		S_StartSound (pos, 0, 0, cl.media.sfxTeleportOut, 1, ATTN_NORM, 0);
		CL_TeleportEffect (pos);
		break;

	case TE_LIGHTNING:
		ent = MSG_ReadShort (&net_message);
		ent2 = MSG_ReadShort (&net_message);

		MSG_ReadPos (&net_message, pos);
		MSG_ReadPos (&net_message, pos2);

		CL_AddLightning ( ent, ent2, pos, pos2, cl.media.modLightning );
		S_StartSound (NULL, ent, CHAN_WEAPON, cl.media.sfxLightning, 1, ATTN_NORM, 0);
		break;

	default:
		Com_Error (ERR_DROP, "CL_ParseTEnt: bad type");
	}
}

/*
=================
CL_AddBeams
=================
*/
void CL_AddBeams (void)
{
	int			i, j;
	beam_t		*b;
	vec3_t		dist, org, angles, angles2;
	float		d;
	entity_t	ent;
	float		len, steps;
	float		model_length;

// update beams
	for (i=0, b=cl_beams ; i< MAX_BEAMS ; i++, b++)
	{
		if (!b->model || b->endtime < cl.time)
			continue;

		// if coming from the player, update the start position
		if (b->entity == cl.playernum+1)	// entity 0 is the world
		{
			VectorCopy (cl.refdef.vieworg, b->start);
			b->start[2] -= 22;	// adjust for view height
		}
		VectorAdd (b->start, b->offset, org);

	// calculate pitch and yaw
		VectorSubtract (b->end, org, dist);
		VecToAngles (dist, angles2);

	// add new entities for the beams
		d = VectorNormalize(dist);

		memset (&ent, 0, sizeof(ent));

		ent.scale = 1.0f;
		ent.color[0] = ent.color[1] = ent.color[2] = ent.color[3] = 255;

		if (b->model == cl.media.modLightning)
		{
			model_length = 35.0;
			d-= 20.0;  // correction so it doesn't end in middle of tesla
		}
		else
		{
			model_length = 30.0;
		}
		steps = ceil(d/model_length);
		len = (d-model_length)/(steps-1);

		// PMM - special case for lightning model .. if the real length is shorter than the model,
		// flip it around & draw it from the end to the start.  This prevents the model from going
		// through the tesla mine (instead it goes through the target)
		if ((b->model == cl.media.modLightning) && (d <= model_length))
		{
			VectorCopy (b->end, ent.origin);
			VectorCopy (b->end, ent.lightingOrigin);
			VectorCopy (b->end, ent.oldorigin);

			// offset to push beam outside of tesla model
			// (negative because dist is from end to start for this beam)
			ent.type = ET_MODEL;
			ent.model = b->model;
			ent.flags = RF_FULLBRIGHT|RF_NOSHADOW;
			angles[0] = angles2[0];
			angles[1] = angles2[1];
			angles[2] = rand()%360;
			AnglesToAxis (angles, ent.axis);
			V_AddEntity (&ent);
			return;
		}

		ent.type = ET_MODEL;
		ent.flags = RF_NOSHADOW;
		ent.model = b->model;

		while (d > 0)
		{
			VectorCopy (org, ent.origin);
			VectorCopy (org, ent.lightingOrigin);
			VectorCopy (org, ent.oldorigin);

			if (b->model == cl.media.modLightning)
			{
				angles[0] = -angles2[0];
				angles[1] = angles2[1] + 180.0;
				angles[2] = rand()%360;
			}
			else
			{
				angles[0] = angles2[0];
				angles[1] = angles2[1];
				angles[2] = rand()%360;
			}

			AnglesToAxis (angles, ent.axis);

			V_AddEntity (&ent);

			for (j=0 ; j<3 ; j++)
				org[j] += dist[j]*len;
			d -= model_length;
		}
	}
}


/*
=================
CL_AddLocalEntities
=================
*/
void CL_AddLocalEntities (void)
{
	int			f;
	lentity_t	*le, *next, *hnode;
	entity_t	*ent;
	float		frac, fade, time;
	float		backlerp;

	time = cls.frametime;
	backlerp = 1.0f - cl.lerpfrac;

	hnode = &cl_localents_headnode;
	for ( le = hnode->next; le != hnode; le = next )
	{
		next = le->next;

		frac = (cl.time - le->start) * 0.01f;
		f = floor (frac);
		f = max ( f, 0.0f );

		// it's time to DIE
		if ( f >= le->frames-1 ) {
			le->type = LE_FREE;
			CL_FreeLocalEntity ( le );
			continue;
		}

		if ( le->frames > 1 ) {
			fade = 1.0 - frac / (le->frames-1);
			fade = max ( fade, 0.0f );
		} else {
			fade = 1.0f;
		}

		ent = &le->ent;

		switch ( le->type )
		{
			case LE_NO_FADE:
				break;

			case LE_RGB_FADE:
				le->ent.color[0] = FloatToByte ( fade );
				le->ent.color[1] = FloatToByte ( fade );
				le->ent.color[2] = FloatToByte ( fade );
				break;

			case LE_ALPHA_FADE:
				le->ent.color[3] = FloatToByte ( fade );
				break;

			case LE_SCALE_ALPHA_FADE:
				le->ent.scale = 1.0 + 1/fade;
				le->ent.scale = min (le->ent.scale, 5.0f);

				le->ent.color[3] = FloatToByte ( fade );

				if (f >= le->frames-1) {
					le->type = LE_FREE;
				}
				break;

			case LE_LASER:
				break;
		}

		if ( le->type == LE_LASER ) {
			CL_AddLaser ( le->ent.origin, le->ent.oldorigin, le->ent.radius, le->ent.skinnum, le->ent.customShader );
			continue;
		}

		if ( le->light && fade ) {
			V_AddLight (ent->origin, le->light*fade,
				le->lightcolor[0], le->lightcolor[1], le->lightcolor[2]);
		}

		ent->frame = f + 1;
		ent->oldframe = f;
		ent->backlerp = backlerp;

		VectorCopy ( ent->origin, ent->oldorigin );
		VectorMA ( ent->origin, time, le->velocity, ent->origin );
		VectorCopy ( ent->origin, ent->lightingOrigin );
		VectorMA ( le->velocity, time, le->accel, le->velocity );

		V_AddEntity (ent);
	}
}

