/*
Copyright (C) 1997-2001 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// cl_view.c -- player rendering positioning

#include "client.h"

//=============

cvar_t		*cl_testparticles;
cvar_t		*cl_testentities;
cvar_t		*cl_testlights;
cvar_t		*cl_testblend;

cvar_t		*cl_stats;

cvar_t		*cl_thirdPerson;
cvar_t		*cl_thirdPersonAngle;
cvar_t		*cl_thirdPersonRange;

int			r_numdlights;
dlight_t	r_dlights[MAX_DLIGHTS];

int			r_numentities;
entity_t	r_entities[MAX_ENTITIES];

int			r_numparticles;
particle_t	r_particles[MAX_PARTICLES];

int			r_numpolys;
poly_t		r_polys[MAX_POLYS];

char		cl_weaponmodels[MAX_CLIENTWEAPONMODELS][MAX_QPATH];
int			num_cl_weaponmodels;

/*
====================
V_ClearScene

Specifies the model that will be used as the world
====================
*/
void V_ClearScene (void)
{
	r_numdlights = 0;
	r_numentities = 0;
	r_numparticles = 0;
	r_numpolys = 0;
}


/*
=====================
V_AddEntity

=====================
*/
void V_AddEntity (entity_t *ent)
{
	if (r_numentities >= MAX_ENTITIES)
		return;
	r_entities[r_numentities++] = *ent;
}


/*
=====================
V_AddParticle

=====================
*/
void V_AddParticle (vec3_t org, float rgb[3], float alpha, float scale)
{
	particle_t	*p;

	if (r_numparticles >= MAX_PARTICLES)
		return;
	if (scale <= 0)
		return;
	p = &r_particles[r_numparticles++];
	VectorCopy (org, p->origin);
	p->color[0] = FloatToByte (bound(0, rgb[0], 1.0f));
	p->color[1] = FloatToByte (bound(0, rgb[1], 1.0f));
	p->color[2] = FloatToByte (bound(0, rgb[2], 1.0f));
	p->color[3] = FloatToByte (bound(0, alpha, 1.0f));
	p->scale = scale;
}

/*
=====================
V_AddLight

=====================
*/
void V_AddLight (vec3_t org, float intensity, float r, float g, float b)
{
	dlight_t	*dl;

	if (r_numdlights >= MAX_DLIGHTS)
		return;
	dl = &r_dlights[r_numdlights++];
	VectorCopy (org, dl->origin);
	dl->intensity = intensity;
	dl->color[0] = r;
	dl->color[1] = g;
	dl->color[2] = b;
}

/*
=====================
V_AddPoly

=====================
*/
void V_AddPoly (poly_t *poly)
{
	if (r_numpolys >= MAX_POLYS)
		return;
	r_polys[r_numpolys++] = *poly;
}

/*
================
V_TestParticles

If cl_testparticles is set, create 4096 particles in the view
================
*/
void V_TestParticles (void)
{
	particle_t	*p;
	int			i, j;
	float		d, r, u;

	r_numparticles = MAX_PARTICLES;
	for (i=0 ; i<r_numparticles ; i++)
	{
		d = i*0.25;
		r = 4*((i&7)-3.5);
		u = 4*(((i>>3)&7)-3.5);
		p = &r_particles[i];

		for (j=0 ; j<3 ; j++)
			p->origin[j] = cl.refdef.vieworg[j] + cl.v_forward[j]*d +
			cl.v_right[j]*r + cl.v_up[j]*u;

		p->scale = 1;
		Vector4Set ( p->color, 127, 127, 127, 255 );
	}
}

/*
================
V_TestEntities

If cl_testentities is set, create 32 player models
================
*/
void V_TestEntities (void)
{
	int			i, j;
	float		f, r;
	entity_t	*ent;

	r_numentities = 32;
	memset (r_entities, 0, sizeof(r_entities));

	for (i=0 ; i<r_numentities ; i++)
	{
		ent = &r_entities[i];

		r = 64 * ( (i%4) - 1.5 );
		f = 64 * (i/4) + 128;

		for (j=0 ; j<3 ; j++)
			ent->origin[j] = cl.refdef.vieworg[j] + cl.v_forward[j]*f +
			cl.v_right[j]*r;

		//splitmodels (skin files)
		ent->model = cl.baseclientinfo.model[UPPER];
		ent->customSkin = cl.baseclientinfo.customSkin[UPPER];
	}
}

/*
================
V_TestLights

If cl_testlights is set, create 32 lights models
================
*/
void V_TestLights (void)
{
	int			i, j;
	float		f, r;
	dlight_t	*dl;

	r_numdlights = 32;
	memset (r_dlights, 0, sizeof(r_dlights));

	for (i=0 ; i<r_numdlights ; i++)
	{
		dl = &r_dlights[i];

		r = 64 * ( (i%4) - 1.5 );
		f = 64 * (i/4) + 128;

		for (j=0 ; j<3 ; j++)
			dl->origin[j] = cl.refdef.vieworg[j] + cl.v_forward[j]*f +
			cl.v_right[j]*r;
		dl->color[0] = ((i%6)+1) & 1;
		dl->color[1] = (((i%6)+1) & 2)>>1;
		dl->color[2] = (((i%6)+1) & 4)>>2;
		dl->intensity = 200;
	}
}

//===================================================================

/*
=================
CL_PrepRefresh

Call before entering a new level, or after changing dlls
=================
*/
void CL_PrepRefresh (void)
{
	char		mapname[32];
	int			i;
	char		*name;
	char		default_clientinfo[MAX_QPATH];//splitmodels

	if (!cl.configstrings[CS_MODELS+1][0])
		return;		// no map loaded

	CL_ClearEffects ();
	CL_ClearLocalEntities ();
	CL_ClearDecals ();
	//splitmodels (skin files)
	R_ClearCustomSkins();

	// let the render dll load the map
	strcpy (mapname, cl.configstrings[CS_MODELS+1] + 5);	// skip "maps/"
	mapname[strlen(mapname)-4] = 0;		// cut off ".bsp"

	// register models, pics, and skins
	R_BeginRegistration (mapname);

	CL_LoadingString ("models");
	CL_RegisterMediaModels ();

	num_cl_weaponmodels = 1;
	strcpy(cl_weaponmodels[0], "weapon.md3");//splitmodels: default replacement weapon.

	for (i=1 ; i<MAX_MODELS && cl.configstrings[CS_MODELS+i][0] ; i++)
	{
		name = cl.configstrings[CS_MODELS+i];
		if (name[0] != '*')
			CL_LoadingFilename (name);
		Sys_SendKeyEvents ();	// pump message loop
		if (name[0] == '#')
		{
			// special player weapon model
			if (num_cl_weaponmodels < MAX_CLIENTWEAPONMODELS)
			{
				Q_strncpyz (cl_weaponmodels[num_cl_weaponmodels], cl.configstrings[CS_MODELS+i]+1,
					sizeof(cl_weaponmodels[num_cl_weaponmodels]));
				num_cl_weaponmodels++;
			}
		}
		else
		{
			cl.model_draw[i] = R_RegisterModel (cl.configstrings[CS_MODELS+i]);
			if (name[0] == '*')
				cl.model_clip[i] = CM_InlineModel (cl.configstrings[CS_MODELS+i]);
			else
				cl.model_clip[i] = NULL;
		}
	}

	CL_LoadingString ("images");
	CL_RegisterMediaShaders ();

	for (i=1 ; i<MAX_IMAGES && cl.configstrings[CS_IMAGES+i][0] ; i++)
	{
		CL_LoadingFilename (cl.configstrings[CS_IMAGES+i]);
		cl.image_precache[i] = R_RegisterPic (cl.configstrings[CS_IMAGES+i]);
		Sys_SendKeyEvents ();	// pump message loop
	}

	// precache status bar pics
	SCR_TouchPics ();

	CL_LoadingFilename ("");

	for (i=0 ; i<MAX_CLIENTS ; i++)
	{
		if (!cl.configstrings[CS_PLAYERSKINS+i][0])
			continue;

		CL_LoadingString (va("client %i", i));
		Sys_SendKeyEvents ();	// pump message loop
		CL_ParseClientinfo (i);
	}

	//splitmodels - DEFAULT_PLAYERMODEL
	Com_sprintf (default_clientinfo, sizeof(default_clientinfo), "unnamed\\%s/%s", DEFAULT_PLAYERMODEL, DEFAULT_PLAYERSKIN);
	CL_LoadClientinfo (&cl.baseclientinfo, default_clientinfo);

	// the renderer can now free unneeded stuff
	R_EndRegistration ();

	// clear any lines of console text
	Con_ClearNotify ();

	SCR_UpdateScreen ();
	cl.refresh_prepped = qtrue;
	cl.force_refdef = qtrue;	// make sure we have a valid refdef
}

//============================================================================
/*
================
Camera_Update
================
*/
void Camera_Update (void)
{
	float	dist, f, r;
	vec3_t	dest, stop;
	vec3_t	chase_dest;
	trace_t	trace;
	static vec3_t mins = { -4, -4, -4 };
	static vec3_t maxs = { 4, 4, 4 };

	// calc exact destination
	VectorCopy ( cl.refdef.vieworg, chase_dest );
	r = DEG2RAD( cl_thirdPersonAngle->value );
	f = -cos( r );
	r = -sin( r );
	VectorMA( chase_dest, cl_thirdPersonRange->value * f, cl.v_forward, chase_dest );
	VectorMA( chase_dest, cl_thirdPersonRange->value * r, cl.v_right, chase_dest );
	chase_dest[2] += 8;

	// find the spot the player is looking at
	VectorMA ( cl.refdef.vieworg, 512, cl.v_forward, dest );
	trace = CL_Trace ( cl.refdef.vieworg, mins, maxs, dest, IGNORE_PLAYER, MASK_SOLID );

	// calculate pitch to look at the same spot from camera
	VectorSubtract ( trace.endpos, cl.refdef.vieworg, stop );
	dist = sqrt ( stop[0] * stop[0] + stop[1] * stop[1] );
	if (dist < 1)
		dist = 1;
	cl.refdef.viewangles[PITCH] = RAD2DEG( -atan2(stop[2], dist) );
	cl.refdef.viewangles[YAW] -= cl_thirdPersonAngle->value;
	AngleVectors ( cl.refdef.viewangles, cl.v_forward, cl.v_right, cl.v_up );

	// move towards destination
	trace = CL_Trace ( cl.refdef.vieworg, mins, maxs, chase_dest, IGNORE_PLAYER, MASK_SOLID );

	if ( trace.fraction != 1.0 ) {
		VectorCopy ( trace.endpos, stop );
		stop[2] += ( 1.0 - trace.fraction ) * 32;
		trace = CL_Trace ( cl.refdef.vieworg, mins, maxs, stop, IGNORE_PLAYER, MASK_SOLID );
		VectorCopy ( trace.endpos, chase_dest );
	}

	VectorCopy ( chase_dest, cl.refdef.vieworg );
}

//============================================================================

/*
==================
V_RenderFlags

==================
*/
int V_RenderFlags (void)
{
	int rdflags, contents;

	rdflags = 0;

	contents = CL_PointContents ( cl.refdef.vieworg );
	if ( contents & (CONTENTS_LAVA|CONTENTS_SLIME|CONTENTS_WATER) )
		rdflags |= RDF_UNDERWATER;
	else
		rdflags &= ~RDF_UNDERWATER;

	return rdflags;
}

/*
==================
V_RenderView

==================
*/
#define	WAVE_AMPLITUDE	1
#define	WAVE_FREQUENCY	0.4

void V_RenderView( float stereo_separation )
{
	if (cls.state != ca_active)
		return;

	if (!cl.refresh_prepped)
		return;			// still loading

	if (cl_timedemo->value)
	{
		if (!cl.timedemo_start)
			cl.timedemo_start = Sys_Milliseconds ();
		cl.timedemo_frames++;
	}

	// an invalid frame will just use the exact previous refdef
	// we can't use the old frame if the video mode has changed, though...
	if ( cl.frame.valid && (cl.force_refdef || !cl_paused->value) )
	{
		cl.force_refdef = qfalse;

		V_ClearScene ();

		// build a refresh entity list and calc cl.sim*
		// this also calls CL_CalcViewValues which loads
		// v_forward, etc.
		CL_AddEntities ();

		if (cl_testparticles->value)
			V_TestParticles ();
		if (cl_testentities->value)
			V_TestEntities ();
		if (cl_testlights->value)
			V_TestLights ();
		if (cl_testblend->value)
		{
			cl.refdef.blend[0] = 1;
			cl.refdef.blend[1] = 0.5;
			cl.refdef.blend[2] = 0.25;
			cl.refdef.blend[3] = 0.5;
		}

		// offset vieworg appropriately if we're doing stereo separation
		if ( stereo_separation != 0 ) {
			VectorMA( cl.refdef.vieworg, stereo_separation, cl.v_right, cl.refdef.vieworg );
		}

		if ( cl.thirdperson ) {
			Camera_Update ();
		}

		// never let it sit exactly on a node line, because a water plane can
		// dissapear when viewed with the eye exactly on it.
		// the server protocol only specifies to 1/8 pixel, so add 1/16 in each axis
		cl.refdef.vieworg[0] += 1.0/16;
		cl.refdef.vieworg[1] += 1.0/16;
		cl.refdef.vieworg[2] += 1.0/16;

		cl.refdef.x = scr_vrect.x;
		cl.refdef.y = scr_vrect.y;
		cl.refdef.width = scr_vrect.width;
		cl.refdef.height = scr_vrect.height;
		cl.refdef.fov_y = CalcFov (cl.refdef.fov_x, cl.refdef.width, cl.refdef.height);

		cl.refdef.time = cl.time*0.001;

		cl.refdef.areabits = cl.frame.areabits;

		if (!cl_add_entities->value)
			r_numentities = 0;
		if (!cl_add_particles->value)
			r_numparticles = 0;
		if (!cl_add_lights->value)
			r_numdlights = 0;
		if (!cl_add_polys->value)
			r_numpolys = 0;
		if (!cl_add_blend->value)
		{
			cl.refdef.blend[0] = cl.refdef.blend[1] = cl.refdef.blend[2] = cl.refdef.blend[3] = 0;
		}

		cl.refdef.num_entities = r_numentities;
		cl.refdef.entities = r_entities;
		cl.refdef.num_particles = r_numparticles;
		cl.refdef.particles = r_particles;
		cl.refdef.num_dlights = r_numdlights;
		cl.refdef.dlights = r_dlights;
		cl.refdef.num_polys = r_numpolys;
		cl.refdef.polys = r_polys;

		cl.refdef.rdflags = V_RenderFlags ();

		// warp if underwater
		if ( cl.refdef.rdflags & RDF_UNDERWATER ) {
			float phase = cl.refdef.time * WAVE_FREQUENCY * M_TWOPI;
			float v = WAVE_AMPLITUDE * sin( phase );
			cl.refdef.fov_x += v;
			cl.refdef.fov_y -= v;
		}
	}

	R_RenderFrame (&cl.refdef);

	if (cl_stats->value)
		Com_Printf ("ent:%i  lt:%i  part:%i pls:%i\n", r_numentities, r_numdlights, r_numparticles, r_numpolys);
	if ( log_stats->value && ( log_stats_file != 0 ) )
		fprintf( log_stats_file, "%i,%i,%i,%i,",r_numentities, r_numdlights, r_numparticles, r_numpolys);
}


/*
=============
V_Viewpos_f
=============
*/
void V_Viewpos_f (void)
{
	Com_Printf ("(%i %i %i) : %i\n", (int)cl.refdef.vieworg[0],
		(int)cl.refdef.vieworg[1], (int)cl.refdef.vieworg[2],
		(int)cl.refdef.viewangles[YAW]);
}

/*
=============
V_Init
=============
*/
void V_Init (void)
{
	Cmd_AddCommand ("viewpos", V_Viewpos_f);

	cl_testblend = Cvar_Get ("cl_testblend", "0", CVAR_CHEAT);
	cl_testparticles = Cvar_Get ("cl_testparticles", "0", CVAR_CHEAT);
	cl_testentities = Cvar_Get ("cl_testentities", "0", CVAR_CHEAT);
	cl_testlights = Cvar_Get ("cl_testlights", "0", CVAR_CHEAT);

	cl_stats = Cvar_Get ("cl_stats", "0", 0);

	cl_thirdPerson = Cvar_Get ( "cl_thirdperson", "0", CVAR_CHEAT );

	cl_thirdPersonAngle = Cvar_Get ( "cl_thirdpersonangle", "0", 0 );
	cl_thirdPersonRange = Cvar_Get ( "cl_thirdpersonrange", "70", 0 );
}
