/*

AI FUNCTION LISTING

FUSION_AI_CheckShot					 Check if shot is blocked or too far
FUSION_AI_Think						 Main think function for bot
FUSION_AI_Determine_LR_Goal			 Evaluate best Long Range goal
FUSION_AI_Determine_SR_Goal			 Evaluate best Short Range goal
FUSION_AI_LocateEnemy				 Locate the enemy
FUSION_AI_ChooseWeapon				 Choose best weapon
FUSION_AI_ChoosePreferredWeapon		 Choose preferred weapon from config

*/
/*
        .__
 _____  |__|
 \__  \ |  |
  / __ \|  |
 (____  /__|
      \/

	fusionbot_ai.c
	AI routines for the FUSION bot.
================================================================
Should you decide to release a modified version of FUSION, you MUST
include the following text in the documentation and source of your
modification.

---===[BEGIN TEXT]===---
================================================================
___________           .__                __________        __
\_   _____/_ __  _____|__| ____   ____   \______   \ _____/  |_
 |    __)|  |  \/  ___/  |/  _ \ /    \   |    |  _//  _ \   __\
 |     \ |  |  /\___ \|  (  <_> )   |  \  |    |   (  <_> )  |
 \___  / |____//____  >__|\____/|___|  /  |______  /\____/|__|
     \/             \/               \/          \/  Version 2.0
================================================================
Fusion v2.0
Copyright (c) 1998-2003 Trenton Shaffer.
All rights reserved.

ACEBOT Copyright(c) 1998, Steve Yeager.
EraserBot Copyright(c) 1998, Ryan Feltrin.
QFusion is Copyright (C) 2002-2003 Victor Luchits.
All other files are Copyright(C) 1997-2001 Id Software, Inc.
================================================================
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
================================================================
---===[END TEXT]===---
*/
#include "..\game\g_local.h"
#include "..\game\m_player.h"

#include "fusionbot.h"

//===========================================================================
// Routine		: FUSION_AI_CheckShot
// Description	: Checks if shot is blocked or if too far to shoot
qboolean FUSION_AI_CheckShot(edict_t *fusionbot)
{
	float	range;		//range
	vec3_t	vector;		//vector
	trace_t tr;			//trace

	//Get enemy range
	//-------------------------------------------------------
	VectorSubtract (fusionbot->s.origin, fusionbot->enemy->s.origin, vector);
	range = VectorLength(vector);
	//-------------------------------------------------------

	//Trace enemy
	//-------------------------------------------------------
	tr = gi.trace (fusionbot->s.origin, tv(-8,-8,-8), tv(8,8,8), fusionbot->enemy->s.origin, fusionbot, MASK_SOLID|MASK_OPAQUE);
	//-------------------------------------------------------

	//See if range too far, if so don't shoot
	//-------------------------------------------------------
	if (range > 2000)
		return qfalse;
	//-------------------------------------------------------

	// Blocked, don't shoot
	//-------------------------------------------------------
	if(tr.fraction < 0.9)
	//if (tr.fraction != 1.0)
		return qfalse;
	//-------------------------------------------------------

	return qtrue; //qbism typical: true --> qtrue, false --> qfalse
}

//===========================================================================
// Routine		: FUSION_AI_Think
// Description	: Main think function for bot
void FUSION_AI_Think (edict_t *fusionbot)
{
	usercmd_t	ucmd;

	// Set up client movement
	//-------------------------------------------------------
	VectorCopy(fusionbot->client->ps.viewangles,fusionbot->s.angles);
	VectorSet (fusionbot->client->ps.pmove.delta_angles, 0, 0, 0);
	memset (&ucmd, 0, sizeof (ucmd));
	fusionbot->enemy = NULL;
	fusionbot->movetarget = NULL;
	//-------------------------------------------------------

	// Force respawn
	//-------------------------------------------------------
	if (fusionbot->deadflag)
	{
		fusionbot->client->buttons = 0;
		ucmd.buttons = BUTTON_ATTACK;
	}
	//-------------------------------------------------------

	// Pick a new long range goal
	//-------------------------------------------------------
	if(fusionbot->state == BOT_STATE_WANDER && fusionbot->wander_timeout < level.time)
		FUSION_AI_Determine_LR_Goal(fusionbot);

	if(fusionbot->state == BOT_STATE_POSITION)
	{
		if(level.time >= fusionbot->teamPauseTime)
			fusionbot->state = BOT_STATE_WANDER;

		if (fusionbot->goal_node == INVALID || fusionbot->wander_timeout < level.time)
			FUSION_AI_Determine_LR_Goal(fusionbot);
			//fusionbot->state = BOT_STATE_WANDER;
	}
	//-------------------------------------------------------

	// Kill the bot if completely stuck somewhere
	//-------------------------------------------------------
	if(VectorLength(fusionbot->velocity) > 37)
		fusionbot->suicide_timeout = level.time + 10.0;

	if(sv_godall->value) //qbism
	{
		if(fusionbot->suicide_timeout < level.time)
		{
			PutClientInServer(fusionbot);
		}
	}
	else if(fusionbot->suicide_timeout < level.time)
	{
		fusionbot->health = 0;
		player_die (fusionbot, fusionbot, fusionbot, 100000, vec3_origin);
	}

	//-------------------------------------------------------

	// Find any short range goal
	//-------------------------------------------------------
	FUSION_AI_Determine_SR_Goal(fusionbot);
	//-------------------------------------------------------

	// Look for enemies
	//-------------------------------------------------------
	if(FUSION_AI_LocateEnemy(fusionbot))
	{
		// Choose best weapon
		FUSION_AI_ChooseWeapon(fusionbot);
		// Attack
		FUSION_MOVEMENT_Attack (fusionbot, &ucmd);
	}
	else
	{
		// Wander
		if(fusionbot->state == BOT_STATE_WANDER)
			FUSION_MOVEMENT_Wander(fusionbot,&ucmd);
		// Move
		else if( (fusionbot->state == BOT_STATE_MOVE) || (fusionbot->state == BOT_STATE_POSITION) )
		{
			FUSION_MOVEMENT_Move(fusionbot,&ucmd);
		}
	}
	//-------------------------------------------------------

	// set approximate ping and show values
	//-------------------------------------------------------
	ucmd.msec = 75 + floor (random () * 25) + 1;
	// show random ping values in scoreboard
	fusionbot->client->ping = ucmd.msec;
	//-------------------------------------------------------

	// set bot's view angle
	//-------------------------------------------------------
	ucmd.angles[PITCH] = ANGLE2SHORT(fusionbot->s.angles[PITCH]);
	ucmd.angles[YAW] = ANGLE2SHORT(fusionbot->s.angles[YAW]);
	ucmd.angles[ROLL] = ANGLE2SHORT(fusionbot->s.angles[ROLL]);
	//-------------------------------------------------------

	// send command through id's code
	//-------------------------------------------------------
	ClientThink (fusionbot, &ucmd);
	fusionbot->nextthink = level.time + FRAMETIME;
	//-------------------------------------------------------
}

//===========================================================================
// Routine		: FUSION_AI_Determine_LR_Goal
// Description	: Evaluate the best long range goal and send the bot on
//				: its way. This is a good time waster, so function needs
//				: to be used sparingly.
//				: DO NOT CALL FOR EVERY THINK CYCLE
void FUSION_AI_Determine_LR_Goal(edict_t *fusionbot)
{
	int		i;
	int		node;
	float	weight,best_weight=0.0;
	int		current_node,goal_node;
	edict_t *goal_ent;
	float	cost;

	// look for a target
	//-------------------------------------------------------
	current_node = FUSION_NAVIGATION_FindClosestReachableNode(fusionbot,NODE_DENSITY,NODE_ALL);
	fusionbot->current_node = current_node;

	// Setup Wander
	if(current_node == -1)
	{
		fusionbot->state = BOT_STATE_WANDER;
		fusionbot->wander_timeout = level.time + 1.0;
		fusionbot->goal_node = -1;
		return;
	}
	//-------------------------------------------------------

	// Teamplay POSITION state
	//-------------------------------------------------------
	if( fusionbot->state == BOT_STATE_POSITION )
	{
		int counter = 0;
		cost = INVALID;
		fusionbot->goal_node = INVALID;

		// Pick a random node to go to
		//-------------------------------------------------------
		while( cost == INVALID && counter < 10)
		{
			counter++;
			// Any current node will do
			i = (int)(random() * number_of_nodes -1);
			cost = FUSION_NAVIGATION_FindCost(current_node, i);

			// Ignore INVALID and very short hops
			if(cost == INVALID || cost < 2)
			{
				cost = INVALID;
				i = INVALID;
				continue;
			}
		}
		//-------------------------------------------------------

		// We have a target node - just go there!
		//-------------------------------------------------------
		if( i != INVALID )
		{
			fusionbot->state = BOT_STATE_MOVE;
			// Reset try count for this goal
			fusionbot->tries = 0;
			FUSION_NAVIGATION_SetGoal(fusionbot,i);
			fusionbot->wander_timeout = level.time + 1.0;
			return;
		}
		//-------------------------------------------------------
	}
	//-------------------------------------------------------

	// Items
	//-------------------------------------------------------
	for(i=0;i<num_items;i++)
	{
		// Ignore items that are not there (solid)
		if(item_table[i].ent == NULL || item_table[i].ent->solid == SOLID_NOT)
			continue;

		cost = FUSION_NAVIGATION_FindCost(current_node,item_table[i].node);

		// Ignore INVALID and very short hops
		if(cost == INVALID || cost < 2)
			continue;

		weight = FUSION_ITEMS_ItemNeed(fusionbot, item_table[i].item);

		// If BOT is on TEAM1 and has the flag for TEAM2, BOT should return it to
		// the TEAM1 base
		//-------------------------------------------------------
		// This really needs its own function, it has basically
		// been stripped down. I have worked out a better method
		// of doing this, but am too lazy to release the code.
		//
		// Try your own variations and see if you can get something
		// good going. I will probably release my code for better
		// CTF Play as soon as I perfect it.
		//-------------------------------------------------------
		if (ctf->value)
		{
			if (item_table[i].item == ITEMLIST_FLAG1 &&
				fusionbot->client->pers.inventory[ITEMLIST_FLAG2])
			{
				weight = 10.0;
			}
			else if (item_table[i].item == ITEMLIST_FLAG2 &&
				fusionbot->client->pers.inventory[ITEMLIST_FLAG1])
			{
				weight = 10.0;
			}
		}
		//-------------------------------------------------------

		// Random Variations
		weight *= random();
		// Check against COST of getting there
		weight /= cost;

		if(weight > best_weight)
		{
			best_weight = weight;
			goal_node = item_table[i].node;
			goal_ent = item_table[i].ent;
		}
	}
	//-------------------------------------------------------

	// Players (NOTE: needs its own function)
	// Find a Player and set as the goal
	//-------------------------------------------------------
	for(i=0;i<num_players;i++)
	{
		if(players[i] == fusionbot)
			continue;

		// If it's dark and he's not already our enemy, ignore him
		if( fusionbot->enemy && players[i] != fusionbot->enemy)
		{
			if( players[i]->light_level < 30)
				continue;
		}


		node = FUSION_NAVIGATION_FindClosestReachableNode(players[i],NODE_DENSITY,NODE_ALL);

		cost = FUSION_NAVIGATION_FindCost(current_node, node);

		// Ignore INVALID and very short hops
		if(cost == INVALID || cost < 3)
			continue;

		// Check if Player is carrying the flag
		if(ctf->value && (players[i]->client->pers.inventory[ITEMLIST_FLAG2] || players[i]->client->pers.inventory[ITEMLIST_FLAG1]))
		  weight = 2.0;
		else
		  weight = 0.2;

		// Random Variations
		weight *= random();
		// Check cost of getting there
		weight /= cost;


		if(weight > best_weight)
		{
			best_weight = weight;
			goal_node = node;
			goal_ent = players[i];
		}
	}
	//-------------------------------------------------------
	// If didn't find a goal, go wandering....
	//-------------------------------------------------------
	if(best_weight == 0.0 || goal_node == INVALID)
	{
		fusionbot->goal_node = INVALID;
		fusionbot->state = BOT_STATE_WANDER;
		fusionbot->wander_timeout = level.time + 1.0;
		if(debug_mode)
			gi.dprintf("%s did not find a LR goal, wandering.\n",fusionbot->client->pers.netname);
		return; // no path?
	}
	//-------------------------------------------------------

	// OK, everything valid, let's start moving to our goal.
	//-------------------------------------------------------
	fusionbot->state = BOT_STATE_MOVE;
	// Reset try counter for this goal
	fusionbot->tries = 0;

	if(goal_ent != NULL && debug_mode)
		gi.dprintf("%s selected a %s at node %d for LR goal.\n",fusionbot->client->pers.netname, goal_ent->classname, goal_node);

	FUSION_NAVIGATION_SetGoal(fusionbot,goal_node);


	//-------------------------------------------------------

}

//===========================================================================
// Routine		: FUSION_AI_Determine_SR_Goal
// Description	: Pick best goal based on importance and range.
//				: This function overrides the long range goal selection
//				: for items that are very close to the bot and are reachable
void FUSION_AI_Determine_SR_Goal(edict_t *fusionbot)
{
	edict_t *target;
	float	weight,best_weight=0.0;
	edict_t *best;
	int		index;

	// Hunt!
	// look for a target
	//-------------------------------------------------------
	target = findradius(NULL, fusionbot->s.origin, 200);

	while(target)
	{
		if(target->classname == NULL)
			return;

		// Missile avoidance code
		// Set our movetarget to be the rocket or grenade fired at us.
		if(strcmp(target->classname,"rocket")==0 || strcmp(target->classname,"grenade")==0)
		{
			if(debug_mode)
				gi.dprintf("ROCKET ALERT!\n");

			fusionbot->movetarget = target;
			return;
		}

		if (FUSION_ITEMS_IsReachable(fusionbot,target->s.origin))
		{
			if (infront(fusionbot, target))
			{
				index = FUSION_ITEMS_ClassnameToIndex(target->classname);
				weight = FUSION_ITEMS_ItemNeed(fusionbot, index);

				if(weight > best_weight)
				{
					best_weight = weight;
					best = target;
				}
			}
		}

		// next target
		target = findradius(target, fusionbot->s.origin, 200);
	}
	//-------------------------------------------------------

	//-------------------------------------------------------
	if(best_weight)
	{
		fusionbot->movetarget = best;

		if(debug_mode && fusionbot->goalentity != fusionbot->movetarget)
			gi.dprintf("%s selected a %s for SR goal.\n",fusionbot->client->pers.netname, fusionbot->movetarget->classname);

		fusionbot->goalentity = best;

	}
	//-------------------------------------------------------

}

//===========================================================================
// Routine		: FUSION_AI_LocateEnemy
// Description	: Scans for enemy (very simple, just picks any visible enemy)
qboolean FUSION_AI_LocateEnemy(edict_t *fusionbot)
{
	int i;

	edict_t		*bestenemy = NULL;
	float		bestweight = 99999;
	float		weight;
	vec3_t		dist;


	//-------------------------------------------------------
	// Find Enemy
	//-------------------------------------------------------
	for(i=0;i<=num_players;i++)
	{
		if(ctf->value && fusionbot->client->pers.inventory[ITEMLIST_FLAG2])
			continue;
		if(ctf->value && fusionbot->client->pers.inventory[ITEMLIST_FLAG1])
			continue;
		if(players[i] == NULL || players[i] == fusionbot || players[i]->solid == SOLID_NOT)
			continue;

		// If dark, and current target is not enemy, IGNORE
		//-------------------------------------------------------
		if( fusionbot->enemy && players[i] != fusionbot->enemy)
		{
			if( players[i]->light_level < 30)
				continue;
		}
		//-------------------------------------------------------

		if(ctf->value && fusionbot->client->resp.ctf_team == players[i]->client->resp.ctf_team)
			continue;

		if(!players[i]->deadflag && visible(fusionbot, players[i]) &&
			gi.inPVS (fusionbot->s.origin, players[i]->s.origin))
		{
			// Assess enemy
			//-------------------------------------------------------
			VectorSubtract(fusionbot->s.origin, players[i]->s.origin, dist);
			weight = VectorLength( dist );
			//-------------------------------------------------------

			// Is enemy visible, or is it too close to ignore
			//-------------------------------------------------------
			if( infront( fusionbot, players[i] ) ||
				(weight < 300 ) )
			{
				// Check if best target, or better than current target
				//-------------------------------------------------------
				if (weight < bestweight)
				{
					bestweight = weight;
					bestenemy = players[i];
				}
				//-------------------------------------------------------
			}
			//-------------------------------------------------------
		}
	}

	// If best enemy, set up
	//-------------------------------------------------------
	if(bestenemy)
	{
		fusionbot->enemy = bestenemy;
		return qtrue;
	}
	//-------------------------------------------------------

	// Check if shot from behind or out of view
	//-------------------------------------------------------
	if( fusionbot->client->attacker )
	{
		// Check if recent
		//-------------------------------------------------------
		if( fusionbot->client->push_timeout > 0)
		{
			if(!fusionbot->client->attacker->deadflag && visible(fusionbot, fusionbot->client->attacker) &&
				gi.inPVS(fusionbot->s.origin, fusionbot->client->attacker->s.origin)	)
			{
				fusionbot->enemy = fusionbot->client->attacker;
				return qtrue;
			}
		}
		//-------------------------------------------------------
	}
	//-------------------------------------------------------

	// NO enemy
	return qfalse;

}

//===========================================================================
// Routine		: FUSION_AI_ChooseWeapon
// Description	: Chooses the best weapon for bot
void FUSION_AI_ChooseWeapon(edict_t *fusionbot)
{
	float	range;
	vec3_t	v;

	// If no enemy, bail
	//-------------------------------------------------------
	if(!fusionbot->enemy)
		return;
	//-------------------------------------------------------

	// Always favor rocket launcher
	if(FUSION_AI_CheckShot(fusionbot) && FUSION_ITEMS_ChangeWeapon(fusionbot, FindItem("rocket launcher")))
		return;

	// Base weapon selection on distance
	//-------------------------------------------------------
	VectorSubtract (fusionbot->s.origin, fusionbot->enemy->s.origin, v);
	range = VectorLength(v);
	//-------------------------------------------------------

	// Long Range
	//-------------------------------------------------------
	if(range > 1000)
	{
		// Railgun
		//-------------------------------------------------------
		if(FUSION_ITEMS_ChangeWeapon(fusionbot,FindItem("railgun")))
			return;
		//-------------------------------------------------------
	}
	//-------------------------------------------------------

	// Medium Range
	//-------------------------------------------------------
	if(range > 750)
	{
		// choose BFG if enough ammo
		//-------------------------------------------------------
		if(fusionbot->client->pers.inventory[ITEMLIST_CELLS] > 50)
			if(FUSION_AI_CheckShot(fusionbot) && FUSION_ITEMS_ChangeWeapon(fusionbot, FindItem("bfg10k")))
				return;
		//-------------------------------------------------------
	}

	// Only use Grenade Launcher in certain ranges and only on targets at or below our level
	//-------------------------------------------------------
	if(range > 100 && range < 500 && fusionbot->enemy->s.origin[2] - 20 < fusionbot->s.origin[2])
	{
		if(FUSION_ITEMS_ChangeWeapon(fusionbot,FindItem("grenade launcher")))
			return;
	}
	//-------------------------------------------------------

	if(range < 500 && range > 300)
	{
		// Use Plasma Gun
		//-------------------------------------------------------
		if(FUSION_AI_CheckShot(fusionbot) && FUSION_ITEMS_ChangeWeapon(fusionbot,FindItem("plasma gun")))
			return;
	}

	// Short Range
	if(range < 300)
	{
		// Only use ChainGun when ammo > 50
		//-------------------------------------------------------
		if(fusionbot->client->pers.inventory[ITEMLIST_BULLETS] >= 50)
		{
			if(FUSION_AI_CheckShot(fusionbot) && FUSION_ITEMS_ChangeWeapon(fusionbot,FindItem("chaingun")))
				return;
		}

		// Use Machine Gun
		//-------------------------------------------------------
		if(fusionbot->client->pers.inventory[ITEMLIST_BULLETS] >= 50)
		{
			if(FUSION_AI_CheckShot(fusionbot) && FUSION_ITEMS_ChangeWeapon(fusionbot,FindItem("machinegun")))
				return;
		}

		// Use Super Shotgun
		//-------------------------------------------------------
		//qbism remove if(FUSION_AI_CheckShot(fusionbot) && FUSION_ITEMS_ChangeWeapon(fusionbot,FindItem("super shotgun")))
		//	return;

		// Use Shotgun
		//-------------------------------------------------------
		if(FUSION_AI_CheckShot(fusionbot) && FUSION_ITEMS_ChangeWeapon(fusionbot,FindItem("shotgun")))
   			return;

		// Use Blaster
		//-------------------------------------------------------
		if(FUSION_AI_CheckShot(fusionbot) && FUSION_ITEMS_ChangeWeapon(fusionbot,FindItem("blaster")))
   			return;
	}


	return;
}

//===========================================================================
// Routine		: FUSION_AI_ChoosePreferredWeapon
// Description	: Chooses the bots preferred weapon
void FUSION_AI_ChoosePreferredWeapon (edict_t *fusionbot, char *s)
{
    if ( stricmp(s, BLASTER_NAME) == 0 )
            fusionbot->client->resp.weapon = FindItem("blaster");
    else if ( stricmp(s, SHOTGUN_NAME) == 0 )
            fusionbot->client->resp.weapon = FindItem("shotgun");
   //qbism remove else if ( stricmp(s, SSHOTGUN_NAME) == 0 )
            //fusionbot->client->resp.weapon = FindItem("super shotgun");
    else if ( stricmp(s, CHAINGUN_NAME) == 0 )
            fusionbot->client->resp.weapon = FindItem("chaingun");
	else if ( stricmp(s, MACHINEGUN_NAME) == 0 )
            fusionbot->client->resp.weapon = FindItem("machinegun");
    else if ( stricmp(s, GRENADELAUNCHER_NAME) == 0 )
            fusionbot->client->resp.weapon = FindItem("grenade launcher");
    else if ( stricmp(s, ROCKETLAUNCHER_NAME) == 0 )
            fusionbot->client->resp.weapon = FindItem("rocket launcher");
    else if ( stricmp(s, HYPERBLASTER_NAME) == 0 )
            fusionbot->client->resp.weapon = FindItem("plasma gun");
    else if ( stricmp(s, RAILGUN_NAME) == 0 )
            fusionbot->client->resp.item = FindItem("railgun");
	else if ( stricmp(s, BFG_NAME) == 0 )
            fusionbot->client->resp.item = FindItem("bfg10k");
}
