/*
   ____  ____   _____ _____________   ____   ______ ______
 _/ ___\/  _ \ /     \\____ \_  __ \_/ __ \ /  ___//  ___/
 \  \__(  <_> )  Y Y  \  |_> >  | \/\  ___/ \___ \ \___ \ 
  \___  >____/|__|_|  /   __/|__|    \___  >____  >____  >
      \/            \/|__|               \/     \/     \/

	fusionbot_compress.c
	Data compress based on LZSS
================================================================
Should you decide to release a modified version of FUSION, you MUST
include the following text in the documentation and source of your 
modification.

---===[BEGIN TEXT]===---
================================================================
___________           .__                __________        __   
\_   _____/_ __  _____|__| ____   ____   \______   \ _____/  |_ 
 |    __)|  |  \/  ___/  |/  _ \ /    \   |    |  _//  _ \   __\
 |     \ |  |  /\___ \|  (  <_> )   |  \  |    |   (  <_> )  |  
 \___  / |____//____  >__|\____/|___|  /  |______  /\____/|__|  
     \/             \/               \/          \/  Version 2.0          
================================================================ 
Fusion v2.0
Copyright (c) 1998-2003 Trenton Shaffer.
All rights reserved.

ACEBOT Copyright(c) 1998, Steve Yeager.
EraserBot Copyright(c) 1998, Ryan Feltrin.
QFusion is Copyright (C) 2002-2003 Victor Luchits.
All other files are Copyright(C) 1997-2001 Id Software, Inc.
================================================================
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
================================================================
---===[END TEXT]===---
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <malloc.h>

// Size of ring buffer
#define N		 4096
// Upper limit for match_length
#define F		   18
// Encode string into position and length if match_length is greater
#define THRESHOLD 2 
// Inded for root of binary search trees
#define NIL			N

unsigned long int	textsize = 0, codesize = 0,	printcount = 0;
unsigned char		text_buf[N + F - 1];
int					match_position, match_length, lson[N + 1], rson[N + 257], dad[N + 1];

//===========================================================================
// Routine		: FUSION_COMPRESS_InitTree
// Description	: Initializes Trees
void FUSION_COMPRESS_InitTree(void)
{
	int  i;
	for (i = N + 1; i <= N + 256; i++) rson[i] = NIL;
	for (i = 0; i < N; i++) dad[i] = NIL;
}

//===========================================================================
// Routine		: FUSION_COMPRESS_InsertNode
// Description	: Inserts string of length F, text_buf[r..r+F-1], into one of the
//				: trees (text_buf[r]'th tree) and returns the longest-match position
//				: and length via the global variables match_position and match_length.	
void FUSION_COMPRESS_InsertNode(int r)
{
	int  i, p, cmp;
	unsigned char  *key;

	cmp = 1;  key = &text_buf[r];  p = N + 1 + key[0];
	rson[r] = lson[r] = NIL;  match_length = 0;
	for ( ; ; ) {
		if (cmp >= 0) {
			if (rson[p] != NIL) p = rson[p];
			else {  rson[p] = r;  dad[r] = p;  return;  }
		} else {
			if (lson[p] != NIL) p = lson[p];
			else {  lson[p] = r;  dad[r] = p;  return;  }
		}
		for (i = 1; i < F; i++)
			if ((cmp = key[i] - text_buf[p + i]) != 0)  break;
		if (i > match_length) {
			match_position = p;
			if ((match_length = i) >= F)  break;
		}
	}
	dad[r] = dad[p];  lson[r] = lson[p];  rson[r] = rson[p];
	dad[lson[p]] = r;  dad[rson[p]] = r;
	if (rson[dad[p]] == p) rson[dad[p]] = r;
	else                   lson[dad[p]] = r;

	// remove p
	//-------------------------------------------------------
	dad[p] = NIL;
}

//===========================================================================
// Routine		: FUSION_COMPRESS_DeleteNode
// Description	: Deletes NODE p from Tree
void FUSION_COMPRESS_DeleteNode(int p)  
{
	int  q;
	
	// not in tree
	//-------------------------------------------------------
	if (dad[p] == NIL) return;
	if (rson[p] == NIL) q = lson[p];
	else if (lson[p] == NIL) q = rson[p];
	else {
		q = lson[p];
		if (rson[q] != NIL) {
			do {  q = rson[q];  } while (rson[q] != NIL);
			rson[dad[q]] = lson[q];  dad[lson[q]] = dad[q];
			lson[q] = lson[p];  dad[lson[p]] = q;
		}
		rson[q] = rson[p];  dad[rson[p]] = q;
	}
	dad[q] = dad[p];
	if (rson[dad[p]] == p) rson[dad[p]] = q;  else lson[dad[p]] = q;
	dad[p] = NIL;
}

//===========================================================================
// Routine		: FUSION_COMPRESS_Encode
// Description	:
int FUSION_COMPRESS_Encode(char *filename, unsigned char *buffer, int bufsize, int version)
{
	int  i, c, len, r, s, last_match_length, code_buf_ptr;
	unsigned char  code_buf[17], mask;
	int bufptr = 0;
	FILE *pOut;
	//int version;
	
	pOut  = fopen(filename, "wb");
	if(pOut == NULL)
		// bail
		//-------------------------------------------------------
		return -1;

	//version = 2; // compressed version

	// Read version info (uneeded here, but moves file ptr)
	//-------------------------------------------------------
	fwrite(&version,sizeof(int),1,pOut); // write version
	fwrite(&bufsize,sizeof(int),1,pOut); // write out the size of the buffer
	
	// initialize trees
	//-------------------------------------------------------
	FUSION_COMPRESS_InitTree();
	
	code_buf[0] = 0;  
	code_buf_ptr = mask = 1;
	s = 0;  r = N - F;
	
	// clear the buffer with any character that will appear often
	//-------------------------------------------------------
	for (i = s; i < r; i++) text_buf[i] = ' ';
	for (len = 0; len < F &&  bufptr < bufsize; len++)
	{	
		c = buffer[bufptr++];
		text_buf[r + len] = c;
	}
	
	// text size zero
	//-------------------------------------------------------
	if ((textsize = len) == 0) return -1;
	
	for (i = 1; i <= F; i++) FUSION_COMPRESS_InsertNode(r - i);

	FUSION_COMPRESS_InsertNode(r);  
	do 
	{
		// match_length may be long near the end of text
		//-------------------------------------------------------
		if (match_length > len) match_length = len;  
		if (match_length <= THRESHOLD) 
		{
			// not long enough match. send one byte.
			//-------------------------------------------------------
			match_length = 1;
			// send one byte flag
			//-------------------------------------------------------
			code_buf[0] |= mask; 
			// send uncoded
			//-------------------------------------------------------
			code_buf[code_buf_ptr++] = text_buf[r];
		} 
		else 
		{
			code_buf[code_buf_ptr++] = (unsigned char) match_position;
			
			// send position and length pair. note match_length > THRESHOLD
			//-------------------------------------------------------
			code_buf[code_buf_ptr++] = (unsigned char)
				(((match_position >> 4) & 0xf0)
			  | (match_length - (THRESHOLD + 1)));
		}
		
		// Shift mask left one bit.
		//-------------------------------------------------------
		if ((mask <<= 1) == 0) 
		{			
			// Send at most 8 units of code together
			//-------------------------------------------------------
			for (i = 0; i < code_buf_ptr; i++)
				putc(code_buf[i], pOut); 
			codesize += code_buf_ptr;
			code_buf[0] = 0;  code_buf_ptr = mask = 1;
		}
		last_match_length = match_length;
		for (i = 0; i < last_match_length &&
				bufptr < bufsize; i++) 
		{
			c = buffer[bufptr++];
			
			// Delete old strings and read new bytes
			//-------------------------------------------------------
			FUSION_COMPRESS_DeleteNode(s);		
			text_buf[s] = c;	
			
			// If the position is near the end of buffer, extend the 
			// buffer to make string comparison easier
			//-------------------------------------------------------
			if (s < F - 1) text_buf[s + N] = c; 
			
			// Since this is a ring buffer, increment the position
			//-------------------------------------------------------
			s = (s + 1) & (N - 1);  r = (r + 1) & (N - 1);

			// Register the string in text_buf[r..r+F-1]
			//-------------------------------------------------------
			FUSION_COMPRESS_InsertNode(r);
		}
		
		while (i++ < last_match_length) 
		{
			FUSION_COMPRESS_DeleteNode(s);				
			s = (s + 1) & (N - 1);  r = (r + 1) & (N - 1);
			if (--len) FUSION_COMPRESS_InsertNode(r);	
		}
	} while (len > 0);	
	
	// Send remaining code
	//-------------------------------------------------------
	if (code_buf_ptr > 1) 
	{
		for (i = 0; i < code_buf_ptr; i++) putc(code_buf[i], pOut);
		codesize += code_buf_ptr;
	}
	fclose(pOut);

	return codesize;
}

//===========================================================================
// Routine		: FUSION_COMPRESS_Decode
// Description	: Be careful with the buffersize, will return an exit of -1 if failure
int FUSION_COMPRESS_Decode(char *filename, unsigned char *buffer, int bufsize)
{
	int  i, j, k, r, c;
	unsigned int  flags;
	int bufptr=0;
	FILE *pIn;
	int version;
	
	pIn  = fopen(filename, "rb");
	if(pIn == NULL)
		return -1; 

	// Read version info (uneeded here, but moves file ptr)
	//-------------------------------------------------------
	fread(&version,sizeof(int),1,pIn); // read version
	fread(&version,sizeof(int),1,pIn); // read buffersize (not needed, so repeat into version)

	
	for (i = 0; i < N - F; i++) text_buf[i] = ' ';
	r = N - F;  flags = 0;
	for ( ; ; ) {
		
		// Uses higher byte cleverly to count eight
		//-------------------------------------------------------
		if (((flags >>= 1) & 256) == 0) 
		{
			if ((c = getc(pIn)) == EOF) break;
			flags = c | 0xff00;
		}
		if (flags & 1) 
		{
			if ((c = getc(pIn)) == EOF) break;
			buffer[bufptr++] = c;	
			// Check for overflow
			//-------------------------------------------------------
			if(bufptr > bufsize)
				return -1;
			text_buf[r++] = c;  
			r &= (N - 1);
		} else {
			if ((i = getc(pIn)) == EOF) break;
			if ((j = getc(pIn)) == EOF) break;
			i |= ((j & 0xf0) << 4);  j = (j & 0x0f) + THRESHOLD;
			for (k = 0; k <= j; k++) {
				c = text_buf[(i + k) & (N - 1)];
				buffer[bufptr++] = c;
				// Check for overflow
				//-------------------------------------------------------
				if(bufptr > bufsize)
					return -1; 
				text_buf[r++] = c;  
				r &= (N - 1);
			}
		}
	}
	
	fclose(pIn);
	// return uncompressed size
	//-------------------------------------------------------
	return bufptr;
}
