/*

ITEMS FUNCTION LISTING

FUSION_ITEMS_FusionBotAdded
FUSION_ITEMS_FusionBotRemoved
FUSION_ITEMS_IsVisible
FUSION_ITEMS_IsReachable
FUSION_ITEMS_ChangeWeapon
FUSION_ITEMS_CanUseArmor
FUSION_ITEMS_ItemNeed
FUSION_ITEMS_ClassnameToIndex
FUSION_ITEMS_BuildItemNodeTable

*/
/*
 .__  __
 |__|/  |_  ____   _____   ______
 |  \   __\/ __ \ /     \ /  ___/
 |  ||  | \  ___/|  Y Y  \\___ \
 |__||__|  \___  >__|_|  /____  >
               \/      \/     \/

	fusionbot_items.c
	Item handling routines, including fact table support
================================================================
Should you decide to release a modified version of FUSION, you MUST
include the following text in the documentation and source of your
modification.

---===[BEGIN TEXT]===---
================================================================
___________           .__                __________        __
\_   _____/_ __  _____|__| ____   ____   \______   \ _____/  |_
 |    __)|  |  \/  ___/  |/  _ \ /    \   |    |  _//  _ \   __\
 |     \ |  |  /\___ \|  (  <_> )   |  \  |    |   (  <_> )  |
 \___  / |____//____  >__|\____/|___|  /  |______  /\____/|__|
     \/             \/               \/          \/  Version 2.0
================================================================
Fusion v2.0
Copyright (c) 1998-2003 Trenton Shaffer.
All rights reserved.

ACEBOT Copyright(c) 1998, Steve Yeager.
EraserBot Copyright(c) 1998, Ryan Feltrin.
QFusion is Copyright (C) 2002-2003 Victor Luchits.
All other files are Copyright(C) 1997-2001 Id Software, Inc.
================================================================
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
================================================================
---===[END TEXT]===---
*/
#include "..\game\g_local.h"
#include "fusionbot.h"

int	num_players = 0;
int num_items = 0;
item_table_t	item_table[MAX_EDICTS];
edict_t *players[MAX_CLIENTS];		// pointers to all players in the game

//===========================================================================
// Routine		: FUSION_ITEMS_FusionBotAdded
// Description	: Add the Player to our list
void FUSION_ITEMS_FusionBotAdded(edict_t *ent)
{
	players[num_players++] = ent;
}

//===========================================================================
// Routine		: FUSION_ITEMS_FusionBotRemoved
// Description	: Remove the Player from our list
void FUSION_ITEMS_FusionBotRemoved(edict_t *ent)
{
	int i;
	int pos;

	// watch for 0 players
	//-------------------------------------------------------
	if(num_players == 0)
		return;
	//-------------------------------------------------------

	// special case for only one player
	//-------------------------------------------------------
	if(num_players == 1)
	{
		num_players = 0;
		return;
	}
	//-------------------------------------------------------

	// Find the player
	//-------------------------------------------------------
	for(i=0;i<num_players;i++)
		if(ent == players[i])
			pos = i;
	//-------------------------------------------------------

	// decrement
	//-------------------------------------------------------
	for(i=pos;i<num_players-1;i++)
		players[i] = players[i+1];

	num_players--;
	//-------------------------------------------------------
}

//===========================================================================
// Routine		: FUSION_ITEMS_IsReachable
// Description	: Determine if Item is reachable
qboolean FUSION_ITEMS_IsReachable(edict_t *fusionbot, vec3_t goal)
{
	trace_t trace;
	vec3_t v;

	VectorCopy(fusionbot->mins,v);
	v[2] += 18; // Stepsize

	trace = gi.trace (fusionbot->s.origin, v, fusionbot->maxs, goal, fusionbot, MASK_SOLID|MASK_OPAQUE);

	// Yes we can see it
	//-------------------------------------------------------
	if (trace.fraction == 1.0)
		return qtrue;
	else
		return qfalse;
	//-------------------------------------------------------

}

//===========================================================================
// Routine		: FUSION_ITEMS_IsVisible
// Description	: Determine if Item is visible
qboolean FUSION_ITEMS_IsVisible(edict_t *fusionbot, vec3_t goal)
{
	trace_t trace;

	trace = gi.trace (fusionbot->s.origin, vec3_origin, vec3_origin, goal, fusionbot, MASK_SOLID|MASK_OPAQUE);

	// Yes we can see it
	//-------------------------------------------------------
	if (trace.fraction == 1.0)
		return qtrue;
	else
		return qfalse;
	//-------------------------------------------------------
}

//===========================================================================
// Routine		: FUSION_ITEMS_ChangeWeapon
// Description	: Weapon changing
qboolean FUSION_ITEMS_ChangeWeapon (edict_t *ent, gitem_t *item)
{
	int			ammo_index;
	gitem_t		*ammo_item;

	// see if we're already using it
	//-------------------------------------------------------
	if (item == ent->client->pers.weapon)
		return qtrue;
	//-------------------------------------------------------

	// Has not picked up weapon yet
	//-------------------------------------------------------
	if(!ent->client->pers.inventory[ITEM_INDEX(item)])
		return qfalse;
	//-------------------------------------------------------

	// Do we have ammo for it?
	//-------------------------------------------------------
	if (item->ammo)
	{
		ammo_item = FindItem(item->ammo);
		ammo_index = ITEM_INDEX(ammo_item);
		if (!ent->client->pers.inventory[ammo_index] && !g_select_empty->value)
			return qfalse;
	}
	//-------------------------------------------------------

	// Change to this weapon
	//-------------------------------------------------------
	ent->client->newweapon = item;

	return qtrue;
	//-------------------------------------------------------
}


extern gitem_armor_t jacketarmor_info;
extern gitem_armor_t combatarmor_info;
extern gitem_armor_t bodyarmor_info;

//===========================================================================
// Routine		: FUSION_ITEMS_CanUseArmor
// Description	: Check if bot can use armor
qboolean FUSION_ITEMS_CanUseArmor (gitem_t *item, edict_t *other)
{
	int				old_armor_index;
	gitem_armor_t	*oldinfo;
	gitem_armor_t	*newinfo;
	int				newcount;
	float			salvage;
	int				salvagecount;

	// get info on new armor
	//-------------------------------------------------------
	newinfo = (gitem_armor_t *)item->info;
	old_armor_index = ArmorIndex (other);
	//-------------------------------------------------------

	// handle armor shards specially
	//-------------------------------------------------------
	if (item->tag == ARMOR_SHARD)
		return qtrue;
	//-------------------------------------------------------

	// get info on old armor
	//-------------------------------------------------------
	if (old_armor_index == ITEM_INDEX(FindItem("Jacket Armor")))
		oldinfo = &jacketarmor_info;
	else if (old_armor_index == ITEM_INDEX(FindItem("Combat Armor")))
		oldinfo = &combatarmor_info;
	else
		oldinfo = &bodyarmor_info;
	//-------------------------------------------------------

	if (newinfo->normal_protection <= oldinfo->normal_protection)
	{
		// calc new armor values
		//-------------------------------------------------------
		salvage = newinfo->normal_protection / oldinfo->normal_protection;
		salvagecount = salvage * newinfo->base_count;
		newcount = other->client->pers.inventory[old_armor_index] + salvagecount;

		if (newcount > oldinfo->max_count)
			newcount = oldinfo->max_count;

		// if we're already maxed out then we don't need the new armor
		//-------------------------------------------------------
		if (other->client->pers.inventory[old_armor_index] >= newcount)
			return qfalse;

	}

	return qtrue;
}


//===========================================================================
// Routine		: FUSION_ITEMS_ItemNeed
// Description	: Determines the need for an Item.
//				: Function can be modified to support new items to pick up.
//				: Any other logic for custom decision making can be added here
float FUSION_ITEMS_ItemNeed(edict_t *fusionbot, int item)
{

	// Make sure item is at least close to being valid
	//-------------------------------------------------------
	if(item < 0 || item > 100)
		return 0.0;
	//-------------------------------------------------------

	switch(item)
	{
		// Health
		//-------------------------------------------------------
		case ITEMLIST_HEALTH_SMALL:
		case ITEMLIST_HEALTH_MEDIUM:
		case ITEMLIST_HEALTH_LARGE:
		case ITEMLIST_HEALTH_MEGA:
			if(fusionbot->health < 100)
				return 1.0 - (float)fusionbot->health/100.0f; // worse off, higher priority
			else
				return 0.0;
		//-------------------------------------------------------

		case ITEMLIST_AMMOPACK:
		case ITEMLIST_QUADDAMAGE:
		case ITEMLIST_INVULNERABILITY:
		case ITEMLIST_SILENCER:
		case ITEMLIST_ADRENALINE:
		case ITEMLIST_BANDOLIER:
			return 0.5;

		// Weapons
		//-------------------------------------------------------
		case ITEMLIST_ROCKETLAUNCHER:
		case ITEMLIST_RAILGUN:
		case ITEMLIST_MACHINEGUN:
		case ITEMLIST_CHAINGUN:
		case ITEMLIST_SHOTGUN:
		//qbism remove case ITEMLIST_SUPERSHOTGUN:
		case ITEMLIST_BFG10K:
		case ITEMLIST_GRENADELAUNCHER:
		case ITEMLIST_HYPERBLASTER:
			if(!fusionbot->client->pers.inventory[item])
				return 0.7;
			else
				return 0.0;
		//-------------------------------------------------------

		// Ammo
		//-------------------------------------------------------
		case ITEMLIST_SLUGS:
			if(fusionbot->client->pers.inventory[ITEMLIST_SLUGS] < fusionbot->client->pers.max_slugs)
				return 0.3;
			else
				return 0.0;

		case ITEMLIST_BULLETS:
			if(fusionbot->client->pers.inventory[ITEMLIST_BULLETS] < fusionbot->client->pers.max_bullets)
				return 0.3;
			else
				return 0.0;

		case ITEMLIST_SHELLS:
		   if(fusionbot->client->pers.inventory[ITEMLIST_SHELLS] < fusionbot->client->pers.max_shells)
				return 0.3;
			else
				return 0.0;

		case ITEMLIST_CELLS:
			if(fusionbot->client->pers.inventory[ITEMLIST_CELLS] <	fusionbot->client->pers.max_cells)
				return 0.3;
			else
				return 0.0;

		case ITEMLIST_ROCKETS:
			if(fusionbot->client->pers.inventory[ITEMLIST_ROCKETS] < fusionbot->client->pers.max_rockets)
				return 1.5;
			else
				return 0.0;

		case ITEMLIST_GRENADES:
			if(fusionbot->client->pers.inventory[ITEMLIST_GRENADES] < fusionbot->client->pers.max_grenades)
				return 0.3;
			else
				return 0.0;
		//-------------------------------------------------------

		// Armor
		//-------------------------------------------------------
		case ITEMLIST_BODYARMOR:
			if(FUSION_ITEMS_CanUseArmor (FindItem("Body Armor"), fusionbot))
				return 0.6;
			else
				return 0.0;

		case ITEMLIST_COMBATARMOR:
			if(FUSION_ITEMS_CanUseArmor (FindItem("Combat Armor"), fusionbot))
				return 0.6;
			else
				return 0.0;

		case ITEMLIST_JACKETARMOR:
			if(FUSION_ITEMS_CanUseArmor (FindItem("Jacket Armor"), fusionbot))
				return 0.6;
			else
				return 0.0;
		//-------------------------------------------------------

		// Power-Ups
		//-------------------------------------------------------
		case ITEMLIST_POWERSCREEN:
		case ITEMLIST_POWERSHIELD:
			return 0.5;
		//-------------------------------------------------------

		// Flags
		//-------------------------------------------------------
		case ITEMLIST_FLAG1:
			// If I am on team one, I want team two's flag
			if(!fusionbot->client->pers.inventory[item] && fusionbot->client->resp.ctf_team == CTF_TEAM2)
				return 10.0;
			else
				return 0.0;

		case ITEMLIST_FLAG2:
			if(!fusionbot->client->pers.inventory[item] && fusionbot->client->resp.ctf_team == CTF_TEAM1)
				return 10.0;
			else
				return 0.0;
		//-------------------------------------------------------

		// Techs
		//-------------------------------------------------------
		case ITEMLIST_RESISTANCETECH:
		case ITEMLIST_STRENGTHTECH:
		case ITEMLIST_HASTETECH:
		case ITEMLIST_REGENERATIONTECH:
			// Check for other tech
			if(!fusionbot->client->pers.inventory[ITEMLIST_RESISTANCETECH] &&
			   !fusionbot->client->pers.inventory[ITEMLIST_STRENGTHTECH] &&
			   !fusionbot->client->pers.inventory[ITEMLIST_HASTETECH] &&
			   !fusionbot->client->pers.inventory[ITEMLIST_REGENERATIONTECH])
			    return 0.4;
			else
				return 0.0;
		//-------------------------------------------------------

		default:
			return 0.0;

	}

}

//===========================================================================
// Routine		: FUSION_ITEMS_ClassnameToIndex
// Description	: Convert a classname to its index value
//				: Integers/Defines can lead to some slowdowns, but makes
//				: dealing with the rest of the code easier
int FUSION_ITEMS_ClassnameToIndex(char *classname)
{
	if(strcmp(classname,"item_armor_body")==0)
		return ITEMLIST_BODYARMOR;

	if(strcmp(classname,"item_armor_combat")==0)
		return ITEMLIST_COMBATARMOR;

	if(strcmp(classname,"item_armor_jacket")==0)
		return ITEMLIST_JACKETARMOR;

	if(strcmp(classname,"item_armor_shard")==0)
		return ITEMLIST_ARMORSHARD;

	if(strcmp(classname,"item_power_screen")==0)
		return ITEMLIST_POWERSCREEN;

	if(strcmp(classname,"item_power_shield")==0)
		return ITEMLIST_POWERSHIELD;

	if(strcmp(classname,"weapon_grapple")==0)
		return ITEMLIST_GRAPPLE;

	if(strcmp(classname,"weapon_blaster")==0)
		return ITEMLIST_BLASTER;

	if(strcmp(classname,"weapon_shotgun")==0)
		return ITEMLIST_SHOTGUN;

	//qbism remove if(strcmp(classname,"weapon_supershotgun")==0)
		//return ITEMLIST_SUPERSHOTGUN;

	if(strcmp(classname,"weapon_machinegun")==0)
		return ITEMLIST_MACHINEGUN;

	if(strcmp(classname,"weapon_chaingun")==0)
		return ITEMLIST_CHAINGUN;

	if(strcmp(classname,"weapon_chaingun")==0)
		return ITEMLIST_CHAINGUN;

	if(strcmp(classname,"ammo_grenades")==0)
		return ITEMLIST_GRENADES;

	if(strcmp(classname,"weapon_grenadelauncher")==0)
		return ITEMLIST_GRENADELAUNCHER;

	if(strcmp(classname,"weapon_rocketlauncher")==0)
		return ITEMLIST_ROCKETLAUNCHER;

	if(strcmp(classname,"weapon_hyperblaster")==0)
		return ITEMLIST_HYPERBLASTER;

	if(strcmp(classname,"weapon_railgun")==0)
		return ITEMLIST_RAILGUN;

	if(strcmp(classname,"weapon_bfg10k")==0)
		return ITEMLIST_BFG10K;

	if(strcmp(classname,"ammo_shells")==0)
		return ITEMLIST_SHELLS;

	if(strcmp(classname,"ammo_bullets")==0)
		return ITEMLIST_BULLETS;

	if(strcmp(classname,"ammo_cells")==0)
		return ITEMLIST_CELLS;

	if(strcmp(classname,"ammo_rockets")==0)
		return ITEMLIST_ROCKETS;

	if(strcmp(classname,"ammo_slugs")==0)
		return ITEMLIST_SLUGS;

	if(strcmp(classname,"item_quad")==0)
		return ITEMLIST_QUADDAMAGE;

	if(strcmp(classname,"item_invunerability")==0)
		return ITEMLIST_INVULNERABILITY;

	if(strcmp(classname,"item_silencer")==0)
		return ITEMLIST_SILENCER;

	if(strcmp(classname,"item_rebreather")==0)
		return ITEMLIST_REBREATHER;

	if(strcmp(classname,"item_enviornmentsuit")==0)
		return ITEMLIST_ENVIRONMENTSUIT;

	if(strcmp(classname,"item_ancienthead")==0)
		return ITEMLIST_ANCIENTHEAD;

	if(strcmp(classname,"item_adrenaline")==0)
		return ITEMLIST_ADRENALINE;

	if(strcmp(classname,"item_bandolier")==0)
		return ITEMLIST_BANDOLIER;

	if(strcmp(classname,"item_pack")==0)
		return ITEMLIST_AMMOPACK;

	if(strcmp(classname,"item_datacd")==0)
		return ITEMLIST_DATACD;

	if(strcmp(classname,"item_powercube")==0)
		return ITEMLIST_POWERCUBE;

	if(strcmp(classname,"item_pyramidkey")==0)
		return ITEMLIST_PYRAMIDKEY;

	if(strcmp(classname,"item_dataspinner")==0)
		return ITEMLIST_DATASPINNER;

	if(strcmp(classname,"item_securitypass")==0)
		return ITEMLIST_SECURITYPASS;

	if(strcmp(classname,"item_bluekey")==0)
		return ITEMLIST_BLUEKEY;

	if(strcmp(classname,"item_redkey")==0)
		return ITEMLIST_REDKEY;

	if(strcmp(classname,"item_commandershead")==0)
		return ITEMLIST_COMMANDERSHEAD;

	if(strcmp(classname,"item_airstrikemarker")==0)
		return ITEMLIST_AIRSTRIKEMARKER;

	if(strcmp(classname,"item_health")==0) // ??
		return ITEMLIST_HEALTH;

	if(strcmp(classname,"team_CTF_redflag")==0)
		return ITEMLIST_FLAG1;

	if(strcmp(classname,"team_CTF_blueflag")==0)
		return ITEMLIST_FLAG2;

	if(strcmp(classname,"item_tech1")==0)
		return ITEMLIST_RESISTANCETECH;

	if(strcmp(classname,"item_tech2")==0)
		return ITEMLIST_STRENGTHTECH;

	if(strcmp(classname,"item_tech3")==0)
		return ITEMLIST_HASTETECH;

	if(strcmp(classname,"item_tech4")==0)
		return ITEMLIST_REGENERATIONTECH;

	if(strcmp(classname,"item_health_small")==0)
		return ITEMLIST_HEALTH_SMALL;

	if(strcmp(classname,"item_health_medium")==0)
		return ITEMLIST_HEALTH_MEDIUM;

	if(strcmp(classname,"item_health_large")==0)
		return ITEMLIST_HEALTH_LARGE;

	if(strcmp(classname,"item_health_mega")==0)
		return ITEMLIST_HEALTH_MEGA;

	return INVALID;
}

//===========================================================================
// Routine		: FUSION_ITEMS_BuildItemNodeTable
// Description	: Build Item Node Table
//				: Only called once per level, when saved will not be called again.
//				:
//				: Downside is items can not move about. If the level has been saved
//				: before and reloaded, it could cause a problem if there are items
//				: that spawn at random locations.
//#define DEBUG // uncomment to write out items to a file.
//========================================================
void FUSION_ITEMS_BuildItemNodeTable (qboolean rebuild)
{
	edict_t *items;
	int i,item_index;
	vec3_t v,v1,v2;

#ifdef DEBUG
	FILE *pOut;
	if((pOut = fopen("items.txt","wt"))==NULL)
		return;
#endif

	num_items = 0;

	// Add game items
	for(items = g_edicts; items < &g_edicts[globals.num_edicts]; items++)
	{
		// filter out crap
		if(items->solid == SOLID_NOT)
			continue;

		if(!items->classname)
			continue;

		// Items
		//-------------------------------------------------------
		item_index = FUSION_ITEMS_ClassnameToIndex(items->classname);
		//-------------------------------------------------------

		// SPECIAL NAV NODE DROPPING CODE
		//-------------------------------------------------------
		// Special node dropping for platforms
		//-------------------------------------------------------
		if(strcmp(items->classname,"func_plat")==0)
		{
			if(!rebuild)
				FUSION_NODES_AddNode(items,NODE_PLATFORM);
			item_index = 99; // to allow to pass the item index test
		}
		//-------------------------------------------------------

		// Special node dropping for teleporters
		//-------------------------------------------------------
		if(strcmp(items->classname,"misc_teleporter_dest")==0 || strcmp(items->classname,"misc_teleporter")==0)
		{
			if(!rebuild)
				FUSION_NODES_AddNode(items,NODE_TELEPORTER);
			item_index = 99;
		}
		//-------------------------------------------------------

		#ifdef DEBUG
		if(item_index == INVALID)
			fprintf(pOut,"Rejected item: %s node: %d pos: %f %f %f\n",items->classname,item_table[num_items].node,items->s.origin[0],items->s.origin[1],items->s.origin[2]);
		else
			fprintf(pOut,"item: %s node: %d pos: %f %f %f\n",items->classname,item_table[num_items].node,items->s.origin[0],items->s.origin[1],items->s.origin[2]);
		#endif

		if(item_index == INVALID)
			continue;

		// add a pointer to the item entity
		//-------------------------------------------------------
		item_table[num_items].ent = items;
		item_table[num_items].item = item_index;
		//-------------------------------------------------------

		// If new, add nodes for items
		//-------------------------------------------------------
		if(!rebuild)
		{
			// Add a new node at the item's location.
			// FIXME!!!!

			// Add Red Flag Node
			if ((ctf->value) && (item_index == 43))
			{
				item_table[num_items].node = FUSION_NODES_AddNode(items,NODE_FLAG1);
			}
			// Add Blue Flag Node
			else if ((ctf->value) && (item_index == 44))
			{
				item_table[num_items].node = FUSION_NODES_AddNode(items,NODE_FLAG2);
			}
			// Add Item Nodes
			else
			{
				item_table[num_items].node = FUSION_NODES_AddNode(items,NODE_ITEM);
			}
			num_items++;

		}
		// Now if rebuilding, just relink ent structures
		else
		{
			// Find stored location
			for(i=0;i<number_of_nodes;i++)
			{
				// valid types
				if(nodes[i].type == NODE_ITEM ||
				   nodes[i].type == NODE_PLATFORM ||
				   nodes[i].type == NODE_TELEPORTER ||
				   nodes[i].type == NODE_FLAG1 ||
				   nodes[i].type == NODE_FLAG2)
				{
					VectorCopy(items->s.origin,v);

					// Add 16 to item type nodes
					if(nodes[i].type == NODE_ITEM)
						v[2] += 16;

					// Add 32 to teleporter
					if(nodes[i].type == NODE_TELEPORTER)
						v[2] += 32;
					// Add 16 to Flag
					if(nodes[i].type == NODE_FLAG1)
						v[2] += 16;
					if(nodes[i].type == NODE_FLAG2)
						v[2] += 16;

					if(nodes[i].type == NODE_PLATFORM)
					{
						VectorCopy(items->maxs,v1);
						VectorCopy(items->mins,v2);

						// To get the center
						v[0] = (v1[0] - v2[0]) / 2 + v2[0];
						v[1] = (v1[1] - v2[1]) / 2 + v2[1];
						v[2] = items->mins[2]+64;
					}

					if(v[0] == nodes[i].origin[0] &&
 					   v[1] == nodes[i].origin[1] &&
					   v[2] == nodes[i].origin[2])
					{
						// found a match now link to facts
						item_table[num_items].node = i;

#ifdef DEBUG
						fprintf(pOut,"Relink item: %s node: %d pos: %f %f %f\n",items->classname,item_table[num_items].node,items->s.origin[0],items->s.origin[1],items->s.origin[2]);
#endif
						num_items++;
					}
				}
			}
		}
		//-------------------------------------------------------

	}

#ifdef DEBUG
	fclose(pOut);
#endif

}
