/*

MAPPING FUNCTION LISTING

FUSION_MAPPING_FollowPath
FUSION_MAPPING_GrapFired
FUSION_MAPPING_CheckForLadder
FUSION_MAPPING_PathMap
FUSION_MAPPING_InitNodes
FUSION_MAPPING_ShowNode
FUSION_MAPPING_DrawPath
FUSION_MAPPING_ShowPath

*/

/*
    _____                       .__
   /     \ _____  ______ ______ |__| ____    ____
  /  \ /  \\__  \ \____ \\____ \|  |/    \  / ___\
 /    Y    \/ __ \|  |_> >  |_> >  |   |  \/ /_/  >
 \____|__  (____  /   __/|   __/|__|___|  /\___  /
         \/     \/|__|   |__|           \//_____/

	fusionbot_mapping.c
	Mapping routines for the FUSION bot.
================================================================
Should you decide to release a modified version of FUSION, you MUST
include the following text in the documentation and source of your
modification.

---===[BEGIN TEXT]===---
================================================================
___________           .__                __________        __
\_   _____/_ __  _____|__| ____   ____   \______   \ _____/  |_
 |    __)|  |  \/  ___/  |/  _ \ /    \   |    |  _//  _ \   __\
 |     \ |  |  /\___ \|  (  <_> )   |  \  |    |   (  <_> )  |
 \___  / |____//____  >__|\____/|___|  /  |______  /\____/|__|
     \/             \/               \/          \/  Version 2.0
================================================================
Fusion v2.0
Copyright (c) 1998-2003 Trenton Shaffer.
All rights reserved.

ACEBOT Copyright(c) 1998, Steve Yeager.
EraserBot Copyright(c) 1998, Ryan Feltrin.
QFusion is Copyright (C) 2002-2003 Victor Luchits.
All other files are Copyright(C) 1997-2001 Id Software, Inc.
================================================================
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
================================================================
---===[END TEXT]===---
*/
#include "..\game\g_local.h"
#include "fusionbot.h"

// Total number of nodes that are items
//-------------------------------------------------------
int number_of_itemnodes;

// Total number of nodes
//-------------------------------------------------------
int number_of_nodes;

// For debugging paths
//-------------------------------------------------------
int show_path_from = -1;
int show_path_to = -1;

// array for node data
//-------------------------------------------------------
node_t nodes[MAX_NODES];
short int path_table[MAX_NODES][MAX_NODES];

//===========================================================================
// Routine		: FUSION_MAPPING_FollowPath
// Description	: Move closer to goal by pointing the bot to the next node
//				: that is closer to the goal
qboolean FUSION_MAPPING_FollowPath(edict_t *fusionbot)
{
	vec3_t			v;

	// Show the path
	//-------------------------------------------------------
	if(debug_mode)
	{
		show_path_from = fusionbot->current_node;
		show_path_to = fusionbot->goal_node;
		if(showpath->value)
			FUSION_MAPPING_DrawPath(fusionbot);
	}
	//-------------------------------------------------------

	// Try again?
	//-------------------------------------------------------
	if(fusionbot->node_timeout ++ > 30)
	{
		if(fusionbot->tries++ > 3)
			return qfalse;
		else
			FUSION_NAVIGATION_SetGoal(fusionbot,fusionbot->goal_node);
	}

	// Checks if bot is off-course
	/* REM New Pathing Algorithm
	if (level.time == (float)((int)level.time))
	{
		if (!FUSION_PA_LinkExists(fusionbot->current_node, FUSION_PA_SLFront(&fusionbot->pathList))
			&& (fusionbot->current_node != FUSION_PA_SLFront(&fusionbot->pathList)))
		{
			FUSION_PA_InitSearch(fusionbot);
		}
	}

	if (FUSION_PA_SLEmpty(&fusionbot->pathList) && (fusionbot->current_node != fusionbot->goal_node))
	{
		if (!FUSION_PA_StartSearch(fusionbot, fusionbot->current_node, fusionbot->goal_node))
		{
			if (debug_mode)
				gi.bprintf(PRINT_HIGH, "%s: Target at(%i) - No Path \n", fusionbot->client->pers.netname, fusionbot->goal_node, fusionbot->next_node);
			return qfalse;
		}
	}
	*/

	// Are we there yet?
	//-------------------------------------------------------
	VectorSubtract(fusionbot->s.origin,nodes[fusionbot->next_node].origin,v);

	if(VectorLength(v) < 32)
	{
		// reset timeout
		//-------------------------------------------------------
		fusionbot->node_timeout = 0;

		if(fusionbot->next_node == fusionbot->goal_node)
		{
			if(debug_mode)
				gi.dprintf("FUSION BOT: %s reached goal!\n",fusionbot->client->pers.netname);

			// Pick a new goal
			FUSION_AI_Determine_LR_Goal(fusionbot);
		}
		else
		{
			fusionbot->current_node = fusionbot->next_node;
			fusionbot->next_node = path_table[fusionbot->current_node][fusionbot->goal_node];
			/* REM New Pathing Algorithm
			FUSION_PA_SLPop_Front(&fusionbot->pathList);

			if (!FUSION_PA_SLEmpty(&fusionbot->pathList))
				fusionbot->next_node = FUSION_PA_SLFront(&fusionbot->pathList);
			else
			{
				if (debug_mode)
					gi.bprintf(PRINT_HIGH, "Trying to read an empty SL nodelist!\n");
				fusionbot->next_node = INVALID;
			}
			*/
		}
	}
	//-------------------------------------------------------

	if(fusionbot->current_node == -1 || fusionbot->next_node ==-1)
		return qfalse;

	// Set bot's movement vector
	//-------------------------------------------------------
	VectorSubtract (nodes[fusionbot->next_node].origin, fusionbot->s.origin , fusionbot->move_vector);

	return qtrue;
}



//===========================================================================
// Routine		: FUSION_MAPPING_GrapFired
// Description	: Capture when grappling hook has been fired for mapping
void FUSION_MAPPING_GrapFired(edict_t *fusionbot)
{
	int			closest_node;

	if(!fusionbot->owner)
		// should not be here
		return;

	// Check to see if the grapple is in pull mode
	//-------------------------------------------------------
	if(fusionbot->owner->client->ctf_grapplestate == CTF_GRAPPLE_STATE_PULL)
	{
		// Look for the closest node of type grapple
		//-------------------------------------------------------
		closest_node = FUSION_NAVIGATION_FindClosestReachableNode(fusionbot,NODE_DENSITY,NODE_GRAPPLE);
		if( (closest_node == -1 ) && (dorebuild) ) // we need to drop a node
		{
			closest_node = FUSION_NODES_AddNode(fusionbot,NODE_GRAPPLE);

			//========================================================
			// Add an edge
			FUSION_NODES_UpdateNodeEdge(fusionbot->owner->last_node,closest_node);

			fusionbot->owner->last_node = closest_node;
		}
		else
			// zero out so other nodes will not be linked
			fusionbot->owner->last_node = closest_node;
		//-------------------------------------------------------
	}
	//-------------------------------------------------------
}

//===========================================================================
// Routine		: FUSION_MAPPING_CheckForLadders
// Description	: Check when adding ladder nodes
qboolean FUSION_MAPPING_CheckForLadder(edict_t *fusionbot)
{
	int			closest_node;

	// If there is a ladder and we are moving up, see if we should add a ladder node
	//-------------------------------------------------------
	if (gi.pointcontents(fusionbot->s.origin) & CONTENTS_LADDER && fusionbot->velocity[2] > 0)
	{
		closest_node = FUSION_NAVIGATION_FindClosestReachableNode(fusionbot,NODE_DENSITY,NODE_LADDER);
		if( (closest_node == -1) && (dorebuild) )
		{
			closest_node = FUSION_NODES_AddNode(fusionbot,NODE_LADDER);

			// Now add link
			//-------------------------------------------------------
			FUSION_NODES_UpdateNodeEdge(fusionbot->last_node,closest_node);

			// Set current to last
			//-------------------------------------------------------
			fusionbot->last_node = closest_node;
		}
		else
		{
			FUSION_NODES_UpdateNodeEdge(fusionbot->last_node,closest_node);
			// set visited to last
			fusionbot->last_node = closest_node;
		}
		return qtrue;
	}
	return qfalse;
}

//===========================================================================
// Routine		: FUSION_MAPPING_PathMap
// Description	: Hooks in the pathing code and sets the current node if valid
void FUSION_MAPPING_PathMap(edict_t *fusionbot)
{
	int			 closest_node;
	// start off low
	static float last_update=0;
	vec3_t		 v;

	if(level.time < last_update)
		return;

	// slow down updates a bit
	last_update = level.time + 0.15;

	// Special node drawing code for debugging
	//-------------------------------------------------------
	if(debug_mode)
	{
		if(show_path_to != -1)
		{
			if(showpath->value)
				FUSION_MAPPING_DrawPath(fusionbot);
		}
	}
	//-------------------------------------------------------

	// Special check for ladder nodes
	//-------------------------------------------------------
	if(FUSION_MAPPING_CheckForLadder(fusionbot))
		return;
	//-------------------------------------------------------

	// Not on ground, and not in the water, so bail
	//-------------------------------------------------------
    if(!fusionbot->groundentity && !fusionbot->waterlevel)
		return;
	//-------------------------------------------------------

	// Lava/Slime
	//-------------------------------------------------------
	VectorCopy(fusionbot->s.origin,v);
	v[2] -= 18;
	if(gi.pointcontents(v) & (CONTENTS_LAVA|CONTENTS_SLIME))
		// no nodes in slime
		return;
	//-------------------------------------------------------

    // Jumping
	//-------------------------------------------------------
	if(fusionbot->is_jumping)
	{
	   	// See if there is a closeby jump landing node (prevent adding too many)
		//-------------------------------------------------------
		closest_node = FUSION_NAVIGATION_FindClosestReachableNode(fusionbot, 64, NODE_JUMP);

		if( (closest_node == INVALID) && (dorebuild) )
			closest_node = FUSION_NODES_AddNode(fusionbot,NODE_JUMP);
		//-------------------------------------------------------

		// Now add link
		//-------------------------------------------------------
		if(fusionbot->last_node != -1)
			FUSION_NODES_UpdateNodeEdge(fusionbot->last_node, closest_node);
		//-------------------------------------------------------

		fusionbot->is_jumping = qfalse;
		return;
	}

	// Grapple
	// Do not add nodes during grapple, added elsewhere manually
	//-------------------------------------------------------
	if(ctf->value && fusionbot->client->ctf_grapplestate == CTF_GRAPPLE_STATE_PULL)
		return;
	//-------------------------------------------------------

	// Iterate through all nodes to make sure far enough apart
	//-------------------------------------------------------
	closest_node = FUSION_NAVIGATION_FindClosestReachableNode(fusionbot, NODE_DENSITY, NODE_ALL);
	//-------------------------------------------------------

	// Special Check for Platforms
	//-------------------------------------------------------
	if(fusionbot->groundentity && fusionbot->groundentity->use == Use_Plat)
	{
		if(closest_node == INVALID)
			return;

		// Here we want to add links
		//-------------------------------------------------------
		if(closest_node != fusionbot->last_node && fusionbot->last_node != INVALID)
			FUSION_NODES_UpdateNodeEdge(fusionbot->last_node,closest_node);

		// set visited to last
		fusionbot->last_node = closest_node;
		return;
	}

	// Add Nodes as needed
	//-------------------------------------------------------
	if( (closest_node == INVALID) && (dorebuild) )
	{
		// Add nodes in the water as needed
		//-------------------------------------------------------
		if(fusionbot->waterlevel)
			closest_node = FUSION_NODES_AddNode(fusionbot,NODE_WATER);
		else
			closest_node = FUSION_NODES_AddNode(fusionbot,NODE_MOVE);

		// Now add link
		//-------------------------------------------------------
		if(fusionbot->last_node != -1)
			FUSION_NODES_UpdateNodeEdge(fusionbot->last_node, closest_node);
	}
	else if(closest_node != fusionbot->last_node && fusionbot->last_node != INVALID)
		FUSION_NODES_UpdateNodeEdge(fusionbot->last_node,closest_node);
	//-------------------------------------------------------
	// set visited to last
	fusionbot->last_node = closest_node;

}

//===========================================================================
// Routine		: FUSION_MAPPING_InitNodes
// Description	: Initialize the node array (set all to INVALID)
void FUSION_MAPPING_InitNodes(void)
{
	number_of_nodes = 1;
	number_of_itemnodes = 1;
	memset(nodes,0,sizeof(node_t) * MAX_NODES);
	memset(path_table,INVALID,sizeof(short int)*MAX_NODES*MAX_NODES);
}

//===========================================================================
// Routine		: FUSION_MAPPING_ShowNode
// Description	: Show the node (debug only) for mapping
void FUSION_MAPPING_ShowNode(int node)
{
	edict_t			*ent;

	//return;

	ent = G_Spawn();

	ent->movetype = MOVETYPE_NONE;
	ent->solid = SOLID_NOT;

	if(nodes[node].type == NODE_MOVE)
		ent->s.renderfx = RF_SHELL_BLUE;
	else if (nodes[node].type == NODE_WATER)
		ent->s.renderfx = RF_SHELL_RED;
	else
		// action nodes
		ent->s.renderfx = RF_SHELL_GREEN;

	ent->s.modelindex = gi.modelindex ("models/ammo/grenade1.md3");  //qbism md3 replace

	ent->owner = ent;
	ent->nextthink = level.time + 200000.0;
	ent->think = G_FreeEdict;
	ent->dmg = 0;

	VectorCopy(nodes[node].origin,ent->s.origin);
	gi.linkentity (ent);

}

//===========================================================================
// Routine		: FUSION_MAPPING_DrawPath
// Description	: Draws the current path (debug only)
void FUSION_MAPPING_DrawPath(edict_t *fusionbot)
{
	int			current_node, goal_node, next_node;

	current_node = show_path_from;
	goal_node = show_path_to;

	/* REM New Pathing Algorithm
	FUSION_PA_StartSearch (fusionbot, current_node, goal_node);
	*/

	next_node = path_table[current_node][goal_node];

	/* REM New Pathing Algorithm
	next_node = FUSION_PA_SLFront(&fusionbot->pathList);
	*/

	// Now set up and display the path
	//-------------------------------------------------------
	while(current_node != goal_node && current_node != -1)
	{
		gi.WriteByte (svc_temp_entity);
		gi.WriteByte (TE_BUBBLETRAIL);
		gi.WritePosition (nodes[current_node].origin);
		gi.WritePosition (nodes[next_node].origin);
		gi.multicast (nodes[current_node].origin, MULTICAST_PVS);
		current_node = next_node;
		next_node = path_table[current_node][goal_node];
		/* REM New Pathing Algorithm
		FUSION_PA_SLPop_Front(&fusionbot->pathList);
		next_node = FUSION_PA_SLFront(&fusionbot->pathList);
		*/
	}
}

//===========================================================================
// Routine		: FUSION_MAPPING_ShowPath
// Description	: Turns on/off showing of path (debug only)
//				: set goal to -1 to shut off
void FUSION_MAPPING_ShowPath(edict_t *fusionbot, int goal_node)
{
	show_path_from = FUSION_NAVIGATION_FindClosestReachableNode(fusionbot, NODE_DENSITY, NODE_ALL);
	show_path_to = goal_node;
}
