/*

MOVEMENT FUNCTION LISTING

FUSION_MOVEMENT_SpecialMove
FUSION_MOVEMENT_Move
FUSION_MOVEMENT_Attack
FUSION_MOVEMENT_Wander

*/
/*
                                                   __
   _____   _______  __ ____   _____   ____   _____/  |_
  /     \ /  _ \  \/ // __ \ /     \_/ __ \ /    \   __\
 |  Y Y  (  <_> )   /\  ___/|  Y Y  \  ___/|   |  \  |
 |__|_|  /\____/ \_/  \___  >__|_|  /\___  >___|  /__|
       \/                 \/      \/     \/     \/

	fusionbot_movement.c
	Movement routines for the FUSION bot
================================================================
Should you decide to release a modified version of FUSION, you MUST
include the following text in the documentation and source of your
modification.

---===[BEGIN TEXT]===---
================================================================
___________           .__                __________        __
\_   _____/_ __  _____|__| ____   ____   \______   \ _____/  |_
 |    __)|  |  \/  ___/  |/  _ \ /    \   |    |  _//  _ \   __\
 |     \ |  |  /\___ \|  (  <_> )   |  \  |    |   (  <_> )  |
 \___  / |____//____  >__|\____/|___|  /  |______  /\____/|__|
     \/             \/               \/          \/  Version 2.0
================================================================
Fusion v2.0
Copyright (c) 1998-2003 Trenton Shaffer.
All rights reserved.

ACEBOT Copyright(c) 1998, Steve Yeager.
EraserBot Copyright(c) 1998, Ryan Feltrin.
QFusion is Copyright (C) 2002-2003 Victor Luchits.
All other files are Copyright(C) 1997-2001 Id Software, Inc.
================================================================
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
================================================================
---===[END TEXT]===---
*/
#include "..\game\g_local.h"
#include "fusionbot.h"

//===========================================================================
// Routine		: FUSION_MOVEMENT_CanMove
// Description	: Checks if bot can move (really just checking ground)
//				: Not an accurate check, but does a decent enough job
//				: and looks for lava/slime.
qboolean FUSION_MOVEMENT_CanMove(edict_t *fusionbot, int direction)
{
	vec3_t forward, right;
	vec3_t offset,start,end;
	vec3_t angles;
	trace_t tr;

	// Now check to see if move will move us off an edge
	//-------------------------------------------------------
	VectorCopy(fusionbot->s.angles,angles);

	if(direction == BOT_MOVE_LEFT)
		angles[1] += 90;
	else if(direction == BOT_MOVE_RIGHT)
		angles[1] -= 90;
	else if(direction == BOT_MOVE_BACK)
		angles[1] -=180;
	//-------------------------------------------------------

	// Set up the vectors
	//-------------------------------------------------------
	AngleVectors (angles, forward, right, NULL);

	VectorSet(offset, 36, 0, 24);
	G_ProjectSource (fusionbot->s.origin, offset, forward, right, start);

	VectorSet(offset, 36, 0, -100);
	G_ProjectSource (fusionbot->s.origin, offset, forward, right, end);

	tr = gi.trace(start, NULL, NULL, end, fusionbot, MASK_SOLID|MASK_OPAQUE);

	if(tr.fraction == 1.0 || tr.contents & (CONTENTS_LAVA|CONTENTS_SLIME))
	{
		if(debug_mode)
			gi.dprintf("%s: move blocked\n",fusionbot->client->pers.netname);
		return qfalse;
	}

	// Can Move
	return qtrue;
}

//===========================================================================
// Routine		: FUSION_MOVEMENT_SpecialMove
// Description	: Handles special cases of crouch/jump
//				: If resolved here, function returns qtrue
qboolean FUSION_MOVEMENT_SpecialMove(edict_t *fusionbot, usercmd_t *ucmd)
{
	vec3_t dir,forward,right,start,end,offset;
	vec3_t top;
	trace_t tr;

	// Get current direction
	//-------------------------------------------------------
	VectorCopy(fusionbot->client->ps.viewangles,dir);
	dir[YAW] = fusionbot->s.angles[YAW];
	AngleVectors (dir, forward, right, NULL);

	VectorSet(offset, 0, 0, 0);
	G_ProjectSource (fusionbot->s.origin, offset, forward, right, start);
	offset[0] += 18;
	G_ProjectSource (fusionbot->s.origin, offset, forward, right, end);
	//-------------------------------------------------------

	// trace it
	//-------------------------------------------------------
	// So bots don't jump all the time
	start[2] += 18;
	end[2] += 18;
	tr = gi.trace (start, fusionbot->mins, fusionbot->maxs, end, fusionbot, MASK_MONSTERSOLID);

	if( tr.fraction < 1.0 )
	{
		// Check for crouching
		//-------------------------------------------------------
		start[2] -= 14;
		end[2] -= 14;
		//-------------------------------------------------------

		// Set up for crouching check
		//-------------------------------------------------------
		VectorCopy(fusionbot->maxs,top);
		 // crouching height
		top[2] = 0.0;
		tr = gi.trace (start, fusionbot->mins, top, end, fusionbot, MASK_MONSTERSOLID);
		//-------------------------------------------------------

		// Crouch
		//-------------------------------------------------------
		if( tr.fraction == 1.0)
		{
			ucmd->forwardmove = 400;
			ucmd->upmove = -400;
			return qtrue;
		}
		//-------------------------------------------------------

		// Check for jump
		//-------------------------------------------------------
		start[2] += 32;
		end[2] += 32;
		tr = gi.trace (start, fusionbot->mins, fusionbot->maxs, end, fusionbot, MASK_MONSTERSOLID);

		if(tr.fraction == 1.0)
		{
			ucmd->forwardmove = 400;

			if(fusionbot->groundentity)
				ucmd->upmove = 400;
			return qtrue;
		}

		//-------------------------------------------------------
	}

	return qfalse; // We did not resolve a move here
}

//===========================================================================
// Routine		: FUSION_MOVEMENT_ChangeBotAngle
// Description	: Make changes in angles a little more gradual, not so snappy
//				: but noticable. Modified from id code ChangeYaw.
void FUSION_MOVEMENT_ChangeBotAngle (edict_t *ent)
{
	float	ideal_yaw;
	float   ideal_pitch;
	float	current_yaw;
	float   current_pitch;
	float	move;
	float	speed;
	vec3_t  ideal_angle;

	// Normalize the move angle first
	//-------------------------------------------------------
	VectorNormalize(ent->move_vector);

	current_yaw = anglemod(ent->s.angles[YAW]);
	current_pitch = anglemod(ent->s.angles[PITCH]);

	VecToAngles (ent->move_vector, ideal_angle); //qbism- switch to VecToAngles

	ideal_yaw = anglemod(ideal_angle[YAW]);
	ideal_pitch = anglemod(ideal_angle[PITCH]);
	//-------------------------------------------------------

	// Yaw
	//-------------------------------------------------------
	if (current_yaw != ideal_yaw)
	{
		move = ideal_yaw - current_yaw;
		speed = ent->yaw_speed;
		if (ideal_yaw > current_yaw)
		{
			if (move >= 180)
				move = move - 360;
		}
		else
		{
			if (move <= -180)
				move = move + 360;
		}
		if (move > 0)
		{
			if (move > speed)
				move = speed;
		}
		else
		{
			if (move < -speed)
				move = -speed;
		}
		ent->s.angles[YAW] = anglemod (current_yaw + move);
	}
	//-------------------------------------------------------

	// Pitch
	//-------------------------------------------------------
	if (current_pitch != ideal_pitch)
	{
		move = ideal_pitch - current_pitch;
		speed = ent->yaw_speed;
		if (ideal_pitch > current_pitch)
		{
			if (move >= 180)
				move = move - 360;
		}
		else
		{
			if (move <= -180)
				move = move + 360;
		}
		if (move > 0)
		{
			if (move > speed)
				move = speed;
		}
		else
		{
			if (move < -speed)
				move = -speed;
		}
		ent->s.angles[PITCH] = anglemod (current_pitch + move);
	}
	//-------------------------------------------------------
}

//===========================================================================
// Routine		: FUSION_MOVEMENT_CheckEyes
// Description	: Checks for obstruction in front of BOT
//				: Created to help BOT steer around obstructions
//				: If the move is resolved here, returns qtrue
//				: NOTE: No longer used ... it's all messed up
qboolean FUSION_MOVEMENT_CheckEyes(edict_t *fusionbot, usercmd_t *ucmd)
{
	vec3_t  forward, right;
	vec3_t  leftstart, rightstart,focalpoint;
	vec3_t	upend;
	vec3_t  dir,offset;

	trace_t traceRight, traceLeft, traceFront;

	// Get current angle and set up "eyes"
	//-------------------------------------------------------
	VectorCopy(fusionbot->s.angles,dir);
	AngleVectors (dir, forward, right, NULL);
	//-------------------------------------------------------

	// Let them move to targets by walls
	//-------------------------------------------------------
	if(!fusionbot->movetarget)
		VectorSet(offset,64,0,4); // focalpoint
	else
		VectorSet(offset,36,0,4); // focalpoint

	G_ProjectSource (fusionbot->s.origin, offset, forward, right, focalpoint);
	//-------------------------------------------------------

	// Check from fusionbot to focalpoint
	// Ladder code
	//-------------------------------------------------------
	VectorSet(offset,36,0,0); // set as high as possible
	G_ProjectSource (fusionbot->s.origin, offset, forward, right, upend);
	traceFront = gi.trace(fusionbot->s.origin, fusionbot->mins, fusionbot->maxs, upend, fusionbot, MASK_SOLID|MASK_OPAQUE);

	// using detail brush here ... sometimes it does not pick up ladders...??
	if(traceFront.contents & 0x8000000)
	{
		ucmd->upmove = 400;
		ucmd->forwardmove = 400;
		return qtrue;
	}
	//-------------------------------------------------------

	// If this check fails we need to continue on with more detailed checks
	//-------------------------------------------------------
	if(traceFront.fraction == 1)
	{
		ucmd->forwardmove = 400;
		return qtrue;
	}

	VectorSet(offset, 0, 18, 4);
	G_ProjectSource (fusionbot->s.origin, offset, forward, right, leftstart);

	offset[1] -= 36; // want to make sure this is correct
	G_ProjectSource (fusionbot->s.origin, offset, forward, right, rightstart);

	traceRight = gi.trace(rightstart, NULL, NULL, focalpoint, fusionbot, MASK_SOLID|MASK_OPAQUE);
	traceLeft = gi.trace(leftstart, NULL, NULL, focalpoint, fusionbot, MASK_SOLID|MASK_OPAQUE);
	//-------------------------------------------------------

	// Wall checking code, this will degenerate progressivly so the least cost
	// check will be done first.
	// If open space move ok
	//-------------------------------------------------------
	if(traceRight.fraction != 1 || traceLeft.fraction != 1 || strcmp(traceLeft.ent->classname,"func_door")!=0)
	{
		{
			if(traceRight.fraction > traceLeft.fraction)
				fusionbot->s.angles[YAW] += (1.0 - traceLeft.fraction) * 45.0;
			else
				fusionbot->s.angles[YAW] += -(1.0 - traceRight.fraction) * 45.0;

			ucmd->forwardmove = 400;

			FUSION_MOVEMENT_ChangeBotAngle(fusionbot);

			return qtrue;
		}
	//-------------------------------------------------------
	}

	return qfalse;
}

//===========================================================================
// Routine		: FUSION_MOVEMENT_MoveToGoal
// Description	: Set BOT to move to its movetarget (following node paths)
void FUSION_MOVEMENT_MoveToGoal(edict_t *fusionbot, usercmd_t *ucmd)
{
	// If a rocket or grenade is around deal with it
	// Simple, but effective (could be rewritten to be more accurate)
	//-------------------------------------------------------
	if(strcmp(fusionbot->movetarget->classname,"rocket")==0 ||
	   strcmp(fusionbot->movetarget->classname,"grenade")==0)
	{
		VectorSubtract (fusionbot->movetarget->s.origin, fusionbot->s.origin, fusionbot->move_vector);
		FUSION_MOVEMENT_ChangeBotAngle(fusionbot);
		if(debug_mode)
			gi.dprintf("%s: Oh crap a rocket!\n",fusionbot->client->pers.netname);

		// strafe left/right
		//-------------------------------------------------------
		if(rand()%1 && FUSION_MOVEMENT_CanMove(fusionbot, BOT_MOVE_LEFT))
				ucmd->sidemove = -400;
		else if(FUSION_MOVEMENT_CanMove(fusionbot, BOT_MOVE_RIGHT))
				ucmd->sidemove = 400;
		return;
		//-------------------------------------------------------

	}
	else
	{
		// Set bot's movement direction
		//-------------------------------------------------------
		VectorSubtract (fusionbot->movetarget->s.origin, fusionbot->s.origin, fusionbot->move_vector);
		FUSION_MOVEMENT_ChangeBotAngle(fusionbot);
		ucmd->forwardmove = 400;
		return;
		//-------------------------------------------------------
	}
}

//===========================================================================
// Routine		: FUSION_MOVEMENT_Move
// Description	: Main movement (following node paths)
void FUSION_MOVEMENT_Move(edict_t *fusionbot, usercmd_t *ucmd)
{
	vec3_t dist;
	int current_node_type=-1;
	int next_node_type=-1;
	int i;
	float	distance;

	// Get current and next node back from nav code.
	//-------------------------------------------------------
	if(!FUSION_MAPPING_FollowPath(fusionbot))
	{

		if(!ctf->value)
		{
			fusionbot->state = BOT_STATE_WANDER;
			fusionbot->wander_timeout = level.time + 1.0;
			return;
		}

		if (ctf->value && level.time >= (fusionbot->teamPauseTime))
		{
			fusionbot->state = BOT_STATE_WANDER;
			fusionbot->wander_timeout = level.time + 1.0;
			return;
		}
		else
		{
			fusionbot->state = BOT_STATE_POSITION;
			fusionbot->goal_node = INVALID;
			fusionbot->wander_timeout = level.time + 1.0;
			return;
		}

	}
	//-------------------------------------------------------

	current_node_type = nodes[fusionbot->current_node].type;
	next_node_type = nodes[fusionbot->next_node].type;

	// Move To Goal
	//-------------------------------------------------------
	if (fusionbot->movetarget)
		FUSION_MOVEMENT_MoveToGoal(fusionbot,ucmd);
	//-------------------------------------------------------

	// Grapple
	//-------------------------------------------------------
	if(next_node_type == NODE_GRAPPLE)
	{
		FUSION_MOVEMENT_ChangeBotAngle(fusionbot);
		FUSION_ITEMS_ChangeWeapon(fusionbot,FindItem("grapple"));
		ucmd->buttons = BUTTON_ATTACK;
		return;
	}
	//-------------------------------------------------------

	// Reset the grapple if hangin on a grapple node
	//-------------------------------------------------------
	if(current_node_type == NODE_GRAPPLE)
	{
		CTFPlayerResetGrapple(fusionbot);
		return;
	}
	//-------------------------------------------------------

	// Platforms
	//-------------------------------------------------------
	if(current_node_type != NODE_PLATFORM && next_node_type == NODE_PLATFORM)
	{
		// check to see if lift is down?
		//-------------------------------------------------------
		for(i=0;i<num_items;i++)
			if(item_table[i].node == fusionbot->next_node)
				if(item_table[i].ent->moveinfo.state != PLATFORM_STATE_BOTTOM)
				    return; // Wait for elevator
		//-------------------------------------------------------
	}
	if(current_node_type == NODE_PLATFORM && next_node_type == NODE_PLATFORM)
	{
		// Move to the center
		fusionbot->move_vector[2] = 0; // kill z movement
		if(VectorLength(fusionbot->move_vector) > 10)
			ucmd->forwardmove = 200; // walk to center

		FUSION_MOVEMENT_ChangeBotAngle(fusionbot);

		return; // No move, riding elevator
	}

	// Jumpto Nodes
	//-------------------------------------------------------
	if(next_node_type == NODE_JUMP ||
	  (current_node_type == NODE_JUMP && next_node_type != NODE_ITEM && nodes[fusionbot->next_node].origin[2] > fusionbot->s.origin[2]))
	{
		// Set up a jump move
		ucmd->forwardmove = 400;
		if(fusionbot->groundentity)
			ucmd->upmove = 400;

		FUSION_MOVEMENT_ChangeBotAngle(fusionbot);

		VectorCopy(fusionbot->move_vector,dist);
		VectorScale(dist,440,fusionbot->velocity);

		return;
	}
	//-------------------------------------------------------

	// Item Nodes
	//-------------------------------------------------------
	if(next_node_type == NODE_ITEM)
	{
		FUSION_MOVEMENT_ChangeBotAngle(fusionbot);
	}
	//-------------------------------------------------------

	// Flag Nodes
	// Experimental at this point
	//-------------------------------------------------------
	if((ctf->value) && fusionbot->client->resp.ctf_team == CTF_TEAM2 && next_node_type == NODE_FLAG1)
	{
		FUSION_MOVEMENT_ChangeBotAngle(fusionbot);
	}

	if((ctf->value) && fusionbot->client->resp.ctf_team == CTF_TEAM1 && next_node_type == NODE_FLAG2)
	{
		FUSION_MOVEMENT_ChangeBotAngle(fusionbot);
	}

	if((ctf->value) && fusionbot->client->pers.inventory[ITEMLIST_FLAG2] && next_node_type == NODE_FLAG1)
	{
		FUSION_MOVEMENT_ChangeBotAngle(fusionbot);
	}

	if((ctf->value) && fusionbot->client->pers.inventory[ITEMLIST_FLAG1] && next_node_type == NODE_FLAG2)
	{
		FUSION_MOVEMENT_ChangeBotAngle(fusionbot);
	}
	//-------------------------------------------------------

	// Ladder Nodes
	//-------------------------------------------------------
	if(next_node_type == NODE_LADDER && nodes[fusionbot->next_node].origin[2] > fusionbot->s.origin[2])
	{
		// Otherwise move as fast as we can
		ucmd->forwardmove = 100; // FIXME: 400
		fusionbot->velocity[2] = 320;
		ucmd->sidemove = 0;

		FUSION_MOVEMENT_ChangeBotAngle(fusionbot);

		return;

	}
	//-------------------------------------------------------

	// If getting off the ladder
	//-------------------------------------------------------
	if(current_node_type == NODE_LADDER && next_node_type != NODE_LADDER &&
	   nodes[fusionbot->next_node].origin[2] > fusionbot->s.origin[2])
	{
		ucmd->forwardmove = 400;
		ucmd->upmove = 200;
		fusionbot->velocity[2] = 200;
		FUSION_MOVEMENT_ChangeBotAngle(fusionbot);
		return;
	}
	//-------------------------------------------------------

	// Find the distance vector to the next node
	VectorSubtract( nodes[fusionbot->next_node].origin, fusionbot->s.origin, dist);
	// Lose the vertical component
	dist[2] = 0;
	// Get the absolute length
	distance = VectorLength(dist);

	// LEDGES etc..
	// If trying to jump up a ledge
	//-------------------------------------------------------
	if(
	   (current_node_type == NODE_MOVE &&
	   nodes[fusionbot->next_node].origin[2] > fusionbot->s.origin[2]+16 && distance < NODE_DENSITY)
	   )
	{
		ucmd->forwardmove = 400;
		if(fusionbot->groundentity)
			ucmd->upmove = 400;
		FUSION_MOVEMENT_ChangeBotAngle(fusionbot);
		return;
	}
	//-------------------------------------------------------

	// Water Nodes
	//-------------------------------------------------------
	if(current_node_type == NODE_WATER)
	{
		// We need to be pointed up/down
		FUSION_MOVEMENT_ChangeBotAngle(fusionbot);

		// If the next node is not in the water, then move up to get out.
		if(next_node_type != NODE_WATER && !(gi.pointcontents(nodes[fusionbot->next_node].origin) & MASK_WATER)) // Exit water
			ucmd->upmove = 400;

		ucmd->forwardmove = 300;
		return;

	}
	//-------------------------------------------------------

	// Falling off ledge?
	//-------------------------------------------------------

	if(!fusionbot->groundentity)
	{
		FUSION_MOVEMENT_ChangeBotAngle(fusionbot);

		fusionbot->velocity[0] = fusionbot->move_vector[0] * 360;
		fusionbot->velocity[1] = fusionbot->move_vector[1] * 360;

		return;
	}

	//-------------------------------------------------------

	// Check to see if stuck, and if so try to free us
	// Also handles crouching
	//-------------------------------------------------------
 	if(VectorLength(fusionbot->velocity) < 37)
	{
		// Keep a random factor just in case....
		if(random() > 0.1 && FUSION_MOVEMENT_SpecialMove(fusionbot, ucmd))
			return;

		fusionbot->s.angles[YAW] += random() * 180 - 90;

		FUSION_MOVEMENT_ChangeBotAngle(fusionbot);

		ucmd->forwardmove = 400;

		return;
	}
	//-------------------------------------------------------

	FUSION_MOVEMENT_ChangeBotAngle(fusionbot);

	// Otherwise move as fast as we can
	// If it's safe to move forward (I can't believe FUSION didn't check this!
	//-------------------------------------------------------
	if( FUSION_MOVEMENT_CanMove( fusionbot, BOT_MOVE_FORWARD) || (current_node_type == NODE_LADDER) )
	{
		ucmd->forwardmove = 400;
	}
	else
	{
		// Forget about this route!
		fusionbot->movetarget = NULL;
	}
	//-------------------------------------------------------
}

//===========================================================================
// Routine		: FUSION_MOVEMENT_Wander
// Description	: Wandering code (needs to be made a little smarter)
void FUSION_MOVEMENT_Wander(edict_t *fusionbot, usercmd_t *ucmd)
{
	vec3_t  temp;

	// Do not move
	//-------------------------------------------------------
	if(fusionbot->next_move_time > level.time)
		return;
	//-------------------------------------------------------

	// Special check for elevators, stand still until the ride comes to a complete stop.
	//-------------------------------------------------------
	if(fusionbot->groundentity != NULL && fusionbot->groundentity->use == Use_Plat)
		if(fusionbot->groundentity->moveinfo.state == PLATFORM_STATE_UP ||
		   fusionbot->groundentity->moveinfo.state == PLATFORM_STATE_DOWN)
		{
			fusionbot->velocity[0] = 0;
			fusionbot->velocity[1] = 0;
			fusionbot->velocity[2] = 0;
			fusionbot->next_move_time = level.time + 0.5;
			return;
		}
	//-------------------------------------------------------

	// Is there a target to move to
	//-------------------------------------------------------
	if (fusionbot->movetarget)
		FUSION_MOVEMENT_MoveToGoal(fusionbot,ucmd);
	//-------------------------------------------------------

	// Swimming?
	//-------------------------------------------------------
	VectorCopy(fusionbot->s.origin,temp);
	temp[2]+=24;

	if(gi.pointcontents (temp) & MASK_WATER)
	{
		// If drowning and no node, move up
		//-------------------------------------------------------
		if(fusionbot->client->next_drown_time > 0)
		{
			ucmd->upmove = 1;
			fusionbot->s.angles[PITCH] = -45;
		}
		else
			ucmd->upmove = 15;

		ucmd->forwardmove = 300;
	}

	// Lava?
	//-------------------------------------------------------
	temp[2]-=48;
	if(gi.pointcontents(temp) & (CONTENTS_LAVA|CONTENTS_SLIME|CONTENTS_DONOTENTER))
	{
		fusionbot->s.angles[YAW] += random() * 360 - 180;
		ucmd->forwardmove = 400;
		if(fusionbot->groundentity)
			ucmd->upmove = 400;
		else
			ucmd->upmove = 0;
		return;
	}
	//-------------------------------------------------------

	// Check for special movement if we have a normal move (have to test)
	//-------------------------------------------------------
 	if(VectorLength(fusionbot->velocity) < 37)
	{
		if(random() > 0.1 && FUSION_MOVEMENT_SpecialMove(fusionbot,ucmd))
			return;

		fusionbot->s.angles[YAW] += random() * 180 - 90;

		// if there is ground continue otherwise wait for next move
		if(!M_CheckBottom && !fusionbot->groundentity)
			ucmd->forwardmove = 0; //0
		else if( FUSION_MOVEMENT_CanMove( fusionbot, BOT_MOVE_FORWARD))
			ucmd->forwardmove = 200;

		return;
	}
	//-------------------------------------------------------

	// Otherwise move as fast as we can
	// If it's safe to move forward (I can't believe FUSION_ didn't check this!
	//-------------------------------------------------------
	if( FUSION_MOVEMENT_CanMove( fusionbot, BOT_MOVE_FORWARD))
	{
		ucmd->forwardmove = 400;
	}
	else
	{
		// Need a "findbestdirection" routine in here
		// Forget about this route!
		ucmd->forwardmove = -100;
		fusionbot->movetarget = NULL;
	}

}

//===========================================================================
// Routine		: FUSION_MOVEMENT_Attack
// Description	: Attack movement
//				: Just basic move and avoidance
// TODO			: Change routine for more advanced attack movement
void FUSION_MOVEMENT_Attack (edict_t *fusionbot, usercmd_t *ucmd)
{
	float c, firedelay;
	vec3_t  target;
	vec3_t  angles;
	vec3_t	attackvector;
	float	dist;

	// Randomly choose a movement direction
	//-------------------------------------------------------
	c = random();

	if(c < 0.2 && FUSION_MOVEMENT_CanMove(fusionbot,BOT_MOVE_LEFT))
		ucmd->sidemove -= 400;
	else if(c < 0.4 && FUSION_MOVEMENT_CanMove(fusionbot,BOT_MOVE_RIGHT))
		ucmd->sidemove += 400;
	else if(c < 0.6 && FUSION_MOVEMENT_CanMove(fusionbot,BOT_MOVE_FORWARD))
		ucmd->forwardmove += 400;
	else if(c < 0.8 && FUSION_MOVEMENT_CanMove(fusionbot,BOT_MOVE_FORWARD))
		ucmd->forwardmove -= 400;
	//-------------------------------------------------------

	VectorSubtract( fusionbot->s.origin, fusionbot->enemy->s.origin, attackvector);
	dist = VectorLength( attackvector);

	if(dist < 200)
	{
		// Randomly choose a vertical movement direction
		//-------------------------------------------------------
		c = (random()-0.5)*5;

		if(fusionbot->groundentity)
		{
			if(c < 3.0) //FIXME: 2.0
				ucmd->upmove += 200;
			else
				ucmd->upmove -= 200;
		}
		//-------------------------------------------------------
	}
	if(dist < 75)
	{
		ucmd->forwardmove -= 200;
	}

	// Set the attack
	//-------------------------------------------------------
	firedelay = (random()-0.5)*5;

	// Added Skill Level Settings for Attack
	if (fusionbot->skillchoice == 0)
	{
		if (firedelay < 0.10)
		{
			ucmd->buttons = BUTTON_ATTACK;
		}
		else
		{
		}
	}
	else if (fusionbot->skillchoice == 1)
	{
		if (firedelay < 0.20)
		{
			ucmd->buttons = BUTTON_ATTACK;
		}
		else
		{
		}
	}
	else if (fusionbot->skillchoice == 2)
	{
		if (firedelay < 0.40)
		{
			ucmd->buttons = BUTTON_ATTACK;
		}
		else
		{
		}
	}
	else if (fusionbot->skillchoice >= 3)
	{
		ucmd->buttons = BUTTON_ATTACK;
	}
	//-------------------------------------------------------

	// Aim
	//-------------------------------------------------------
	VectorCopy(fusionbot->enemy->s.origin,target);
	//-------------------------------------------------------

	// Modify attack angles based on accuracy
	// added skill level
	//-------------------------------------------------------

	if (fusionbot->skillchoice == 0)
	{
		// Can't hit a good strafing player
		target[0] += (random()-0.5) * 20;
		target[1] += (random()-0.5) * 20;
	}
	else if (fusionbot->skillchoice == 1)
	{
		// Will get ya every now and then
		target[0] += (random()-0.5) * 10;
		target[1] += (random()-0.5) * 10;
	}
	else if (fusionbot->skillchoice == 2)
	{
		// Good competetion for some of the best players
		target[0] += (random()-0.5) * 5;
		target[1] += (random()-0.5) * 5;
	}
	else if (fusionbot->skillchoice >= 3)
	{
		// Never misses, unloads on ya!
		target[0] = target[0];
		target[1] = target[1];
	}
	//-------------------------------------------------------

	// Set direction
	//-------------------------------------------------------
	VectorSubtract (target, fusionbot->s.origin, fusionbot->move_vector);
	VecToAngles (fusionbot->move_vector, angles); //qbism- switch to VecToAngles
	VectorCopy(angles,fusionbot->s.angles);
	//-------------------------------------------------------

	fusionbot->teamPauseTime = level.time;
}

/* REM New Pathing Algorithm
//==========================
// FUSION_PA_PathMove
//==========================
//
qboolean	FUSION_PA_PathMove( edict_t *self )
{
	node_t *temp = &nodes[self->current_node];	// For checking our position

	if( level.time == (float)((int)level.time) )
	{
		if(
			!FUSION_PA_LinkExists( self->current_node, FUSION_PA_SLFront(&self->pathList) )
			&& ( self->current_node != FUSION_PA_SLFront(&self->pathList) )
			)
		{
			// We are off the path - clear out the lists
			FUSION_PA_InitSearch( self );
		}
	}

	// Boot in our new pathing algorithm
	// This will fill ai.pathList with the information we need
	if(
		FUSION_PA_SLEmpty(&self->pathList)				// We have no path and
		&& (self->current_node != self->goal_node)	// we're not at our destination
		)
	{
		if( !FUSION_PA_StartSearch( self, self->current_node, self->goal_node))	// Set up our pathList
		{
			// Failed to find a path
//			gi.bprintf(PRINT_HIGH,"%s: Target at(%i) - No Path \n",
//				self->client->pers.netname, self->goal_node, self->next_node);
			return qfalse;
		}
		return qtrue;
	}

	// And change our pathfinding to use it
	if ( !FUSION_PA_SLEmpty(&self->pathList) )								// We have a path
	{
//		if( AtNextNode( self->s.origin, FUSION_PA_SLFront(&self->pathList))	// we're at the nextnode
		if( FUSION_NAVIGATION_FindClosestReachableNode(self,NODE_DENSITY*0.66,NODE_ALL) == FUSION_PA_SLFront(&self->pathList)	// we're at the nextnode
			|| self->current_node == FUSION_PA_SLFront(&self->pathList)
		)
		{
			self->current_node = FUSION_PA_SLFront(&self->pathList);		// Show we're there
			FUSION_PA_SLPop_Front(&self->pathList);				// Remove the top node
		}
		if( !FUSION_PA_SLEmpty(&self->pathList) )
			self->next_node = FUSION_PA_SLFront(&self->pathList);	// Set our next destination
		else
			self->next_node = INVALID;				// We're at the target location
		return qtrue;
	}
	return qtrue;	// Pathlist is emptyand we are at our destination
}
*/
