/*

NAVIGATION FUNCTION LISTING

FUSION_NAVIGATION_FindCost
FUSION_NAVIGATION_FindCloseReachableNode
FUSION_NAVIGATION_FindClosestReachableNode
FUSION_NAVIGATION_SetGoal
FUSION_NAVIGATION_SetFlagGoal
FUSION_NAVIGATION_FindFlagNode

*/
/*
  _______              .__              __  .__
  \      \ _____ ___  _|__| _________ _/  |_|__| ____   ____
  /   |   \\__  \\  \/ /  |/ ___\__  \\   __\  |/  _ \ /    \
 /    |    \/ __ \\   /|  / /_/  > __ \|  | |  (  <_> )   |  \
 \____|__  (____  /\_/ |__\___  (____  /__| |__|\____/|___|  /
         \/     \/       /_____/     \/                    \/
	fusionbot_navigation.c
	Fusion Navigation
================================================================
Should you decide to release a modified version of FUSION, you MUST
include the following text in the documentation and source of your
modification.

---===[BEGIN TEXT]===---
================================================================
___________           .__                __________        __
\_   _____/_ __  _____|__| ____   ____   \______   \ _____/  |_
 |    __)|  |  \/  ___/  |/  _ \ /    \   |    |  _//  _ \   __\
 |     \ |  |  /\___ \|  (  <_> )   |  \  |    |   (  <_> )  |
 \___  / |____//____  >__|\____/|___|  /  |______  /\____/|__|
     \/             \/               \/          \/  Version 2.0
================================================================
Fusion v2.0
Copyright (c) 1998-2003 Trenton Shaffer.
All rights reserved.

ACEBOT Copyright(c) 1998, Steve Yeager.
EraserBot Copyright(c) 1998, Ryan Feltrin.
QFusion is Copyright (C) 2002-2003 Victor Luchits.
All other files are Copyright(C) 1997-2001 Id Software, Inc.
================================================================
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
================================================================
---===[END TEXT]===---
*/

#include "..\game\g_local.h"
#include "fusionbot.h"

// Total number of nodes that are items
//-------------------------------------------------------
int number_of_itemnodes;

// Total number of nodes
//-------------------------------------------------------
int number_of_nodes;

// array for node data
//-------------------------------------------------------
node_t nodes[MAX_NODES];
short int path_table[MAX_NODES][MAX_NODES];

//===========================================================================
// Routine		: FUSION_NAVIGATION_FindCost
// Description	: Determine cost of moving from one node to another
int FUSION_NAVIGATION_FindCost(int from, int to)
{
	int curnode;
	int cost=1; // Shortest possible is 1

	// If we can't get there then return INVALID
	//-------------------------------------------------------
	if (path_table[from][to] == INVALID)
		return INVALID;
	//-------------------------------------------------------

	// Check the path and return the cost
	//-------------------------------------------------------
	curnode = path_table[from][to];
	//-------------------------------------------------------

	// Find a path (linear time, very fast)
	//-------------------------------------------------------
	while(curnode != to)
	{
		curnode = path_table[curnode][to];
		// Something has corrupted the path ... abort
		if(curnode == INVALID)
			return INVALID;
		cost++;
	}
	//-------------------------------------------------------

	return cost;
}

//===========================================================================
// Routine		: FUSION_NAVIGATION_FindCloseReachableNode
// Description	: Find a close node to the player within given distance
//				: Faster than looking for the closest node, but not very accurate
int FUSION_NAVIGATION_FindCloseReachableNode(edict_t *fusionbot, int range, int type)
{
	vec3_t		v;
	int			i;
	trace_t		tr;
	float		dist;

	vec3_t	maxs, mins;

	VectorCopy(fusionbot->mins,mins);
	mins[2] += 16;
	VectorCopy(fusionbot->maxs,maxs);
	maxs[2] -= 16;

	range *= range;

	for(i=0;i<number_of_nodes;i++)
	{
		// Check Node Type
		if(type == NODE_ALL || type == nodes[i].type)
		{

			VectorSubtract(nodes[i].origin,fusionbot->s.origin,v);

			dist = v[0]*v[0] + v[1]*v[1] + v[2]*v[2];

			// Square Range instead of sqrt
			if(dist < range)
			{
				// make sure it is visible
				//-------------------------------------------------------
				//tr = gi.trace (fusionbot->s.origin, fusionbot->mins, fusionbot->maxs, nodes[i].origin, fusionbot, MASK_SOLID|MASK_OPAQUE);
				tr = gi.trace (fusionbot->s.origin, mins, maxs, nodes[i].origin, fusionbot, MASK_ALL);

				if(tr.fraction == 1.0)
					return i;
				//-------------------------------------------------------
			}
		}
	}

	return -1;
}

//===========================================================================
// Routine		: FUSION_NAVIGATION_FindClosestReachableNode
// Description	: Find the closest node to the player within a certain range
int FUSION_NAVIGATION_FindClosestReachableNode(edict_t *fusionbot, int range, int type)
{
	int			i;
	float		closest = 99999;
	float		dist;
	int			node=-1;
	vec3_t		v;
	trace_t		tr;
	float		rng;
	vec3_t		maxs,mins;

	VectorCopy(fusionbot->mins,mins);
	VectorCopy(fusionbot->maxs,maxs);

	// For Ladders, do not worry so much about reachability
	//-------------------------------------------------------
	if(type == NODE_LADDER)
	{
		VectorCopy(vec3_origin,maxs);
		VectorCopy(vec3_origin,mins);
	}
	else
	{
		// Stepsize
		mins[2] += 18;
		// Duck slightly
		maxs[2] -= 16;
	}
	//-------------------------------------------------------

	// Square Range for distance comparison (eliminated sqrt)
	rng = (float)(range * range);

	for(i=0;i<number_of_nodes;i++)
	{
		// Check Node Type
		if(type == NODE_ALL || type == nodes[i].type)
		{
			VectorSubtract(nodes[i].origin, fusionbot->s.origin,v);

			dist = v[0]*v[0] + v[1]*v[1] + v[2]*v[2];

			if(dist < closest && dist < rng)
			{
				// make sure it is visible
				//-------------------------------------------------------
				tr = gi.trace(fusionbot->s.origin, mins, maxs, nodes[i].origin, fusionbot, MASK_SOLID|MASK_OPAQUE);
				if(tr.fraction == 1.0)
				{
					node = i;
					closest = dist;
				}
				//-------------------------------------------------------
			}
		}
	}

	return node;
}

//===========================================================================
// Routine		: FUSION_NAVIGATION_SetGoal
// Description	: Setup the goal
void FUSION_NAVIGATION_SetGoal(edict_t *fusionbot, int goal_node)
{
	int			node;

	fusionbot->goal_node = goal_node;

	node = FUSION_NAVIGATION_FindClosestReachableNode(fusionbot, NODE_DENSITY*3, NODE_ALL);

	if(node == -1)
		return;

	if(debug_mode)
		gi.dprintf("%s new start node selected %d\n",fusionbot->client->pers.netname,node);


	fusionbot->current_node = node;
	 // make sure bot gets to nearest node first
	fusionbot->next_node = fusionbot->current_node;
	fusionbot->node_timeout = 0;

}

//===========================================================================
// Routine		: FUSION_NAVIGATION_SetFlagGoal
// Description	: Setup the goal
void FUSION_NAVIGATION_SetFlagGoal(edict_t *fusionbot, int goal_node)
{
	int			node;

	fusionbot->goal_node = goal_node;
	if((ctf->value) && fusionbot->client->pers.inventory[ITEMLIST_FLAG2])
	{
		node = FUSION_NAVIGATION_FindClosestReachableNode(fusionbot, NODE_DENSITY*10, NODE_FLAG1);
	}
	else if((ctf->value) && fusionbot->client->pers.inventory[ITEMLIST_FLAG1])
	{
		node = FUSION_NAVIGATION_FindClosestReachableNode(fusionbot, NODE_DENSITY*10, NODE_FLAG2);
	}
	else
		node = FUSION_NAVIGATION_FindClosestReachableNode(fusionbot, NODE_DENSITY, NODE_ALL);

	if(node == -1)
		return;

	if(debug_mode)
		gi.dprintf("%s new start node selected %d\n",fusionbot->client->pers.netname,node);


	fusionbot->current_node = node;
	 // make sure bot gets to nearest node first
	fusionbot->next_node = fusionbot->current_node;
	fusionbot->node_timeout = 0;

}

//===========================================================================
// Routine		: FUSION_NAVIGATION_FindFlagNode
// Description	: Find the closest node to the player within a certain range
int FUSION_NAVIGATION_FindFlagNode(edict_t *fusionbot, int range, int type)
{
	int			i;
	float		closest = 99999;
	float		dist;
	int			node=-1;
	vec3_t		v;
	trace_t		tr;
	float		rng;
	vec3_t		maxs,mins;

	VectorCopy(fusionbot->mins,mins);
	VectorCopy(fusionbot->maxs,maxs);

	// Square Range for distance comparison (eliminated sqrt)
	rng = (float)(range * range);

	for(i=0;i<number_of_nodes;i++)
	{
		// Check Node Type
		if((type == NODE_FLAG1) || (type == NODE_FLAG2))
		{
			VectorSubtract(nodes[i].origin, fusionbot->s.origin,v);

			dist = v[0]*v[0] + v[1]*v[1] + v[2]*v[2];

			if(dist < closest && dist < rng)
			{
				// make sure it is visible
				//-------------------------------------------------------
				tr = gi.trace(fusionbot->s.origin, mins, maxs, nodes[i].origin, fusionbot, MASK_SOLID|MASK_OPAQUE);
				if(tr.fraction == 1.0)
				{
					node = i;
					closest = dist;
				}
				//-------------------------------------------------------
			}
		}

	}

	return node;
}
