/*

NODES FUNCTION LISTING

FUSION_NODES_AddNode
FUSION_NODES_UpdateNodeEdge
FUSION_NODES_RemoveNodeEdge
FUSION_NODES_ResolveAllPaths
FUSION_NODES_SaveNodes
FUSION_NODES_SaveNodesAgain
FUSION_NODES_LoadNodes

*/
/*
                   .___
   ____   ____   __| _/____   ______
  /    \ /  _ \ / __ |/ __ \ /  ___/
 |   |  (  <_> ) /_/ \  ___/ \___ \
 |___|  /\____/\____ |\___  >____  >
      \/            \/    \/     \/

	fusionbot_nodes.c
	Pathing routines for the FUSION bot.
================================================================
Should you decide to release a modified version of FUSION, you MUST
include the following text in the documentation and source of your
modification.

---===[BEGIN TEXT]===---
================================================================
___________           .__                __________        __
\_   _____/_ __  _____|__| ____   ____   \______   \ _____/  |_
 |    __)|  |  \/  ___/  |/  _ \ /    \   |    |  _//  _ \   __\
 |     \ |  |  /\___ \|  (  <_> )   |  \  |    |   (  <_> )  |
 \___  / |____//____  >__|\____/|___|  /  |______  /\____/|__|
     \/             \/               \/          \/  Version 2.0
================================================================
Fusion v2.0
Copyright (c) 1998-2003 Trenton Shaffer.
All rights reserved.

ACEBOT Copyright(c) 1998, Steve Yeager.
EraserBot Copyright(c) 1998, Ryan Feltrin.
QFusion is Copyright (C) 2002-2003 Victor Luchits.
All other files are Copyright(C) 1997-2001 Id Software, Inc.
================================================================
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
================================================================
---===[END TEXT]===---
*/
#include "..\game\g_local.h"
#include "fusionbot.h"

// Flags
//-------------------------------------------------------
qboolean newmap=qtrue;

// Total number of nodes that are items
//-------------------------------------------------------
int number_of_itemnodes;

// Total number of nodes
//-------------------------------------------------------
int number_of_nodes;

// array for node data
//-------------------------------------------------------
node_t nodes[MAX_NODES];
short int path_table[MAX_NODES][MAX_NODES];

//===========================================================================
// Routine		: FUSION_NODES_AddNode
// Description	: Adds a node of specified type
//
// Types		: 0 - Move
//				: 1 - Ladder
//				: 2 - Platform
//				: 3 - Teleporter
//				: 4 - Item/Weapon
//				: 5 - Water
//				: 6 - Grapple
//				: 7 - Jump
int FUSION_NODES_AddNode(edict_t *fusionbot, int type)
{
	vec3_t		v1,v2;
	int			i;

	// Block if we exceed maximum
	//-------------------------------------------------------
	if (number_of_nodes + 1 > MAX_NODES)
		return qfalse;
	//-------------------------------------------------------

	// Set location
	//-------------------------------------------------------
	VectorCopy(fusionbot->s.origin,nodes[number_of_nodes].origin);
	//-------------------------------------------------------

	// Set type and node number
	//-------------------------------------------------------
	nodes[number_of_nodes].type = type;
	nodes[number_of_nodes].nodenum = number_of_nodes;
	//-------------------------------------------------------

	for( i = 0; i<MAXLINKS; i++)
	{
		nodes[number_of_nodes].links[i].targetNode = INVALID;
	}

	// ITEMS
	// Move the z location up just a bit.
	//-------------------------------------------------------
	if(type == NODE_ITEM)
	{
		nodes[number_of_nodes].origin[2] += 16;
		number_of_itemnodes++;
	}
	//-------------------------------------------------------

	// TELEPORTERS
	//-------------------------------------------------------
	if(type == NODE_TELEPORTER)
	{
		// Up 32
		nodes[number_of_nodes].origin[2] += 32;
	}

	if(type == NODE_FLAG1)
	{
		// Up 8
		nodes[number_of_nodes].origin[2] += 8;
	}

	if(type == NODE_FLAG2)
	{
		// Up 8
		nodes[number_of_nodes].origin[2] += 8;
	}
	//-------------------------------------------------------

	// JUMP PAD
	//-------------------------------------------------------
	if(type == NODE_JUMP)
	{
		// Up 32
		nodes[number_of_nodes].origin[2] += 16;
	}
	//-------------------------------------------------------

	// LADDERS
	//-------------------------------------------------------
	if(type == NODE_LADDER)
	{
		nodes[number_of_nodes].type = NODE_LADDER;

		number_of_nodes++;
		// Return the Node added
		return number_of_nodes-1;

	}
	//-------------------------------------------------------

	// PLATFORMS
	// For platforms drop two nodes one at top, one at bottom
	//-------------------------------------------------------
	if(type == NODE_PLATFORM)
	{
		VectorCopy(fusionbot->maxs,v1);
		VectorCopy(fusionbot->mins,v2);

		// To get the center
		nodes[number_of_nodes].origin[0] = (v1[0] - v2[0]) / 2 + v2[0];
		nodes[number_of_nodes].origin[1] = (v1[1] - v2[1]) / 2 + v2[1];
		nodes[number_of_nodes].origin[2] = fusionbot->maxs[2];

		if(debug_mode)
			FUSION_MAPPING_ShowNode(number_of_nodes);

		number_of_nodes++;

		nodes[number_of_nodes].origin[0] = nodes[number_of_nodes-1].origin[0];
		nodes[number_of_nodes].origin[1] = nodes[number_of_nodes-1].origin[1];
		nodes[number_of_nodes].origin[2] = fusionbot->mins[2]+64;

		nodes[number_of_nodes].type = NODE_PLATFORM;

		// Add a link
		FUSION_NODES_UpdateNodeEdge(number_of_nodes,number_of_nodes-1);

		number_of_nodes++;

		return number_of_nodes -1;
	}
	//-------------------------------------------------------

	// DEBUG
	//-------------------------------------------------------
	if(debug_mode)
	{
		if(nodes[number_of_nodes].type == NODE_MOVE)
			gi.dprintf("Node added %d type: Move\n",number_of_nodes);
		else if(nodes[number_of_nodes].type == NODE_TELEPORTER)
			gi.dprintf("Node added %d type: Teleporter\n",number_of_nodes);
		else if(nodes[number_of_nodes].type == NODE_LADDER)
			gi.dprintf("Node added %d type: Ladder\n",number_of_nodes);
		else if(nodes[number_of_nodes].type == NODE_PLATFORM)
			gi.dprintf("Node added %d type: Platform\n",number_of_nodes);
		else if(nodes[number_of_nodes].type == NODE_JUMP)
			gi.dprintf("Node added %d type: Jump\n",number_of_nodes);
		else if(nodes[number_of_nodes].type == NODE_ITEM)
			gi.dprintf("Node added %d type: Item\n",number_of_nodes);
		else if(nodes[number_of_nodes].type == NODE_FLAG1)
			gi.dprintf("Node added %d type: Red Flag\n",number_of_nodes);
		else if(nodes[number_of_nodes].type == NODE_FLAG2)
			gi.dprintf("Node added %d type: Blue Flag\n",number_of_nodes);
		else if(nodes[number_of_nodes].type == NODE_WATER)
			gi.dprintf("Node added %d type: Water\n",number_of_nodes);
		else if(nodes[number_of_nodes].type == NODE_GRAPPLE)
			gi.dprintf("Node added %d type: Grapple\n",number_of_nodes);

		FUSION_MAPPING_ShowNode(number_of_nodes);
	}
	//-------------------------------------------------------

	number_of_nodes++;

	return number_of_nodes-1; // return the node added
}

//===========================================================================
// Routine		: FUSION_NODES_UpdateNodeEdge
// Description	: Add/Update node connections (paths)
void FUSION_NODES_UpdateNodeEdge(int from, int to)
{
	int			i;

	if(from == -1 || to == -1 || from == to)
		return; // safety

	for( i=0; i<MAXLINKS; i++)
	{
		if (nodes[from].links[i].targetNode == to)
		{
			if(debug_mode)
			{
				gi.dprintf("ABORTED!!!!  Link %d -> %d\n", from, to);
			}
			return;
		}
	}

	// do not allow creation of nodes wehre falling distance will kill
	if( (nodes[from].origin[2]) > (nodes[to].origin[2] + 180) )
		return;

	// Add the link
	//-------------------------------------------------------
	path_table[from][to] = to;

	for( i=0; i<MAXLINKS; i++)
	{
		if( nodes[from].links[i].targetNode == to)
			break;
		if( nodes[from].links[i].targetNode == INVALID)
		{
			vec3_t		v;
			float		thisCost;

			VectorSubtract (nodes[from].origin, nodes[to].origin, v);
			thisCost = VectorLength(v);
			nodes[from].links[i].targetNode = to;
			nodes[from].links[i].cost = thisCost;
			break;
		}
	}

	// Now for the fusionbot-referencing part, linear time for each link added
	//-------------------------------------------------------
	for(i=0;i<number_of_nodes;i++)
		if(path_table[i][from] != INVALID)
			if(i == to)
				// Make sure we terminate
				path_table[i][to] = INVALID;
			else
				path_table[i][to] = path_table[i][from];

	if(debug_mode)
		gi.dprintf("Link %d -> %d\n", from, to);

	// REVERSE -- Can't believe Riever, nor anyone else figured out how
	// to properly reverse nodes this way. Much easier and smarter algorithm.

	// Now that we have a link, reverse it
	if(from == -1 || to == -1 || to == from)
		return; // safety


	for( i=0; i<MAXLINKS; i++)
	{
		if (nodes[to].links[i].targetNode == from)
		{
			if(debug_mode)
			{
				gi.dprintf("ABORTED!!!!  ReverseLink %d -> %d\n", from, to);
			}
			return;
		}
	}

	// do not allow creation of nodes wehre falling distance will kill
	if( (nodes[to].origin[2]) > (nodes[from].origin[2] + 180) )
		return;

	// Add the link
	//-------------------------------------------------------
	path_table[to][from] = from;

	for( i=0; i<MAXLINKS; i++)
	{
		if( nodes[to].links[i].targetNode == from)
			break;
		if( nodes[to].links[i].targetNode == INVALID)
		{
			vec3_t		v;
			float		thisCost;

			VectorSubtract (nodes[to].origin, nodes[from].origin, v);
			thisCost = VectorLength(v);
			nodes[to].links[i].targetNode = from;
			nodes[to].links[i].cost = thisCost;
			break;
		}
	}


	// Now for the fusionbot-referencing part, linear time for each link added
	//-------------------------------------------------------
	for(i=0;i<number_of_nodes;i++)
		if(path_table[i][to] != INVALID)
			if(i == from)
				// Make sure we terminate
				path_table[i][from] = INVALID;
			else
				path_table[i][from] = path_table[i][to];

	if(debug_mode)
		gi.dprintf("ReverseLink %d -> %d\n", to, from);
}

//===========================================================================
// Routine		: FUSION_NODES_RemoveNodeEdge
// Description	: Remove a node edge (debug only)
void FUSION_NODES_RemoveNodeEdge(edict_t *fusionbot, int from, int to)
{
	int			i;

	if(debug_mode)
		gi.dprintf("%s: Removing Edge %d -> %d\n", fusionbot->client->pers.netname, from, to);

	// Set to invalid
	path_table[from][to] = INVALID;

	// Make sure this gets updated in our path array
	//-------------------------------------------------------
	for(i=0;i<number_of_nodes;i++)
		if(path_table[from][i] == to)
			path_table[from][i] = INVALID;
}

//===========================================================================
// Routine		: FUSION_NODES_ResolveAllPaths
// Description	: Resolves all paths that are incomplete
//				: Call before saving to disk
void FUSION_NODES_ResolveAllPaths()
{
	int			i, from, to;
	int			num=0;

	gi.bprintf(PRINT_HIGH,"Resolving all paths...");

	for(from=0;from<number_of_nodes;from++)
	{
		for(to=0;to<number_of_nodes;to++)
		{
			// update unresolved paths
			// Not equal to itself, not equal to -1 and equal to the last link
			//-------------------------------------------------------
			if(from != to && path_table[from][to] == to)
			{
				num++;

				// Now for the self-referencing part linear time for each link added
				//-------------------------------------------------------
				for(i=0;i<number_of_nodes;i++)
					if(path_table[i][from] != -1)
						if(i == to)
							// Make sure we terminate
							path_table[i][to] = -1;
						else
							path_table[i][to] = path_table[i][from];
			}
			//-------------------------------------------------------
		}
	}

	gi.bprintf(PRINT_MEDIUM,"done (%d updated)\n",num);
}

//===========================================================================
// Routine		: FUSION_NODES_SaveNodes
// Description	: Save nodes to .nod file on disk
void FUSION_NODES_SaveNodes()
{
	FILE		*pOut;
	char		filename[60];
	int			i,j;
	int			version = 1;
	cvar_t	*game; //qbism use game or base dir

	game = gi.cvar ("fs_game", "", 0);
	if (!*game->string) game = gi.cvar ("fs_basedir", "", 0);

	if(!dorebuild)
	{
		gi.bprintf(PRINT_MEDIUM, "Nodes not saved, Node Table already exists!");
		return;
	}

	// Resolve paths
	//-------------------------------------------------------
	FUSION_NODES_ResolveAllPaths();

	gi.bprintf(PRINT_MEDIUM,"Saving node table...");

	sprintf (filename, "%s/bot_config/%s.nod", game->string,level.mapname);

	if((pOut = fopen(filename, "wb" )) == NULL)
		return; // bail

	// Write Version
	fwrite(&version,sizeof(int),1,pOut);
	// Write Count
	fwrite(&number_of_nodes,sizeof(int),1,pOut);
	// Write Facts Count
	fwrite(&num_items,sizeof(int),1,pOut);

	// Write Nodes
	fwrite(nodes,sizeof(node_t),number_of_nodes,pOut);

	for(i=0;i<number_of_nodes;i++)
		for(j=0;j<number_of_nodes;j++)
			// Write Count
			fwrite(&path_table[i][j],sizeof(short int),1,pOut);

	// Write out fact table
	fwrite(item_table,sizeof(item_table_t),num_items,pOut);

	fclose(pOut);

	gi.bprintf(PRINT_MEDIUM,"done.\n");
}

//===========================================================================
// Routine		: FUSION_NODES_SaveNodesAgain
// Description	: Save nodes to .nod file on disk
void FUSION_NODES_SaveNodesAgain()
{
	FILE		*pOut;
	char		filename[60];
	int			i,j;
	int			version = 1;
	cvar_t	*game; //qbism use game or base dir

	game = gi.cvar ("fs_game", "", 0);
	if (!*game->string) game = gi.cvar ("fs_basedir", "", 0);

	if(!dorebuild)
	{
		gi.bprintf(PRINT_MEDIUM, "Node Table already exists!\nRe-Writing Node Table");
	}

	// Resolve paths
	//-------------------------------------------------------
	FUSION_NODES_ResolveAllPaths();

	gi.bprintf(PRINT_MEDIUM,"Saving node table...");

	sprintf (filename, "%s/bot_config/%s.nod", game->string,level.mapname);

	// Bail
	if((pOut = fopen(filename, "wb" )) == NULL)
		return;

	// Write Version
	fwrite(&version,sizeof(int),1,pOut);
	// Write Count
	fwrite(&number_of_nodes,sizeof(int),1,pOut);
	// Write Facts Count
	fwrite(&num_items,sizeof(int),1,pOut);

	// Write Nodes
	fwrite(nodes,sizeof(node_t),number_of_nodes,pOut);

	for(i=0;i<number_of_nodes;i++)
		for(j=0;j<number_of_nodes;j++)
			// Write Count
			fwrite(&path_table[i][j],sizeof(short int),1,pOut);

	// Write out the fact table
	fwrite(item_table,sizeof(item_table_t),num_items,pOut);

	fclose(pOut);

	gi.bprintf(PRINT_MEDIUM,"done.\n");
}

//===========================================================================
// Routine		: FUSION_NODES_LoadNodes
// Description	: Loads nodes from .nod file on disk
void FUSION_NODES_LoadNodes(void)
{
	FILE		*pIn;
	int			i,j;
	char		filename[60];
	int			version;
	cvar_t	*game; //qbism use game or base dir

	game = gi.cvar ("fs_game", "", 0);
	if (!*game->string) game = gi.cvar ("fs_basedir", "", 0);

	sprintf (filename, "%s/bot_config/%s.nod", game->string,level.mapname);


	if((pIn = fopen(filename, "rb" )) == NULL)
    {
		// Create item table
		//-------------------------------------------------------
		gi.bprintf(PRINT_MEDIUM, "FUSION BOT: No node file found, creating new one...");
		FUSION_ITEMS_BuildItemNodeTable(qfalse);
		dorebuild = qtrue;
		gi.bprintf(PRINT_MEDIUM, "done.\n");
		//-------------------------------------------------------
		return;
	}

	// determine version
	//-------------------------------------------------------
	// Read Version
	fread(&version,sizeof(int),1,pIn);

	if(version == 1)
	{
		gi.bprintf(PRINT_MEDIUM,"FUSION BOT: Loading node table...");

		// Read Count
		fread(&number_of_nodes,sizeof(int),1,pIn);
		// Read Facts Count
		fread(&num_items,sizeof(int),1,pIn);

		fread(nodes,sizeof(node_t),number_of_nodes,pIn);

		for(i=0;i<number_of_nodes;i++)
			for(j=0;j<number_of_nodes;j++)
				// Write Count
				fread(&path_table[i][j],sizeof(short int),1,pIn);

		fread(item_table,sizeof(item_table_t),num_items,pIn);
		fclose(pIn);
	}
	else
	{
		// Create item table
		//-------------------------------------------------------
		gi.bprintf(PRINT_MEDIUM, "FUSION BOT: No node file found, creating new one...");
		FUSION_ITEMS_BuildItemNodeTable(qfalse);
		dorebuild = qtrue;
		gi.bprintf(PRINT_MEDIUM, "done.\n");
		//-------------------------------------------------------
		return; // bail
	}

	gi.bprintf(PRINT_MEDIUM, "done.\n");

	FUSION_ITEMS_BuildItemNodeTable(qtrue);
	dorebuild = qfalse;
}

