/*

PATHING FUNCTION LISTING

FUSION_PA_InitSearch
FUSION_PA_StartSearch
FUSION_PA_QuickPath
FUSION_PA_FindPath
FUSION_PA_LinkExists
FUSION_PA_SLPush_Front
FUSION_PA_SLPop_Front
FUSION_PA_SLFront
FUSION_PA_SLPush_Back
FUSION_PA_SLEmpty
FUSION_PA_SLDelete

*/
/*
   _____       ___   _____   _   _   _   __   _   _____
  |  _  \     /   | |_   _| | | | | | | |  \ | | /  ___|
  | |_| |    / /| |   | |   | |_| | | | |   \| | | |
  |  ___/   / / | |   | |   |  _  | | | | |\   | | |  _
  | |      / /  | |   | |   | | | | | | | | \  | | |_| |
  |_|     /_/   |_|   |_|   |_| |_| |_| |_|  \_| \_____/

	fusionbot_pathing.c
	NEW pathing routines for the FUSION bot
================================================================
Should you decide to release a modified version of FUSION, you MUST
include the following text in the documentation and source of your
modification.

---===[BEGIN TEXT]===---
================================================================
___________           .__                __________        __
\_   _____/_ __  _____|__| ____   ____   \______   \ _____/  |_
 |    __)|  |  \/  ___/  |/  _ \ /    \   |    |  _//  _ \   __\
 |     \ |  |  /\___ \|  (  <_> )   |  \  |    |   (  <_> )  |
 \___  / |____//____  >__|\____/|___|  /  |______  /\____/|__|
     \/             \/               \/          \/  Version 2.0
================================================================
Fusion v2.0
Copyright (c) 1998-2003 Trenton Shaffer.
All rights reserved.

ACEBOT Copyright(c) 1998, Steve Yeager.
EraserBot Copyright(c) 1998, Ryan Feltrin.
QFusion is Copyright (C) 2002-2003 Victor Luchits.
All other files are Copyright(C) 1997-2001 Id Software, Inc.
================================================================
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
================================================================
---===[END TEXT]===---
*/
#include "..\game\g_local.h"
#include "fusionbot.h"

int		FusionSearch;

extern	node_t		nodes[MAX_NODES];
extern short path_table[MAX_NODES][MAX_NODES];

qboolean nodeused[MAX_NODES];
short int nodefrom[MAX_NODES];

void FUSION_PA_InitSearch(edict_t *fusionbot)
{
	memset(nodeused, 0, sizeof(nodeused));
	memset(nodefrom, INVALID, sizeof(nodefrom));
	while (!FUSION_PA_SLEmpty(&fusionbot->pathList))
	{
		FUSION_PA_SLPop_Front(&fusionbot->pathList);
	}
}

qboolean	FUSION_PA_StartSearch(edict_t *fusionbot, int from, int to )
{
	// Safety first!
	if( from==INVALID || to==INVALID)
		return qfalse;

	//@@ TESTING ONLY!! FusionSearch always available
	FusionSearch = 1;
	// Check we're allowed to search - if so, do it
	if( (FusionSearch > 0) && (fusionbot->FUSION_PA_LastCallTime < level.time - PA_FREQ) )
	{
		// Decrement the semaphore to limit calls to this function
		//@@ If we ever get multithreading then we can increment later
		FusionSearch--;
		// make a note of when this bot last made a path call
		fusionbot->FUSION_PA_LastCallTime = level.time;
		// Set up the lists
		FUSION_PA_InitSearch(fusionbot);
		// If we found a path
		if( FUSION_PA_FindPath( fusionbot, from, to) )
		{
			// pathList now contains the links in reverse order
			return qtrue;
		}
	}
	// We can use the quick node search method here to get our path and put it in the pathList
	// the same way we do with the FUSION_PA_Search mode. This will have the side effect of finding
	// bad paths and removing them.
	if( FUSION_PA_QuickPath(fusionbot, from, to) )
	{
		return qtrue;
	}
	// If not allowed to search and no path
	FUSION_PA_InitSearch(fusionbot);	// Clear out the path storage
	return qfalse;
}

//=================================
// QuickPath
//=================================
//
// Uses the old path array to get a quick answer and removes bad paths
//

qboolean	FUSION_PA_QuickPath(edict_t *fusionbot, int from, int to)
{
	int	newNode = from;
	int	oldNode;

	// Clean out the arrays, etc.
	FUSION_PA_InitSearch(fusionbot);
	nodeused[from] = qtrue;
	// Check we can get from->to and that the path is complete
	while( newNode != INVALID )
	{
		oldNode = newNode;
		// get next node
		newNode = path_table[newNode][to];
		if( newNode == to )
		{
			// We're there - store it then build the path
			nodeused[newNode] = qtrue;
			nodefrom[newNode] = oldNode;
			break;
		}
		else if( newNode == INVALID )
		{
			// We have a bad path
			break;
		}
		else if( !nodeused[newNode] )
		{
			// Not been here yet - store it!
			nodeused[newNode] = qtrue;
			nodefrom[newNode] = oldNode;
		}
		else
			break; // LOOP encountered
	}

	// If successful, build the pathList
	if( newNode == to )
	{
		FUSION_PA_SLPush_Front(&fusionbot->pathList, to );
		while( newNode != from)
		{
			// Push the
			FUSION_PA_SLPush_Front(&fusionbot->pathList, nodefrom[newNode]);
			newNode = nodefrom[newNode];
		}
		return qtrue;
	}
	// else wipe out the bad path!
	else
	{
		newNode = oldNode;
		while( newNode != from)
		{
			path_table[ nodefrom[newNode] ][ to ] = INVALID;
		}
		path_table[ from ][ to ] = INVALID;
	}
	return qfalse;
}

//=======================
// FindPath
//=======================
//
// Uses OPEN and CLOSED lists to conduct a search
// Many refinements planned
//
qboolean	FUSION_PA_FindPath( edict_t *fusionbot, int from, int to)
{
	int counter = 0;
	int	newNode = INVALID; // Stores the node being tested
	node_t	*tempNode = NULL; // Pointer to a real NODE
	int workingNode,atNode; // Structures for search

	// Locally declared OPEN list
	fusion_link_t	openList;
	openList.head = openList.tail = NULL; // MUST do this!!

	// Safety first again - we don't want crashes!
	if( from==INVALID || to==INVALID )
		return qfalse;

	// Put startnode on the OPEN list
	atNode = from;
	nodefrom[atNode] = INVALID;
	FUSION_PA_SLPush_Back(&openList, from );
	nodeused[from] = qtrue;

	// While there are nodes on the OPEN list AND we are not at destNode
	while( !FUSION_PA_SLEmpty(&openList) && newNode != to )
	{
		counter = 0;

		// Where we are
		atNode = FUSION_PA_SLFront(&openList);

		// Safety check
		assert( atNode > INVALID);

		// Get a pointer to all the node information
		tempNode = &nodes[atNode];
		// Using an array for FAST access to the path ratrher than a CLOSED list
		newNode = tempNode->links[counter].targetNode;

		// Process this node putting linked nodes on the OPEN list
		while( newNode != INVALID)
		{
			// If newNode NOT on open or closed list
			if( !nodeused[newNode])
			{
				// Mark node as used
				nodeused[newNode] = qtrue;
				// Set up working node for storage on OPEN list
				workingNode = newNode;
				nodefrom[newNode] = atNode;
				// Store it
				FUSION_PA_SLPush_Back(&openList, workingNode );
			}
			// If node being linked is destNode then quit
			if( newNode == to)
			{
				break;
			}
			// Check we aren't trying to read out of range
			if( ++counter >= MAXLINKS )
				break;
			else
				newNode = tempNode->links[counter].targetNode;
		}

		// ... and remove atNode from the OPEN List
		FUSION_PA_SLPop_Front(&openList);
	}	// END While

	// Free up the memory we allocated
	FUSION_PA_SLDelete (&openList);

	// Optimise stored path with this new information
	if( newNode == to)
	{
		// Make the path using the fromnode array pushing node numbers on in reverse order
		// so we can FUSION_PA_SLPop_Front them back later
		FUSION_PA_SLPush_Front(&fusionbot->pathList, newNode );
		// Set the to path in node array because this is shortest path
		path_table[ nodefrom[to] ][ to ] = to;
		// We earlier set our start node to INVALID to set up the termination
		while(
			(newNode=nodefrom[newNode])!=INVALID // there is a path and
			&& (newNode != from)	// it's not the node we're standing on (safety check)
			)
		{
			// Push it onto the pathlist
			FUSION_PA_SLPush_Front(&fusionbot->pathList, newNode );
			// Set the path in the node array to match this shortest path
			path_table[ nodefrom[newNode] ][ to ] = newNode;
		}
		return qtrue;
	}
	// else
	return qfalse;
}

//=============================
// LinkExists
//=============================
//
// Used to check we haven't wandered off path!
//
qboolean	FUSION_PA_LinkExists( int from, int to)
{
	int counter =0;
	int	testnode;
	node_t	*tempNode = &nodes[from];

	if( from==INVALID || to==INVALID )
		return qfalse;
	// Check if the link exists
	while( counter < MAXLINKS)
	{
		testnode = tempNode->links[counter].targetNode;
		if( testnode == to)
		{
			// A path exists from,to
			return qtrue;
		}
		else if( testnode == INVALID )
		{
			// No more links and no path found
			return qfalse;
		}
		counter++;
	}
	// Didn't find it!
	return qfalse;
}

// ****************************************************** //
//=======================================================
// FUSION_PA_SL functions used by search code
//=======================================================

//==============================
// FUSION_PA_SLPush_Front
//==============================
// Add to the front of the list
//
void		FUSION_PA_SLPush_Front( fusion_link_t *list, int nodedata )
{
	single_link_t	*temp;

	// Store the current head pointer
	temp = list->head;
	// allocate memory for the new data (LEVEL tagged)
	list->head = gi.TagMalloc( sizeof(single_link_t), TAG_LEVEL);
	// Set up the data and pointer
	list->head->nodedata = nodedata;
	list->head->next = temp;
	// Check if there;'s a next item
	if( !list->head->next)
	{
		// Set the tail pointer = head
		list->tail = list->head;
	}
}

//==============================
// FUSION_PA_SLPop_Front
//==============================
// Remove the iten from the front of the list
//
void		FUSION_PA_SLPop_Front( fusion_link_t *list )
{
	single_link_t *temp;

	// Store the head pointer
	temp = list->head;
	// Check if there's a next item
	if( list && list->head )
	{
		if( list->head == list->tail )
		{
			// List is now emptying
			list->tail = list->head = NULL;
		}
		else
		{
			// Move head to point to next item
			list->head = list->head->next;
		}
		// Free the memory (LEVEL tagged)
		gi.TagFree( temp );
	}
	else
	{
		gi.bprintf( PRINT_HIGH, "Attempting to POP an empty list!\n");
	}
}

//==============================
// FUSION_PA_SLFront
//==============================
// Get the integer value from the front of the list
// without removing the item (Query the list)
//
int			FUSION_PA_SLFront( fusion_link_t *list )
{
	if( list && !FUSION_PA_SLEmpty( list) )
		return( list->head->nodedata);
	else
		return INVALID;
}

//==============================
// FUSION_PA_SLPush_Front
//==============================
// Add to the back of the list
//
void		FUSION_PA_SLPush_Back( fusion_link_t *list, int nodedata )
{
	single_link_t	*temp;

	// Allocate memory for the new item (LEVEL tagged)
	temp = (single_link_t *)gi.TagMalloc( sizeof(single_link_t), TAG_LEVEL);
	// Store the data
	temp->nodedata = nodedata;
	temp->next = NULL;		// End of the list
	// Store the new item in the list
	// Is the list empty?
	if( !list->head )
	{
		// Yes - add as a new item
		list->head = temp;
		list->tail = temp;
	}
	else
	{
		// No make this the new tail item
		list->tail->next = temp;
		list->tail = temp;
	}
}

//==============================
// FUSION_PA_SLEmpty
//==============================
// See if the list is empty (qfalse if not empty)
//
qboolean	FUSION_PA_SLEmpty( fusion_link_t *list )
{
	// If there is any item in the list then it is NOT empty...
	// If there is a list
	if(list)
		return (list->head == NULL);
	else	// No list so return empty
		return (qtrue);
}

//===============================
// Delete the list
//===============================
// Avoids memory leaks
//
void	FUSION_PA_SLDelete( fusion_link_t *list )
{
	single_link_t	*temp;

	while( !FUSION_PA_SLEmpty( list ))
	{
		temp = list->head;
		list->head = list->head->next;
		gi.TagFree( temp );
	}
}
