/*

SPAWN FUNCTION LISTING

FUSION_SPAWN_SaveBots
FUSION_SPAWN_LoadBots
FUSION_SPAWN_HoldSpawn
FUSION_SPAWN_LoadBotConfig
FUSION_SPAWN_SpawnBotConfig
FUSION_SPAWN_PutClientInServer
FUSION_SPAWN_Respawn
FUSION_SPAWN_FindFreeClient
FUSION_SPAWN_SetName
FUSION_SPAWN_SpawnBot
FUSION_SPAWN_ReAddBots
FUSION_SPAWN_RemoveBot

*/
/*
   _________________ __  _  ______
  /  ___/\____ \__  \\ \/ \/ /    \
  \___ \ |  |_> > __ \\     /   |  \
 /____  >|   __(____  /\/\_/|___|  /
      \/ |__|       \/           \/

	fusionbot_spawn.c
	Spawing support routines for the FUSION bot.
================================================================
Should you decide to release a modified version of FUSION, you MUST
include the following text in the documentation and source of your
modification.

---===[BEGIN TEXT]===---
================================================================
___________           .__                __________        __
\_   _____/_ __  _____|__| ____   ____   \______   \ _____/  |_
 |    __)|  |  \/  ___/  |/  _ \ /    \   |    |  _//  _ \   __\
 |     \ |  |  /\___ \|  (  <_> )   |  \  |    |   (  <_> )  |
 \___  / |____//____  >__|\____/|___|  /  |______  /\____/|__|
     \/             \/               \/          \/  Version 2.0
================================================================
Fusion v2.0
Copyright (c) 1998-2003 Trenton Shaffer.
All rights reserved.

ACEBOT Copyright(c) 1998, Steve Yeager.
EraserBot Copyright(c) 1998, Ryan Feltrin.
QFusion is Copyright (C) 2002-2003 Victor Luchits.
All other files are Copyright(C) 1997-2001 Id Software, Inc.
================================================================
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
================================================================
---===[END TEXT]===---
*/
#include "..\game\g_local.h"
#include "..\game\m_player.h"
#include "fusionbot.h"
#include "fusionbot_chat.h"
#include "fusionbot_scan.h"

#define	CONFIG_FILE_VERSION 1

//===========================================================================
// Routine		: FUSION_SPAWN_SpawnBotFromConfig
// Description	: Get the number of the next bots CTF team
int		FUSION_SPAWN_NextCTFTeam()
{
	int	i;
	int	onteam1 = 0;
	int	onteam2 = 0;
	edict_t		*fusionbot;

	// Only use in CTF games
	if (!ctf->value)
		return 0;

	for (i = 1; i < maxclients->value + 1; i++)
	{
		fusionbot = g_edicts + 1;
		if (fusionbot->inuse)
		{
			if (fusionbot->client->resp.ctf_team == CTF_TEAM1)
				onteam1++;
			else if (fusionbot->client->resp.ctf_team == CTF_TEAM2)
				onteam2++;
		}
	}

	if (onteam1 > onteam2)
		return (2);
	else if (onteam2 >= onteam1)
		return (1);

	return (1);
}

//===========================================================================
// Routine		: FUSION_SPAWN_SpawnBotFromConfig
// Description	: Input string is from the config file and should have all
//				: data we need to spawn a bot
void	FUSION_SPAWN_SpawnBotFromConfig( char *inString )
{
	edict_t	*fusionbot;
	char	userinfo[MAX_INFO_STRING];
	int		count=1;
	char	name[32];
	char	modelskin[80];
	int		team=0, weaponchoice=0, skillchoice=0;

	// Scanner stuff
	char	tokenString[81];
	char	*sp, *tp;
	int		ttype;

	sp = inString;
	ttype = UNDEF;

	while( ttype != EOL )
	{
		tp = tokenString;
		FUSION_SCAN_Scanner(&sp, tp, &ttype);

		// Check for comments
		if( ttype == HASH || ttype == LEXERR)
			return;

		// Check for semicolon (end of input marker)
		if( ttype == SEMIC || ttype == EOL)
			continue;

		// Keep track of which parameter we are reading
		if( ttype == COMMA )
		{
			count++;
			continue;
		}

		// NAME (parameter 1)
		if(count == 1 && ttype == STRLIT)
		{
			strcpy( name, tokenString);
			continue;
		}

		// MODELSKIN (parameter 2)
		if(count == 2 && ttype == STRLIT)
		{
			strcpy( modelskin, tokenString);
			continue;
		}

		// TEAM (parameter 3)
		if(count == 3 && ttype == INTLIT )
		{
			team = atoi(tokenString);
			continue;
		}

		// Weapon (parameter 4)
		if(count == 4 && ttype == INTLIT )
		{
			weaponchoice = atoi(tokenString);
			continue;
		}

		// Skill (parameter 5)
		if(count == 5 && ttype == INTLIT )
		{
			skillchoice = atoi(tokenString);
			continue;
		}

	}// End while

	fusionbot = FUSION_SPAWN_FindFreeClient ();

	if (!fusionbot)
	{
		gi.bprintf (PRINT_MEDIUM, "FUSION BOT: Server is full, increase Maxclients.\n");
		return;
	}

	// To allow bots to respawn
	// initialize userinfo
	memset (userinfo, 0, sizeof(userinfo));

	// add bot's name/skin/hand to userinfo
	Info_SetValueForKey (userinfo, "name", name);
	Info_SetValueForKey (userinfo, "skin", modelskin);
	// Bot is center handed for now
	Info_SetValueForKey (userinfo, "hand", "2");

	ClientConnect (fusionbot, userinfo);

	fusionbot->yaw_speed = 100; // yaw speed
	fusionbot->inuse = qtrue;
	fusionbot->svflags |= SVF_BOT; //qbism must be set AFTER ClientConnect


	G_InitEdict (fusionbot);

	if(ctf->value)
	{
		if(!team)
			team = FUSION_SPAWN_NextCTFTeam();
	}

	// Set up the preferred weapon & skill
	fusionbot->weaponchoice = weaponchoice;
	fusionbot->skillchoice = skillchoice;

	InitClientResp (fusionbot->client);

	// locate ent at a spawn point
	if(ctf->value)
	{
		if (team == 1)
			team = CTF_TEAM1;
		else
			team = CTF_TEAM2;

		FUSION_SPAWN_PutClientInServer (fusionbot,qtrue, team);
	}
	else
		FUSION_SPAWN_PutClientInServer (fusionbot,qtrue,0);

	// make sure all view stuff is valid
	ClientEndServerFrame (fusionbot);

	// Let the world know we added another
	FUSION_ITEMS_FusionBotAdded (fusionbot);

	// Pick a new goal
	FUSION_AI_Determine_LR_Goal(fusionbot);

	if( random() < 0.33)
	{
		// Store current enemies available
		int		i, counter = 0;
		//edict_t *myplayer[num_players]; //qbism was MAX_BOTS
		edict_t *myplayer[MAX_BOTS]; //at qbism was num_players

		for(i=0;i<=num_players;i++)
		{
			// Find all available enemies to insult
			if(players[i] == NULL || players[i] == fusionbot ||
			   players[i]->solid == SOLID_NOT)
			   continue;

			if(ctf->value && OnSameTeam( fusionbot, players[i]) )
				continue;

			myplayer[counter++] = players[i];
		}
		if(counter > 0)
		{
			// Say greeting to them!
			if(botchat->value)
				FUSION_CHAT_Chat( fusionbot, myplayer[rand()%counter], DBC_WELCOME);
		}
	}
}
void FUSION_SPAWN_LoadBotConfig(char *cfgfile) //qbism load bot config from command line
{
    FILE	*pIn;
	int		i;
	char	filename[60];
	// Scanner stuff
	int		fileVersion = 0;
	char	inString[81];
	char	tokenString[81];
	char	*sp, *tp;
	int		ttype;
	cvar_t	*game; //qbism use game or base dir

	game = gi.cvar ("fs_game", "", 0);
	if (!*game->string) game = gi.cvar ("fs_basedir", "", 0);

	sprintf (filename, "%s/bot_config/", game->string);

	//qbism look for file name provided, or level file if not.


	if (cfgfile) strcat(filename, cfgfile);
		else strcat(filename, level.mapname);

	strcat(filename,".cfg");
	// If there's no specific file for this level then get the normal one
	if((pIn = fopen(filename, "rb" )) == NULL)
	{
		sprintf (filename, "%s/bot_config/bot_data.cfg", game->string);
		// No bot file available, get out of here!
		if((pIn = fopen(filename, "rb" )) == NULL)
		{
			gi.bprintf(PRINT_MEDIUM,"Bot config file not found.");
			return; // bail
		}
	}

	// Now scan each line for information
	// First line should be the file version number
	fgets( inString, 80, pIn );
	sp = inString;
	tp = tokenString;
	ttype = UNDEF;

	// Scan it for the version number
	FUSION_SCAN_Scanner( &sp, tp, &ttype );
	if(ttype == BANG)
	{
		FUSION_SCAN_Scanner( &sp, tp, &ttype );
		if(ttype == INTLIT)
		{
			fileVersion = atoi( tokenString );
		}
		if( fileVersion != CONFIG_FILE_VERSION )
		{
			// ERROR!
			gi.bprintf(PRINT_HIGH, "FUSION BOT: Bot Config file is out of date!\n");
			fclose(pIn);
			return;
		}
	}

	// Now process each line of the config file
	while( fgets(inString, 80, pIn) )
	{
		FUSION_SPAWN_SpawnBotFromConfig( inString );
	}

    fclose(pIn);
}

//===========================================================================
// Routine		: FUSION_SPAWN_SaveBots
// Description	: Save bots out to a file to respawn on next level
//
// NOTE			: Bug when the bots are saved for a DM game and reloaded into CTF game
void FUSION_SPAWN_SaveBots()
{
    edict_t		*fusionbot;
    FILE		*pOut;
	int			i,count = 0;

	if((pOut = fopen("fusion\\bots.tmp", "wb" )) == NULL)
		return; // bail

	// Get number of bots
	//-------------------------------------------------------
	for (i = maxclients->value; i > 0; i--)
	{
		fusionbot = g_edicts + i + 1;

		if (fusionbot->inuse && (fusionbot->svflags & SVF_BOT))
			count++;
	}
	//-------------------------------------------------------

	fwrite(&count,sizeof (int),1,pOut); // Write number of bots

	for (i = maxclients->value; i > 0; i--)
	{
		fusionbot = g_edicts + i + 1;

		if (fusionbot->inuse && (fusionbot->svflags & SVF_BOT))
			fwrite(fusionbot->client->pers.userinfo,sizeof (char) * MAX_INFO_STRING,1,pOut);
	}

    fclose(pOut);
}

//===========================================================================
// Routine		: FUSION_SPAWN_LoadBots
// Description	: Load bots after level change
//
// NOTE			: Bots are persistant between games
void FUSION_SPAWN_LoadBots()
{
    FILE		*pIn;
	char		userinfo[MAX_INFO_STRING];
	int			i, count;

	if((pIn = fopen("fusion\\bots.tmp", "rb" )) == NULL)
		return; // bail

	fread(&count,sizeof (int),1,pIn);

	for(i=0;i<count;i++)
	{
		fread(userinfo,sizeof(char) * MAX_INFO_STRING,1,pIn);
		FUSION_SPAWN_SpawnBot (NULL, NULL, NULL, userinfo);
	}

    fclose(pIn);
}

//===========================================================================
// Routine		: FUSION_SPAWN_HoldSpawn
// Description	: Called by PutClient in Server to actually release the bot into
//				: the game, keeps from telefragging
void FUSION_SPAWN_HoldSpawn(edict_t *fusionbot)
{
	if (!KillBox (fusionbot))
	{	// could't spawn in?
	}

	gi.linkentity (fusionbot);

	fusionbot->think = FUSION_AI_Think;
	fusionbot->nextthink = level.time + FRAMETIME;

	// send effect
	//-------------------------------------------------------
	gi.WriteByte (svc_muzzleflash);
	gi.WriteShort (fusionbot-g_edicts);
	gi.WriteByte (MZ_BLASTER);
	gi.multicast (fusionbot->s.origin, MULTICAST_PVS);
	//-------------------------------------------------------

	if(ctf->value)
	gi.bprintf(PRINT_MEDIUM, "%s joined the %s team.\n",
		fusionbot->client->pers.netname, CTFTeamName(fusionbot->client->resp.ctf_team));
	else
		gi.bprintf (PRINT_MEDIUM, "%s entered the game\n", fusionbot->client->pers.netname);

}

//===========================================================================
// Routine		: FUSION_SPAWN_PutClientInServer
// Description	: Just a modified version of id's code
void FUSION_SPAWN_PutClientInServer (edict_t *fusionbot, qboolean respawn, int team)
{
	vec3_t				mins = {-16, -16, -24};
	vec3_t				maxs = {16, 16, 32};
	int					index;
	vec3_t				spawn_origin, spawn_angles;
	gclient_t			*client;
	int					i, counter;
	client_persistant_t	saved;
	client_respawn_t	resp;
	char				*s;

	// find a spawn point
	// do it before setting health back up, so farthest
	// ranging doesn't count this client
	//-------------------------------------------------------


	SelectSpawnPoint (fusionbot, spawn_origin, spawn_angles);

	index = fusionbot-g_edicts-1;
	client = fusionbot->client;

	// deathmatch wipes most client data every spawn
	//-------------------------------------------------------
	if (deathmatch->value)
	{
		char userinfo[MAX_INFO_STRING];

		resp = fusionbot->client->resp;
		memcpy (userinfo, client->pers.userinfo, sizeof(userinfo));
		InitClientPersistant (client);
		ClientUserinfoChanged (fusionbot, userinfo);
	}
	else
		memset (&resp, 0, sizeof(resp));
	//-------------------------------------------------------

	// clear everything but the persistant data
	//-------------------------------------------------------
	saved = client->pers;
	memset (client, 0, sizeof(*client));
	client->pers = saved;
	client->resp = resp;
	//-------------------------------------------------------

	// copy some data from the client to the entity
	//-------------------------------------------------------
	FetchClientEntData (fusionbot);

	// clear entity values
	//-------------------------------------------------------
	fusionbot->groundentity = NULL;
	fusionbot->client = &game.clients[index];
	fusionbot->takedamage = DAMAGE_AIM;
	fusionbot->movetype = MOVETYPE_WALK;
	fusionbot->viewheight = 24;
	fusionbot->classname = "bot";
	fusionbot->mass = 200;
	fusionbot->solid = SOLID_BBOX;
	fusionbot->deadflag = DEAD_NO;
	fusionbot->air_finished = level.time + 12;
	fusionbot->clipmask = MASK_PLAYERSOLID;
	//JAL splitmodels

	fusionbot->pain = player_pain;//FUSION_AI_BotPain;
	fusionbot->die = player_die;
	fusionbot->waterlevel = 0;
	fusionbot->watertype = 0;
	fusionbot->flags &= ~FL_NO_KNOCKBACK;
	fusionbot->svflags &= ~SVF_CORPSE;
	fusionbot->is_jumping = qfalse;
	//-------------------------------------------------------

	if(ctf->value)
	{
		client->resp.ctf_team = team;
		client->resp.ctf_state = 1;
		s = Info_ValueForKey (client->pers.userinfo, "skin");
		CTFAssignSkin(fusionbot, s);
	}

	VectorCopy (mins, fusionbot->mins);
	VectorCopy (maxs, fusionbot->maxs);
	VectorClear (fusionbot->velocity);

	// clear playerstate values
	//-------------------------------------------------------
	memset (&fusionbot->client->ps, 0, sizeof(client->ps));

	client->ps.pmove.origin[0] = spawn_origin[0]*8;
	client->ps.pmove.origin[1] = spawn_origin[1]*8;
	client->ps.pmove.origin[2] = spawn_origin[2]*8;

//ZOID
	client->ps.pmove.pm_flags &= ~PMF_NO_PREDICTION;
//ZOID

	if (deathmatch->value && ((int)dmflags->value & DF_FIXED_FOV))
	{
		client->ps.fov = 90;
	}
	else
	{
		client->ps.fov = atoi(Info_ValueForKey(client->pers.userinfo, "fov"));
		if (client->ps.fov < 1)
			client->ps.fov = 90;
		else if (client->ps.fov > 160)
			client->ps.fov = 160;
	}

	client->ps.gunindex = gi.modelindex(client->pers.weapon->view_model);

	// clear entity state values
	//-------------------------------------------------------
	fusionbot->s.effects = 0;
	fusionbot->s.skinnum = fusionbot - g_edicts - 1;
	// Will use the skin specified model
	fusionbot->s.modelindex = 255;
	// Custom gun model
	fusionbot->s.modelindex2 = 255;

	//JAL splitmodels (Put-bot-client in server)
	//fusionbot->s.frame = 0;
	fusionbot->client->anim_priority = ANIM_BASIC;
	fusionbot->client->anim_event_time = 0;
	fusionbot->client->anim_upper = TORSO_STAND;
	fusionbot->client->anim_lower = LEGS_STAND;
	fusionbot->client->anim_head = 0;



	VectorCopy (spawn_origin, fusionbot->s.origin);
	// Make sure off ground
	fusionbot->s.origin[2] += 1;
	//-------------------------------------------------------

	// set the delta angle
	//-------------------------------------------------------
	for (i=0 ; i<3 ; i++)
		client->ps.pmove.delta_angles[i] = ANGLE2SHORT(spawn_angles[i] - client->resp.cmd_angles[i]);

	fusionbot->s.angles[PITCH] = 0;
	fusionbot->s.angles[YAW] = spawn_angles[YAW];
	fusionbot->s.angles[ROLL] = 0;
	VectorCopy (fusionbot->s.angles, client->ps.viewangles);
	VectorCopy (fusionbot->s.angles, client->v_angle);
	//-------------------------------------------------------

	// force the current weapon up
	//-------------------------------------------------------
	client->newweapon = client->pers.weapon;
	ChangeWeapon (fusionbot);
	//-------------------------------------------------------

	fusionbot->enemy = NULL;
	fusionbot->movetarget = NULL;
	if(!ctf->value)
		fusionbot->state = BOT_STATE_MOVE;
	else
		fusionbot->state = BOT_STATE_POSITION;

	// Set the current node
	//-------------------------------------------------------
	fusionbot->current_node = FUSION_NAVIGATION_FindClosestReachableNode(fusionbot,NODE_DENSITY, NODE_ALL);
	fusionbot->goal_node = fusionbot->current_node;
	fusionbot->next_node = fusionbot->current_node;
	fusionbot->next_move_time = level.time;
	fusionbot->suicide_timeout = level.time + 15.0;
	//-------------------------------------------------------

	// If we are not respawning hold off for up to three seconds before releasing into game
	//-------------------------------------------------------
    if(!respawn)
	{
		fusionbot->think = FUSION_SPAWN_HoldSpawn;
		fusionbot->nextthink = level.time + 0.1;
		// Up to 3 seconds
		fusionbot->nextthink = level.time + random()*3.0;
	}
	else
	{

		if (!KillBox (fusionbot))
		{	// could't spawn in?
		}

		gi.linkentity (fusionbot);

		fusionbot->think = FUSION_AI_Think;
		fusionbot->nextthink = level.time + FRAMETIME;

		if( fusionbot->weaponchoice == 0)
			counter = rand() % 4;
		else
			counter = fusionbot->weaponchoice -1;

		switch(counter)
		{
		case 0:
			FUSION_AI_ChoosePreferredWeapon( fusionbot, BLASTER_NAME);
			break;
		case 1:
			FUSION_AI_ChoosePreferredWeapon( fusionbot, SHOTGUN_NAME);
			break;
		//case 2:
			//qbism remove FUSION_AI_ChoosePreferredWeapon( fusionbot, SSHOTGUN_NAME);
			break;
		case 2:
			FUSION_AI_ChoosePreferredWeapon( fusionbot, MACHINEGUN_NAME);
			break;
		case 3:
			FUSION_AI_ChoosePreferredWeapon( fusionbot, GRENADELAUNCHER_NAME);
			break;
		case 4:
			FUSION_AI_ChoosePreferredWeapon( fusionbot, ROCKETLAUNCHER_NAME);
			break;
		case 5:
			FUSION_AI_ChoosePreferredWeapon( fusionbot, HYPERBLASTER_NAME);
			break;
		case 6:
			FUSION_AI_ChoosePreferredWeapon( fusionbot, RAILGUN_NAME);
			break;
		case 7:
			FUSION_AI_ChoosePreferredWeapon( fusionbot, BFG_NAME);
			break;
		default:
			FUSION_AI_ChoosePreferredWeapon( fusionbot, ROCKETLAUNCHER_NAME);
			break;
		}

		client->newweapon = client->pers.weapon;
		ChangeWeapon (fusionbot);

		if (ctf->value)
		{
			fusionbot->solid = SOLID_BBOX;
			gi.linkentity(fusionbot);
		}

		// send effect
		//-------------------------------------------------------
		gi.WriteByte (svc_muzzleflash);
		gi.WriteShort (fusionbot-g_edicts);
		gi.WriteByte (MZ_BLASTER);
		gi.multicast (fusionbot->s.origin, MULTICAST_PVS);
	}


	if(ctf->value)
	{
		int		randomnode;
		fusionbot->client->resp.ctf_team = team;
		s = Info_ValueForKey (fusionbot->client->pers.userinfo, "skin");
		CTFAssignSkin(fusionbot, s);
		fusionbot->teamPauseTime = level.time + 3.0 + (rand() % 7);
		randomnode = (int)(num_players * random() );
		VectorSubtract (nodes[randomnode].origin, fusionbot->s.origin, fusionbot->move_vector);
		fusionbot->move_vector[2] = 0;
	}
	else
		fusionbot->teamPauseTime = level.time;

	memset(&fusionbot->pathList, 0, sizeof(fusionbot->pathList));
	fusionbot->pathList.head = fusionbot->pathList.tail = NULL;


	//-------------------------------------------------------

}

//===========================================================================
// Routine		: FUSION_SPAWN_Respawn
// Description	: Respawn the bot
void FUSION_SPAWN_Respawn (edict_t *fusionbot)
{
	int		i, counter = 0;
	edict_t *myplayer[MAX_BOTS];


	//qbism attempt to fix floating bot corpse
	if (fusionbot->movetype != MOVETYPE_NOCLIP)
			CopyToBodyQue (fusionbot);
	fusionbot->svflags &= ~SVF_NOCLIENT;

	//if (fusionbot->solid != SOLID_NOT || fusionbot->deadflag == DEAD_DEAD)
	//{
	//	CopyToBodyQue (fusionbot);
	//}


	if(ctf->value)
		FUSION_SPAWN_PutClientInServer (fusionbot,qtrue, fusionbot->client->resp.ctf_team);
	else
		FUSION_SPAWN_PutClientInServer (fusionbot,qtrue,0);

	// add a teleportation effect
	//-------------------------------------------------------
	fusionbot->s.event = EV_TELEPORT;

	// hold in place briefly (prevent telefrag)
	//-------------------------------------------------------
	fusionbot->client->ps.pmove.pm_flags = PMF_TIME_TELEPORT;
	fusionbot->client->ps.pmove.pm_time = 14;

	fusionbot->client->respawn_time = level.time;

	if( random() < 0.15)
	{

			// Store current enemies available
		int	counter = 0;

		if( fusionbot->lastkilledby)
		{
			// Have a comeback line!
			if(botchat->value)
				FUSION_CHAT_Chat( fusionbot, fusionbot->lastkilledby, DBC_KILLED);
			fusionbot->lastkilledby = NULL;
		}
	}
	else
	{
		// Pick someone at random to insult
		for(i=0;i<=num_players;i++)
		{
			// Find all available enemies to insult
			if(players[i] == NULL || players[i] == fusionbot || players[i]->solid == SOLID_NOT)
				continue;
			if(ctf->value && OnSameTeam( fusionbot, players[i]) )
				continue;
			myplayer[counter++] = players[i];
		}

		if(counter > 0)
		{
			if (random() < 0.175)
			{
				// Say something insulting to them!
				if(botchat->value)
					FUSION_CHAT_Chat( fusionbot, myplayer[rand()%counter], DBC_INSULT);


			}
		}
	}
}

//===========================================================================
// Routine		: FUSION_SPAWN_FindFreeClient
// Description	: Find and Free the client spot
edict_t *FUSION_SPAWN_FindFreeClient (void)
{
	edict_t		*fusionbot;
	int			i;
	int			max_count=0;

	// This is for the naming of the bots
	//-------------------------------------------------------
	for (i = maxclients->value; i > 0; i--)
	{
		fusionbot = g_edicts + i + 1;

		if(fusionbot->count > max_count)
			max_count = fusionbot->count;
	}
	//-------------------------------------------------------

	// Check for free spot
	//-------------------------------------------------------
	for (i = maxclients->value; i > 0; i--)
	{
		fusionbot = g_edicts + i + 1;

		if (!fusionbot->inuse)
			break;
	}
	//-------------------------------------------------------

	// Will become bot name
	fusionbot->count = max_count + 1;

	if (fusionbot->inuse)
		fusionbot = NULL;

	return fusionbot;
}

//===========================================================================
// Routine		: FUSION_SPAWN_SetName
// Description	: Set the name of the bot and update the userinfo
void FUSION_SPAWN_SetName(edict_t *fusionbot, char *name, char *skin, char *team)
{
	float	rnd;
	char	userinfo[MAX_INFO_STRING];
	char	bot_skin[MAX_INFO_STRING];
	char	bot_name[MAX_INFO_STRING];

	// Set the name for the bot.
	// name
	//-------------------------------------------------------
	if(strlen(name) == 0)
		sprintf(bot_name,"FusionBot_%d",fusionbot->count);
	else
		strcpy(bot_name,name);
	//-------------------------------------------------------

	// skin
	//-------------------------------------------------------
	if(strlen(skin) == 0)
	{
		// randomly choose skin
		rnd = random();
		//qbism FIXME- set this up with a cfg or script?
		if(rnd  < 0.1)
			sprintf(bot_skin,"grunt/default");
		else if(rnd < 0.2)
			sprintf(bot_skin,"grunt/stripe");
		else if(rnd < 0.4)
			sprintf(bot_skin,"major/daemia");
		else if(rnd < 0.6)
			sprintf(bot_skin,"major/default");
		else if(rnd < 0.7)
			sprintf(bot_skin,"sarge/default");
		else if(rnd < 0.8)
			sprintf(bot_skin,"sarge/krusade");
		else
			sprintf(bot_skin,"visor/default");
	}
	else
		strcpy(bot_skin,skin);
	//-------------------------------------------------------

	// initialize userinfo
	//-------------------------------------------------------
	memset (userinfo, 0, sizeof(userinfo));

	// add bot's name/skin/hand to userinfo
	//-------------------------------------------------------
	Info_SetValueForKey (userinfo, "name", bot_name);
	Info_SetValueForKey (userinfo, "skin", bot_skin);
	// bot is center handed for now!
	Info_SetValueForKey (userinfo, "hand", "2");
	//-------------------------------------------------------

	ClientConnect (fusionbot, userinfo);

	fusionbot->svflags |= SVF_BOT; //qbism ClientConnect wipes this out.
}

extern char current_map[55];
char	*LocalTeamNames[3] = {"spectator", "red", "blue"};

//===========================================================================
// Routine		: FUSION_SPAWN_SpawnBot
// Description	: Spawn the bot
void FUSION_SPAWN_SpawnBot (char *team, char *name, char *skin, char *userinfo)
{
	edict_t		*fusionbot;
	int			counter = 0;


	fusionbot = FUSION_SPAWN_FindFreeClient ();
	if (!fusionbot)
	{
		gi.bprintf (PRINT_MEDIUM, "FUSION BOT: Server is full, increase Maxclients.\n");
		return;
	}

	// To allow bots to respawn
	//-------------------------------------------------------
	if(userinfo == NULL)
		FUSION_SPAWN_SetName(fusionbot, name, skin, team);
	else
		ClientConnect (fusionbot, userinfo);

	// Yaw speed
	fusionbot->yaw_speed = 100; // yaw speed
	fusionbot->inuse = qtrue;
	fusionbot->svflags |= SVF_BOT; //qbism must be set AFTER ClientConnect

	//-------------------------------------------------------

	G_InitEdict (fusionbot);

		// locate ent at a spawn point
	//-------------------------------------------------------
	if(ctf->value)
	{
		if( (team == NULL) || (strlen(team) < 1) )
		{
			if( FUSION_SPAWN_NextCTFTeam() == 1 )
			{
				team = "blue";
			}
			else
			{
				team = "red";
			}
		}
	}

	InitClientResp (fusionbot->client);

	//-------------------------------------------------------

	fusionbot->weaponchoice = 0;
	fusionbot->skillchoice = 0;

	if (ctf->value)
	{
		if ((team !=NULL) && (strcmp(team, "blue") == 0))
			FUSION_SPAWN_PutClientInServer (fusionbot,qfalse, CTF_TEAM2);
		else
			FUSION_SPAWN_PutClientInServer (fusionbot,qfalse, CTF_TEAM1);
	}
	else
 		FUSION_SPAWN_PutClientInServer (fusionbot,qfalse,0);
	// make sure all view stuff is valid
	//-------------------------------------------------------
	ClientEndServerFrame (fusionbot);

	// Let the world know we added another
	FUSION_ITEMS_FusionBotAdded (fusionbot);

	// pick a new goal
	FUSION_AI_Determine_LR_Goal(fusionbot);



	if( random() < 0.33)
	{
		// Store current enemies available
		int		i, counter = 0;
		edict_t *myplayer[MAX_BOTS];

		for(i=0;i<=num_players;i++)
		{
			// Find all available enemies to insult
			if(players[i] == NULL || players[i] == fusionbot ||
			   players[i]->solid == SOLID_NOT)
			   continue;
			if(ctf->value && OnSameTeam( fusionbot, players[i]))
				continue;
			myplayer[counter++] = players[i];
		}
		if(counter > 0)
		{
			// Say something insulting to them!
			if(botchat->value)
				FUSION_CHAT_Chat( fusionbot, myplayer[rand()%counter], DBC_WELCOME);
		}
	}
}

//===========================================================================
// Routine		: FUSION_SPAWN_RemoveBot
// Description	: Remove a bot by name or all bots
void FUSION_SPAWN_RemoveBot(char *name)
{
	int			i;
	qboolean	freed=qfalse;
	edict_t		*fusionbot;

	for(i=0;i<maxclients->value;i++)
	{
		fusionbot = g_edicts + i + 1;
		if(fusionbot->inuse)
		{
			if((fusionbot->svflags & SVF_BOT) && (strcmp(fusionbot->client->pers.netname,name)==0 || strcmp(name,"all")==0))
			{
				fusionbot->health = 0;
				player_die (fusionbot, fusionbot, fusionbot, 100000, vec3_origin);
				// don't even bother waiting for death frames
				//-------------------------------------------------------
				fusionbot->deadflag = DEAD_DEAD;
				fusionbot->inuse = qfalse;
				freed = qtrue;
				FUSION_ITEMS_FusionBotRemoved (fusionbot);
				gi.bprintf (PRINT_MEDIUM, "FUSION BOT: %s removed\n", fusionbot->client->pers.netname);
			}
		}
	}

	if(!freed)
		gi.bprintf (PRINT_MEDIUM, "FUSION BOT: %s not found\n", name);

	FUSION_SPAWN_SaveBots(); // Save them again
}

