/*
Copyright (C) 1997-2001 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
#include "g_local.h"
#include "m_player.h"

void SP_misc_teleporter_dest (edict_t *ent);


/*QUAKED info_player_start (1 0 0) (-16 -16 -24) (16 16 32)
The normal starting point for a level.
*/
void SP_info_player_start(edict_t *self)
{
}

/*QUAKED info_player_deathmatch (1 0 1) (-16 -16 -24) (16 16 32)
potential spawning position for deathmatch games
*/
void SP_info_player_deathmatch(edict_t *self)
{
	if (!deathmatch->value)
	{
		G_FreeEdict (self);
		return;
	}
	SP_misc_teleporter_dest (self);
}

/*QUAKED info_player_coop (1 0 1) (-16 -16 -24) (16 16 32)
potential spawning position for coop games
*/

void SP_info_player_coop(edict_t *self)
{
	if (!coop->value)
		G_FreeEdict (self);
}


/*QUAKED info_player_intermission (1 0 1) (-16 -16 -24) (16 16 32)
The deathmatch intermission point will be at one of these
Use 'angles' instead of 'angle', so you can set pitch or roll as well as yaw.  'pitch yaw roll'
*/
void SP_info_player_intermission(edict_t *ent)
{
}


//=======================================================================


void player_pain (edict_t *self, edict_t *other, float kick, int damage)
{
	// player pain is handled at the end of the frame in P_DamageFeedback
}

void ClientObituary (edict_t *self, edict_t *inflictor, edict_t *attacker)
{
	int			mod;
	char		message[64];
	char		message2[64];
	qboolean	ff;

	if (coop->value && attacker->client)
		meansOfDeath |= MOD_FRIENDLY_FIRE;

	if ( deathmatch->value || coop->value )
	{
		ff = meansOfDeath & MOD_FRIENDLY_FIRE;
		mod = meansOfDeath & ~MOD_FRIENDLY_FIRE;

		Q_Obituary ( self, G_PlayerGender ( self ), attacker, mod, message, message2 );

		// duplicate message at server console for logging
		if ( attacker && attacker->client ) {
			if ( attacker != self ) {		// regular death message
				if ( deathmatch->value ) {
					if ( ff ) {
						attacker->client->resp.score--;
					} else {
						attacker->client->resp.score++;
					}
				}

				self->enemy = attacker;

				if ( dedicated->value ) {
					gi.dprintf ( "%s %s %s%s\n", self->client->pers.netname, message, attacker->client->pers.netname, message2 );
				}
			} else {			// suicide
				if ( deathmatch->value ) {
					self->client->resp.score--;
				}

				self->enemy = NULL;

				if ( dedicated->value ) {
					gi.dprintf ( "%s %s\n", self->client->pers.netname, message );
				}
			}

			gi.obituary ( self - g_edicts, attacker - g_edicts, mod, PRINT_MEDIUM );
		} else {		// wrong place, suicide, etc.
			if ( deathmatch->value ) {
				self->client->resp.score--;
			}

			self->enemy = NULL;

			if ( dedicated->value ) {
				gi.dprintf ( "%s %s\n", self->client->pers.netname, message );
			}

			gi.obituary ( self - g_edicts, (attacker == self) ? self - g_edicts : 0, mod, PRINT_MEDIUM );
		}
	}
}

void Touch_Item (edict_t *ent, edict_t *other, cplane_t *plane, csurface_t *surf);

void TossClientWeapon (edict_t *self)
{
	gitem_t		*item;
	edict_t		*drop;
	qboolean	quad;
	float		spread;

	if (!deathmatch->value)
		return;

	item = self->client->pers.weapon;
	if (! self->client->pers.inventory[self->client->ammo_index] )
		item = NULL;
	if (item && (strcmp (item->pickup_name, "Blaster") == 0))
		item = NULL;

	if (!((int)(dmflags->value) & DF_QUAD_DROP))
		quad = qfalse;
	else
		quad = (self->client->quad_framenum > (level.framenum + 10));

	if (item && quad)
		spread = 22.5;
	else
		spread = 0.0;

	if (item)
	{
		self->client->v_angle[YAW] -= spread;
		drop = Drop_Item (self, item);
		self->client->v_angle[YAW] += spread;
		drop->spawnflags = DROPPED_PLAYER_ITEM;
	}

	if (quad)
	{
		self->client->v_angle[YAW] += spread;
		drop = Drop_Item (self, FindItemByClassname ("item_quad"));
		self->client->v_angle[YAW] -= spread;
		drop->spawnflags |= DROPPED_PLAYER_ITEM;

		drop->touch = Touch_Item;
		drop->nextthink = level.time + (self->client->quad_framenum - level.framenum) * FRAMETIME;
		drop->think = G_FreeEdict;
	}
}


/*
==================
LookAtKiller
==================
*/
void LookAtKiller (edict_t *self, edict_t *inflictor, edict_t *attacker)
{
	vec3_t		dir;

	if (attacker && attacker != world && attacker != self)
	{
		VectorSubtract (attacker->s.origin, self->s.origin, dir);
	}
	else if (inflictor && inflictor != world && inflictor != self)
	{
		VectorSubtract (inflictor->s.origin, self->s.origin, dir);
	}
	else
	{
		self->client->killer_yaw = self->s.angles[YAW];
		return;
	}

	if (dir[0])
		self->client->killer_yaw = RAD2DEG ( atan2(dir[1], dir[0]) );
	else {
		self->client->killer_yaw = 0;
		if (dir[1] > 0)
			self->client->killer_yaw = 90;
		else if (dir[1] < 0)
			self->client->killer_yaw = -90;
	}
	if (self->client->killer_yaw < 0)
		self->client->killer_yaw += 360;
}

/*
==================
player_die
==================
*/
void player_die (edict_t *self, edict_t *inflictor, edict_t *attacker, int damage, vec3_t point)
{
	int		n;
	int		contents;

	VectorClear (self->avelocity);

	self->takedamage = DAMAGE_YES;
	self->movetype = MOVETYPE_TOSS;

	self->s.modelindex2 = 0;	// remove linked weapon model
//ZOID
	self->s.modelindex3 = 0;	// remove linked ctf flag
//ZOID

	self->s.angles[0] = 0;
	self->s.angles[2] = 0;

	self->s.sound = 0;
	self->client->weapon_sound = 0;

	self->maxs[2] = -8;

//	self->solid = SOLID_NOT;
	self->svflags |= SVF_CORPSE;

	if (!self->deadflag)
	{
		self->client->respawn_time = level.time + 1.0;
		LookAtKiller (self, inflictor, attacker);
		self->client->ps.pmove.pm_type = PM_DEAD;
		ClientObituary (self, inflictor, attacker);
//ZOID
		// if at start and same team, clear
		if (ctf->value && meansOfDeath == MOD_TELEFRAG &&
			self->client->resp.ctf_state < 2 &&
			self->client->resp.ctf_team == attacker->client->resp.ctf_team) {
			attacker->client->resp.score--;
			self->client->resp.ctf_state = 0;
		}

		CTFFragBonuses(self, inflictor, attacker);

//ZOID
		CTFPlayerResetGrapple(self);
//ZOID
		contents = gi.pointcontents ( self->s.origin );

		// check if player is in a nodrop area and
		// reset flags and techs, otherwise drop items
		if ( !(contents & CONTENTS_NODROP) )
		{
//ZOID
			TossClientWeapon ( self );
//ZOID
			CTFDeadDropFlag ( self );
			CTFDeadDropTech ( self );
//ZOID
		} else {		// reset flags and tech
			edict_t *tech = CTFDeadDropTech ( self );

			if ( tech ) {
				CTFRespawnTech ( tech );
			}

			if ( self->client->pers.inventory[ITEM_INDEX(FindItemByClassname("team_CTF_redflag"))] ) {
				CTFResetFlag ( CTF_TEAM1 );
			} else if ( self->client->pers.inventory[ITEM_INDEX(FindItemByClassname("team_CTF_blueflag"))] ) {
				CTFResetFlag ( CTF_TEAM2 );
			}
		}

		if (deathmatch->value && !self->client->showscores)
			Cmd_Help_f (self);		// show scores
	}

	// remove powerups
	self->client->quad_framenum = 0;
	self->client->invincible_framenum = 0;
	self->client->breather_framenum = 0;
	self->client->enviro_framenum = 0;
	self->flags &= ~FL_POWER_ARMOR;

	// clear inventory
	memset(self->client->pers.inventory, 0, sizeof(self->client->pers.inventory));

	if ( self->health < GIB_HEALTH && !(contents & CONTENTS_NODROP) )
	{	// gib

		for (n= 0; n < 4; n++)
			ThrowGib (self, "models/gibs/forearm.md3", damage, GIB_ORGANIC); //qbism md3 replace

		self->s.event = EV_GIB;

		ThrowClientHead (self, damage);

		//SPLITMODELS (death animations) [begin]
//ZOID
		self->client->anim_priority = ANIM_DEATH;
//		self->client->anim_end = 0;
//ZOID
	}
	else
	{	// normal death
		if (!self->deadflag)
		{
			static int i;

			i = (i+1)%3;
			// start a death animation
			self->client->anim_priority = ANIM_DEATH;
			if (self->client->ps.pmove.pm_flags & PMF_DUCKED)
			{
				//crouched death removed. Using death3
				self->s.frame = (((BOTH_DEATH3 & 0x3F) << 0)|((BOTH_DEATH3 & 0x3F) << 6));
				self->client->anim_event_time = BOTH_DEATH3_time;

			}
			else switch (i)
			{
			case 0:
				self->s.frame = (((BOTH_DEATH1 & 0x3F) << 0)|((BOTH_DEATH1 & 0x3F) << 6));
				self->client->anim_event_time = BOTH_DEATH1_time;
				break;

			case 1:
				self->s.frame = (((BOTH_DEATH2 & 0x3F) << 0)|((BOTH_DEATH2 & 0x3F) << 6));
				self->client->anim_event_time = BOTH_DEATH2_time;
				break;

			case 2:
				self->s.frame = (((BOTH_DEATH3 & 0x3F) << 0)|((BOTH_DEATH3 & 0x3F) << 6));
				self->client->anim_event_time = BOTH_DEATH3_time;
				break;
			}

			//SPLITMODELS (death animations) [end]

			if ( self->health <= GIB_HEALTH ) {
				self->health = GIB_HEALTH + 1;
			}

			self->s.event = EV_DIE;
		}
	}

	self->deadflag = DEAD_DEAD;

	gi.linkentity (self);
}

//=======================================================================

/*
==============
InitClientPersistant

This is only called when the game first initializes in single player,
but is called after each death and level change in deathmatch
==============
*/
void InitClientPersistant (gclient_t *client)
{
	gitem_t		*item;

	memset (&client->pers, 0, sizeof(client->pers));

	item = FindItem("Blaster");
	client->pers.inventory[ITEM_INDEX(item)] = 1;

	item = FindItem("Bullets");
	client->pers.inventory[ITEM_INDEX(item)] = 50;

	item = FindItem("Machinegun");
	client->pers.selected_item = ITEM_INDEX(item);
	client->pers.inventory[client->pers.selected_item] = 1;

	client->pers.weapon = item;
//ZOID
	client->pers.lastweapon = item;
//ZOID

//ZOID
	if ( ctf->value ) {
		item = FindItem("Grapple");
		client->pers.inventory[ITEM_INDEX(item)] = 1;
	}
//ZOID

	client->pers.health			= 100;
	client->pers.max_health		= 100;

	client->pers.max_bullets	= 200;
	client->pers.max_shells		= 100;
	client->pers.max_rockets	= 50;
	client->pers.max_grenades	= 50;
	client->pers.max_cells		= 200;
	client->pers.max_slugs		= 50;

	client->pers.connected = qtrue;
}


void InitClientResp (gclient_t *client)
{
//ZOID
	int ctf_team = client->resp.ctf_team;
	qboolean id_state = client->resp.id_state;
//ZOID

	memset (&client->resp, 0, sizeof(client->resp));

//ZOID
	client->resp.ctf_team = ctf_team;
	client->resp.id_state = id_state;
//ZOID

	client->resp.enterframe = level.framenum;
	client->resp.coop_respawn = client->pers;

//ZOID
	if (ctf->value && client->resp.ctf_team < CTF_TEAM1)
		CTFAssignTeam(client);
//ZOID
}

/*
==================
SaveClientData

Some information that should be persistant, like health,
is still stored in the edict structure, so it needs to
be mirrored out to the client structure before all the
edicts are wiped.
==================
*/
void SaveClientData (void)
{
	int		i;
	edict_t	*ent;

	for (i=0 ; i<game.maxclients ; i++)
	{
		ent = &g_edicts[1+i];
		if (!ent->inuse)
			continue;
		game.clients[i].pers.health = ent->health;
		game.clients[i].pers.max_health = ent->max_health;
		game.clients[i].pers.savedFlags = (ent->flags & (FL_GODMODE|FL_NOTARGET|FL_POWER_ARMOR));
		if (coop->value)
			game.clients[i].pers.score = ent->client->resp.score;
	}
}

void FetchClientEntData (edict_t *ent)
{
	ent->health = ent->client->pers.health;
	ent->max_health = ent->client->pers.max_health;
	ent->flags |= ent->client->pers.savedFlags;
	if (coop->value)
		ent->client->resp.score = ent->client->pers.score;
}



/*
=======================================================================

  SelectSpawnPoint

=======================================================================
*/

/*
================
PlayersRangeFromSpot

Returns the distance to the nearest player from the given spot
================
*/
float	PlayersRangeFromSpot (edict_t *spot)
{
	edict_t	*player;
	float	bestplayerdistance;
	vec3_t	v;
	int		n;
	float	playerdistance;


	bestplayerdistance = 9999999;

	for (n = 1; n <= maxclients->value; n++)
	{
		player = &g_edicts[n];

		if (!player->inuse)
			continue;

		if (player->health <= 0)
			continue;

		VectorSubtract (spot->s.origin, player->s.origin, v);
		playerdistance = VectorLength (v);

		if (playerdistance < bestplayerdistance)
			bestplayerdistance = playerdistance;
	}

	return bestplayerdistance;
}

/*
================
SelectRandomDeathmatchSpawnPoint

go to a random point, but NOT the two points closest
to other players
================
*/
edict_t *SelectRandomDeathmatchSpawnPoint (void)
{
	edict_t	*spot, *spot1, *spot2;
	int		count = 0;
	int		selection;
	float	range, range1, range2;

	spot = NULL;
	range1 = range2 = 99999;
	spot1 = spot2 = NULL;

	while ((spot = G_Find (spot, FOFS(classname), "info_player_deathmatch")) != NULL)
	{
		count++;
		range = PlayersRangeFromSpot(spot);
		if (range < range1)
		{
			range1 = range;
			spot1 = spot;
		}
		else if (range < range2)
		{
			range2 = range;
			spot2 = spot;
		}
	}

	if (!count)
		return NULL;

	if (count <= 2)
	{
		spot1 = spot2 = NULL;
	}
	else
	{
		if (spot1)
			count--;
		if (spot2 && spot2 != spot1)
			count--;
	}

	selection = rand() % count;

	spot = NULL;
	do
	{
		spot = G_Find (spot, FOFS(classname), "info_player_deathmatch");
		if (spot == spot1 || spot == spot2)
			selection++;
	} while(selection--);

	return spot;
}

/*
================
SelectFarthestDeathmatchSpawnPoint

================
*/
edict_t *SelectFarthestDeathmatchSpawnPoint (void)
{
	edict_t	*bestspot;
	float	bestdistance, bestplayerdistance;
	edict_t	*spot;


	spot = NULL;
	bestspot = NULL;
	bestdistance = 0;
	while ((spot = G_Find (spot, FOFS(classname), "info_player_deathmatch")) != NULL)
	{
		bestplayerdistance = PlayersRangeFromSpot (spot);

		if (bestplayerdistance > bestdistance)
		{
			bestspot = spot;
			bestdistance = bestplayerdistance;
		}
	}

	if (bestspot)
	{
		return bestspot;
	}

	// if there is a player just spawned on each and every start spot
	// we have no choice to turn one into a telefrag meltdown
	spot = G_Find (NULL, FOFS(classname), "info_player_deathmatch");

	return spot;
}

edict_t *SelectDeathmatchSpawnPoint (void)
{
	if ( (int)(dmflags->value) & DF_SPAWN_FARTHEST)
		return SelectFarthestDeathmatchSpawnPoint ();
	else
		return SelectRandomDeathmatchSpawnPoint ();
}


edict_t *SelectCoopSpawnPoint (edict_t *ent)
{
	int		index;
	edict_t	*spot = NULL;
	char	*target;

	index = ent->client - game.clients;

	// player 0 starts in normal player spawn point
	if (!index)
		return NULL;

	spot = NULL;

	// assume there are four coop spots at each spawnpoint
	while (1)
	{
		spot = G_Find (spot, FOFS(classname), "info_player_coop");
		if (!spot)
			return NULL;	// we didn't have enough...

		target = spot->targetname;
		if (!target)
			target = "";
		if ( Q_stricmp(game.spawnpoint, target) == 0 )
		{	// this is a coop spawn point for one of the clients here
			index--;
			if (!index)
				return spot;		// this is it
		}
	}


	return spot;
}


/*
===========
SelectSpawnPoint

Chooses a player start, deathmatch start, coop start, etc
============
*/
void	SelectSpawnPoint (edict_t *ent, vec3_t origin, vec3_t angles)
{
	edict_t	*spot = NULL;

	if (deathmatch->value)
//ZOID
		if (ctf->value)
			spot = SelectCTFSpawnPoint(ent);
		else
//ZOID
			spot = SelectDeathmatchSpawnPoint ();
	else if (coop->value)
		spot = SelectCoopSpawnPoint (ent);

	// find a single player start spot
	if (!spot)
	{
		while ((spot = G_Find (spot, FOFS(classname), "info_player_start")) != NULL)
		{
			if (!game.spawnpoint[0] && !spot->targetname)
				break;

			if (!game.spawnpoint[0] || !spot->targetname)
				continue;

			if (Q_stricmp(game.spawnpoint, spot->targetname) == 0)
				break;
		}

		if (!spot)
		{
			if (!game.spawnpoint[0])
			{	// there wasn't a spawnpoint without a target, so use any
				spot = G_Find (spot, FOFS(classname), "info_player_start");
			}
			if (!spot)
				gi.error ("Couldn't find spawn point %s\n", game.spawnpoint);
		}
	}

	VectorCopy (spot->s.origin, origin);
	origin[2] += 9;
	VectorCopy (spot->s.angles, angles);
}

//======================================================================


void InitBodyQue (void)
{
	int		i;
	edict_t	*ent;

	level.body_que = 0;
	for (i=0; i<BODY_QUEUE_SIZE ; i++)
	{
		ent = G_Spawn();
		ent->classname = "bodyque";
	}
}

void body_die (edict_t *self, edict_t *inflictor, edict_t *attacker, int damage, vec3_t point)
{
	int n;

	if (self->health >= GIB_HEALTH)
		return;

	for (n= 0; n < 4; n++)
		ThrowGib (self, "models/gibs/forearm.md3", damage, GIB_ORGANIC); //qbism md3 replace

	self->s.event = EV_GIB;
	self->s.origin[2] -= 48;
	ThrowClientHead (self, damage);
}

/*
=============
body_think - SPLITMODELS

  autodestruct body. Body self removal after some time.
  fixme: make a better effect for the body when dissapearing
=============
*/
void body_think (edict_t *self)
{
	int n, damage;

	//temporary gibbed.
	self->health = GIB_HEALTH - 1;
	damage = 25;

	//effect: small gibs, and only when it is still a body, not a gibbed head.
	if (self->s.modelindex == 255){

		self->s.event = EV_GIB;

		for (n= 0; n < 4; n++)
			ThrowGib (self, "models/gibs/forearm.md3", damage, GIB_ORGANIC); //qbism md3 replace
	}

	//remove the model
	self->s.modelindex = 0;

	//disallow interaction with the world.
	self->takedamage = DAMAGE_NO;
	self->solid = SOLID_NOT;
	self->s.sound = 0;
	self->flags |= FL_NO_KNOCKBACK;

	gi.linkentity (self);
}

void CopyToBodyQue (edict_t *ent)
{
	edict_t		*body;
	int			contents;
	int			animboth;//splitmodels

	gi.unlinkentity (ent);

	contents = gi.pointcontents ( ent->s.origin );

	if ( contents & CONTENTS_NODROP ) {
		return;
	}

	// grab a body que and cycle to the next one
	body = &g_edicts[(int)maxclients->value + level.body_que + 1];
	level.body_que = (level.body_que + 1) % BODY_QUEUE_SIZE;

	// FIXME: send an effect on the removed body

	gi.unlinkentity (body);
	body->s = ent->s;
	body->s.number = body - g_edicts;

	VectorCopy (ent->mins, body->mins);
	VectorCopy (ent->maxs, body->maxs);
	VectorCopy (ent->absmin, body->absmin);
	VectorCopy (ent->absmax, body->absmax);
	VectorCopy (ent->size, body->size);
	body->solid = ent->solid;
	body->clipmask = CONTENTS_SOLID | CONTENTS_PLAYERCLIP;
	body->svflags = SVF_CORPSE;
	body->owner = ent->owner;
	body->movetype = ent->movetype;

	//splitmodels (deadbody fix) jal [begin]
	body->s.renderfx |= RF_DEADBODY;

	//advance animations by one. Not really needed but lets use the DEATH -> DEAD animations as at Q3 scripts.
	animboth = ent->client->upper_lastsent+1;
	body->s.frame = ( (animboth&0x3F) | ((animboth&0x3F)<<6) | ((0&0xF)<<12) );

	body->think = body_think;//body self destruction
	body->nextthink = level.time + (FRAMETIME*400);

	//floating corpse fix
	M_CheckGround(body);
	M_CategorizePosition (body);

	//emulate underwater physics.
	if (body->waterlevel > 1)
		body->gravity = 0.1;
	else {
		body->s.angles[PITCH]=0;
		body->gravity = 1.0;
	}
	//jal [end]

	body->die = body_die;
	body->takedamage = DAMAGE_YES;


	gi.linkentity (body);
}


void respawn (edict_t *self)
{
	if (deathmatch->value || coop->value)
	{
		//qbism- from Fusion Bot
		//FB[start]
		if (self->svflags & SVF_BOT)
		{
			FUSION_SPAWN_Respawn (self);
			return;
		}
		//FB[end]

		if (self->movetype != MOVETYPE_NOCLIP)
			CopyToBodyQue (self);
		self->svflags &= ~SVF_NOCLIENT;
		PutClientInServer (self);
		self->s.event = EV_TELEPORT;

		// add a teleportation effect
		gi.WriteByte ( svc_temp_entity );
		gi.WriteByte ( TE_PLAYER_TELEPORT_IN );
		gi.WritePosition ( self->s.origin );
		gi.multicast ( self->s.origin, MULTICAST_PVS );

		// hold in place briefly
		self->client->ps.pmove.pm_flags = PMF_TIME_TELEPORT;
		self->client->ps.pmove.pm_time = 14;

		self->client->respawn_time = level.time;

		return;
	}

	// restart the entire server
	gi.AddCommandString ("menu_loadgame\n");
}

//==============================================================


/*
===========
PutClientInServer

Called when a player connects to a server or respawns in
a deathmatch.
============
*/
void PutClientInServer (edict_t *ent)
{
	vec3_t	mins = {-15, -15, -24};
	vec3_t	maxs = {15, 15, 32};
	int		index;
	vec3_t	spawn_origin, spawn_angles;
	gclient_t	*client;
	int		i;
	client_persistant_t	saved;
	client_respawn_t	resp;

	// find a spawn point
	// do it before setting health back up, so farthest
	// ranging doesn't count this client
	SelectSpawnPoint (ent, spawn_origin, spawn_angles);

	index = ent-g_edicts-1;
	client = ent->client;

	// deathmatch wipes most client data every spawn
	if (deathmatch->value)
	{
		char		userinfo[MAX_INFO_STRING];

		resp = client->resp;
		memcpy (userinfo, client->pers.userinfo, sizeof(userinfo));
		InitClientPersistant (client);
		ClientUserinfoChanged (ent, userinfo);
	}
	else if (coop->value)
	{
		char		userinfo[MAX_INFO_STRING];

		resp = client->resp;
		memcpy (userinfo, client->pers.userinfo, sizeof(userinfo));
		client->pers = resp.coop_respawn;
		ClientUserinfoChanged (ent, userinfo);
		if (resp.score > client->pers.score)
			client->pers.score = resp.score;
	}
	else
	{
		memset (&resp, 0, sizeof(resp));
	}

	// clear everything but the persistant data
	saved = client->pers;
	memset (client, 0, sizeof(*client));
	client->pers = saved;
	if (client->pers.health <= 0)
		InitClientPersistant(client);
	client->resp = resp;

	// copy some data from the client to the entity
	FetchClientEntData (ent);

	// clear entity values
	ent->groundentity = NULL;
	ent->client = &game.clients[index];
	ent->takedamage = DAMAGE_AIM;
	ent->movetype = MOVETYPE_WALK;
	ent->viewheight = 22;
	ent->inuse = qtrue;
	ent->classname = "player";
	ent->mass = 200;
	ent->solid = SOLID_BBOX;
	ent->deadflag = DEAD_NO;
	ent->air_finished = level.time + 12;
	ent->clipmask = MASK_PLAYERSOLID;
	//splitmodels (PutClientInServer1)
	//ent->model = "models/players/male/upper.md3";
	ent->pain = player_pain;
	ent->die = player_die;
	ent->waterlevel = 0;
	ent->watertype = 0;
	ent->flags &= ~FL_NO_KNOCKBACK;
	ent->svflags &= ~SVF_CORPSE;

	VectorCopy (mins, ent->mins);
	VectorCopy (maxs, ent->maxs);
	VectorClear (ent->velocity);

	// clear playerstate values
	memset (&ent->client->ps, 0, sizeof(client->ps));

	client->ps.pmove.origin[0] = spawn_origin[0]*8;
	client->ps.pmove.origin[1] = spawn_origin[1]*8;
	client->ps.pmove.origin[2] = spawn_origin[2]*8;
//ZOID
	client->ps.pmove.pm_flags &= ~PMF_NO_PREDICTION;
//ZOID

	if (deathmatch->value && ((int)dmflags->value & DF_FIXED_FOV))
	{
		client->ps.fov = 90;
	}
	else
	{
		client->ps.fov = atoi(Info_ValueForKey(client->pers.userinfo, "fov"));
		if (client->ps.fov < 1)
			client->ps.fov = 90;
		else if (client->ps.fov > 160)
			client->ps.fov = 160;
	}

	client->ps.gunindex = gi.modelindex(client->pers.weapon->view_model);

	// clear entity state values
	ent->s.effects = 0;
	ent->s.light = 0;
	ent->s.modelindex = 255;		// will use the skin specified model
	ent->s.modelindex2 = 255;		// custom gun model
	// sknum is player num and weapon number
	// weapon number will be added in changeweapon
	ent->s.skinnum = index;

	//splitmodels (PutClientInServer)
	ent->client->anim_priority = ANIM_BASIC;
	ent->client->anim_event_time = 0;
	ent->client->anim_upper = TORSO_STAND;
	ent->client->anim_lower = LEGS_STAND;
	ent->client->anim_head = 0;
	ent->client->lower_lastsent = 0;//force resend
	ent->client->upper_lastsent = 0;
	ent->client->head_lastsent = 0;
	//ent->s.frame = 0;

	VectorCopy (spawn_origin, ent->s.origin);
	ent->s.origin[2] += 1;	// make sure off ground
	VectorCopy (ent->s.origin, ent->s.old_origin);

	// set the delta angle
	for (i=0 ; i<3 ; i++)
		client->ps.pmove.delta_angles[i] = ANGLE2SHORT(spawn_angles[i] - client->resp.cmd_angles[i]);

	ent->s.angles[PITCH] = 0;
	ent->s.angles[YAW] = spawn_angles[YAW];
	ent->s.angles[ROLL] = 0;
	VectorCopy (ent->s.angles, client->ps.viewangles);
	VectorCopy (ent->s.angles, client->v_angle);

//ZOID
	if (ctf->value && CTFStartClient(ent))
		return;
//ZOID

	if (!KillBox (ent))
	{	// could't spawn in?
	}

	gi.linkentity (ent);

	// force the current weapon up
	client->newweapon = client->pers.weapon;
	ChangeWeapon (ent);
}

/*
=====================
ClientBeginDeathmatch

A client has just connected to the server in
deathmatch mode, so clear everything out before starting them.
=====================
*/
void ClientBeginDeathmatch (edict_t *ent)
{
	//qbism- from Fusion Bot
	//FB[start]
	static char current_map[55];
	//FB[end]

	G_InitEdict (ent);

	InitClientResp (ent->client);

	//qbism- from Fusion Bot
	//FB[start]
	FUSION_ITEMS_FusionBotAdded(ent);
	//FB[end]

	// locate ent at a spawn point
	PutClientInServer (ent);

	if (level.intermissiontime)
	{
		MoveClientToIntermission (ent);
	}
	else
	{
		// send effect
		gi.WriteByte ( svc_temp_entity );
		gi.WriteByte ( TE_PLAYER_TELEPORT_IN );
		gi.WritePosition ( ent->s.origin );
		gi.multicast ( ent->s.origin, MULTICAST_PVS );
	}

	gi.bprintf (PRINT_HIGH, "%s %sentered the game\n", ent->client->pers.netname, S_COLOR_WHITE);

	//qbism- from Fusion Bot
	//FB[start]
	gi.centerprintf(ent,"\n======================================\nFUSION BOT\nby Trenton Shaffer\n======================================\n'sv addbot' -------------> add a bot  \n'sv removebot <name>' ---> remove bot \n'sv savenodes' ----------> save nodes \n\n'sv fusiondebug on' -----> debug on   \n'sv fusiondebug off' ----> debug off  \n======================================\n\n");

	// If the map changes on us, init and reload the nodes
	if(strcmp(level.mapname,current_map))
	{

		FUSION_MAPPING_InitNodes();
		FUSION_NODES_LoadNodes();
		FUSION_SPAWN_LoadBotConfig(NULL); //qbism added NULL- load default config
		strcpy(current_map,level.mapname);
	}
	//FB[end]

// make sure all view stuff is valid
	ClientEndServerFrame (ent);
}


/*
===========
ClientBegin

called when a client has finished connecting, and is ready
to be placed into the game.  This will happen every level load.
============
*/
void ClientBegin (edict_t *ent)
{
	int		i;

	ent->client = game.clients + (ent - g_edicts - 1);

	if (deathmatch->value)
	{
		ClientBeginDeathmatch (ent);
		return;
	}

	// if there is already a body waiting for us (a loadgame), just
	// take it, otherwise spawn one from scratch
	if (ent->inuse == qtrue)
	{
		// the client has cleared the client side viewangles upon
		// connecting to the server, which is different than the
		// state when the game is saved, so we need to compensate
		// with deltaangles
		for (i=0 ; i<3 ; i++)
			ent->client->ps.pmove.delta_angles[i] = ANGLE2SHORT(ent->client->ps.viewangles[i]);
	}
	else
	{
		// a spawn point will completely reinitialize the entity
		// except for the persistant data that was initialized at
		// ClientConnect() time
		G_InitEdict (ent);
		ent->classname = "player";
		InitClientResp (ent->client);
		PutClientInServer (ent);
	}

	if (level.intermissiontime)
	{
		MoveClientToIntermission (ent);
	}
	else
	{
		// send effect if in a multiplayer game
		if (game.maxclients > 1)
		{
			gi.WriteByte ( svc_temp_entity );
			gi.WriteByte ( TE_PLAYER_TELEPORT_IN );
			gi.WritePosition ( ent->s.origin );
			gi.multicast ( ent->s.origin, MULTICAST_PVS );

			gi.bprintf ( PRINT_HIGH, "%s %sentered the game\n", ent->client->pers.netname, S_COLOR_WHITE );
		}
	}

	// make sure all view stuff is valid
	ClientEndServerFrame (ent);
}

/*
===========
ClientUserInfoChanged

called whenever the player updates a userinfo variable.

The game can override any of the settings in place
(forcing skins or names, etc) before copying it off.
============
*/
void ClientUserinfoChanged (edict_t *ent, char *userinfo)
{
	char	*s;
	int		playernum;

	// check for malformed or illegal info strings
	if (!Info_Validate(userinfo))
	{
		//splitmodels: use default_playermodel
		Com_sprintf (userinfo, sizeof(userinfo), "\\name\\badinfo\\skin\\%s/%s", DEFAULT_PLAYERMODEL, DEFAULT_PLAYERSKIN);
	}

	// set name
	s = Info_ValueForKey (userinfo, "name");
	Q_strncpyz (ent->client->pers.netname, s, sizeof(ent->client->pers.netname));

	// set skin
	s = Info_ValueForKey (userinfo, "skin");

	playernum = ent-g_edicts-1;

	// combine name and skin into a configstring
//ZOID
	if (ctf->value)
		CTFAssignSkin(ent, s);
	else
//ZOID
		gi.configstring (CS_PLAYERSKINS+playernum, va("%s\\%s", ent->client->pers.netname, s) );

//ZOID
	// set player name field (used in id_state view)
	gi.configstring (CS_GENERAL+playernum, ent->client->pers.netname);
//ZOID

	// fov
	if (deathmatch->value && ((int)dmflags->value & DF_FIXED_FOV))
	{
		ent->client->ps.fov = 90;
	}
	else
	{
		ent->client->ps.fov = atoi(Info_ValueForKey(userinfo, "fov"));
		if (ent->client->ps.fov < 1)
			ent->client->ps.fov = 90;
		else if (ent->client->ps.fov > 160)
			ent->client->ps.fov = 160;
	}

	// handedness
	s = Info_ValueForKey (userinfo, "hand");
	if (strlen(s))
	{
		ent->client->pers.hand = atoi(s);
	}

	// save off the userinfo in case we want to check something later
	Q_strncpyz (ent->client->pers.userinfo, userinfo, sizeof(ent->client->pers.userinfo));
}


/*
===========
ClientConnect

Called when a player begins connecting to the server.
The game can refuse entrance to a client by returning false.
If the client is allowed, the connection process will continue
and eventually get to ClientBegin()
Changing levels will NOT cause this to be called again, but
loadgames will.
============
*/
qboolean ClientConnect (edict_t *ent, char *userinfo)
{
	char	*value;

	// check to see if they are on the banned IP list
	value = Info_ValueForKey (userinfo, "ip");
	if (SV_FilterPacket(value)) {
		Info_SetValueForKey(userinfo, "rejmsg", "Banned.");
		return qfalse;
	}


	// check for a password
	value = Info_ValueForKey (userinfo, "password");
	if (*password->string && strcmp(password->string, "none") &&
		strcmp(password->string, value)) {
		Info_SetValueForKey(userinfo, "rejmsg", "Password required or incorrect.");
		return qfalse;
	}

	// they can connect
	ent->client = game.clients + (ent - g_edicts - 1);

	// if there is already a body waiting for us (a loadgame), just
	// take it, otherwise spawn one from scratch
	if (ent->inuse == qfalse)
	{
		// clear the respawning variables
//ZOID -- force team join
		ent->client->resp.ctf_team = -1;
		ent->client->resp.id_state = qtrue;
//ZOID
		InitClientResp (ent->client);
		if (!game.autosaved || !ent->client->pers.weapon)
			InitClientPersistant (ent->client);
	}

	ClientUserinfoChanged (ent, userinfo);

	if (game.maxclients > 1)
		gi.dprintf ("%s %sconnected\n", ent->client->pers.netname, S_COLOR_WHITE);

	ent->svflags = 0; // make sure we start with known default
	ent->client->pers.connected = qtrue;
	return qtrue;
}

/*
===========
ClientDisconnect

Called when a player drops from the server.
Will not be called between levels.
============
*/
void ClientDisconnect (edict_t *ent)
{
	int		playernum;

	if (!ent->client)
		return;

	gi.bprintf (PRINT_HIGH, "%s %sdisconnected\n", ent->client->pers.netname, S_COLOR_WHITE);

	//qbism- from Fusion Bot
	//FB[start]
	FUSION_ITEMS_FusionBotRemoved (ent);
	//FB[end]

//ZOID
	CTFDeadDropFlag(ent);
	CTFDeadDropTech(ent);
//ZOID

	// send effect
	gi.WriteByte ( svc_temp_entity );
	gi.WriteByte ( TE_PLAYER_TELEPORT_OUT );
	gi.WritePosition ( ent->s.origin );
	gi.multicast ( ent->s.origin, MULTICAST_PVS );

	gi.unlinkentity (ent);
	ent->s.modelindex = 0;
	ent->solid = SOLID_NOT;
	ent->inuse = qfalse;
	ent->classname = "disconnected";
	ent->client->pers.connected = qfalse;

	playernum = ent-g_edicts-1;
	gi.configstring (CS_PLAYERSKINS+playernum, "");
}


//==============================================================


edict_t	*pm_passent;

// pmove doesn't need to know about passent and contentmask
trace_t	PM_trace (vec3_t start, vec3_t mins, vec3_t maxs, vec3_t end)
{
	if (pm_passent->health > 0)
		return gi.trace (start, mins, maxs, end, pm_passent, MASK_PLAYERSOLID);
	else
		return gi.trace (start, mins, maxs, end, pm_passent, MASK_DEADSOLID);
}

/*
==============
ClientThink

This will be called once for each client frame, which will
usually be a couple times for each server frame.
==============
*/
void ClientThink (edict_t *ent, usercmd_t *ucmd)
{
	gclient_t	*client;
	edict_t	*other;
	int		i, j, xyspeedcheck;//splitmodels
	pmove_t	pm;

	level.current_entity = ent;
	client = ent->client;

	if (level.intermissiontime)
	{
		client->ps.pmove.pm_type = PM_FREEZE;
		// can exit intermission after five seconds
		if (level.time > level.intermissiontime + 5.0
			&& (ucmd->buttons & BUTTON_ANY) )
			level.exitintermission = qtrue;
		return;
	}

	pm_passent = ent;

//ZOID
	if (ent->client->chase_target) {
		client->resp.cmd_angles[0] = SHORT2ANGLE(ucmd->angles[0]);
		client->resp.cmd_angles[1] = SHORT2ANGLE(ucmd->angles[1]);
		client->resp.cmd_angles[2] = SHORT2ANGLE(ucmd->angles[2]);
		return;
	}
//ZOID

	// set up for pmove
	memset (&pm, 0, sizeof(pm));

	if (ent->movetype == MOVETYPE_NOCLIP)
		client->ps.pmove.pm_type = PM_SPECTATOR;
	else if (ent->s.modelindex != 255)
		client->ps.pmove.pm_type = PM_GIB;
	else if (ent->deadflag)
		client->ps.pmove.pm_type = PM_DEAD;
	else
		client->ps.pmove.pm_type = PM_NORMAL;

	client->ps.pmove.gravity = sv_gravity->value;
	pm.s = client->ps.pmove;

	for (i=0 ; i<3 ; i++)
	{
		pm.s.origin[i] = ent->s.origin[i]*8;
		pm.s.velocity[i] = ent->velocity[i]*8;
	}

	if (memcmp(&client->old_pmove, &pm.s, sizeof(pm.s)))
	{
		pm.snapinitial = qtrue;
//		gi.dprintf ("pmove changed!\n");
	}

	pm.cmd = *ucmd;

	pm.trace = PM_trace;	// adds default parms
	pm.pointcontents = gi.pointcontents;

	// perform a pmove
	gi.Pmove (&pm);

	// save results of pmove
	client->ps.pmove = pm.s;
	client->old_pmove = pm.s;

	for (i=0 ; i<3 ; i++)
	{
		ent->s.origin[i] = pm.s.origin[i]*0.125;
		ent->velocity[i] = pm.s.velocity[i]*0.125;
	}

	VectorCopy (pm.mins, ent->mins);
	VectorCopy (pm.maxs, ent->maxs);

	client->resp.cmd_angles[0] = SHORT2ANGLE(ucmd->angles[0]);
	client->resp.cmd_angles[1] = SHORT2ANGLE(ucmd->angles[1]);
	client->resp.cmd_angles[2] = SHORT2ANGLE(ucmd->angles[2]);

	if (ent->groundentity && !pm.groundentity && (pm.cmd.upmove >= 10) && (pm.waterlevel == 0))
	{
		ent->s.event = EV_JUMP;
		PlayerNoise(ent, ent->s.origin, PNOISE_SELF);

		//JAL splitmodels (LAUNCH JUMP animation)

		//vertical velocity:
		//a simple strafe jump gives a velocity of 2083
		//double jumps & ramp jumps are around 2900/3200
		//strafe running rebouncing is around 2100/2200
		//
		//note: this checks may not work at altering gravity or the
		//values defined in pmove.c

		if ((pm.s.velocity[2] > 2600)){//is double or ramp jump

			if (!ent->client->anim_jump)
			//is a rampjump and isn't rebouncing
			{
				ent->client->anim_jump_style = 0;
				ent->client->anim_jump = qtrue;
				ent->client->anim_jump_thunk = qfalse;
			}
		}
		else
		{// is a simple jump or strafing/bunnyhopping.

			//set the new animation style

			//see if it has speed of any kind
			xyspeedcheck = sqrt(ent->velocity[0]*ent->velocity[0] + ent->velocity[1]*ent->velocity[1]);
			if (xyspeedcheck < 50)
				xyspeedcheck = 0;

			if (ent->client->anim_jump == qtrue && ent->client->anim_jump_thunk == qtrue)
			{	//REBOUNCE
				if (xyspeedcheck){

					if (ent->client->anim_jump_style < 2){
						ent->client->anim_jump_style = 2;
					}
					else
						ent->client->anim_jump_style = 1;
				} else
					ent->client->anim_jump_style = 0;

			} else {
				//player wasn't rebouncing
				if (pm.cmd.forwardmove >= 20 && xyspeedcheck)
				{
					if (ent->client->anim_jump_style < 2)
						ent->client->anim_jump_style = 2;
					else
						ent->client->anim_jump_style = 1;
				} else
					ent->client->anim_jump_style = 0;
			}

			//set the jump as active and launch the animation change process.
			ent->client->anim_jump_thunk = qfalse;
			ent->client->anim_jump = qtrue;
		}//jal [end]
	}

	ent->viewheight = pm.viewheight;
	ent->waterlevel = pm.waterlevel;
	ent->watertype = pm.watertype;
	ent->groundentity = pm.groundentity;
	if (pm.groundentity)
		ent->groundentity_linkcount = pm.groundentity->linkcount;

	if (ent->deadflag)
	{
		client->ps.viewangles[ROLL] = 40;
		client->ps.viewangles[PITCH] = -15;
		client->ps.viewangles[YAW] = client->killer_yaw;
	}
	else
	{
		VectorCopy (pm.viewangles, client->v_angle);
		VectorCopy (pm.viewangles, client->ps.viewangles);
	}
	//ent->client->backmove = (pm.cmd.forwardmove < 0);//splitmodels

//ZOID
	if (client->ctf_grapple)
		CTFGrapplePull(client->ctf_grapple);
//ZOID

	gi.linkentity (ent);

	if (ent->movetype != MOVETYPE_NOCLIP)
		G_TouchTriggers (ent);

	// touch other objects
	for (i=0 ; i<pm.numtouch ; i++)
	{
		other = pm.touchents[i];
		for (j=0 ; j<i ; j++)
			if (pm.touchents[j] == other)
				break;
		if (j != i)
			continue;	// duplicated
		if (!other->touch)
			continue;
		other->touch (other, ent, NULL, NULL);
	}


	client->oldbuttons = client->buttons;
	client->buttons = ucmd->buttons;
	client->latched_buttons |= client->buttons & ~client->oldbuttons;

	//splitmodels: set movement flags for legs animations jal[begin]

	//start from 0
	ent->client->anim_moveflags = 0;

	if (ucmd->forwardmove < -1)
		ent->client->anim_moveflags |= ANIMMOVE_BACK;
	else if (ucmd->forwardmove > 1)
		ent->client->anim_moveflags |= ANIMMOVE_FRONT;

	if (ucmd->sidemove < -1)
		ent->client->anim_moveflags |= ANIMMOVE_LEFT;
	else if (ucmd->sidemove > 1)
		ent->client->anim_moveflags |= ANIMMOVE_RIGHT;

	if (ucmd->buttons & BUTTON_WALK)
		ent->client->anim_moveflags |= ANIMMOVE_WALK;
//jal[end]

	// fire weapon from final position if needed
	if (client->latched_buttons & BUTTON_ATTACK
//ZOID
		&& ent->movetype != MOVETYPE_NOCLIP
//ZOID
		)
	{
		if (!client->weapon_thunk)
		{
			client->weapon_thunk = qtrue;
			Think_Weapon (ent);
		}
	}

//ZOID
//regen tech
	CTFApplyRegeneration(ent);
//ZOID

//ZOID
	for (i = 1; i <= maxclients->value; i++) {
		other = g_edicts + i;
		if (other->inuse && other->client->chase_target == ent)
			UpdateChaseCam(other);
	}

	//qbism- from Fusion Bot
	//FB[start]
	if (!(ent->svflags & SVF_BOT) && !ent->deadflag)
		FUSION_MAPPING_PathMap(ent);
	//FB[end]

	if (client->menudirty && client->menutime <= level.time) {
		PMenu_Do_Update(ent);
		gi.unicast (ent, qtrue);
		client->menutime = level.time;
		client->menudirty = qfalse;
	}
//ZOID
}


/*
==============
ClientBeginServerFrame

This will be called once for each server frame, before running
any other entities in the world.
==============
*/
void ClientBeginServerFrame (edict_t *ent)
{
	gclient_t	*client;
	int			buttonMask;

	if (level.intermissiontime)
		return;

	client = ent->client;

	// run weapon animations if it hasn't been done by a ucmd_t
	if (!client->weapon_thunk
//ZOID
		&& ent->movetype != MOVETYPE_NOCLIP
//ZOID
		)
		Think_Weapon (ent);
	else
		client->weapon_thunk = qfalse;

	if (ent->deadflag)
	{
		// wait for any button just going down
		if ( level.time > client->respawn_time )
		{
			//splitmodels: wait for the death animation to end before respawning.
			if ((ent->movetype == MOVETYPE_NOCLIP) || (client->anim_event_time == 0))
			{
				// in deathmatch, only wait for attack button
				if (deathmatch->value)
					buttonMask = BUTTON_ATTACK;
				else
					buttonMask = -1;

				if ( ( client->latched_buttons & buttonMask ) ||
					(deathmatch->value && ((int)dmflags->value & DF_FORCE_RESPAWN) ) ||
					CTFMatchOn())
				{
					respawn(ent);
					client->latched_buttons = 0;
				}
			}
		}
		return;
	}

	// add player trail so monsters can follow
	if (!deathmatch->value)
		if (!visible (ent, PlayerTrail_LastSpot() ) )
			PlayerTrail_Add (ent->s.old_origin);

	client->latched_buttons = 0;
}
