/*
Copyright (C) 1997-2001 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
#include "q_shared.h"

//============================================================================

/*
============
COM_SkipPath
============
*/
char *COM_SkipPath (char *pathname)
{
	char	*last;

	last = pathname;
	while (*pathname)
	{
		if (*pathname=='/')
			last = pathname+1;
		pathname++;
	}
	return last;
}

/*
============
COM_StripExtension
============
*/
void COM_StripExtension (char *in, char *out)
{
	while (*in && *in != '.')
		*out++ = *in++;
	*out = 0;
}

/*
============
COM_FileExtension
============
*/
char *COM_FileExtension (char *in)
{
	static char exten[8];
	int		i;

	while (*in && *in != '.')
		in++;
	if (!*in)
		return "";
	in++;
	for (i=0 ; i<7 && *in ; i++,in++)
		exten[i] = *in;
	exten[i] = 0;
	return exten;
}

/*
============
COM_FileBase
============
*/
void COM_FileBase (char *in, char *out)
{
	char *p, *start, *end;
	int length;

	end = in + strlen(in);
	start = in;

	for (p = end-1 ; p > in ; p--)
	{
		if (*p == '.')
			end = p;
		else if (*p == '/')
		{
			start = p + 1;
			break;
		}
	}

	length = end - start;
	if (length <= 0)
		strcpy (out, "?empty filename?");
	else
	{
		if (length > 31)
			length = 31;
		memcpy (out, start, length);
		out[length] = 0;
	}
}

/*
============
COM_FilePath

Returns the path up to, but not including the last /
============
*/
void COM_FilePath (char *in, char *out)
{
	char *s;

	s = in + strlen(in) - 1;

	while (s != in && *s != '/')
		s--;

	strncpy (out,in, s-in);
	out[s-in] = 0;
}


/*
==================
COM_DefaultExtension
==================
*/
void COM_DefaultExtension (char *path, char *extension)
{
	char    *src;
//
// if path doesn't have a .EXT, append extension
// (extension should include the .)
//
	src = path + strlen(path) - 1;

	while (*src != '/' && src != path)
	{
		if (*src == '.')
			return;                 // it has an extension
		src--;
	}

	strcat (path, extension);
}

/*
============================================================================

					BYTE ORDER FUNCTIONS

============================================================================
*/

#if !defined(ENDIAN_LITTLE) && !defined(ENDIAN_BIG)
short   (*BigShort) (short l);
short   (*LittleShort) (short l);
int     (*BigLong) (int l);
int     (*LittleLong) (int l);
float   (*BigFloat) (float l);
float   (*LittleFloat) (float l);
#endif

short   ShortSwap (short l)
{
	qbyte    b1, b2;

	b1 = l&255;
	b2 = (l>>8)&255;

	return (b1<<8) + b2;
}

#if !defined(ENDIAN_LITTLE) && !defined(ENDIAN_BIG)
short   ShortNoSwap (short l)
{
	return l;
}
#endif

int    LongSwap (int l)
{
	qbyte    b1, b2, b3, b4;

	b1 = l&255;
	b2 = (l>>8)&255;
	b3 = (l>>16)&255;
	b4 = (l>>24)&255;

	return ((int)b1<<24) + ((int)b2<<16) + ((int)b3<<8) + b4;
}

#if !defined(ENDIAN_LITTLE) && !defined(ENDIAN_BIG)
int     LongNoSwap (int l)
{
	return l;
}
#endif

float FloatSwap (float f)
{
	union
	{
		float   f;
		qbyte   b[4];
	} dat1, dat2;


	dat1.f = f;
	dat2.b[0] = dat1.b[3];
	dat2.b[1] = dat1.b[2];
	dat2.b[2] = dat1.b[1];
	dat2.b[3] = dat1.b[0];
	return dat2.f;
}

#if !defined(ENDIAN_LITTLE) && !defined(ENDIAN_BIG)
float FloatNoSwap (float f)
{
	return f;
}
#endif


/*
================
Swap_Init
================
*/
void Swap_Init (void)
{
#if !defined(ENDIAN_LITTLE) && !defined(ENDIAN_BIG)
	qbyte    swaptest[2] = {1,0};

// set the byte swapping variables in a portable manner
	if ( *(short *)swaptest == 1)
	{
		BigShort = ShortSwap;
		LittleShort = ShortNoSwap;
		BigLong = LongSwap;
		LittleLong = LongNoSwap;
		BigFloat = FloatSwap;
		LittleFloat = FloatNoSwap;
	}
	else
	{
		BigShort = ShortNoSwap;
		LittleShort = ShortSwap;
		BigLong = LongNoSwap;
		LittleLong = LongSwap;
		BigFloat = FloatNoSwap;
		LittleFloat = FloatSwap;
	}
#endif
}



/*
============
va

does a varargs printf into a temp buffer, so I don't need to have
varargs versions of all text functions.
FIXME: make this buffer size safe someday
============
*/
char	*va(char *format, ...)
{
	va_list		argptr;
	static char		string[1024];

	va_start (argptr, format);
	vsnprintf (string, sizeof(string), format, argptr);
	va_end (argptr);

	return string;
}


char	com_token[MAX_TOKEN_CHARS];

/*
==============
COM_ParseExt

Parse a token out of a string
==============
*/
char *COM_ParseExt (char **data_p, qboolean nl)
{
	int		c;
	int		len;
	char	*data;
	qboolean newlines = qfalse;

	data = *data_p;
	len = 0;
	com_token[0] = 0;

	if (!data)
	{
		*data_p = NULL;
		return "";
	}

// skip whitespace
skipwhite:
	while ( (c = *data) <= ' ')
	{
		if (c == 0)
		{
			*data_p = NULL;
			return "";
		}
		if (c == '\n')
			newlines = qtrue;
		data++;
	}

	if ( newlines && !nl ) {
		*data_p = data;
		return com_token;
	}

// skip // comments
	if (c == '/' && data[1] == '/')
	{
		while (*data && *data != '\n')
			data++;
		goto skipwhite;
	}

// handle quoted strings specially
	if (c == '\"')
	{
		data++;
		while (1)
		{
			c = *data++;
			if (c=='\"' || !c)
			{
				com_token[len] = 0;
				*data_p = data;
				return com_token;
			}
			if (len < MAX_TOKEN_CHARS)
			{
				com_token[len] = c;
				len++;
			}
		}
	}

// parse a regular word
	do
	{
		if (len < MAX_TOKEN_CHARS)
		{
			com_token[len] = c;
			len++;
		}
		data++;
		c = *data;
	} while (c>32);

	if (len == MAX_TOKEN_CHARS)
	{
//		Com_Printf ("Token exceeded %i chars, discarded.\n", MAX_TOKEN_CHARS);
		len = 0;
	}
	com_token[len] = 0;

	*data_p = data;
	return com_token;
}

/*
===============
Com_PageInMemory

===============
*/
int	paged_total;

void Com_PageInMemory (qbyte *buffer, int size)
{
	int		i;

	for (i=size-1 ; i>0 ; i-=4096)
		paged_total += buffer[i];
}



/*
============================================================================

					LIBRARY REPLACEMENT FUNCTIONS

============================================================================
*/

void Q_strncpyz (char *dest, char *src, int size)
{
	strncpy (dest, src, size-1);
	dest[size-1] = 0;
}


void Com_sprintf (char *dest, int size, char *fmt, ...)
{
	int		len;
	va_list	argptr;
	char	bigbuffer[0x10000];

	va_start (argptr,fmt);
	len = vsnprintf (bigbuffer, sizeof(bigbuffer), fmt, argptr);
	va_end (argptr);
	if (len >= size)
		Com_Printf ("Com_sprintf: overflow of %i in %i\n", len, bigbuffer, size);
	strncpy (dest, bigbuffer, size-1);
}

/*
=====================================================================

  INFO STRINGS

=====================================================================
*/

/*
===============
Info_ValueForKey

Searches the string for the given
key and returns the associated value, or an empty string.
===============
*/
char *Info_ValueForKey (char *s, char *key)
{
	char	pkey[512];
	static	char value[2][512];	// use two buffers so compares
								// work without stomping on each other
	static	int	valueindex;
	char	*o;

	valueindex ^= 1;
	if (*s == '\\')
		s++;
	while (1)
	{
		o = pkey;
		while (*s != '\\')
		{
			if (!*s)
				return "";
			*o++ = *s++;
		}
		*o = 0;
		s++;

		o = value[valueindex];

		while (*s != '\\' && *s)
		{
			if (!*s)
				return "";
			*o++ = *s++;
		}
		*o = 0;

		if (!strcmp (key, pkey) )
			return value[valueindex];

		if (!*s)
			return "";
		s++;
	}
}

void Info_RemoveKey (char *s, char *key)
{
	char	*start;
	char	pkey[512];
	char	value[512];
	char	*o;

	if (strstr (key, "\\"))
	{
//		Com_Printf ("Can't use a key with a \\\n");
		return;
	}

	while (1)
	{
		start = s;
		if (*s == '\\')
			s++;
		o = pkey;
		while (*s != '\\')
		{
			if (!*s)
				return;
			*o++ = *s++;
		}
		*o = 0;
		s++;

		o = value;
		while (*s != '\\' && *s)
		{
			if (!*s)
				return;
			*o++ = *s++;
		}
		*o = 0;

		if (!strcmp (key, pkey) )
		{
			strcpy (start, s);	// remove this part
			return;
		}

		if (!*s)
			return;
	}

}


/*
==================
Info_Validate

Some characters are illegal in info strings because they
can mess up the server's parsing
==================
*/
qboolean Info_Validate (char *s)
{
	if (strstr (s, "\""))
		return qfalse;
	if (strstr (s, ";"))
		return qfalse;
	return qtrue;
}

void Info_SetValueForKey (char *s, char *key, char *value)
{
	char	newi[MAX_INFO_STRING], *v;
	int		c;
	int		maxsize = MAX_INFO_STRING;

	if (strstr (key, "\\") || strstr (value, "\\") )
	{
		Com_Printf ("Can't use keys or values with a \\\n");
		return;
	}

	if (strstr (key, ";") )
	{
		Com_Printf ("Can't use keys or values with a semicolon\n");
		return;
	}

	if (strstr (key, "\"") || strstr (value, "\"") )
	{
		Com_Printf ("Can't use keys or values with a \"\n");
		return;
	}

	if (strlen(key) > MAX_INFO_KEY-1 || strlen(value) > MAX_INFO_KEY-1)
	{
		Com_Printf ("Keys and values must be < 64 characters.\n");
		return;
	}
	Info_RemoveKey (s, key);
	if (!value || !strlen(value))
		return;

	Com_sprintf (newi, sizeof(newi), "\\%s\\%s", key, value);

	if (strlen(newi) + strlen(s) > maxsize)
	{
		Com_Printf ("Info string length exceeded\n");
		return;
	}

	// only copy ascii values
	s += strlen(s);
	v = newi;
	while (*v)
	{
		c = *v++;
		c &= 127;		// strip high bits
		if (c >= 32 && c < 127)
			*s++ = c;
	}
	*s = 0;
}

//====================================================================

void Q_Obituary ( void *victim, int gender, void *attacker, int mod, char *message, char *message2 )
{
	message [0] = 0;
	message2[0] = 0;

	switch (mod)
	{
		case MOD_SUICIDE:
			strcpy ( message, "suicides" );
			break;
		case MOD_FALLING:
			strcpy ( message, "cratered" );
			break;
		case MOD_CRUSH:
			strcpy ( message, "was squished" );
			break;
		case MOD_WATER:
			strcpy ( message, "sank like a rock" );
			break;
		case MOD_SLIME:
			strcpy ( message, "melted" );
			break;
		case MOD_LAVA:
			strcpy ( message, "does a back flip into the lava" );
			break;
		case MOD_EXPLOSIVE:
		case MOD_BARREL:
			strcpy ( message, "blew up" );
			break;
		case MOD_EXIT:
			strcpy ( message, "found a way out" );
			break;
		case MOD_TARGET_LASER:
			strcpy ( message, "saw the light" );
			break;
		case MOD_TARGET_BLASTER:
			strcpy ( message, "got blasted" );
			break;
		case MOD_BOMB:
		case MOD_SPLASH:
		case MOD_TRIGGER_HURT:
			strcpy ( message, "was in the wrong place" );
			break;
	}

	if ( attacker == victim )
	{
		switch (mod)
		{
		case MOD_HELD_GRENADE:
			strcpy ( message, "tried to put the pin back in" );
			break;
		case MOD_LASER_TRAP:
			if ( gender == GENDER_FEMALE )
				strcpy ( message, "didn't see her own laser trap" );
			else if ( gender == GENDER_MALE )
				strcpy ( message, "didn't see his own laser trap" );
			else
				strcpy ( message, "didn't see its own laser trap" );
			break;
		case MOD_HG_SPLASH:
		case MOD_G_SPLASH:
			if ( gender == GENDER_FEMALE )
				strcpy ( message, "tripped on her own grenade" );
			else if ( gender == GENDER_MALE )
				strcpy ( message, "tripped on his own grenade" );
			else
				strcpy ( message, "tripped on its own grenade" );
			break;
		case MOD_R_SPLASH:
			if ( gender == GENDER_FEMALE )
				strcpy ( message, "blew herself up" );
			else if ( gender == GENDER_MALE )
				strcpy ( message, "blew himself up" );
			else
				strcpy ( message, "blew itself up" );
			break;
		case MOD_BFG_BLAST:
			strcpy ( message, "should have used a smaller gun" );
			break;
		default:
			if ( gender == GENDER_FEMALE )
				strcpy ( message, "killed herself" );
			else if ( gender == GENDER_MALE )
				strcpy ( message, "killed himself" );
			else
				strcpy ( message, "killed itself" );
			break;
		}
	}

	// we already have one
	if ( message[0] ) {
		return;
	}

	if ( attacker ) {
		switch (mod)
		{
		case MOD_BLASTER:
			strcpy ( message, "was blasted by" );
			break;
		case MOD_SHOTGUN:
			strcpy ( message, "was gunned down by" );
			break;
	/*qbism remove case MOD_SSHOTGUN:
			strcpy ( message, "was blown away by" );
			strcpy ( message2, "'s super shotgun" );
			break; */
		case MOD_MACHINEGUN:
			strcpy ( message, "was machinegunned by" );
			break;
		case MOD_CHAINGUN:
			strcpy ( message, "was cut in half by" );
			strcpy ( message2, "'s chaingun" );
			break;
		case MOD_GRENADE:
			strcpy ( message, "was popped by" );
			strcpy ( message2, "'s grenade" );
			break;
		case MOD_G_SPLASH:
			strcpy ( message, "was shredded by" );
			strcpy ( message, "'s shrapnel" );
			break;
		case MOD_ROCKET:
			strcpy ( message, "ate" );
			strcpy ( message2, "'s rocket" );
			break;
		case MOD_R_SPLASH:
			strcpy ( message, "almost dodged" );
			strcpy ( message2, "'s rocket" );
			break;
		case MOD_HYPERBLASTER:
			strcpy ( message, "was melted by" );
			strcpy ( message2,"'s hyperblaster" );
			break;
		case MOD_RAILGUN:
			strcpy ( message, "was railed by" );
			break;
		case MOD_BFG_LASER:
			strcpy ( message, "saw the pretty lights from" );
			strcpy ( message2, "'s BFG" );
			break;
		case MOD_BFG_BLAST:
			strcpy ( message, "was disintegrated by" );
			strcpy ( message2, "'s BFG blast" );
			break;
		case MOD_BFG_EFFECT:
			strcpy ( message, "couldn't hide from" );
			strcpy ( message2, "'s BFG" );
			break;
		case MOD_HANDGRENADE:
			strcpy ( message, "caught" );
			strcpy ( message2, "'s handgrenade" );
			break;
		case MOD_HG_SPLASH:
			strcpy ( message, "didn't see" );
			strcpy ( message2, "'s handgrenade" );
			break;
		case MOD_HELD_GRENADE:
			strcpy ( message, "feels" );
			strcpy ( message2, "'s pain" );
			break;
		case MOD_TELEFRAG:
			strcpy ( message, "tried to invade" );
			strcpy ( message2, "'s personal space" );
			break;
		case MOD_LASER_TRAP:
			strcpy ( message, "got into" );
			strcpy ( message2, "'s laser trap" );
			break;
			//ZOID
		case MOD_GRAPPLE:
			strcpy ( message, "was caught by" );
			strcpy ( message2, "'s grapple" );
			break;
			//ZOID

		default:
			strcpy ( message, "was killed by" );
			break;
		}

		if ( message[0] ) {
			return;
		}
	}

	strcpy ( message, " died." );
}
