/*
Copyright (C) 1997-2001 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "ui_local.h"

void M_SetMultiplayerStatusBar ( const char *string );

/*
=============================================================================

PLAYER CONFIG MENU

=============================================================================
*/
static menuframework_s	s_player_config_menu;
static menufield_s		s_player_name_field;
static menulist_s		s_player_model_box;
static menulist_s		s_player_skin_box;
static menulist_s		s_player_handedness_box;
static menulist_s		s_player_rate_box;
static menuseparator_s	s_player_skin_title;
static menuseparator_s	s_player_model_title;
static menuseparator_s	s_player_hand_title;
static menuseparator_s	s_player_rate_title;

#define MAX_DISPLAYNAME 16
#define MAX_PLAYERMODELS 1024

//splitmodels
typedef struct
{
	int		nskins;
	char	**skindisplaynames;
	char	displayname[MAX_DISPLAYNAME];
	char	directory[MAX_QPATH];
	int		upperframe;
	int		lowerframe;
} playermodelinfo_s;

static playermodelinfo_s s_pmi[MAX_PLAYERMODELS];
static char *s_pmnames[MAX_PLAYERMODELS];
static int s_numplayermodels;

static int rate_tbl[] = { 2500, 3200, 5000, 10000, 25000, 0 };
static const char *rate_names[] = { "28.8 Modem", "33.6 Modem", "Single ISDN",
	"Dual ISDN/Cable", "T1/LAN", "User defined", 0 };

static void HandednessCallback( void *unused )
{
	trap_Cvar_SetValue( "hand", s_player_handedness_box.curvalue );
}

static void RateCallback( void *unused )
{
	if (s_player_rate_box.curvalue != sizeof(rate_tbl) / sizeof(*rate_tbl) - 1)
		trap_Cvar_SetValue( "rate", rate_tbl[s_player_rate_box.curvalue] );
}

static void ModelCallback( void *unused )
{
	s_player_skin_box.itemnames = s_pmi[s_player_model_box.curvalue].skindisplaynames;
	s_player_skin_box.curvalue = 0;
}

//jal: removed
/*
static qboolean IconOfSkinExists( char *skin, char *pcxfiles, int npcxfiles )
{
	int i;
	char scratch[1024];
	char *ptr, len = 0;

	strcpy( scratch, skin );
	*strrchr( scratch, '.' ) = 0;
	strcat( scratch, "_i.pcx" );

	ptr = pcxfiles;

	for ( i = 0; i < npcxfiles; i++, ptr += len+1 )
	{
		len = strlen (ptr);

		if ( strcmp( ptr, scratch ) == 0 )
			return qtrue;
	}

	return qfalse;
}
*/


/*
//---------------
LoadSplitSkinFile - SPLITMODELS (jal removed)
//---------------
*/

	// bye, bye SPS


/*
================
ui_PickPresentationFrames - splitmodels

  All this work is just to extract the LEGS_STAND first frame and
  the TORSO_STAND first frame.

  NOTE: This read begins the animation count from 0, so
  the animation number in here is one value less as in the
  client equivalent code (cl_parse).
================
*/
qboolean ui_PickPresentationFrames (char *model_name, int playernumber)
{
	qbyte *buf;
	char configName[MAX_QPATH];
	int 	anim_data[4][PM_MAX_ANIMS];
	//0 = first frame
	//1 = number of frames
	//2 = looping frames
	//3 = fps
	char *ptr, *token;
	int 	rounder, counter, i, offset, nooffset;

	if (!model_name)
		return qfalse;

	rounder = counter = nooffset = 0;

	//read a script
	Com_sprintf( configName, sizeof( configName ), "models/players/%s/animation.cfg", model_name);

	if (!trap_FS_FileExists(configName)){
		return qfalse;
	}

	trap_FS_LoadFile ( configName, (void **)&buf );
	if ( !buf ) {
		trap_FS_FreeFile ( buf );
		return qfalse;
	}

	//proceed
	ptr = ( char * )buf;
	while ( ptr )
	{
		token = COM_ParseExt ( &ptr, qtrue );
		if ( !token ) {
			break;

		} else if (*token < '0' || *token > '9') {

			//string

			//headoffset (ignored)
			if ( !Q_stricmp (token, "headoffset") )
			{
				token = COM_ParseExt ( &ptr, qfalse );
				if ( !token )
					break;
				token = COM_ParseExt ( &ptr, qfalse );
				if ( !token )
					break;
				token = COM_ParseExt ( &ptr, qfalse );
				if ( !token )
					break;

			} else if ( !Q_stricmp (token, "nooffset") )
				nooffset = 1;

		} else {

			//values
			i = (int)atoi(token);

			anim_data[rounder][counter] = i;

			rounder++;
			if (rounder > 3){
				rounder = 0;
				counter++;
			}
		}
	}

	trap_FS_FreeFile ( buf );

	//reorganize to make my life easier
	for (i=0 ; i<counter ; i++){
		anim_data[1][i] += anim_data[0][i];
	}

	// Fix the rates to fit my looping calculations
	for (i = 0; i < counter; ++i){
		if (anim_data[2][i])
			anim_data[2][i] -= 1;

		anim_data[1][i] -= 1;
	}

	if (!nooffset){
		//fix offset:  LEGS_CRWALK = 14-1, BOTH_CRDEATH = 7-1
		offset = anim_data[0][13] - anim_data[0][6];

		// Remove offset from the lower animations
		for (i = 13; i < counter; ++i){
			anim_data[0][i] -= offset;
			anim_data[1][i] -= offset;
		}
	}

	//grab the frames I want: TORSO_STAND = 12-1,	LEGS_STAND = 23-1
	s_pmi[playernumber].upperframe = anim_data[0][11];
	s_pmi[playernumber].lowerframe = anim_data[0][22];

	return qtrue;
}

/*
==============================
//	FolderIsDuplicated - SPLITMODELS
==============================
*/
static qboolean FolderIsDuplicated( char *folder, int numplayermodels )
{
	int k;

	//check if this folder has already been registered

	for ( k = 0; k < numplayermodels; k++){
		if (!strcmp(s_pmi[k].directory, folder)){
			return qtrue;
		}
	}

	return qfalse;
}

/*
==============================
//	SkinIsDuplicated - SPLITMODELS
==============================
*/
static qboolean SkinIsRegistered( char *skinptr, int numskins, char **skinnames )
{
	int k;
	char scratch[1024];
	char skinname[1024];
	strcpy( scratch, skinptr );
	strcpy( skinname, skinptr );
	*strstr(skinname, ".") = 0;
	*strstr(scratch, "_") = 0;

	strcpy(skinname, skinname + strlen(scratch) + 1);

	//check if this skin has already been registered

	for ( k = 0; k < numskins; k++){

		if (!strcmp(skinnames[k], skinname)){
			return qtrue;
		}
	}

	return qfalse;
}

/*
==============================
//PlayerConfig_ScanDirectories - SPLITMODELS
==============================
*/
static void PlayerConfig_ScanDirectories( void )
{
	int		ndirs, dirlen;
	char	dirnames[1024];
	char	scratch[1024];
	char	*path = NULL, *dirptr;
	int i;

	s_numplayermodels = 0;

	// get a list of directories
	//--------------------------

	do
	{
		if ( ndirs = trap_FS_ListFiles( "models/players", "/", dirnames, sizeof(dirnames) ) ){
			break;
		}
		//qbism added "\\"
		if ( ndirs = trap_FS_ListFiles( "models/players", "\\", dirnames, sizeof(dirnames) ) ){
			break;
		}
	} while ( path );

	if ( !ndirs || !dirnames ) {
		return;
	}

	// go through the subdirectories
	//------------------------------

	ndirs = ndirs;
	if ( ndirs > MAX_PLAYERMODELS )
		ndirs = MAX_PLAYERMODELS;
	dirptr = dirnames;

	for ( i = 0; i < ndirs; i++, dirptr += dirlen+1 )
	{
		int		k;
		char	pcxnames[1024];
		char	**skinnames, *skinptr;
		char	**skinfilenames;
		int		npcxfiles;
		int		nskins = 0, skinlen;

		dirlen = strlen(dirptr);

		if (dirlen && dirptr[dirlen-1]=='/')
			dirptr[dirlen-1]='\0';

		if (!strcmp(dirptr,".") || !strcmp(dirptr,".."))
			continue;

		//check for duplicated folders
		if (FolderIsDuplicated(dirptr, s_numplayermodels))
			continue;

		// verify the existence of at least one skin file
		npcxfiles = trap_FS_ListFiles( va("models/players/%s",dirptr), ".skin", pcxnames, sizeof(pcxnames) );
		if ( !npcxfiles )
			continue;

		skinptr = pcxnames;
		skinlen = 0;

		//there are valid skins. So:
		skinnames = UI_malloc( sizeof( char * ) * ( npcxfiles + 1 ) );
		skinptr = pcxnames;

		//copy the valid skins
		for ( nskins = 0, k = 0; k < npcxfiles; k++, skinptr += skinlen+1 )
		{
			skinlen = strlen (skinptr);

			if (!SkinIsRegistered(skinptr, nskins, skinnames))
			{
				skinnames[nskins] = strdup( skinptr );
				*strstr(skinnames[nskins], ".") = 0;

				strcpy (scratch, skinnames[nskins]);
				*strstr(scratch, "_") = 0;

				//add & count
				strcpy (skinnames[nskins], skinnames[nskins] + strlen(scratch) + 1);
				nskins++;
			}
		}

		//Ignore the model if none of the skin files were valid
		if (!nskins)
			continue;

		//read the upper & lower frames for the presentation:
		if (!ui_PickPresentationFrames (dirptr, s_numplayermodels)){
			s_pmi[s_numplayermodels].upperframe = 100;
			s_pmi[s_numplayermodels].lowerframe = 100;
		}

		//free
		skinfilenames = UI_malloc( sizeof( char * ) * ( nskins + 1 ) );

		for ( k = 0; k < nskins; k++)
			skinfilenames[k] = strdup( skinnames[k] );
			//strcpy (skinfilenames[k], skinnames[k]);

		for ( k = 0; k < npcxfiles; k++)
			skinnames[k] = 0;

		UI_free( skinnames );


		// at this point we have a valid player model
		s_pmi[s_numplayermodels].nskins = nskins;
		s_pmi[s_numplayermodels].skindisplaynames = skinfilenames;

		// make short name for the model
		Q_strncpyz( s_pmi[s_numplayermodels].displayname, dirptr, sizeof(s_pmi[s_numplayermodels].displayname) );
		strcpy( s_pmi[s_numplayermodels].directory, dirptr );

		s_numplayermodels++;
	}
}

static int pmicmpfnc( const void *_a, const void *_b )
{
	const playermodelinfo_s *a = ( const playermodelinfo_s * ) _a;
	const playermodelinfo_s *b = ( const playermodelinfo_s * ) _b;


	//sort by male, female, then alphabetical
	//splitmodels: ...

	/*
	if ( strcmp( a->directory, "male" ) == 0 )
		return -1;
	else if ( strcmp( b->directory, "male" ) == 0 )
		return 1;

	if ( strcmp( a->directory, "female" ) == 0 )
		return -1;
	else if ( strcmp( b->directory, "female" ) == 0 )
		return 1;
	*/
	return strcmp( a->directory, b->directory );
}


qboolean PlayerConfig_MenuInit( void )
{
	char currentdirectory[1024];
	char currentskin[1024];
	int i = 0;
	int currentdirectoryindex = 0;
	int currentskinindex = 0;
	int	hand = trap_Cvar_VariableValue( "hand" );
	char *name = trap_Cvar_VariableString ( "name" );
	char *skin = trap_Cvar_VariableString ( "skin" );
	int y = 0;
	int y_offset = UI_StringHeightOffset ( 0 );

	static const char *handedness[] = { "right", "left", "center", 0 };

	PlayerConfig_ScanDirectories();

	if (s_numplayermodels <= 0)
		return qfalse;

	if ( hand < 0 || hand > 2 )
		trap_Cvar_SetValue( "hand", 0 );

	strcpy( currentdirectory, skin );

	if ( strchr( currentdirectory, '/' ) )
	{
		strcpy( currentskin, strchr( currentdirectory, '/' ) + 1 );
		*strchr( currentdirectory, '/' ) = 0;
	}
	else if ( strchr( currentdirectory, '\\' ) )
	{
		strcpy( currentskin, strchr( currentdirectory, '\\' ) + 1 );
		*strchr( currentdirectory, '\\' ) = 0;
	}
	else
	{
		strcpy( currentdirectory, DEFAULT_PLAYERMODEL );//splitmodels
		strcpy( currentskin, DEFAULT_PLAYERSKIN );
	}

	qsort( s_pmi, s_numplayermodels, sizeof( s_pmi[0] ), pmicmpfnc );

	memset( s_pmnames, 0, sizeof( s_pmnames ) );
	for ( i = 0; i < s_numplayermodels; i++ )
	{
		s_pmnames[i] = s_pmi[i].displayname;
		if ( Q_stricmp( s_pmi[i].directory, currentdirectory ) == 0 )
		{
			int j;

			currentdirectoryindex = i;

			for ( j = 0; j < s_pmi[i].nskins; j++ )
			{
				if ( Q_stricmp( s_pmi[i].skindisplaynames[j], currentskin ) == 0 )
				{
					currentskinindex = j;
					break;
				}
			}
		}
	}

	s_player_config_menu.x = uis.vidWidth / 2 - 95;
	s_player_config_menu.y = uis.vidHeight / 2 - 97;
	s_player_config_menu.nitems = 0;

	s_player_name_field.generic.type = MTYPE_FIELD;
	s_player_name_field.generic.name = "name";
	s_player_name_field.generic.callback = 0;
	s_player_name_field.generic.x		= 0;
	s_player_name_field.generic.y		= y;
	s_player_name_field.length	= 20;
	s_player_name_field.visible_length = 20;
	strcpy( s_player_name_field.buffer, name );
	s_player_name_field.cursor = strlen( name );
	y += y_offset*3;

	s_player_model_title.generic.type = MTYPE_SEPARATOR;
	s_player_model_title.generic.name = "model";
	s_player_model_title.generic.x    = -8;
	s_player_model_title.generic.y	 = y+=y_offset;

	s_player_model_box.generic.type = MTYPE_SPINCONTROL;
	s_player_model_box.generic.x	= -56;
	s_player_model_box.generic.y	= y+=y_offset;
	s_player_model_box.generic.callback = ModelCallback;
	s_player_model_box.generic.cursor_offset = -72;
	s_player_model_box.curvalue = currentdirectoryindex;
	s_player_model_box.itemnames = s_pmnames;

	s_player_skin_title.generic.type = MTYPE_SEPARATOR;
	s_player_skin_title.generic.name = "skin";
	s_player_skin_title.generic.x    = -16;
	s_player_skin_title.generic.y	 = y+=y_offset;

	s_player_skin_box.generic.type			= MTYPE_SPINCONTROL;
	s_player_skin_box.generic.x				= -56;
	s_player_skin_box.generic.y				= y+=y_offset;
	s_player_skin_box.generic.name			= 0;
	s_player_skin_box.generic.callback		= 0;
	s_player_skin_box.generic.cursor_offset = -72;
	s_player_skin_box.curvalue				= currentskinindex;
	s_player_skin_box.itemnames				= s_pmi[currentdirectoryindex].skindisplaynames;
	y+=y_offset;

	s_player_hand_title.generic.type		= MTYPE_SEPARATOR;
	s_player_hand_title.generic.name		= "handedness";
	s_player_hand_title.generic.x			= 32;
	s_player_hand_title.generic.y			= y+=y_offset;

	s_player_handedness_box.generic.type	= MTYPE_SPINCONTROL;
	s_player_handedness_box.generic.x		= -56;
	s_player_handedness_box.generic.y		= y+=y_offset;
	s_player_handedness_box.generic.name	= 0;
	s_player_handedness_box.generic.cursor_offset = -72;
	s_player_handedness_box.generic.callback = HandednessCallback;
	s_player_handedness_box.curvalue		= trap_Cvar_VariableValue( "hand" );
	s_player_handedness_box.itemnames		= handedness;
	y+=y_offset;

	for (i = 0; i < sizeof(rate_tbl) / sizeof(*rate_tbl) - 1; i++)
		if (trap_Cvar_VariableValue("rate") == rate_tbl[i])
			break;

	s_player_rate_title.generic.type = MTYPE_SEPARATOR;
	s_player_rate_title.generic.name = "connect speed";
	s_player_rate_title.generic.x    = 56;
	s_player_rate_title.generic.y	 = y+=y_offset;

	s_player_rate_box.generic.type		= MTYPE_SPINCONTROL;
	s_player_rate_box.generic.x			= -56;
	s_player_rate_box.generic.y			= y+=y_offset;
	s_player_rate_box.generic.name		= 0;
	s_player_rate_box.generic.cursor_offset = -72;
	s_player_rate_box.generic.callback = RateCallback;
	s_player_rate_box.curvalue			= i;
	s_player_rate_box.itemnames			= rate_names;

	Menu_AddItem( &s_player_config_menu, &s_player_name_field );
	Menu_AddItem( &s_player_config_menu, &s_player_model_title );
	Menu_AddItem( &s_player_config_menu, &s_player_model_box );
	if ( s_player_skin_box.itemnames )
	{
		Menu_AddItem( &s_player_config_menu, &s_player_skin_title );
		Menu_AddItem( &s_player_config_menu, &s_player_skin_box );
	}
	Menu_AddItem( &s_player_config_menu, &s_player_hand_title );
	Menu_AddItem( &s_player_config_menu, &s_player_handedness_box );
	Menu_AddItem( &s_player_config_menu, &s_player_rate_title );
	Menu_AddItem( &s_player_config_menu, &s_player_rate_box );

	Menu_Init ( &s_player_config_menu );

	return qtrue;
}

/*
=================
PlayerConfig_MenuDraw SPLITMODELS (ui rendering playermodel)
=================
*/
typedef enum {
	LOWER,
	UPPER,
	HEAD,

	PLAYERMODEL_PARTS
} playermodelparts_t;

void PlayerConfig_MenuDraw( void )
{
	refdef_t refdef;
	char scratch[MAX_QPATH];
	int w, h, i;
	int	entcounter;
	entity_t entities[PLAYERMODEL_PARTS];

	memset( &refdef, 0, sizeof( refdef ) );

	trap_Vid_GetCurrentInfo ( &w, &h );

	//model window size
	refdef.width = 300;
	refdef.height = 400;

	//placement
	refdef.x = (w / 2 - (refdef.width/2)) + 80;
	refdef.y = (h / 2 - (refdef.height/2)) + 80;

	//placement: center of screen
	//refdef.x = (w / 2 - (refdef.width/2));
	//refdef.y = (h / 2 - (refdef.height/2));

	refdef.fov_x = 30;
	refdef.fov_y = 30;
	refdef.time = trap_CL_GetTime() * 0.001f;
	refdef.areabits = 0;
	refdef.rdflags = RDF_NOWORLDMODEL;

	entcounter = 0;

	if (( s_pmi[s_player_model_box.curvalue].skindisplaynames ) && (s_pmi[s_player_model_box.curvalue].directory))
	{
		orientation_t	headlerped, torsolerped;
		static			vec3_t angles;

		memset( &entities, 0, sizeof( entities ) );

		//-----
		//legs
		entities[LOWER].type = ET_MODEL;
		entities[LOWER].flags = RF_FULLBRIGHT | RF_NOSHADOW;

		Com_sprintf( scratch, sizeof( scratch ), "models/players/%s/lower.md3", s_pmi[s_player_model_box.curvalue].directory);
		entities[LOWER].model = trap_RegisterModel( scratch );

		Com_sprintf( scratch, sizeof( scratch ), "models/players/%s/lower_%s.skin", s_pmi[s_player_model_box.curvalue].directory, s_pmi[s_player_model_box.curvalue].skindisplaynames[s_player_skin_box.curvalue] );
		entities[LOWER].customShader = NULL;
		entities[LOWER].customSkin = trap_RegisterSkinFile( scratch );

		entities[LOWER].origin[0] = 180;
		entities[LOWER].origin[2] = 0;
		entities[LOWER].scale = 1.0f;
		VectorCopy( entities[LOWER].origin, entities[LOWER].oldorigin );

		//rotate
		angles[1] += 1.0f;

		if ( angles[1] > 360 )
			angles[1] -= 360;

		AnglesToAxis ( angles, entities[LOWER].axis );

		entities[LOWER].frame = s_pmi[s_player_model_box.curvalue].lowerframe;
		entities[LOWER].oldframe = entities[LOWER].frame;

		if (entities[LOWER].model){

			entcounter = 1;

			//------
			//torso

			if (trap_R_LerpAttachment( &torsolerped, entities[LOWER].model, entities[LOWER].frame, entities[LOWER].oldframe, entities[LOWER].backlerp, "tag_torso") )
			{
				entities[UPPER] = entities[LOWER];

				Com_sprintf( scratch, sizeof( scratch ), "models/players/%s/upper.md3", s_pmi[s_player_model_box.curvalue].directory );
				entities[UPPER].model = trap_RegisterModel( scratch );

				Com_sprintf( scratch, sizeof( scratch ), "models/players/%s/upper_%s.skin", s_pmi[s_player_model_box.curvalue].directory, s_pmi[s_player_model_box.curvalue].skindisplaynames[s_player_skin_box.curvalue] );
				entities[UPPER].customShader = NULL;
				entities[UPPER].customSkin = trap_RegisterSkinFile( scratch );

				entities[UPPER].frame = s_pmi[s_player_model_box.curvalue].upperframe;
				entities[UPPER].oldframe = entities[UPPER].frame;

				//torso tag (move torso to legs)
				if (entities[UPPER].model){

					for ( i = 0 ; i < 3 ; i++ )
						VectorMA( entities[UPPER].origin, torsolerped.origin[i], entities[LOWER].axis[i], entities[UPPER].origin );

					VectorCopy (entities[UPPER].origin, entities[UPPER].oldorigin);

					Matrix3_Multiply( torsolerped.axis, entities[LOWER].axis, entities[UPPER].axis );

					entcounter++;

					//-------
					//head

					if (trap_R_LerpAttachment( &headlerped, entities[UPPER].model, entities[UPPER].frame, entities[UPPER].oldframe, entities[UPPER].backlerp, "tag_head" ))
					{
						entities[HEAD] = entities[UPPER];

						Com_sprintf( scratch, sizeof( scratch ), "models/players/%s/head.md3", s_pmi[s_player_model_box.curvalue].directory );
						entities[HEAD].model = trap_RegisterModel( scratch );

						Com_sprintf( scratch, sizeof( scratch ), "models/players/%s/head_%s.skin", s_pmi[s_player_model_box.curvalue].directory, s_pmi[s_player_model_box.curvalue].skindisplaynames[s_player_skin_box.curvalue] );
						entities[HEAD].customSkin = trap_RegisterSkinFile( scratch );
						entities[HEAD].customShader = NULL;

						entities[HEAD].frame = 0;
						entities[HEAD].oldframe = entities[HEAD].frame;

						//head tag (move head to torso)
						if (entities[HEAD].model){
							for ( i = 0 ; i < 3 ; i++ )
								VectorMA( entities[HEAD].origin, headlerped.origin[i], entities[UPPER].axis[i], entities[HEAD].origin );

							VectorCopy (entities[HEAD].origin, entities[HEAD].oldorigin);

							Matrix3_Multiply( headlerped.axis, entities[UPPER].axis, entities[HEAD].axis );

							entcounter++;

						}
					}
				}
			}
		}

		if (entcounter == PLAYERMODEL_PARTS){

			refdef.num_entities = PLAYERMODEL_PARTS;
			refdef.entities = entities;
			trap_RenderFrame( &refdef );
		}

		Menu_Draw( &s_player_config_menu );

		//icon
		Com_sprintf( scratch, sizeof( scratch ), "models/players/%s/icon_%s", s_pmi[s_player_model_box.curvalue].directory, s_pmi[s_player_model_box.curvalue].skindisplaynames[s_player_skin_box.curvalue]);
		trap_DrawStretchPic( s_player_config_menu.x - 40, s_player_config_menu.y + 32, 32, 32, 0, 0, 1, 1, colorWhite, trap_RegisterPic(scratch) );
	}
}


const char *PlayerConfig_MenuKey (int key)
{
	int i;
	menucommon_s *item;

	item = Menu_ItemAtCursor ( &s_player_config_menu );

	if ( key == K_ESCAPE || ( (key == K_MOUSE2) && (item->type != MTYPE_SPINCONTROL) &&
		(item->type != MTYPE_SLIDER)) )
	{
		char scratch[1024];

		trap_Cvar_Set( "name", s_player_name_field.buffer );

		Com_sprintf( scratch, sizeof( scratch ), "%s/%s",
			s_pmi[s_player_model_box.curvalue].directory,
			s_pmi[s_player_model_box.curvalue].skindisplaynames[s_player_skin_box.curvalue] );

		trap_Cvar_Set( "skin", scratch );

		for ( i = 0; i < s_numplayermodels; i++ )
		{
			int j;

			for ( j = 0; j < s_pmi[i].nskins; j++ )
			{
				UI_free( s_pmi[i].skindisplaynames[j] );
				s_pmi[i].skindisplaynames[j] = 0;
			}
			UI_free( s_pmi[i].skindisplaynames );
			s_pmi[i].skindisplaynames = 0;
			s_pmi[i].nskins = 0;
		}
	}

	return Default_MenuKey( &s_player_config_menu, key );
}


void M_Menu_PlayerConfig_f (void)
{
	if (!PlayerConfig_MenuInit())
	{
		M_SetMultiplayerStatusBar ( "No valid player models found" );
		return;
	}

	M_SetMultiplayerStatusBar ( NULL );
	M_PushMenu( &s_player_config_menu, PlayerConfig_MenuDraw, PlayerConfig_MenuKey );
}

